%-----------------------------------------------------------------------------%
% Copyright (C) 1996-1997 The University of Melbourne.
% This file may only be copied under the terms of the GNU Library General
% Public License - see the file COPYING.LIB in the Mercury distribution.
%-----------------------------------------------------------------------------%
%
% cfloat_float - part of the interface with the CLP(R) constraint solver. 
%
%	This module defines the operators which mix cfloats and floats.
%	(The reason this is in a different module than the stuff in `cfloat'
%	is because Mercury doesn't support overloading operators with
%	the same module, name, and arity.)
%
% Main author: fjh.
% Stability: low.
%
%-----------------------------------------------------------------------------%

:- module cfloat_float.

:- interface.
:- import_module cfloat, float.

%-----------------------------------------------------------------------------%

	% cfloat-float equality
	%
:- pred ==(cfloat, float).
:- mode ==(ca, in) is semidet.
:- mode ==(co, in) is det.

	% cfloat-float disequality
	%
:- pred \==(cfloat, float).
:- mode \==(ca, in) is semidet.
:- mode \==(co, in) is det.

	% addition
	%
:- func '+'(cfloat, float) = cfloat.
:- mode '+'(ca, in) = ca is semidet.
:- mode '+'(co, in) = ca is det.
:- mode '+'(ca, in) = co is det.
:- mode '+'(co, in) = co is det.

	% subtraction
	%
:- func '-'(cfloat, float) = cfloat.
:- mode '-'(ca, in) = ca is semidet.
:- mode '-'(co, in) = ca is det.
:- mode '-'(ca, in) = co is det.
:- mode '-'(co, in) = co is det.

	% multiplication
	%
:- func '*'(cfloat, float) = cfloat.
:- mode '*'(ca, in) = ca is semidet.
:- mode '*'(co, in) = ca is semidet.	% semidet since eg. X*0=1 fails
:- mode '*'(ca, in) = co is det.
:- mode '*'(co, in) = co is det.

	% division
	% note that division by a zero float results in a runtime error
	% (whereas division by a zero cfloat just fails)
	%
:- func '/'(cfloat, float) = cfloat.
:- mode '/'(ca, in) = ca is semidet.
:- mode '/'(co, in) = ca is det.
:- mode '/'(ca, in) = co is det.
:- mode '/'(co, in) = co is det.

	% X > Y
	%
:- pred '>'(cfloat, float).
:- mode '>'(ca, in) is semidet.
:- mode '>'(co, in) is det.

	% X >= Y
	%
:- pred '>='(cfloat, float).
:- mode '>='(ca, in) is semidet.
:- mode '>='(co, in) is det.

	% X < Y
	%
:- pred '<'(cfloat, float).
:- mode '<'(ca, in) is semidet.
:- mode '<'(co, in) is det.

	% X =< Y
	%
:- pred '=<'(cfloat, float).
:- mode '=<'(ca, in) is semidet.
:- mode '=<'(co, in) is det.

%-----------------------------------------------------------------------------%
%-----------------------------------------------------------------------------%

:- implementation.

X == Y :- cfloat__eq_float(X, Y).
X \== Y :- cfloat__diseq_float(X, Y).

X + Y = Z :- cfloat__plus_float(X, Y, Z).
X - Y = Z :- cfloat__minus_float(X, Y, Z).
X * Y = Z :- cfloat__mult_float(X, Y, Z).
X / Y = Z :- cfloat__div_float(X, Y, Z).

X > Y :- cfloat__gt_float(X, Y).
X >= Y :- cfloat__ge_float(X, Y).
X < Y :- cfloat__lt_float(X, Y).
X =< Y :- cfloat__le_float(X, Y).

%-----------------------------------------------------------------------------%
:- end_module cfloat_float.
%-----------------------------------------------------------------------------%
