#ifndef CLPR_MISC_H
#define CLPR_MISC_H

#include <stdio.h>
#include "standard.h"
#include "solver.h"

	/* some CLP(R) internals */
extern CLPR_int CLPR_solver_id;
extern CLPR_int CLPR_slack_id;
extern CLPR_int CLPR_stamp;
extern CLPR_int CLPR_trtop;
extern CLPR_int **CLPR_trail;

typedef CLPR_int CLPR_svar;

	/*
	** CLP(R) streams - these might need to be
	** used if the solver gets into trouble
	*/
extern FILE *CLPR_error_stream;
extern FILE *CLPR_outfile;
extern FILE *CLPR_dumpF;

	/* some CLP(R) functions */
void CLPR_init_solver(void);
void CLPR_init_solver_goal(void);
void CLPR_solver_backtrack(CLPR_int);

	/* a CLP(R) function - used if there is a solver error */
void CLPR_fatal(const char *);

	/* linear equations */
	/*
	** Call CLPR_init_eqn(),
	** then call CLPR_s_eval() any number of times and/or
	** add a constant to CLPR_eqn_scratch->constant,
	** then call CLPR_s_econ(), CLPR_s_ge0(), or CLPR_s_gt0(),
	** which will return TRUE/FALSE.
	*/ 
void CLPR_init_eqn(void);
void CLPR_s_evar(double, CLPR_svar);
CLPR_int CLPR_s_econ(double);
CLPR_int CLPR_s_ge0(double);
CLPR_int CLPR_s_gt0(double);

extern EQN_ptr CLPR_eqn_scratch;

	/* non-linear equations */
CLPR_int CLPR_s_mult(double, CLPR_svar, CLPR_svar, CLPR_svar);
CLPR_int CLPR_s_min(CLPR_svar, CLPR_svar, CLPR_svar);
CLPR_int CLPR_s_max(CLPR_svar, CLPR_svar, CLPR_svar);
CLPR_int CLPR_s_abs(CLPR_svar, CLPR_svar);
CLPR_int CLPR_s_sin(CLPR_svar, CLPR_svar);
CLPR_int CLPR_s_cos(CLPR_svar, CLPR_svar);
CLPR_int CLPR_s_arcsin(CLPR_svar, CLPR_svar);
CLPR_int CLPR_s_arccos(CLPR_svar, CLPR_svar);

	/* pointer to list of currently delayed non-linear equations */
extern NL_EQN_ptr CLPR_nl_eqn_top;

void CLPR_print_nl_eqn(NL_EQN_ptr);

	/* check if a variable is ground, and if so, find its value */
CLPR_int CLPR_is_ground(CLPR_svar, double *);

void CLPR_init(void);

#endif /* CLPR_MISC_H */
