/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: codegen.c                                            *
 ***************************************************************/

#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include "standard.h"
#include "config.h"
#include "parse.h"
#include "symbols.h"
#include "lib.h"
#include "compile.h"
#include "opcodes.h"
#include "emul.h"
#include "main.h"
#include "codegen.h"

#define MAX_VAR_TO_VREC 256
#define MAX_DUMP_VARSTRING 4096

#define typecheck_error() \
	CLPR_type_error("(error) arithmetic/functor type clash\n", 0)
#define typecheck_arith(t1, t2) \
	if (t1 == FUNC || t2 == FUNC) \
		CLPR_type_error("functor in arithmetic term\n", 0)
#define typecheck_clash(t1, t2) \
	if ((t1 == REAL && t2 == FUNC) || (t1 == FUNC && t2 == REAL)) \
		CLPR_type_error("arithmetic/functor type clash\n", 0)
#define check_reg_range(reg) \
	if (reg <= 0 || reg > MAX_REGS) CLPR_fatal("23863");

#define is_void_var(v) (v->occ_top == 1 && v->occ[1].atom == 0)
#define is_single_occ(v) (v->occ_top == 1)
#define is_head_var(v) (v->occ[1].atom == 0)
#define is_headonly_var(v) (v->occ[v->occ_top].atom == 0)
#define is_chunk01only_var(v) \
	(v->occ[1].atom <= 1 && v->occ[v->occ_top].atom <= 1)
#define is_in1atomonly_var(v) (v->occ[1].atom == v->occ[v->occ_top].atom)
#define is_1stinfunc_var(v) (v->occ[1].level > 1)
#define is_toplevel_firstvar(v) (v->occ[1].level == 1)
#define is_lastatomonly_var(v) (v->occ[1].atom == CLPR_n_chunks)

static PTERM *cur_rule_pterm, *cur_head, *cur_atom;
static char *p_name;
static CLPR_int p_val, p_arity, p_r_number;
static CLPR_int cur_atom_no = 0;
static CLPR_int cutreg;
static CLPR_int has_fastcut;
RULE_ptr CLPR_current_rule;
CLPR_int CLPR_reg_lowmark, CLPR_reg_highmark;
CLPR_int CLPR_hascut; 
CLPR_int CLPR_CODEGEN_DEBUG = FALSE;
VREC *CLPR_varlist = NULL, *CLPR_end_varlist = NULL;
VREC *CLPR_vrec_freelist = NULL, *CLPR_vrec_endfreelist = NULL;
CLPR_int CLPR_n_chunks;
VREC CLPR_dummy_vrec;
CLPR_int CLPR_live_regs_top = 0;
CLPR_int CLPR_regused[MAX_REGS + 1];
VREC *CLPR_live_regs[MAX_REGS + 1]; /*** 0 = dead, -1 = non-var ***/
VREC *CLPR_var_to_vrec[MAX_VAR_TO_VREC + 1];

VREC *CLPR_notevar(CLPR_int val);
VREC *CLPR_get_vrec(void);
char *CLPR_maketag(CLPR_int n, char *extra, char *str, CLPR_int arity);
char *CLPR_smalltag(char *str, CLPR_int arity);
char *CLPR_dumpvarnames(PTERM *t);
PTERM *CLPR_makedumpvars(VREC *v);
PTERM *CLPR_makedumpvars0(VREC *v);
PTERM *CLPR_makedumpvars1(VREC *v, PTERM *prev);

static VREC *val_info(CLPR_int val, CLPR_int *reg, CLPR_int *stack_loc, CLPR_int *is_new_reg, CLPR_int *is_new_var);

extern CLPR_int CLPR_CODE_SZ;
extern CLPR_unsigned *CLPR_code;
extern FILE *CLPR_outfile;
extern FILE *CLPR_error_stream;
extern FILE *CLPR_var_file;
extern CLPR_int CLPR_pcode;
extern CLPR_int CLPR_epcode;
extern CLPR_int CLPR_debug_flg;
extern CLPR_int CLPR_must_do_save;
extern CLPR_unsigned CLPR_bit1vec[];
extern SPNODE *CLPR_spred[];
extern HASH_P_NODE_ptr CLPR_hash_p_table[];
extern CLPR_int CLPR_startcode[], CLPR_endcode[];
extern FUNCTREC CLPR_functors[];
extern CLPR_int CLPR_trtop, **CLPR_trail;
extern CLPR_int CLPR_STYLE_CHECK_FLAG;

extern HASH_P_NODE_ptr CLPR_find_hash_p_node(CLPR_int name);
extern HASH_P_NODE_ptr CLPR_new_hash_p_node(CLPR_int name);
extern RULE_ptr CLPR_p_first_rule(CLPR_int name);
extern RULE_ptr CLPR_p_last_rule(CLPR_int name);
extern char *CLPR_make_ftag(char *name, CLPR_int arity, CLPR_int rule_number);
extern RULE_ptr CLPR_p_last_compiled_rule(CLPR_int name);
extern PTERM *CLPR_new_pterm(void);
extern char *CLPR_char_malloc(CLPR_int n);
extern struct HASHNODE *CLPR_get_similar_func(char *name);

extern CLPR_int CLPR_s_count[];
 
/*------------------------------------------------------ compile's ----------*/


extern void CLPR_init_lable (void);
extern CLPR_int CLPR_getmaxfunc (void);
extern CLPR_int CLPR_p_added (CLPR_int name);
void CLPR_check_multiple_defs (char *name);
extern CLPR_int CLPR_p_dynamic (CLPR_int name);
extern CLPR_int CLPR_p_compiled (CLPR_int name);
extern void CLPR_assign_p_start_hdrcode (CLPR_int name, CLPR_int i);
extern void CLPR_emit_dynamic_try (CLPR_int val);
extern void CLPR_assign_p_end_hdrcode (CLPR_int name, CLPR_int i);
void CLPR_compile_rule (PTERM *t, CLPR_int num);
extern void CLPR_assign_p_last_compiled_rule (CLPR_int name, RULE_ptr rule);
extern void CLPR_assign_p_compiled (CLPR_int name, CLPR_int flag);
extern void CLPR_delete_lable (CLPR_int val);
extern void CLPR_gen_index_code (RULE *r, char *str, CLPR_int a, CLPR_int nclauses);
extern CLPR_int CLPR_p_n_rules (CLPR_int name);
extern void CLPR_init_emit (void);
void CLPR_init_live_regs (CLPR_int head_arity);
void CLPR_lookatvars (PTERM *t);
void CLPR_check_rule (PTERM *r, CLPR_int num);
CLPR_int CLPR_alocvar (PTERM *t, CLPR_int CLPR_hascut, CLPR_int all_perm);
void CLPR_make_live_reg (CLPR_int reg, VREC *v);
CLPR_int CLPR_checksave (PTERM *t, CLPR_int CLPR_hascut);
extern void CLPR_emit_maybesave (CLPR_int nvars);
void CLPR_compile_head (PTERM *t, CLPR_int num, CLPR_int nvars);
extern void CLPR_emit_realsave (void);
extern void CLPR_emit_savecut (CLPR_int R, CLPR_int Y);
void CLPR_compile_tail (PTERM *t, CLPR_int sav, CLPR_int nvars);
extern void CLPR_emit_proceed (void);
void CLPR_free_varlist (void);
extern void CLPR_emit_special_savecut (CLPR_int r);
void CLPR_compile_body (PTERM *t, CLPR_int last, CLPR_int sav, CLPR_int nvars);
extern void CLPR_emit_dyn_call (CLPR_int val);
extern void CLPR_emit_callp (char *s, CLPR_int val);
void CLPR_put_arg (PTERM *t, CLPR_int reg, CLPR_int last);
extern void CLPR_emit_dump (char *str, CLPR_int *CLPR_regs);
extern void CLPR_emit_restore (void);
extern void CLPR_emit_halt (void);
extern void CLPR_resolve (CLPR_int warn);
extern void CLPR_storename (char *s, CLPR_int val, CLPR_int narg);
extern CLPR_int CLPR_p_cut (CLPR_int name);
extern void CLPR_emit_savecurcp (CLPR_int R);
extern void CLPR_emit_dbg_entry (PTERM *t, CLPR_int num);
void CLPR_fetch_arg (PTERM *t, CLPR_int reg);
extern void CLPR_emit_inst (CLPR_int w);
extern void CLPR_emit_crestore (void);
extern void CLPR_emit_dyn_jump (CLPR_int val);
extern void CLPR_emit_jump (char *s, CLPR_int val);
CLPR_int CLPR_arith_tree (PTERM *t);
CLPR_int CLPR_getreg (void);
extern void CLPR_emit_gettval (CLPR_int R1, CLPR_int R2);
extern void CLPR_full_constraint (PTERM *t);
extern void CLPR_emit_fastcut (CLPR_int R);
extern void CLPR_emit_docut (CLPR_int Y);
CLPR_int CLPR_set_arg (PTERM *t, CLPR_int *free);
extern void CLPR_emit_printf (CLPR_int *CLPR_regs);
void CLPR_freereg (CLPR_int i);
extern void CLPR_emit_print (CLPR_int *CLPR_regs);
extern void CLPR_emit_read (char *str, CLPR_int num, CLPR_int *CLPR_regs);
CLPR_int CLPR_check_dump_args (PTERM *t);
extern void CLPR_emit_retract (CLPR_int *CLPR_regs);
extern void CLPR_emit_roole (CLPR_int *CLPR_regs);
extern CLPR_int CLPR_hashbuiltina (char *s, CLPR_int n);
extern void CLPR_emit_spcall (CLPR_int val, CLPR_int *CLPR_regs);
extern void CLPR_show_pterm (PTERM *t);
extern void CLPR_print_term (PTERM *t, CLPR_int priority);
void CLPR_cleanup_rulebase (void);
extern CLPR_int CLPR_close_input (FILE *stream);
void CLPR_vars_in_atom (PTERM *t, CLPR_int atom);
CLPR_int CLPR_vars_in_term (PTERM *t, CLPR_int atom, CLPR_int arg, CLPR_int constr, CLPR_int arith, CLPR_int level);
CLPR_int CLPR_find_target_reg (VREC *v, CLPR_int atom);
extern void CLPR_emit_gettnum (double f, CLPR_int R);
extern void CLPR_emit_getnil (CLPR_int R);
extern void CLPR_emit_getcon (CLPR_int val, CLPR_int R);
extern void CLPR_half_constraint (CLPR_int is_new_reg, CLPR_int i, PTERM *t);
void CLPR_fetch_func (PTERM *t, CLPR_int reg);
extern void CLPR_emit_getpval (CLPR_int S1, CLPR_int R2);
extern void CLPR_emit_getpvar (CLPR_int S1, CLPR_int R2);
extern void CLPR_emit_getlist (CLPR_int R);
extern void CLPR_emit_getstr (CLPR_int val, CLPR_int R);
extern void CLPR_emit_unilist (void);
extern void CLPR_emit_unistr (CLPR_int val);
CLPR_int CLPR_can_gen_uni2var (PTERM *t);
CLPR_int CLPR_gen_uni2var (PTERM *t, CLPR_int fmask, CLPR_int *r1, CLPR_int *r2);
void CLPR_fetch_uni_nonfunc (PTERM *t);
extern void CLPR_emit_unitvar (CLPR_int R);
extern void CLPR_emit_uninum (double f);
extern void CLPR_emit_uninil (void);
extern void CLPR_emit_unicon (CLPR_int val);
extern void CLPR_emit_univoid (CLPR_int n);
extern void CLPR_emit_unipval (CLPR_int R);
extern void CLPR_emit_unipvar (CLPR_int R);
extern void CLPR_emit_unitval (CLPR_int R);
void CLPR_put_arg_nonfunc (PTERM *t, CLPR_int reg, CLPR_int last);
void CLPR_check_if_save_reg (CLPR_int reg);
CLPR_int CLPR_set_rr_info (PTERM *t);
extern void CLPR_emit_putlist (CLPR_int reg);
extern void CLPR_emit_putstr (CLPR_int val, CLPR_int R);
extern void CLPR_emit_bldtval (CLPR_int R);
void CLPR_put_last_func (PTERM *t);
void CLPR_put_bld_nonfunc (PTERM *t);
extern void CLPR_emit_puttnum (double f, CLPR_int R);
extern void CLPR_emit_putnil (CLPR_int reg);
extern void CLPR_emit_putcon (CLPR_int val, CLPR_int reg);
extern void CLPR_emit_putuval (CLPR_int S1, CLPR_int R2);
extern void CLPR_emit_putpval (CLPR_int S1, CLPR_int R2);
extern void CLPR_emit_putpvar (CLPR_int S1, CLPR_int R2);
extern void CLPR_emit_gettvar (CLPR_int R1, CLPR_int R2);
extern void CLPR_emit_puttvar (CLPR_int R);
extern void CLPR_emit_bldtvar (CLPR_int R);
extern void CLPR_emit_bldlist (void);
extern void CLPR_emit_bldstr (CLPR_int val);
extern void CLPR_emit_bldnum (double f);
extern void CLPR_emit_bldnil (void);
extern void CLPR_emit_bldcon (CLPR_int val);
extern void CLPR_emit_bldpvar (CLPR_int R);
extern void CLPR_emit_bldpval (CLPR_int R);
extern void CLPR_emit_bldvoid (void);
extern void CLPR_emit_uni2ttvar (CLPR_int R1, CLPR_int R2);
extern void CLPR_emit_uni2tpvar (CLPR_int R1, CLPR_int R2);
extern void CLPR_emit_uni2tvvar (CLPR_int R);
extern void CLPR_emit_uni2ptvar (CLPR_int R1, CLPR_int R2);
extern void CLPR_emit_uni2vtvar (CLPR_int R);
extern void CLPR_emit_uni2ppvar (CLPR_int R1, CLPR_int R2);
extern void CLPR_emit_uni2pvvar (CLPR_int S);
extern void CLPR_emit_uni2vpvar (CLPR_int S);
CLPR_int CLPR_find_last_arg (VREC *v, CLPR_int atom);
CLPR_int CLPR_set_inactive_rule (CLPR_int val, CLPR_int rule_no, CLPR_int max);
extern void CLPR_free_rule (RULE_ptr r);
extern void CLPR_assign_p_first_rule (CLPR_int name, RULE_ptr rule);
extern void CLPR_assign_p_last_rule (CLPR_int name, RULE_ptr rule);
extern CLPR_int CLPR_find_loc_for_lable (CLPR_int val, char *s);
extern void CLPR_assign_p_added (CLPR_int name, CLPR_int flag);

void CLPR_compile(char clobber)
{
CLPR_int i, j, k;
char *s;
CLPR_int count;
RULE *r, *r2, *last_compiled_rule, dummy;
CLPR_int compiled_flag;
	CLPR_init_lable();
	CLPR_pcode = CLPR_epcode;
	for (i = 1; i < CLPR_getmaxfunc(); i++) { /*** procedure i ***/

		if (!(r = CLPR_p_first_rule(i))) continue;
		if ((CLPR_STYLE_CHECK_FLAG & OVERLOAD_STYLE) && CLPR_p_added(i)) 
			CLPR_check_multiple_defs(string_of(i));
		compiled_flag = FALSE;

		if (CLPR_p_dynamic(i)) {
			if (CLPR_p_compiled(i)) {
				last_compiled_rule = CLPR_p_last_compiled_rule(i);
				r = last_compiled_rule->next;
			} else {
				last_compiled_rule = (RULE *) NULL;
				r = CLPR_p_first_rule(i);
				CLPR_assign_p_start_hdrcode(i, CLPR_pcode);
				CLPR_emit_dynamic_try(i);
				last_compiled_rule = r;
				CLPR_assign_p_end_hdrcode(i, CLPR_pcode-1);
			}
			while (r) {
				r->startcode = CLPR_pcode;
				CLPR_current_rule = r;
/*
printf("@@@ COMPILING: ");CLPR_show_pterm(r->wt);printf("\n");
*/
				CLPR_compile_rule(r->wt, -1);
				r->endcode = CLPR_pcode - 1;
				last_compiled_rule = r;
				r = r->next;
				compiled_flag = TRUE;
			}
			if (compiled_flag) {
				CLPR_assign_p_last_compiled_rule(i, last_compiled_rule);
				CLPR_assign_p_compiled(i, TRUE);
				CLPR_epcode = CLPR_pcode;
			}
			continue;
		} 

		if (!CLPR_p_compiled(i)) { /*** static procedure ***/
			if (clobber) CLPR_delete_lable(i);
			CLPR_assign_p_start_hdrcode(i, CLPR_pcode);
			if (r->next) {
				CLPR_gen_index_code(r, string_of(i), arity_of(i), CLPR_p_n_rules(i));
				CLPR_assign_p_end_hdrcode(i, CLPR_pcode - 1);
				for (count = 0; r; r = r->next) {
					r->startcode = CLPR_pcode;
					CLPR_compile_rule(r->wt, ++count);
					r->endcode = CLPR_pcode - 1;
				}
			} else {
				r->startcode = CLPR_pcode;
				CLPR_compile_rule(r->wt, 0);
				r->endcode = CLPR_pcode - 1;
				CLPR_assign_p_end_hdrcode(i, 0);
			}
			CLPR_assign_p_last_compiled_rule(i, CLPR_p_last_rule(i));
			CLPR_assign_p_compiled(i, TRUE);
			CLPR_epcode = CLPR_pcode;
		}
	}
}
 
void CLPR_compile_rule(PTERM *t, CLPR_int num) /*** num == -1 for dynamic ***/
         
        
{
CLPR_int nvars;
CLPR_int sav;

	CLPR_init_emit();
	CLPR_must_do_save = FALSE;
	cur_rule_pterm = t;
	cur_head = (is_implies(t) ? t->first : t);
	CLPR_hascut = FALSE;
	CLPR_init_live_regs(cur_head->narg);
	CLPR_lookatvars(t);
	if (CLPR_STYLE_CHECK_FLAG & SINGLE_STYLE) CLPR_check_rule(t, num);
	nvars = CLPR_alocvar(t, CLPR_hascut, FALSE);
	if (is_implies(t)) {
		if (CLPR_hascut) CLPR_make_live_reg((cutreg=cur_head->narg + 1), &CLPR_dummy_vrec);
		sav = CLPR_checksave(t->first->next, CLPR_hascut);
		if (sav || nvars) {
			if (has_fastcut) CLPR_emit_maybesave(nvars);
			else if (CLPR_hascut) CLPR_emit_maybesave(nvars+1);
			else CLPR_emit_maybesave(nvars);
		}
		CLPR_compile_head(t->first, num, nvars);
		if (sav) CLPR_emit_realsave();
		if (CLPR_hascut && !has_fastcut) 
			CLPR_emit_savecut(cutreg, nvars+1);
		CLPR_compile_tail(t->first->next, sav, nvars);
	} else {
		CLPR_compile_head(t, num, nvars);
		CLPR_emit_proceed();
	}
	CLPR_free_varlist();
}
 
void CLPR_compile_goal(PTERM *t) /*** command-line goal ***/
         
{
CLPR_int nvars, sav;
VREC *v;
PTERM *arg1, *t2;
CLPR_int dumpreg;
	CLPR_init_emit();
	cur_rule_pterm = t;
	CLPR_init_lable();
	CLPR_must_do_save = FALSE;
	cur_atom_no = 0;
	CLPR_pcode = GOAL_START;
	CLPR_init_live_regs(0);
	CLPR_lookatvars(t);
	nvars = CLPR_alocvar(t, TRUE, TRUE);
	sav = TRUE;
	CLPR_emit_maybesave(nvars+1);
	CLPR_emit_realsave();
	CLPR_emit_special_savecut(nvars+1);
	while (is_comma(t)) {
		arg1 = t->first;
		CLPR_compile_body(arg1, FALSE, sav, nvars);
		if (CLPR_p_dynamic(arg1->val)) CLPR_emit_dyn_call(arg1->val);
		else if (is_userpred(arg1))
			CLPR_emit_callp(CLPR_maketag(0, "", arg1->str, arg1->narg), arg1->val);
		t = arg1->next;
		check_goalspace(CLPR_pcode);
	}
	CLPR_compile_body(t, FALSE, sav, nvars);
	check_goalspace(CLPR_pcode);
	if (CLPR_p_dynamic(t->val)) CLPR_emit_dyn_call(t->val);
	else if (is_userpred(t)) 
		CLPR_emit_callp(CLPR_maketag(0, "", t->str, t->narg), t->val);
	t2 = CLPR_makedumpvars(CLPR_varlist);
	CLPR_put_arg(t2, 1, TRUE);
	dumpreg=1;
	CLPR_emit_dump((char *) CLPR_dumpvarnames(t2), &dumpreg);
	CLPR_emit_restore();
	CLPR_emit_halt();
	check_goalspace(CLPR_pcode);
	CLPR_resolve(TRUE);
	CLPR_free_varlist();
}

CLPR_int CLPR_compile_fgoal(PTERM *t)
{
CLPR_int sav;
VREC *v;
CLPR_int vnum = 1;
PTERM *tt, *arg1;
CLPR_int start;
CLPR_int nvars;

	CLPR_init_emit();
	cur_rule_pterm = t;
	CLPR_init_lable();
	CLPR_pcode = CLPR_epcode;
	start = CLPR_pcode;
	CLPR_init_live_regs(0);
	CLPR_lookatvars(t);
	nvars = CLPR_alocvar(t, TRUE, TRUE);
	sav = TRUE;
	CLPR_emit_maybesave(nvars+1);
	CLPR_emit_realsave();
	CLPR_emit_special_savecut(nvars+1);
	while (is_comma(t)) {
		arg1 = t->first;
		CLPR_compile_body(arg1, FALSE, sav, nvars);
		if (CLPR_p_dynamic(arg1->val)) CLPR_emit_dyn_call(arg1->val);
		else if (is_userpred(arg1))
			CLPR_emit_callp(CLPR_maketag(0, "", arg1->str, arg1->narg), arg1->val);
		t = arg1->next;
		check_codespace(CLPR_pcode);
	}
	CLPR_compile_body(t, FALSE, sav, nvars);
	if (CLPR_p_dynamic(t->val)) CLPR_emit_dyn_call(t->val);
	else if (is_userpred(t)) 
		CLPR_emit_callp(CLPR_maketag(0, "", t->str, t->narg), t->val);
	CLPR_emit_restore();
	CLPR_emit_halt();
	CLPR_resolve(TRUE);
	CLPR_epcode = CLPR_pcode;
	CLPR_free_varlist();
	return start;
}

void CLPR_compile_head(PTERM *t, CLPR_int num, CLPR_int nvars)
{
PTERM *it;
CLPR_int i;
	cur_atom_no = 0;
	p_name = t->str;
	p_arity = t->narg;
	p_val = t->val;
	p_r_number = num;
	if (num >= 0) {
		CLPR_storename(CLPR_maketag(p_r_number, "", p_name, p_arity), p_val, p_arity);
		if (num == 0 && CLPR_p_cut(p_val)) CLPR_emit_savecurcp(p_arity+1);
	} /* else if (CLPR_p_cut(p_val)) CLPR_emit_saveprevcp(p_arity+1); */
	if (CLPR_CODEGEN_DEBUG) 
		CLPR_emit_dbg_entry(t, num);
	for (i = 1, it = t->first; it; it = it->next, i++) CLPR_fetch_arg(it, i);
}

void CLPR_compile_tail(PTERM *t, CLPR_int sav, CLPR_int nvars)
{
PTERM *arg1;
CLPR_int last;
	while (is_comma(t)) {
		arg1 = t->first;
		CLPR_compile_body(arg1, FALSE, sav, nvars);
		if (CLPR_p_dynamic(arg1->val)) CLPR_emit_dyn_call(arg1->val);
		else if (is_userpred(arg1))
			CLPR_emit_callp(CLPR_maketag(0, "", arg1->str, arg1->narg), arg1->val);
		else if (is_callx(arg1)) CLPR_emit_inst(CALLX);
		t = arg1->next; 
		check_codespace(CLPR_pcode);
	}
	CLPR_compile_body(t, TRUE, sav, nvars); /* CHECK - was last */
	if (sav) CLPR_emit_crestore();
	if (CLPR_p_dynamic(t->val)) CLPR_emit_dyn_jump(t->val);
	else if (is_userpred(t)) 
		CLPR_emit_jump(CLPR_maketag(0, "", t->str, t->narg), t->val);
	else if (is_callx(t)) CLPR_emit_inst(JUMPX);
	else CLPR_emit_proceed();
}
 
void CLPR_compile_body(PTERM *t, CLPR_int last, CLPR_int sav, CLPR_int nvars)
{
PTERM *it;
CLPR_int libval, i, r, r2;
CLPR_int n_dump;
CLPR_int temp_regs[MAX_SYSPRED_ARITY], free_flags[MAX_SYSPRED_ARITY], flag1, flag2;

	if (is_userpred(t)) {
		cur_atom_no++;
		for (i = 1, it = t->first; it != NULL; it = it->next, i++)
			CLPR_put_arg(it, i, last);
		CLPR_init_live_regs(0);	
		return;
	}
	if (is_equation(t) &&
		(!CLPR_arith_tree(t->first) || !CLPR_arith_tree(t->first->next) || 
		(t->first->ty == VAR && (t->first->next->ty == VAR)))) {
		CLPR_put_arg(t->first, (r = CLPR_getreg()), FALSE);
		CLPR_put_arg(t->first->next, (r2 = CLPR_getreg()), FALSE);
		CLPR_emit_gettval(r, r2);
		return;
	}
	if (is_constraint(t)) { /*** this comes after is_eq test ***/
		CLPR_full_constraint(t);
		return;
	}
	if (is_cut(t)) {
		if (has_fastcut) CLPR_emit_fastcut(cutreg);
		else CLPR_emit_docut(nvars+1);
		return;
	}
	if (is_printf(t)) {
		temp_regs[0] = CLPR_set_arg(t->first, &flag1);
		temp_regs[1] = CLPR_set_arg(t->first->next, &flag2);
		CLPR_emit_printf(temp_regs);
		if (flag1) CLPR_freereg(temp_regs[0]);
		if (flag2) CLPR_freereg(temp_regs[1]);
		return;
	}
	if (is_print(t) || is_write(t)) {
		temp_regs[0] = CLPR_set_arg(t->first, &flag1);
		CLPR_emit_print(temp_regs);
		if (flag1) CLPR_freereg(temp_regs[0]);
		return;
	}
	if (is_read(t)) {
		if (t->first->ty != VAR) CLPR_type_error("read/1: variable needed", 0);
		temp_regs[0] = CLPR_set_arg(t->first, &flag1);
		CLPR_emit_read(t->first->str,  t->first->val, temp_regs);
		if (flag1) CLPR_freereg(temp_regs[0]);
		return;
	}
	if (is_dump(t)) {
		n_dump = CLPR_check_dump_args(t->first);
		temp_regs[0] = CLPR_set_arg(t->first, &flag1);
		CLPR_emit_dump((char *) CLPR_dumpvarnames(t->first), temp_regs);
		if (flag1) CLPR_freereg(temp_regs[0]);
		return;
	}
	if (is_retract(t)) {
		temp_regs[0] = CLPR_set_arg(t->first, &flag1);
		CLPR_emit_retract(temp_regs);
		CLPR_emit_inst(RETRACT_AGAIN);
		if (flag1) CLPR_freereg(temp_regs[0]);
		return;
	}
	if (is_rule(t)) { 
		temp_regs[0] = CLPR_set_arg(t->first, &flag1);
		temp_regs[1] = CLPR_set_arg(t->first->next, &flag2);
		CLPR_emit_roole(temp_regs);
		CLPR_emit_inst(ROOLE_AGAIN);
		if (flag1) CLPR_freereg(temp_regs[0]);
		if (flag2) CLPR_freereg(temp_regs[1]);
		return;
	}
	if (is_callx(t)) {
		CLPR_put_arg(t->first, 1, FALSE);
		CLPR_init_live_regs(0);
		return;
	}
	if (is_fail(t)) {
		CLPR_emit_inst(FFAIL);
		return;
	}
	if (t->val == TRUEVAL) {
		CLPR_emit_inst(NOOP);
		return;
	}
	libval = CLPR_hashbuiltina(t->str, t->narg);
	if (CLPR_spred[libval] && (CLPR_spred[libval]->arity == t->narg)) {
		if (CLPR_spred[libval]->arity > MAX_SYSPRED_ARITY) 
			CLPR_fatal("arity of system predicate too high");
		for (i = 0, it = t->first; i < t->narg; i++,it = it->next)
			temp_regs[i] = CLPR_set_arg(it, &free_flags[i]);
		CLPR_emit_spcall(libval, temp_regs);
		for (i=0; i < t->narg; i++)
			if (free_flags[i]) CLPR_freereg(temp_regs[i]);
		return;
	}
	CLPR_show_pterm(t);
	CLPR_fatal("34783");
}
 
/*------------------------------------------------------ reg handling -------*/

#define is_live_reg_bit(v, reg) \
	(v->live_regs[((reg) - 1) / CLPR_WORDSIZE] & \
		CLPR_bit1vec[CLPR_WORDSIZE - ((reg) % CLPR_WORDSIZE) + 1])
#define set_live_reg_bit(v, reg) \
	v->live_regs[((reg) - 1) / CLPR_WORDSIZE] |= \
		CLPR_bit1vec[CLPR_WORDSIZE - ((reg) % CLPR_WORDSIZE) + 1]
#define unset_live_reg_bit(v, reg) \
	v->live_regs[((reg) - 1) / CLPR_WORDSIZE] &= \
		~(CLPR_bit1vec[CLPR_WORDSIZE - ((reg) % CLPR_WORDSIZE) + 1])

void CLPR_init_live_regs(CLPR_int head_arity)
{
CLPR_int i;
VREC *v;
	for (i = 1; i < CLPR_reg_lowmark; i++) {
		v = CLPR_live_regs[i];
		if (v) {
			unset_live_reg_bit(v, i);
			v->n_live_regs = 0;
		}
		CLPR_live_regs[i] = NULL;
		CLPR_regused[i] = FALSE;
	}
	for (i = CLPR_reg_highmark + 1; i <= MAX_REGS; i++) {
		v = CLPR_live_regs[i];
		if (v) {
			unset_live_reg_bit(v, i);
			v->n_live_regs = 0;
		}
		CLPR_live_regs[i] = NULL;
		CLPR_regused[i] = FALSE;
	}
	for (i = 1; i <= head_arity; i++) {
		CLPR_live_regs[i] = &CLPR_dummy_vrec;
		CLPR_regused[i] = TRUE;
	}
	CLPR_reg_lowmark = head_arity + 1;
	CLPR_reg_highmark = MAX_REGS;
}

void CLPR_make_dead_reg(CLPR_int reg)
{
VREC *v;
	check_reg_range(reg);
	if ((v = CLPR_live_regs[reg]) && (v != &CLPR_dummy_vrec)) {
		if (is_live_reg_bit(v, reg)) {
			unset_live_reg_bit(v, reg);
			v->n_live_regs--;
		} else CLPR_fatal("49432");
	}
	CLPR_live_regs[reg] = NULL;	
	CLPR_regused[reg] = FALSE;
	if (reg == CLPR_reg_highmark + 1) {
		while (!CLPR_regused[CLPR_reg_highmark + 1] && CLPR_reg_highmark < MAX_REGS) 
			CLPR_reg_highmark++;
	} else if (reg == CLPR_reg_lowmark - 1) {
		while (!CLPR_regused[CLPR_reg_lowmark - 1] && CLPR_reg_lowmark > 0) 
			CLPR_reg_lowmark--;
	}
}

void CLPR_make_live_reg(CLPR_int reg, VREC *v)
{
	check_reg_range(reg);
	CLPR_regused[reg] = TRUE;
	CLPR_live_regs[reg] = v;
	if (v != &CLPR_dummy_vrec) {
		if (!is_live_reg_bit(v, reg)) {
			set_live_reg_bit(v, reg);
			v->n_live_regs++;
		} else CLPR_fatal("49433");
	}
	if (reg <= CLPR_reg_highmark && reg >= CLPR_reg_lowmark) {
		if (CLPR_reg_highmark - reg <= reg - CLPR_reg_lowmark) CLPR_reg_highmark = reg - 1;
		else CLPR_reg_lowmark = reg + 1;
	}
}

CLPR_int CLPR_var_has_reg(CLPR_int reg, VREC *v)
{
	check_reg_range(reg);
	if (CLPR_live_regs[reg] == v) return TRUE;
	else return FALSE;
}

void CLPR_make_change_reg(CLPR_int reg, VREC *v)
{
	if (CLPR_live_regs[reg] != v) {
		CLPR_make_dead_reg(reg);
		CLPR_make_live_reg(reg, v);
	}
}

CLPR_int CLPR_has_live_reg(VREC *v)
{
CLPR_int i, k; 
CLPR_unsigned j;
	if (!v->n_live_regs) return 0;
	for (i = 0; i < (MAX_REGS / CLPR_WORDSIZE); i++) {
		if (j = v->live_regs[i]) {
			for (k = 1; k <= CLPR_WORDSIZE; k++) {
				if (j & CLPR_bit1vec[k])
					return ((CLPR_WORDSIZE * i) + (CLPR_WORDSIZE - k + 1));
			}
		}
	}
	CLPR_fatal("49434");
}

CLPR_int CLPR_getreg(void)
{
CLPR_int i;
	if (CLPR_reg_lowmark <= CLPR_reg_highmark) {
		i = CLPR_reg_highmark--;
		CLPR_regused[i] = TRUE;
		CLPR_live_regs[i] = &CLPR_dummy_vrec;
		return i;
	}
	CLPR_fatal("no more registers (compile)");
}
 
void CLPR_freereg(CLPR_int i)
{
	CLPR_regused[i] = FALSE;
	CLPR_make_dead_reg(i);
	if (i == CLPR_reg_highmark + 1) {
		while (!CLPR_regused[CLPR_reg_highmark + 1] && CLPR_reg_highmark < MAX_REGS)
			CLPR_reg_highmark++;
	} else if (i == CLPR_reg_lowmark - 1) {
		while (!CLPR_regused[CLPR_reg_lowmark - 1] && CLPR_reg_lowmark > 0) CLPR_reg_lowmark--;
	}
}
 
void CLPR_print_live_regs(void)
{
CLPR_int i;
	for (i = 1; i <= MAX_REGS; i++) 
		if (CLPR_live_regs[i] && CLPR_live_regs[i] != &CLPR_dummy_vrec)
			 printf("reg %d var %s\n", i, CLPR_live_regs[i]->str);
}

extern FILE *CLPR_cur_stream;
extern char *CLPR_cur_filename;

void CLPR_type_error(const char *s, ...)
{
char buf[256];
va_list l;
	if (cur_rule_pterm) {
		fprintf(CLPR_error_stream, ">>> ");
		CLPR_print_term(cur_rule_pterm, MAX_OP_PRECEDENCE+1);
		fprintf(CLPR_outfile, ".\n");
	}
	va_start(l, s);
	vsprintf(buf, s, l);
	va_end(l);
	fprintf(CLPR_error_stream, "*** Type error: %s\n", buf);
	CLPR_cleanup_rulebase();

        if (CLPR_cur_stream != stdin) {
                fclose(CLPR_cur_stream);
                CLPR_close_input(CLPR_cur_stream);
	}

	CLPR_top_level();
}
	
/*------------------------------------------------------ var analysis -------*/

void CLPR_lookatvars(PTERM *t)
{
PTERM *it, *atom;
CLPR_int n;

	CLPR_varlist = CLPR_end_varlist = NULL;
	for (n = 0; n <= MAX_VAR_TO_VREC; n++) CLPR_var_to_vrec[n] = NULL;
	n = 0;
	if (!is_implies(t)) CLPR_vars_in_atom(t, 0);
	else {
		CLPR_vars_in_atom(t->first, 0);
		for (it = t->first->next, n = 0; is_comma(it); it = it->first->next) {
			atom = it->first;
			if (n == 0 && (is_userpred(atom) || is_backtrackable(atom))) n = 1;
			CLPR_vars_in_atom(atom, n);
			if (is_userpred(atom) || is_backtrackable(atom)) n++;
		}
		if (n == 0 && (is_userpred(it) || is_backtrackable(it))) n = 1;
		CLPR_vars_in_atom(it, n);
	}
	CLPR_n_chunks = n;
}

CLPR_int CLPR_alocvar(PTERM *t, CLPR_int CLPR_hascut, CLPR_int all_perm)
{
VREC *v;
CLPR_int vnum;
	vnum = 0;
	for (v = CLPR_varlist; v; v = v->next) {
		if (!all_perm && (
			is_chunk01only_var(v) || 
			(is_in1atomonly_var(v) && is_1stinfunc_var(v)) ||
			is_lastatomonly_var(v)
		)) {
			v->stack_loc = 0;
			if (is_void_var(v) && v->occ[1].level == 1) 
				CLPR_make_dead_reg(v->occ[1].arg);
		} else {
			v->stack_loc = ++vnum;
			v->saved_stack_loc = FALSE;
		}
	}
	return vnum;
}

void CLPR_vars_in_atom(PTERM *t, CLPR_int atom)
{
CLPR_int arg;
PTERM *it;
CLPR_int type1, type2;
CLPR_int tmp1, tmp2;
	if (!t->narg) {
		if (!CLPR_hascut) CLPR_hascut = is_cut(t);
		return;
  	} 
	if (is_constraint(t)) {
		tmp1 = (is_ineq(t) || is_arith_term(t->first));
		tmp2 = (is_ineq(t) || is_arith_term(t->first->next));
		type1 = CLPR_vars_in_term(t->first, atom, 1, TRUE, tmp1, 1);
		type2 = CLPR_vars_in_term(t->first->next, atom, 2, TRUE, tmp2, 1);
		if (is_ineq(t)) typecheck_arith(type1, type2);
		else typecheck_clash(type1, type2);
		return;
	}
	for (it = t->first, arg = 1; it != NULL; it = it->next, arg++) {
		CLPR_vars_in_term(it, atom, arg, FALSE, is_arith_term(it), 1);
	}
}
 
CLPR_int CLPR_vars_in_term(PTERM *t, CLPR_int atom, CLPR_int arg, CLPR_int constr, CLPR_int arith, CLPR_int level)
{
PTERM *it, *arg1, *arg2;
VREC *v;
CLPR_int n, type1, type2;
	if (is_p_var(t)) {
		v = CLPR_notevar(t->val);
		v->str = t->str;
		if ((n = v->occ_top + 1) <= MAX_VAR_OCC) {
			v->occ[n].atom = atom;
			v->occ[n].arg = arg;
			v->occ[n].level = level;
			v->occ_top = n;		
		}
		return (arith ? REAL : 0);
	}
	if (is_p_func(t)) {
		if (is_arith_term(t)) {
			arg1 = t->first;
			type1 = CLPR_vars_in_term(arg1, atom, arg, constr, 
								 !is_eval(t), level + 1);
			if (t->narg >= 2) {
				arg2 = arg1->next;
				type2 = CLPR_vars_in_term(arg2, atom, arg, constr, TRUE, level + 1);
			} else type2 = type1; 
			if (!is_eval(t)) typecheck_arith(type1, type2);
			return REAL;
		}
		if (arith) typecheck_error();
		if (!t->narg) return FUNC;
		for (it = t->first; it != NULL; it = it->next) {
			CLPR_vars_in_term(it, atom, arg, constr, FALSE, level + 1);
		}
		return FUNC;
	}
	if (t->ty == REAL) return REAL;
	if (arith) typecheck_error();
	else return FUNC;
}

VREC *CLPR_notevar(CLPR_int val) /*** creates/finds entry for a particular variable ***/
        
{
VREC *v;

	if (val <= MAX_VAR_TO_VREC) {
		if (v = CLPR_var_to_vrec[val]) return v;
		if (CLPR_end_varlist) {
			CLPR_end_varlist->next = CLPR_get_vrec();
			CLPR_end_varlist = CLPR_end_varlist->next;
		} else CLPR_varlist = CLPR_end_varlist = CLPR_get_vrec(); 
		CLPR_var_to_vrec[val] = CLPR_end_varlist;
	} else {
		for (v = CLPR_varlist; v; v = v->next) if (v->val == val) return v;
		if (CLPR_end_varlist) {
			CLPR_end_varlist->next = CLPR_get_vrec();
			CLPR_end_varlist = CLPR_end_varlist->next;
		} else CLPR_varlist = CLPR_end_varlist = CLPR_get_vrec(); 
	}
	CLPR_end_varlist->val = val;
	return CLPR_end_varlist;
}
 
static VREC *val_info(CLPR_int val, CLPR_int *reg, CLPR_int *stack_loc, CLPR_int *is_new_reg, CLPR_int *is_new_var)
{
VREC *v;
CLPR_int i;
	*reg = *stack_loc = 0;
	v = CLPR_notevar(val);
	if (v->stack_loc) {
		*stack_loc = v->stack_loc;
		*is_new_var = (!v->saved_stack_loc);
	} else {
		if (i = CLPR_has_live_reg(v)) {
			*reg = i;
			*is_new_reg = FALSE;
		} else if ((*reg = CLPR_find_target_reg(v, 1))) *is_new_reg = TRUE;
		else {
			*reg = CLPR_getreg();
			*is_new_reg = TRUE;
		}
	}
	return v;
}

VREC *CLPR_get_vrec(void)
{
VREC *v;
CLPR_int i;
	if (!CLPR_vrec_freelist) v = (VREC *) CLPR_emalloc(sizeof(VREC));
	else {
		v = CLPR_vrec_freelist;
		CLPR_vrec_freelist = CLPR_vrec_freelist->next;
		if (!CLPR_vrec_freelist) CLPR_vrec_endfreelist = NULL;
	}
	for (i = 0; i < (MAX_REGS / CLPR_WORDSIZE); i++) v->live_regs[i] = 0;
	v->val = v->occ_top = v->n_live_regs = v->stack_loc = 0;
	v->next = NULL;
	v->str = NULL;
	v->saved_stack_loc = FALSE;
	return v;
}

void CLPR_free_varlist(void)
{
	if (!CLPR_varlist) return;
	if (CLPR_vrec_endfreelist) {
		CLPR_vrec_endfreelist->next = CLPR_varlist;
		CLPR_vrec_endfreelist = CLPR_end_varlist;
	} else {
		CLPR_vrec_freelist = CLPR_varlist;
		CLPR_vrec_endfreelist = CLPR_end_varlist;
	}
	CLPR_varlist = CLPR_end_varlist = NULL;
}

/*---------------------------------------------------------- get's ----------*/

void CLPR_fetch_arg(PTERM *t, CLPR_int reg)
{
PTERM *it;
VREC *v;
CLPR_int tmp;
	check_codespace(CLPR_pcode);
	if (is_p_real(t)) {
		CLPR_emit_gettnum(t->rval, reg);
		if (reg == 1 && !CLPR_CODEGEN_DEBUG && p_r_number != -1)
			CLPR_storename(CLPR_make_ftag(p_name, p_arity, p_r_number), p_val, p_arity);
		CLPR_make_dead_reg(reg);
	} else if (is_p_emptylist(t)) {
		CLPR_emit_getnil(reg);
		if (reg == 1 && !CLPR_CODEGEN_DEBUG && p_r_number != -1)
			CLPR_storename(CLPR_make_ftag(p_name, p_arity, p_r_number), p_val, p_arity);
		CLPR_make_dead_reg(reg);
	} else if (is_p_constant(t)) {
		CLPR_emit_getcon(t->val, reg);
		if (reg == 1 && !CLPR_CODEGEN_DEBUG && p_r_number != -1)
			CLPR_storename(CLPR_make_ftag(p_name, p_arity, p_r_number), p_val, p_arity);
		CLPR_make_dead_reg(reg);
	} else if (is_arith_term(t)) {
		CLPR_half_constraint(FALSE, -reg, t);
		CLPR_make_dead_reg(reg);
	} else if (is_p_func(t)) {
		CLPR_fetch_func(t, reg);
	} else if (is_p_var(t)) {
		v = CLPR_notevar(t->val);
		if (is_void_var(v)) return;
		if (v->stack_loc) {
			if (v->saved_stack_loc) CLPR_emit_getpval(v->stack_loc, reg);
			else {
				CLPR_emit_getpvar(v->stack_loc, reg);
				v->saved_stack_loc = TRUE;
			}
			return;
		}
		if (tmp = CLPR_has_live_reg(v)) CLPR_emit_gettval(tmp, reg);
		CLPR_make_live_reg(reg, v);
	} else if (is_p_string(t)) /*** nothing ***/ ;
	else CLPR_fatal("86654");
}

void CLPR_fetch_func(PTERM *t, CLPR_int reg) /*** reg == 0 means deep level ***/
         
        
{
CLPR_int i, j, r1, r2, tmp, arity, fmask;
CLPR_int count, save_trtop, tmp_trtop;
PTERM *it;
	if (reg) {
		if (is_p_list(t)) CLPR_emit_getlist(reg);
		else CLPR_emit_getstr(t->val, reg);
		if (reg == 1 && !CLPR_CODEGEN_DEBUG && p_r_number != -1)
			CLPR_storename(CLPR_make_ftag(p_name, p_arity, p_r_number), p_val, p_arity);
		CLPR_make_dead_reg(reg);
	} else {
		if (is_p_list(t)) CLPR_emit_unilist();
		else CLPR_emit_unistr(t->val);
	}
	arity = t->narg;
	count = 0;
	save_trtop = CLPR_trtop;
	for (j = 1, it = t->first; it; j++, it = it->next) {
		if (j == arity && is_complex_func(it) && !count) {
			CLPR_fetch_func(it, 0);
			break;
		} 
		if ((fmask = CLPR_can_gen_uni2var(it)) && 
			!(j == arity-1 && is_p_func(it->next) && !count)
			&& CLPR_gen_uni2var(it, fmask, &r1, &r2)) {
			if (bit1(fmask)) {
				count++;
				CLPR_trail[CLPR_trtop++] = (CLPR_int *) it;
				CLPR_trail[CLPR_trtop++] = (CLPR_int *) r1; 
			}
			if (bit2(fmask)) {
				count++;
				CLPR_trail[CLPR_trtop++] = (CLPR_int *) it->next;
				CLPR_trail[CLPR_trtop++] = (CLPR_int *) r2; 
			}
			it = it->next;
			j++;
			continue;
		}
		if ((is_arith_term(it) && !is_eval(it))
				|| !is_p_func(it)) CLPR_fetch_uni_nonfunc(it);
		else {
			count++;
			CLPR_trail[CLPR_trtop++] = (CLPR_int *) it;
			tmp = CLPR_getreg();
			CLPR_trail[CLPR_trtop++] = (CLPR_int *) tmp;
			CLPR_emit_unitvar(tmp);
		}
	}
	tmp_trtop = save_trtop;  /*** dont reset CLPR_trtop yet! ***/
	for (; count; count--) {
		it = (PTERM *) CLPR_trail[tmp_trtop++];
		tmp = (CLPR_int) CLPR_trail[tmp_trtop++];
		if (is_eval(it)) CLPR_half_constraint(FALSE, -tmp, it); /* CHECK */
		else CLPR_fetch_func(it, tmp);
	}
	CLPR_trtop = save_trtop;
}

void CLPR_fetch_uni_nonfunc(PTERM *t)
{
VREC *v;
CLPR_int tmp;
	if (is_p_real(t)) CLPR_emit_uninum(t->rval);
	else if (is_p_emptylist(t)) CLPR_emit_uninil();
	else if (is_p_constant(t)) CLPR_emit_unicon(t->val);
	else if (is_arith_term(t)) {
		CLPR_emit_unitvar(tmp = CLPR_getreg());
		CLPR_half_constraint(FALSE, -tmp, t);
	} else if (is_p_var(t)) {
		v = CLPR_notevar(t->val);
		if (is_void_var(v)) {
			CLPR_emit_univoid(1); 
			return;
		}
		if (v->stack_loc) {
			if (v->saved_stack_loc) CLPR_emit_unipval(v->stack_loc);
			else {
				CLPR_emit_unipvar(v->stack_loc);
				v->saved_stack_loc = TRUE;
			}
			return;
		}
		if (tmp = CLPR_has_live_reg(v)) {
			CLPR_emit_unitval(tmp);
			return;
		}
		if (tmp = CLPR_find_target_reg(v, 1)) {
			CLPR_make_live_reg(tmp, v);
			CLPR_emit_unitvar(tmp);
			return;
		}
		CLPR_emit_unitvar(tmp = CLPR_getreg());
		CLPR_make_live_reg(tmp, v);
	} else if (is_p_string(t)) /*** nothing ***/ ;
	else CLPR_fatal("86652");
}

/*---------------------------------------------------------- put's ----------*/

CLPR_int CLPR_set_arg(PTERM *t, CLPR_int *free)
{
CLPR_int reg;
VREC *v;	

	if (is_p_var(t)) {
		v = CLPR_notevar(t->val);
		if (reg = CLPR_has_live_reg(v)) {
			*free = FALSE;
			return reg;
		}
	}
	reg = CLPR_getreg();
	CLPR_put_arg(t, reg, FALSE);
	if (is_p_var(t)) *free = FALSE;
	else *free = TRUE;
	return reg;
}
	
void CLPR_put_arg(PTERM *t, CLPR_int reg, CLPR_int last) /*** top-level ***/
         
              
{
PTERM *it;
VREC *v;
CLPR_int j, arity, tmp;
CLPR_int save_trtop;
	check_codespace(CLPR_pcode);
	if (!is_complex_func(t) && !is_eval(t)) {
		CLPR_put_arg_nonfunc(t, reg, last);
		return;
	}
	if (is_arith_term(t)) { /* this is to compile eval */
		CLPR_check_if_save_reg(reg);
		CLPR_make_change_reg(reg, &CLPR_dummy_vrec);
		CLPR_half_constraint(TRUE, -reg, t);
		return;
	}
	CLPR_set_rr_info(t);
	arity = t->narg;
	save_trtop = CLPR_trtop;
	for (j = 1, it = t->first; it; j++, it = it->next) {
		if ((is_complex_func(it) || is_eval(it)) && (j < arity || !is_rr(it))) 
			CLPR_trail[CLPR_trtop++] = (CLPR_int *) CLPR_put_func(it);
	} 
	CLPR_check_if_save_reg(reg);
	if (is_p_list(t)) CLPR_emit_putlist(reg);
	else if (is_complex_func(t)) CLPR_emit_putstr(t->val, reg);
	else CLPR_fatal("51232");
	CLPR_make_change_reg(reg, &CLPR_dummy_vrec);
	CLPR_trtop = save_trtop;
	for (j = 1, it = t->first; it; j++, it = it->next) {
		if (is_complex_func(it) || is_eval(it)) {
			if (j < arity || !is_rr(it)) {
				tmp = (CLPR_int) CLPR_trail[CLPR_trtop++];
				CLPR_emit_bldtval(tmp);
				CLPR_freereg(tmp);
			} else CLPR_put_last_func(it);
		} else CLPR_put_bld_nonfunc(it);
	}
	CLPR_trtop = save_trtop;
}

void CLPR_put_arg_nonfunc(PTERM *t, CLPR_int reg, CLPR_int last) /*** top-level ***/
         
              
{
CLPR_int tmp;
VREC *v, *v2;
	if (is_p_real(t)) {
		CLPR_check_if_save_reg(reg);
		CLPR_emit_puttnum(t->rval, reg);
		CLPR_make_change_reg(reg, &CLPR_dummy_vrec);
	} else if (is_p_emptylist(t)) {
		CLPR_check_if_save_reg(reg);
		CLPR_emit_putnil(reg);
		CLPR_make_change_reg(reg, &CLPR_dummy_vrec);
	} else if (is_p_constant(t)) {
		CLPR_check_if_save_reg(reg);
		CLPR_emit_putcon(t->val, reg);
		CLPR_make_change_reg(reg, &CLPR_dummy_vrec);
	} else if (is_arith_term(t)) {
		CLPR_check_if_save_reg(reg);
/*** BUG1 ***/
		CLPR_make_change_reg(reg, &CLPR_dummy_vrec); /* clobber reg */
		CLPR_half_constraint(TRUE, -reg, t);
		CLPR_make_change_reg(reg, &CLPR_dummy_vrec);
	} else if (is_p_var(t)) {
		v = CLPR_notevar(t->val);
		if (v->stack_loc) {
			CLPR_check_if_save_reg(reg);
			if (v->saved_stack_loc)	{
				if (last && !is_head_var(v)) CLPR_emit_putuval(v->stack_loc, reg); 
				else CLPR_emit_putpval(v->stack_loc, reg); 
			} else {
				if (last) CLPR_fatal("54126");
				CLPR_emit_putpvar(v->stack_loc, reg); 
				v->saved_stack_loc = TRUE;
			}
			CLPR_make_change_reg(reg, v);
		} else {
			if (tmp = CLPR_has_live_reg(v)) {
				if (!CLPR_var_has_reg(reg,v) &&  tmp != reg) { /* CHECK */
					CLPR_check_if_save_reg(reg);
					CLPR_emit_gettvar(reg, tmp);
					CLPR_make_dead_reg(tmp);
					CLPR_make_change_reg(reg, v);
				}
			} else { 
				CLPR_check_if_save_reg(reg);
				if (last) {
					CLPR_emit_puttvar(reg);
					CLPR_make_change_reg(reg, v);
				} else {
					/*** CLPR_emit_reginit(reg); ***/
					CLPR_emit_puttvar(reg);
					CLPR_make_change_reg(reg, v);
				}
			}
		}
	}
}

CLPR_int CLPR_put_func(PTERM *t) /*** t is complex */
         
{
PTERM *it;
CLPR_int i, arity;
CLPR_int tmp, tmp0, flag;
CLPR_int save_trtop;
	if (is_arith_term(t)) { /* this is to compile eval */
        CLPR_int reg;
        reg = CLPR_getreg();
		CLPR_half_constraint(TRUE, -reg, t);
		CLPR_make_change_reg(reg, &CLPR_dummy_vrec);
        return reg;
    }
	arity = t->narg;
	save_trtop = CLPR_trtop;
	for (i = 1, it = t->first; it; i++, it = it->next) {
		if ((is_complex_func(it) || is_eval(it)) && 
				(i < arity || !is_rr(it))) 
			CLPR_trail[CLPR_trtop++] = (CLPR_int *) CLPR_put_func(it);
	}
	if (is_p_list(t)) CLPR_emit_putlist(tmp0 = CLPR_getreg());
	else if (is_complex_func(t)) CLPR_emit_putstr(t->val, tmp0 = CLPR_getreg());
	else CLPR_fatal("54127");
	CLPR_trtop = save_trtop;
	for (i = 1, it = t->first; it; i++, it = it->next) {
		if (is_complex_func(it) || is_eval(it)) { 
			if (i < arity || !is_rr(it)) {
				CLPR_emit_bldtval(tmp = (CLPR_int) CLPR_trail[CLPR_trtop++]);
				CLPR_freereg(tmp);
			} else CLPR_put_last_func(it);
		} else CLPR_put_bld_nonfunc(it);
	}
	CLPR_trtop = save_trtop;
	return tmp0;
}

void CLPR_put_last_func(PTERM *t) /*** emit bldstr/bldlist? ***/
         
{
PTERM *it;
CLPR_int tmp;
	if (is_arith_term(t)) { /* this is to compile eval */
                CLPR_int reg;
                reg = CLPR_getreg();
                CLPR_emit_bldtvar(reg);
		CLPR_check_if_save_reg(reg);
		CLPR_half_constraint(FALSE, -reg, t);
		CLPR_make_change_reg(reg, &CLPR_dummy_vrec);
                return;
        }
	if (is_p_list(t)) CLPR_emit_bldlist();
	else CLPR_emit_bldstr(t->val);
	for (it = t->first; it; it = it->next) {
		if (is_complex_func(it)) CLPR_put_last_func(it);
		else CLPR_put_bld_nonfunc(it);
	}
}

void CLPR_put_bld_nonfunc(PTERM *t)
{
VREC *v;
CLPR_int tmp;
	if (is_p_real(t)) CLPR_emit_bldnum(t->rval);
	else if (is_p_emptylist(t)) CLPR_emit_bldnil();
	else if (is_p_constant(t)) CLPR_emit_bldcon(t->val);
	else if (is_arith_term(t)) {
		CLPR_half_constraint(TRUE, -(tmp = CLPR_getreg()), t);
		CLPR_emit_bldtval(tmp);
		CLPR_freereg(tmp);
	} else if (is_p_var(t)) {
		v = CLPR_notevar(t->val);
		if (v->stack_loc) {
			if (!v->saved_stack_loc) {
				CLPR_emit_bldpvar(v->stack_loc);
				v->saved_stack_loc = TRUE;
			} else CLPR_emit_bldpval(v->stack_loc);
		} else {
			if (tmp = CLPR_has_live_reg(v)) CLPR_emit_bldtval(tmp);
			else {
				if (is_single_occ(v)) CLPR_emit_bldvoid();
				else if (tmp = CLPR_find_target_reg(v, cur_atom_no)) {
					CLPR_emit_bldtvar(tmp);
					CLPR_make_live_reg(tmp, v);
				} else { 
					CLPR_emit_bldtvar(tmp = CLPR_getreg());
					CLPR_make_live_reg(tmp, v);
				}
			}
		}
	} else if (is_p_string(t)) 
		CLPR_type_error("strings only in printf/2 (for now)", 0);
	else CLPR_fatal("23842");
}

CLPR_int CLPR_can_gen_uni2var(PTERM *t) /* 0x1:fv 0x2:vf 0x3:ff 0x4:vv 0:false */
         
{
PTERM *sib;
CLPR_int fmask; 
	fmask = 0; 
	if (!(sib = t->next)) return 0;
	if (is_complex_func(t)) set_bit1(fmask);
	else if (!is_p_var(t)) return 0;
	if (is_complex_func(sib)) set_bit2(fmask);
	else if (!is_p_var(sib)) return 0;
	if (!fmask) return 4;
	else return fmask;
}

CLPR_int CLPR_gen_uni2var(PTERM *t, CLPR_int fmask, CLPR_int *r1, CLPR_int *r2)
{
VREC *v1, *v2;
CLPR_int i1, i2, is_new_reg, is_new_var;
CLPR_int trail_reg;

	trail_reg = 0;
	*r2 = *r1 = i1 = i2 = 0;
	if (bit1(fmask)) *r1 = CLPR_getreg();
	else {
		v1 = CLPR_notevar(t->val);
		if (!is_void_var(v1)) {
			v1 = val_info(t->val, r1, &i1, &is_new_reg, &is_new_var);
			if (*r1) {
				if (!is_new_reg) return FALSE;
				CLPR_make_live_reg(*r1, v1);
				trail_reg = *r1;
			} else {
				if (!is_new_var) return FALSE;
				v1->saved_stack_loc = TRUE;
			}
		} 
	}
	if (bit2(fmask)) *r2 = CLPR_getreg();
	else {
		v2 = CLPR_notevar(t->next->val);
		if (!is_void_var(v2)) {
			v2 = val_info(t->next->val, r2, &i2, &is_new_reg, &is_new_var);
			if (*r2) {
				if (!is_new_reg) goto UNDO;
				CLPR_make_live_reg(*r2, v2);
			} else {
				if (!is_new_var) goto UNDO;
				v2->saved_stack_loc = TRUE;
			}
		}
	}
	if (*r1) {
		if (*r2) CLPR_emit_uni2ttvar(*r1, *r2);
		else if (i2) CLPR_emit_uni2tpvar(*r1, i2);
		else CLPR_emit_uni2tvvar(*r1);
	} else if (*r2) {
		if (i1) CLPR_emit_uni2ptvar(i1, *r2);
		else CLPR_emit_uni2vtvar(*r2);
	} else if (i1) {
		if (i2) CLPR_emit_uni2ppvar(i1, i2);
		else CLPR_emit_uni2pvvar(i1);
	} else if (i2) CLPR_emit_uni2vpvar(i2);
	else CLPR_emit_univoid(2);
	return TRUE;
UNDO:
	if (bit1(fmask)) CLPR_freereg(*r1);
	else if (!is_void_var(v1)) {
		if (trail_reg) CLPR_make_dead_reg(trail_reg);
		else v1->saved_stack_loc = FALSE;
	}
	return FALSE;
}





/* OLD */
/*
CLPR_int CLPR_gen_uni2var(t, fmask, r1, r2)
PTERM *t;
CLPR_int fmask, *r1, *r2;
{
VREC *v1, *v2;
CLPR_int i1, i2, is_new_reg, is_new_var;
CLPR_int trail_reg;

	trail_reg = 0;
	if (bit1(fmask)) *r1 = CLPR_getreg();
	else {
		v1 = val_info(t->val, r1, &i1, &is_new_reg, &is_new_var);
		if (*r1) {
			if (!is_new_reg) return FALSE;
			CLPR_make_live_reg(*r1, v1);
			trail_reg = *r1;
		} else {
			if (!is_new_var) return FALSE;
			v1->saved_stack_loc = TRUE;
		}
	}
	if (bit2(fmask)) *r2 = CLPR_getreg();
	else {
		v2 = val_info(t->next->val, r2, &i2, &is_new_reg, &is_new_var);
		if (*r2) {
			if (!is_new_reg) goto UNDO;
			CLPR_make_live_reg(*r2, v2);
		} else {
			if (!is_new_var) goto UNDO;
			v2->saved_stack_loc = TRUE;
		}
	}
	if (*r1) {
		if (*r2) CLPR_emit_uni2ttvar(*r1, *r2);
		else CLPR_emit_uni2tpvar(*r1, i2); 
	} else {
		if (*r2) CLPR_emit_uni2ptvar(i1, *r2);
		else CLPR_emit_uni2ppvar(i1, i2);
	}
	return TRUE;
UNDO:
	if (bit1(fmask)) CLPR_freereg(*r1);
	else if (trail_reg) CLPR_make_dead_reg(trail_reg);
	else v1->saved_stack_loc = FALSE;
	return FALSE;
}
*/

/*---------------------------------------------------------------------------*/

CLPR_int CLPR_checksave(PTERM *t, CLPR_int CLPR_hascut) /* CHECK */
         
           
{
PTERM *it, *atom;
CLPR_int early_cut;

	early_cut = FALSE;
	has_fastcut = FALSE;
	for (it = t; is_comma(it); it = it->first->next) {
		atom = it->first;
		if (is_userpred(atom) || is_backtrackable(atom)) break;
		else if (is_cut(atom)) early_cut = TRUE;
	}
	if (!is_comma(it)) {
		if (is_cut(it) || early_cut) has_fastcut = TRUE;
		return FALSE;
	} else has_fastcut = FALSE;
	return TRUE;
}

void CLPR_check_if_save_reg(CLPR_int reg)
{
VREC *v;
CLPR_int tmp;
	if ((v = CLPR_live_regs[reg]) && 
	    (v->n_live_regs == 1) &&
	    (reg <= CLPR_find_last_arg(v, cur_atom_no))) {
		tmp = CLPR_find_target_reg(v, cur_atom_no);
		if (!tmp) CLPR_emit_gettvar((tmp = CLPR_getreg()), reg);
		else CLPR_emit_gettvar(tmp, reg);
		CLPR_make_live_reg(tmp, v);
	}		
}

CLPR_int CLPR_find_last_arg(VREC *v, CLPR_int atom)
{
CLPR_int i, n, arg;
	arg = 0;
	n = v->occ_top;
	for (i = 1; i <= n; i++) if (v->occ[i].atom >= atom) break;
	for (; i <= n && v->occ[i].atom == atom; i++) arg = v->occ[i].arg;
	return arg;
}

CLPR_int CLPR_find_target_reg(VREC *v, CLPR_int atom) /*** find toplevel var in atom ***/
        
         
{
CLPR_int i, n;
	n = v->occ_top;
	for (i = 1; i <= n; i++) if (v->occ[i].atom >= atom) break;
	for (; i <= n && v->occ[i].atom == atom; i++)
		if (v->occ[i].level == 1 && !CLPR_live_regs[v->occ[i].arg])
			return v->occ[i].arg;
	return 0;
}

CLPR_int CLPR_set_rr_info(PTERM *t) /*** is t right-recursive? ***/
         
{
PTERM *it;
CLPR_int flag, flag0;
	if (!is_complex_func(t)) {
		set_rr(t);
		return TRUE;
	} else {
		flag0 = TRUE;
		for (it = t->first; it; it = it->next) {
			flag = CLPR_set_rr_info(it);
			if (!it->next) { 
				if (!flag0 || !flag) {
					unset_rr(t);
					return FALSE;
				}
			} else if (is_complex_func(it)) flag0 = FALSE;
		}
		set_rr(t);
		return TRUE;
	}
}

/*------------------------------------------------------ aux stuff ----------*/

char *CLPR_maketag(CLPR_int n, char *extra, char *str, CLPR_int arity)
{
char buf[256], *s;
	sprintf(buf, "E$%d%s_%s/%d", n, extra, str, arity);
	s = CLPR_char_malloc(strlen(buf) + 1);
	strcpy(s, buf);
	return s;
}
 
char *CLPR_smalltag(char *str, CLPR_int arity)
{
char CLPR_ts[256];
char *s;
	s = CLPR_char_malloc(80);
	strcat(s, str);
	strcat(s, "/");
	sprintf(CLPR_ts, "%d", arity);
	strcat(s, CLPR_ts);
	return s;
}

CLPR_int CLPR_check_dump_args(PTERM *t) /*** list of vars? ***/
         
{
	CLPR_int n;

	for (n=0; !is_emptylist(t); t=t->first->next, n++) {
		if (!is_p_list(t)) CLPR_type_error("Bad argument to dump/1", 0);
		if (!is_p_var(t->first)) CLPR_type_error("Bad argument to dump/1", 0);
	}
	return n;
}

PTERM *CLPR_makedumpvars(VREC *v) /*** command line goal only, right-left priority ***/
        
{
PTERM *t;
	t = CLPR_new_pterm();
	t->ty = FUNC;
	t->val = EMPTYLIST;
	t->str = "[]";
	t->narg = 0;
	t->first = t->next = NULL;
	return CLPR_makedumpvars1(v, t);
}

PTERM *CLPR_makedumpvars1(VREC *v, PTERM *prev) /*** command line goal only ***/
        
            
{
PTERM *t;
	if (v == NULL) return prev;
	if (v->str[0] == '_') return CLPR_makedumpvars1(v->next, prev);
	t = CLPR_new_pterm();
	t->ty = FUNC;
	t->val = DOT;
	t->str = ".";
	t->narg = 2;
	t->first = CLPR_new_pterm();
	t->first->ty = VAR;
	t->first->val = v->val;
	t->first->str = v->str;
	t->first->narg = 0;
	t->first->next = prev;
	return CLPR_makedumpvars1(v->next, t);
}

PTERM *CLPR_makedumpvars0(VREC *v) /*** command line goal only, left-right priority ***/
        
{
PTERM *t;
	if (v == NULL) {
		t = CLPR_new_pterm();
		t->ty = FUNC;
		t->val = EMPTYLIST;
		t->str = "[]";
		t->narg = 0;
		t->first=t->next=NULL;
	} else {
		t = CLPR_new_pterm();
		t->ty = FUNC;
		t->val = DOT;
		t->str = ".";
		t->narg = 2;
		t->first = CLPR_new_pterm();
		t->first->ty = VAR;
		t->first->val = v->val;
		t->first->str = v->str;
		t->first->narg = 0;
		t->first->next = CLPR_makedumpvars0(v->next);
	}
	return t;
}

char CLPR_dumpvarstring[MAX_DUMP_VARSTRING];
char *CLPR_dumpvarnames(PTERM *t)
{
char *c, *c1;
	c = CLPR_dumpvarstring;
	loop {
		if (t->ty == FUNC && t->narg == 0) {
			*c = '\0';
			break;
		}
		c1 = t->first->str;
		while (*c1 != '\0') *(c++) = *(c1++);
		*(c++) = DUMP_SEPARATOR;
		t = t->first->next;
	}
	*c = '\0';
	c = (char *) malloc(strlen(CLPR_dumpvarstring) + 1);
	strcpy(c, CLPR_dumpvarstring);
	return c;
}
 
CLPR_int CLPR_arith_tree(PTERM *t)
{
PTERM *it;
	if (t->ty == REAL) return TRUE;
	else if (t->ty == FUNC && !is_arith_term(t)) return FALSE;
	else if (is_arith_term(t)) return TRUE;
	else if (t->ty == VAR) return TRUE;
	CLPR_fatal("43892");
}

CLPR_int CLPR_val_of_cur_rule(void)
{
	return cur_head->val;
}

void CLPR_cleanup_rulebase(void)
{
CLPR_int i, j, k;
char *s;
CLPR_int count;
RULE *r, *tmp, *prev, *first_rule, *last_rule;
CLPR_int found_first, max_loc = 0;

	for (i = 1; i < CLPR_getmaxfunc(); i++) { /*** procedure i ***/
		if (!(r = CLPR_p_first_rule(i))) continue;
		last_rule = prev = (RULE *) NULL;
		found_first = FALSE; 
		first_rule = (RULE *) NULL; /* not neccesary but tricky */
		while (r) {
			if (r->endcode == -1) {
				tmp = r;
				max_loc = CLPR_set_inactive_rule(i, r->id, max_loc);
				r = r->next;
				CLPR_free_rule(tmp);
				if (found_first) { /* now delete the rest */
					while (r) {
						max_loc = CLPR_set_inactive_rule(i, r->id, max_loc);
						tmp = r;
						r = r->next;
						CLPR_free_rule(tmp);
					}
					last_rule = prev;
					break;
				}
				if (prev) prev->next = NULL;
			} else {
				if (!found_first) {
					first_rule = r;
					found_first = TRUE;
				}
				prev = r;
				r = r->next;
			} 
		}
		if (!found_first) {
			CLPR_assign_p_compiled(i, FALSE);
			CLPR_assign_p_first_rule(i, NULL);
			CLPR_assign_p_last_rule(i, NULL);
#ifdef FREE_CODE_SPACE
			CLPR_free_hdrcode(i);
#else
			CLPR_assign_p_start_hdrcode(i, 0);
			CLPR_assign_p_end_hdrcode(i, 0);
#endif
		} else {
			CLPR_assign_p_first_rule(i, first_rule);
			CLPR_assign_p_compiled(i, TRUE);
			if (prev) {
				CLPR_assign_p_last_rule(i, prev);
				prev->next = (RULE *) NULL;
			} else {
				CLPR_assign_p_last_rule(i, first_rule);
				first_rule->next = (RULE *) NULL;
			}
		}
	}
	CLPR_resolve(FALSE);
	if (max_loc && CLPR_pcode > max_loc + 1) {
/* CLEAN UP */
/*		printf("@@@ CLPR_pcode %d CLPR_epcode %d reset to %d\n", CLPR_pcode,CLPR_epcode,max_loc);*/
		CLPR_epcode = CLPR_pcode = max_loc + 1;
	}
	CLPR_init_lable(); 
}

CLPR_int CLPR_set_inactive_rule(CLPR_int val, CLPR_int rule_no, CLPR_int max)
{
CLPR_int loc;

	if ((loc = CLPR_find_loc_for_lable(val, 
			CLPR_maketag(rule_no, "", string_of(val), arity_of(val)))) > 0) {
		CLPR_code[loc] = FFAIL;
		if ((loc = CLPR_find_loc_for_lable(val, 
				CLPR_make_ftag(string_of(val), arity_of(val), rule_no))) > 0) {
			CLPR_code[loc] = FFAIL;
		}
		return (max > loc ? max : loc);
	} else return max;
}

void CLPR_check_rule(PTERM *r, CLPR_int num)
{
VREC *v;
CLPR_int n, i, flag = FALSE;
PTERM *head;

	for (n=0,v = CLPR_varlist; v; v = v->next) 
		if (v->occ_top == 1 && v->str[0] != '_') n++;
	if (!n) return;
	if (num == 0) num = 1;
	head = (is_implies(r) ? r->first : r);
	fprintf(CLPR_error_stream, 
		"Warning: Style check, singleton variables, rule %d of %s/%d\n +++ ", 
		num, head->str, head->narg);
	
	for (i=0, v = CLPR_varlist; v; v = v->next, i++) 
		if (v->occ_top == 1 && v->str[0] != '_') {
			n--;
			if ((i ==  6) && (n > 0)) {
				i = 0; 
				fprintf(stderr, "\n +++ ");
			}
			if (n > 0) fprintf(CLPR_error_stream, " %s,", v->str);
			else fprintf(CLPR_error_stream, " %s", v->str);
		}
	fprintf(stderr, "\n");
}

void CLPR_check_multiple_defs(char *name)
{
	PTERM *list;
	struct HASHNODE *hnode;
	CLPR_int i, n;

	if (!(hnode = CLPR_get_similar_func(name))) CLPR_fatal("31778");
	for (n=0,list=hnode->t; list; list=list->next)
		if (CLPR_p_first_rule(list->val)) {
			n++;
			CLPR_assign_p_added(list->val, FALSE);
		}

	if (n > 1) {
		fprintf(stderr, "Warning: rule overloading, same name, different arity:\n +++ ");
		for (i=0,list=hnode->t,i=0; list; list=list->next,i++) {
			if (!CLPR_p_first_rule(list->val)) continue; 
			n--;
			if ((i == 6) && (n > 0)) {
				i = 0;
				fprintf(stderr, "\n +++ ");
			}
			if (n > 0) fprintf(stderr, " %s/%d,", name, list->narg);
			else fprintf(stderr, " %s/%d", name, list->narg);
		}
		fprintf(stderr, "\n");
	}
}

CLPR_int CLPR_do_clear_style(void)
{
	CLPR_int i;

	for (i = 1; i < CLPR_getmaxfunc(); i++) 
		if (CLPR_p_added(i)) CLPR_assign_p_added(i, FALSE);
	return TRUE;
}

CLPR_int CLPR_compiling_r_val(void)
{
	return p_val;
}
