/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: CLPR_compile.c                                            *
 ***************************************************************/

#include <string.h>
#include <stdlib.h>
#include "standard.h"
#include "config.h"
#include "parse.h"
#include "symbols.h"
#include "lib.h"
#include "compile.h"
#include "emul.h"
#include "opcodes.h"
#include "main.h"

#define MIN_CODE_SZ (24000)

CLPR_int CLPR_retract_low_mark, CLPR_free_code_n=0;
extern CLPR_int CLPR_epcode;

CLPR_int CLPR_CODE_SZ = DEF_CODE_SZ;
code_t *CLPR_code;

char *CLPR_out_name;
FILE *CLPR_var_file;
char *CLPR_var_name = NULL;
bool CLPR_read_stdin = TRUE;
HASH_P_NODE_ptr CLPR_hash_p_table[MAX_HASH_P + 1];
HASH_L_NODE_ptr CLPR_hash_l_table[MAX_HASH_L + 1];
CLPR_unsigned CLPR_bit1vec[CLPR_WORDSIZE + 2];
LABLEREC CLPR_unresolved[MAXUNRESOLVED];
CLPR_int CLPR_punresolved;
FUNCTREC CLPR_functors[MAXFUNCTORS];
CLPR_int CLPR_startcode[MAXNAMES], CLPR_endcode[MAXNAMES];

HASH_P_NODE_ptr CLPR_find_hash_p_node(CLPR_int name);
HASH_P_NODE_ptr CLPR_new_hash_p_node(CLPR_int name);
RULE_ptr CLPR_p_first_rule(CLPR_int name), CLPR_p_last_rule(CLPR_int name);
RULE_ptr CLPR_store_rule(PTERM *to, PTERM *t, CLPR_int clobber);
PTERM *CLPR_old_quote_expand(PTERM *t, char quoted), *CLPR_quote_expand(PTERM *t, char quoted);
RULE_ptr CLPR_new_rule(void);
static CLPR_int check_quote(PTERM *pt);
HASH_L_NODE_ptr CLPR_p_jump_occ(CLPR_int name);

extern FILE *CLPR_outfile;
extern CLPR_int *CLPR_curar, *CLPR_stacktop, *CLPR_retreg;
extern CLPR_int *CLPR_regs;
extern char *CLPR_CLPRLIB;
extern FILE *CLPR_cur_stream, *CLPR_error_stream;	
extern CLPR_int CLPR_pcode;
extern CLPR_int CLPR_punresolved;
extern VREC *CLPR_varlist;
extern CLPR_int CLPR_reg_lowmark, CLPR_reg_highmark;
extern SPNODE *CLPR_spred[];
extern CLPR_int CLPR_REDEFINE_FLAG;
extern CLPR_int CLPR_WARN_ABORT;
extern CLPR_int CLPR_STYLE_CHECK_FLAG;

extern code_t *CLPR_alloc_mem(CLPR_int n);
extern FILE *CLPR_my_open(char *name, char *mode);
extern CLPR_int CLPR_the_lineno(void);
extern char *CLPR_the_filename(void);

static CLPR_int check_clause(PTERM *pt, CLPR_int *found_cut, CLPR_int *found_quote);
static CLPR_int check_pred(PTERM *pt, CLPR_int *found_cut, CLPR_int *found_quote);

/*---------------------------------------------------------------------------*/


CLPR_int CLPR_readfile (char *s, CLPR_int clobber);
extern void CLPR_compile (char clobber);
void CLPR_resolve (CLPR_int warn);
extern void CLPR_new_input (FILE *stream, char *name);
void CLPR_readfile_low (CLPR_int clobber);
extern CLPR_int CLPR_close_input (FILE *stream);
void CLPR_assign_p_cut (CLPR_int name, CLPR_int flag);
extern void CLPR_save_context (void);
extern void CLPR_lookatvars (PTERM *t);
extern CLPR_int CLPR_rungoal (CLPR_int ipc, CLPR_int interactive);
extern CLPR_int CLPR_compile_fgoal (PTERM *t);
extern void CLPR_restore_context (void);
CLPR_int CLPR_p_compiled (CLPR_int name);
extern void CLPR_warn_header (void);
void CLPR_assign_p_added (CLPR_int name, CLPR_int flag);
CLPR_int CLPR_p_protected (CLPR_int name);
extern void CLPR_free_pterm (PTERM *p);
void CLPR_delete_all_rules (CLPR_int val);
void CLPR_assign_p_compiled (CLPR_int name, CLPR_int flag);
void CLPR_assign_p_first_rule (CLPR_int name, RULE_ptr rule);
void CLPR_relink (HASH_L_NODE_ptr addr_link);
CLPR_int CLPR_p_dynamic (CLPR_int name);
void CLPR_assign_p_last_rule (CLPR_int name, RULE_ptr rule);
void CLPR_assign_p_n_rules (CLPR_int name, CLPR_int i);
CLPR_int CLPR_p_n_rules (CLPR_int name);
CLPR_int CLPR_p_start_hdrcode (CLPR_int name);
CLPR_int CLPR_p_end_hdrcode (CLPR_int name);
void CLPR_assign_p_start_hdrcode (CLPR_int name, CLPR_int i);
void CLPR_assign_p_end_hdrcode (CLPR_int name, CLPR_int i);
void CLPR_really_free_rules (void);
extern void CLPR_display_pterm (FILE *F, PTERM *t);
extern CLPR_int CLPR_sum_str (register char *s);
extern CLPR_int CLPR_get_quoted_func (CLPR_int v);

void CLPR_init_compiler(void)
{
CLPR_int i;
CLPR_unsigned bi;
	if (CLPR_CODE_SZ < MIN_CODE_SZ) CLPR_fatal("code size too small");
	CLPR_code = CLPR_alloc_mem(CLPR_CODE_SZ);
	CLPR_retract_low_mark = CLPR_CODE_SZ;
	if (!CLPR_code) CLPR_fatal("not enough memory to initialize clpr (calloc)");

	CLPR_punresolved = 0;
	for (i = 1, bi = 0x1; i <= CLPR_WORDSIZE; i++, bi <<= 1) {
		CLPR_bit1vec[i] = bi;
	}
	CLPR_bit1vec[CLPR_WORDSIZE + 1] = 0x1;
	CLPR_reg_lowmark = 1;
	CLPR_reg_highmark = MAX_REGS;
}


void CLPR_init_lable(void)
{
	CLPR_punresolved = 0;
}

static char s1[MAX_MESSAGE];
void CLPR_bootstrap(void) /*** load init file (clam first) ***/
{
	/*******
	strcpy(s1, CLPR_CLPRLIB);
	strcat(s1, "/");
	strcat(s1, "init.clam");
	if (CLPR_internal_consult(s1)) return;
	********/
	sprintf(s1, "%s/%s", CLPR_CLPRLIB, INITFNAME);
	if (!CLPR_readfile(s1, TRUE)) {
		fprintf(CLPR_error_stream, 
			"Error: unable to reconsult file [%s]:\n", s1);
		fprintf(CLPR_error_stream, 
			"Try setting the environment variable CLPR_CLPRLIB to the\n");
		fprintf(CLPR_error_stream, 
			"name of the directory containing the file: init.clpr\n");
		fprintf(CLPR_error_stream, 
			"Bye ...\n");
		exit(1);
	}
	CLPR_compile(TRUE);
	CLPR_resolve(TRUE);
}

CLPR_int CLPR_readfile(char *s, CLPR_int clobber) /*** file s, clobber iff reconsulting ***/
        
            
{
FILE *o_cur_stream, *CLPR_input_fd;
PTERM *tt;
CLPR_int i = 2;
	o_cur_stream = CLPR_cur_stream;
	if (strcmp(s, "user")) {
		if (!(CLPR_input_fd = CLPR_my_open(s, "r"))) return FALSE;
		CLPR_new_input(CLPR_input_fd, s);
	} else
		CLPR_new_input(stdin, "<stdin>");
	CLPR_readfile_low(clobber);
	if (CLPR_cur_stream != stdin) {
		fclose(CLPR_cur_stream);
		CLPR_close_input(CLPR_cur_stream);
	} else
		rewind(stdin);
	CLPR_new_input(o_cur_stream, "");
	return TRUE;
}

/***
	QUERY: 				?-		(run at CLPR_compile AND run time)
	DIRECTIVE:			:-		(CLPR_compile time only)
	COMMAND:			::-		(run time only)
***/

void CLPR_readfile_low(CLPR_int clobber) /*** low-level read ***/
            
{
PTERM *tt, *ttex;
CLPR_int i, *tmp1;
VREC *v;
CLPR_int found_a_cut, found_a_quote;
CLPR_int prev_rule_val = 0, cur_rule_val;
CLPR_int old_epcode;
/*
CLPR_int save_regs[MAX_REGS+1];
*/

	while (tt = (PTERM *) CLPR_readterm()) { 
/***
		if (CLPR_START_USER_SESSION) {
		CLPR_outfile = stdout;
		printf("COMP (%d) ",tt->val); CLPR_show_pterm(tt);
		fprintf(CLPR_outfile, ".\n");
		}
***/	
		found_a_cut = FALSE;
		if (!check_clause(tt, &found_a_cut, &found_a_quote))
			break;
		ttex = (found_a_quote ? CLPR_quote_expand(tt, FALSE) : tt);
		if (found_a_cut) 
			CLPR_assign_p_cut((is_implies(tt) ? tt->first->val : tt->val), TRUE);

		/*** no need to distinguish three kinds of goals here ***/
		if (is_directive(tt) || is_command(tt) || is_query(tt)) {
#if FREE_CODE_SPACE
			CLPR_reclaim_code_space();
#endif
			if (clobber) CLPR_compile(TRUE);
			else CLPR_compile(FALSE);
			CLPR_resolve(TRUE);
			CLPR_save_context();
/*
			for (i=0; i <= MAX_REGS; i++) save_regs[i] = CLPR_regs[i];
*/
			CLPR_lookatvars(ttex->first); 
			for (i = 1, v = CLPR_varlist; v != NULL; i++, v = v->next) {}
			tmp1 = CLPR_curar;
			CLPR_curar = CLPR_stacktop + 2;
			CLPR_stacktop = CLPR_curar + i - 1;
			*(CLPR_curar + FATHER) = (CLPR_int) tmp1;
			*(CLPR_curar + RETPT) = (CLPR_int) CLPR_retreg;
			old_epcode = CLPR_epcode;
			if (!CLPR_rungoal(CLPR_compile_fgoal(ttex->first), FALSE)) {
#ifdef FREE_CODE_SPACE
				CLPR_epcode = old_epcode;
#endif
				CLPR_warning("%s:%d, goal failed within consult", CLPR_the_filename(), CLPR_the_lineno());
				if (CLPR_WARN_ABORT) {		
					fprintf(CLPR_error_stream, "Aborting\n");
					CLPR_interrupt(0);
				}
			}
#ifdef FREE_CODE_SPACE
			CLPR_epcode = old_epcode;
			CLPR_reclaim_code_space();
#endif

			CLPR_restore_context();
/*
			for (i=0; i <= MAX_REGS; i++) CLPR_regs[i] = save_regs[i];
*/
			prev_rule_val = 0;
		} else {
			if (CLPR_STYLE_CHECK_FLAG & CONTIG_STYLE) {
				cur_rule_val = head_of_rule(ttex)->val;
				if (prev_rule_val != cur_rule_val) {
					if (!clobber) {
						if (CLPR_p_first_rule(cur_rule_val) && !CLPR_p_compiled(cur_rule_val)) {	
							CLPR_warn_header();
							fprintf(CLPR_error_stream,
								"Style check, %s/%d is not contiguous\n",
								string_of(cur_rule_val), arity_of(cur_rule_val));
						}
					} else if (!CLPR_p_compiled(cur_rule_val) && CLPR_p_first_rule(cur_rule_val)) {
						CLPR_warn_header();
						fprintf(CLPR_error_stream,
							"Style check, %s/%d is not contiguous\n",
							string_of(cur_rule_val), arity_of(cur_rule_val));
					}
					if (CLPR_store_rule(tt, ttex, clobber)) 
						prev_rule_val = cur_rule_val;
				} else CLPR_store_rule(tt, ttex, clobber);
			} else CLPR_store_rule(tt, ttex, clobber);
			CLPR_assign_p_added(tt->val, TRUE);
		}
	}
}

RULE_ptr CLPR_store_rule(PTERM *to, PTERM *t, CLPR_int clobber) /*** free to eventually ***/
              
            
{
CLPR_int val;
RULE *r, *tr;
HASH_L_NODE_ptr jump_occ;
	val = (is_implies(t) ? t->first->val : t->val);
	if (CLPR_p_protected(val)) {
		fprintf(CLPR_error_stream,
			"Error: %s/%d is protected ... ignored\n",
				string_of(val), arity_of(val));
		CLPR_free_pterm(t);
		return NULL;
	}
	if (CLPR_p_compiled(val)) {
		if (clobber) {
			CLPR_delete_all_rules(val);
			CLPR_assign_p_compiled(val, FALSE);
			CLPR_assign_p_first_rule(val, NULL);
			if (CLPR_REDEFINE_FLAG) 
				CLPR_warning("%s/%d has been redefined", string_of(val), arity_of(val));
			if (jump_occ = CLPR_p_jump_occ(val)) CLPR_relink(jump_occ);
		} else if (!CLPR_p_dynamic(val)) {
			fprintf(CLPR_error_stream,
				"Warning: %s/%d already compiled ... ignored\n",
					string_of(val), arity_of(val));
			CLPR_free_pterm(t);
			return NULL;
		}
	}
	r = CLPR_new_rule();
	r->next = NULL;
	r->t = to;
	r->wt = t;
	if (!CLPR_p_first_rule(val)) {	/* n_rules not used for dynamic */
		CLPR_assign_p_first_rule(val, r);
		CLPR_assign_p_last_rule(val, r);
		CLPR_assign_p_n_rules(val, 1);
		r->id = 1;
	} else {
		tr = CLPR_p_last_rule(val);
		tr->next = r;
		if (tr->id == -1) r->id = 1;
		else r->id = tr->id + 1;
		CLPR_assign_p_n_rules(val, CLPR_p_n_rules(val) + 1);
		CLPR_assign_p_last_rule(val, r);
	}
	return r;
}

#define RULE_BLOCK (80)

static RULE_ptr free_rule_list = (RULE_ptr) NULL;
static RULE_ptr tmp_free_rule_list = (RULE_ptr) NULL;
static char *start_rule = (char *) NULL;
static CLPR_int unused_rule = 0;

RULE_ptr CLPR_new_rule(void)
{
RULE_ptr r;
	if (free_rule_list) {
		r = free_rule_list;
		free_rule_list = r->next;
	} else if (unused_rule > 0) {
		unused_rule--;
		r = (RULE_ptr) start_rule;
		start_rule += sizeof(RULE);
	} else {
		unused_rule = RULE_BLOCK;
		start_rule = (char *) CLPR_emalloc(RULE_BLOCK * sizeof(RULE));
		unused_rule--;
		r = (RULE_ptr) start_rule;
		start_rule += sizeof(RULE);
	}
	r->next = NULL;
	r->endcode = -1; /* uncompiled state */
	return r;
}

void CLPR_free_rule(RULE_ptr r)
{
	CLPR_free_pterm(r->t); /*** r->wt? ***/
	r->id = 0;	/** 0 is special for deleted */
	r->next = tmp_free_rule_list;
	tmp_free_rule_list = r;
#ifdef FREE_CODE_SPACE
	CLPR_free_code_space(r->startcode,r->endcode);
	r->startcode = r->endcode = 0;
#endif
}

void CLPR_free_code_space(CLPR_int start, CLPR_int end)
{
#ifdef FREE_CODE_SPACE
	if (!start || end < start) return;
	if (start < CLPR_retract_low_mark) CLPR_retract_low_mark = start;
	CLPR_free_code_n += end - start + 1;
#endif
}
	
void CLPR_reclaim_code_space(void)
{
    if (CLPR_epcode - CLPR_retract_low_mark == CLPR_free_code_n) {
        CLPR_epcode = CLPR_retract_low_mark;
        CLPR_retract_low_mark = CLPR_CODE_SZ;
/* printf("freeed CS %d\n", CLPR_free_code_n); */
        CLPR_free_code_n  = 0;
    }
}

void CLPR_free_hdrcode(CLPR_int val)
{
    CLPR_free_code_space(CLPR_p_start_hdrcode(val), CLPR_p_end_hdrcode(val));
    CLPR_assign_p_start_hdrcode(val, 0);
    CLPR_assign_p_end_hdrcode(val, 0);
}

void CLPR_delete_all_rules(CLPR_int val)
{
RULE_ptr r, r99;
	for (r = CLPR_p_first_rule(val); r; r = r99) {
		r99 = r->next;
		CLPR_free_rule(r);
	}
#ifdef FREE_CODE_SPACE
	CLPR_free_hdrcode(val);
#endif
}

void CLPR_top_free_things(void)
{
	CLPR_really_free_rules();
#ifdef FREE_CODE_SPACE
	CLPR_reclaim_code_space();
#endif
}

void CLPR_really_free_rules(void)
{
	RULE_ptr prev = (RULE_ptr) NULL, r;

	for (r = tmp_free_rule_list; r; prev = r,r = r->next);
	if (prev) {
		prev->next = free_rule_list;
		free_rule_list = tmp_free_rule_list;
		tmp_free_rule_list = (RULE_ptr) NULL;
	}
}

static CLPR_int check_clause(PTERM *pt, CLPR_int *found_cut, CLPR_int *found_quote)
{
PTERM *body;
CLPR_int dummy;
	*found_quote = FALSE;
	if (is_p_funcval(pt, IMPL)) {
		if (!check_pred(pt->first, &dummy, found_quote))
			return FALSE;
		for (body = pt->first->next; is_p_funcval(body, COMMA); 
				body = body->first->next)
			if (!check_pred(body->first, found_cut, found_quote))
				return FALSE;
		if (!check_pred(body, found_cut, found_quote)) return FALSE;
	} else if (!check_pred(pt, found_cut, found_quote)) return FALSE;
	return TRUE;
}

static CLPR_int check_pred(PTERM *pt, CLPR_int *found_cut, CLPR_int *found_quote)
{
	if (is_p_real(pt) || is_p_var(pt) || is_p_string(pt)) {
		fprintf(CLPR_outfile, "Syntax error in ");
		CLPR_display_pterm(CLPR_outfile, pt);
		fprintf(CLPR_outfile, ".\n");
		return FALSE;
	} else if (is_cut(pt) && !(*found_cut)) *found_cut = TRUE;
	else if (!*found_quote) *found_quote = check_quote(pt);
	return TRUE;
}

static CLPR_int check_quote(PTERM *pt)
{
CLPR_int n;
PTERM *arg;
	if (is_p_func(pt)) {
		if (is_quote(pt)) {
			return TRUE;
		}
		n = pt->narg;
		for (arg = pt->first; n > 0; n--, arg = arg->next)
			if (check_quote(arg)) return TRUE;
	}
	return FALSE;
}
	
void CLPR_relink(HASH_L_NODE_ptr addr_link)
{
CLPR_int op, pc;
	/*
	if (CLPR_REDEFINE_FLAG) CLPR_warning("%s/%d has been redefined", string_of(addr_link->val), arity_of(addr_link->val));
	*/
	for (; addr_link; addr_link = addr_link->next) {
		pc = addr_link->loc;
		if (pc < MAX_GOAL_CODE) continue;
		if ((op = CLPR_code[pc-1]) == CALLP)	CLPR_code[pc-1] = CALLI;
		else if (op == JUMP) CLPR_code[pc-1] = JUMPI;
		else if (op != JUMPI && op != CALLI) CLPR_fatal("71896");
		CLPR_code[pc] = addr_link->val;
	}
}

/*--------------------------------------- procedure info --------------------*/

HASH_P_NODE_ptr CLPR_malloc_null_HASH_P_NODE(void)
{
HASH_P_NODE_ptr p;
	p = (HASH_P_NODE_ptr) calloc(1, sizeof(HASH_P_NODE));
	p->counter = 0.0;
	return p;
}

void CLPR_init_hash_p_table(void)
{
CLPR_int i;
	for (i = 0; i <= MAX_HASH_P; i++) CLPR_hash_p_table[i] = NULL;
}
		
CLPR_int CLPR_hash_p(CLPR_int name)
{
	return ((name % MAX_HASH_P) + 1); /*** 1 <= hash_value <= MAX_HASH_P ***/ 
}

static HASH_P_NODE_ptr last_p_node = NULL;
static CLPR_int last_name = -1;

HASH_P_NODE_ptr CLPR_find_hash_p_node(CLPR_int name)
{
register HASH_P_NODE_ptr hp;
CLPR_int i;
	if (name == last_name) return last_p_node;
	i = CLPR_hash_p(name);
	hp = CLPR_hash_p_table[i];
	while (hp) {
		if (hp->name == name) {
			last_name = name;
			last_p_node = hp;
			return hp;
		}
		hp = hp->next;
	}
	return hp;		
}

HASH_P_NODE_ptr CLPR_new_hash_p_node(CLPR_int name) /*** may or may not malloc ***/
         
{
HASH_P_NODE_ptr hp0, hp;
CLPR_int i;
	if (name == last_name) return last_p_node;
	i = CLPR_hash_p(name);
	hp = hp0 = CLPR_hash_p_table[i];
	while (hp) {
		if (hp->name == name) {
			last_name = name;
			last_p_node = hp;
			return hp;
		}
		hp = hp->next;
	}
	hp = (HASH_P_NODE_ptr) CLPR_malloc_null_HASH_P_NODE();
	hp->next = hp0;
	hp->name = name;
	CLPR_hash_p_table[i] = hp; 
	return hp;		
}

/*---------------------------------------------------------------------------*/
	
RULE_ptr CLPR_p_first_rule(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->first_rule;
	else return (RULE_ptr) NULL;
}	

RULE_ptr CLPR_p_last_rule(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->last_rule;
	else return (RULE_ptr) NULL;
}	

RULE_ptr CLPR_p_last_compiled_rule(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->last_compiled_rule;
	else return (RULE_ptr) NULL;
}	

CLPR_int CLPR_p_n_rules(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->n_rules;
	else return 0;
}	

CLPR_int CLPR_p_start_hdrcode(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->start_hdrcode;
	else return 0;
}	

CLPR_int CLPR_p_end_hdrcode(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->end_hdrcode;
	else return 0;
}	

CLPR_int CLPR_p_compiled(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->flags & H_COMPILED;
	else return FALSE;
}	

CLPR_int CLPR_p_dynamic(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->flags & H_DYNAMIC;
	else return FALSE;
}	

CLPR_int CLPR_p_protected(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->flags & H_PROTECTED;
	else return FALSE;
}	

CLPR_int CLPR_p_spy(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->flags & H_SPY;
	else return FALSE;
}	

CLPR_int CLPR_p_cut(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->flags & H_CUT;
	else return FALSE;
}	

CLPR_int CLPR_p_added(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->flags & H_ADDED;
	else return FALSE;
}	

HASH_L_NODE_ptr CLPR_p_jump_occ(CLPR_int name)
{
HASH_P_NODE_ptr hp;
CLPR_int i;
	hp = CLPR_find_hash_p_node(name);
	if (hp) return hp->jump_occ;
	else return (HASH_L_NODE_ptr) NULL;
}	

/*---------------------------------------------------------------------------*/

void CLPR_assign_p_first_rule(CLPR_int name, RULE_ptr rule)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	hp->first_rule = rule;
}	

void CLPR_assign_p_last_rule(CLPR_int name, RULE_ptr rule)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	hp->last_rule = rule;
}	

void CLPR_assign_p_last_compiled_rule(CLPR_int name, RULE_ptr rule)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	hp->last_compiled_rule = rule;
}	

void CLPR_assign_p_n_rules(CLPR_int name, CLPR_int i)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	hp->n_rules = i;
}	

void CLPR_assign_p_start_hdrcode(CLPR_int name, CLPR_int i)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	hp->start_hdrcode = i;
}	

void CLPR_assign_p_end_hdrcode(CLPR_int name, CLPR_int i)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	hp->end_hdrcode = i;
}	

void CLPR_assign_p_compiled(CLPR_int name, CLPR_int flag)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	if (flag) hp->flags |= H_COMPILED;
	else hp->flags &= ~H_COMPILED;
}	

void CLPR_assign_p_dynamic(CLPR_int name, CLPR_int flag)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	if (flag) hp->flags |= H_DYNAMIC;
	else hp->flags &= ~H_DYNAMIC;
}	

void CLPR_assign_p_protected(CLPR_int name, CLPR_int flag)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	if (flag) hp->flags |= H_PROTECTED;
	else hp->flags &= ~H_PROTECTED;
}	

void CLPR_assign_p_spy(CLPR_int name, CLPR_int flag)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	if (flag) hp->flags |= H_SPY;
	else hp->flags &= ~H_SPY;
}	

void CLPR_assign_p_cut(CLPR_int name, CLPR_int flag)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	if (flag) hp->flags |= H_CUT;
	else hp->flags &= ~H_CUT;
}

void CLPR_assign_p_added(CLPR_int name, CLPR_int flag)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	if (flag) hp->flags |= H_ADDED;
	else hp->flags &= ~H_ADDED;
}

void CLPR_assign_p_jump_occ(CLPR_int name, HASH_L_NODE_ptr occ)
{
HASH_P_NODE_ptr hp;
	hp = CLPR_new_hash_p_node(name);
	hp->jump_occ = occ;
}
	
/*------------------------------------------- label info --------------------*/

void CLPR_init_hash_l_table(void)
{
CLPR_int i;
	for (i = 0; i <= MAX_HASH_L; i++) CLPR_hash_l_table[i] = NULL;
}
		
CLPR_int CLPR_hash_l(CLPR_int val)
{
	return ((val % MAX_HASH_L) + 1); /*** 1 <= hash_value <= MAX_HASH_L ***/ 
}

void CLPR_store_lable(CLPR_int val, char *s)
{
register HASH_L_NODE_ptr hl;
CLPR_int i, fast;
	i = CLPR_hash_l(val);
	hl = CLPR_hash_l_table[i];
	fast = CLPR_sum_str(s);
	while (hl) {
		if (hl->str == NULL) {
			hl->val = val;
			hl->str = s;
			hl->loc = CLPR_pcode;
			hl->quick_hash = fast;
			return;
		} 
		if (fast == hl->quick_hash && hl->val != FAILVAL && !strcmp(hl->str, s)) {
			CLPR_warning("duplicate label: %s", s);
			return;
		}
		hl = hl->next;
	}
	hl = (HASH_L_NODE_ptr) CLPR_emalloc(sizeof(HASH_L_NODE));
	hl->next = CLPR_hash_l_table[i];
	hl->val = val;
	hl->str = s;
	hl->loc = CLPR_pcode;
	hl->quick_hash = fast;
	CLPR_hash_l_table[i] = hl;
}

void CLPR_delete_lable(CLPR_int val)
{
HASH_L_NODE_ptr hl;
CLPR_int i;
	i = CLPR_hash_l(val);
	hl = CLPR_hash_l_table[i];
	while (hl) {
		if (hl->val == val) hl->str = NULL;
		hl = hl->next;
	}
}

CLPR_int CLPR_find_loc_for_lable(CLPR_int val, char *s)
{
HASH_L_NODE_ptr hl;
CLPR_int i, fast;
	i = CLPR_hash_l(val);
	hl = CLPR_hash_l_table[i];
	fast = CLPR_sum_str(s);	
	while (hl) {
		if (hl->str && fast == hl->quick_hash && !strcmp(hl->str, s)) 
			return hl->loc;
		hl = hl->next;
	}
	return (-1);
}

char *CLPR_find_lable_for_addr(CLPR_int *addr)
{
HASH_L_NODE_ptr hl;
CLPR_int i, loc;
	loc = intpc(addr);
	for (i = 0; i <= MAX_HASH_L; i++) {
		if (hl = CLPR_hash_l_table[i]) {
			while (hl) {
				if (hl->str && hl->loc == loc) return hl->str;
				hl = hl->next;
			}
		}
	}
	return NULL;
}
		
/*---------------------------------------------------------------------------*/

void CLPR_storename(char *s, CLPR_int val, CLPR_int narg)
{
	CLPR_store_lable(val, s);
}

void CLPR_add_unres_lable(char *s, CLPR_int val, CLPR_int pc, CLPR_int off)
{
    CLPR_code[pc] = CLPR_punresolved;
    CLPR_unresolved[CLPR_punresolved].lable = s;
    CLPR_unresolved[CLPR_punresolved].val = val;
    CLPR_unresolved[CLPR_punresolved].loc = pc;
    CLPR_unresolved[CLPR_punresolved++].off = off;
}

void CLPR_copy_unres_lable(CLPR_int from_pc, CLPR_int pc)
{
    CLPR_int i;
    if (((i = CLPR_code[from_pc]) < 0) || (i >= CLPR_punresolved))
        CLPR_fatal("791237");
    CLPR_code[pc] = CLPR_punresolved;
    CLPR_unresolved[CLPR_punresolved].lable = CLPR_unresolved[i].lable;
    CLPR_unresolved[CLPR_punresolved].val = CLPR_unresolved[i].val;
    CLPR_unresolved[CLPR_punresolved].loc = pc;
    CLPR_unresolved[CLPR_punresolved++].off = 0;
}

void CLPR_resolve(CLPR_int warn)
{
CLPR_int i, j, k;
	for (i = 0; i < CLPR_punresolved; i++) {
		k = CLPR_unresolved[i].loc;
		j = CLPR_find_loc_for_lable(CLPR_unresolved[i].val, CLPR_unresolved[i].lable);
		if (j != -1) CLPR_code[k] = (CLPR_int) (CLPR_code + j + CLPR_unresolved[i].off);
		else {
			if (warn) CLPR_warning("unresolved name %s - treating as fail/0",
					 CLPR_unresolved[i].lable);
			CLPR_code[k] = (CLPR_int) (CLPR_code + FAIL_ADDR);
		}
	}
}

PTERM *CLPR_duplicate_pterm(PTERM *t)
{
PTERM *nt;
CLPR_int i;
char *c1, *c2;
	nt = (PTERM *) CLPR_new_pterm();
#ifdef BSD
	bcopy(t, nt, sizeof(PTERM));
#else
	memcpy(nt, t, sizeof(PTERM));
#endif
	return nt;
}

PTERM *CLPR_unpack_string(PTERM *t)
{
	return t; /* just a stub for now */
}
			
PTERM *CLPR_old_quote_expand(PTERM *t, char quoted)
{
PTERM *nt;
PTERM *next;
if (!t) return NULL;
if (is_p_string(t))
         {
         nt = CLPR_unpack_string(t);
         nt->next = CLPR_quote_expand(t->next,quoted);
         }
else if (quoted && is_arith_term(t))
         {
         nt = CLPR_duplicate_pterm(t);
         nt->val = (t->val - ARITH_FUNC_0 + ARITH_FUNC_4);
         /* do we also want to fix the string? */
         nt->next = CLPR_quote_expand(t->next,quoted);
         nt->first = CLPR_quote_expand(t->first,quoted);
         }
else if (is_quote(t))
         {
         /* invariant: t->first->next == NULL */
         next = t->next;
         nt = CLPR_quote_expand(t->first,TRUE);
         nt->next = CLPR_quote_expand(next,quoted);
         }
else
         {
         nt = CLPR_duplicate_pterm(t);
         nt->first = CLPR_quote_expand(t->first,quoted);
         nt->next = CLPR_quote_expand(t->next,quoted);
         }
return nt;
}

PTERM *CLPR_quote_expand(PTERM *t, char quoted)
{
CLPR_int v, n, *start, *s;
PTERM *arg, *newt, *newarg, *prev;
	switch (t->ty) {
	case REAL:
	case STR:
	case VAR:
		return CLPR_duplicate_pterm(t);
	case FUNC:
		if (is_quote(t) && !quoted) return CLPR_quote_expand(t->first, TRUE);
		if (t->narg == 0) return CLPR_duplicate_pterm(t);
		newt = CLPR_duplicate_pterm(t);
		newt->first = CLPR_quote_expand(t->first, quoted);
		arg = t->first;
		newarg = newt->first;
		for (n = 2;	n <= t->narg; n++) {
			newarg->next = CLPR_quote_expand(arg->next, quoted);
			arg = arg->next;
			newarg = newarg->next;
		}
		if (is_arith_term(t) && quoted)	newt->val = CLPR_get_quoted_func(t->val);
		return newt;
	default:
		CLPR_fatal("47812");
		return (PTERM *) NULL; /* quiet gcc */
	}
}

