/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: compile.h                                            *
 ***************************************************************/

#ifndef CLPR_COMPILE_H
#define CLPR_COMPILE_H

#include "standard.h"
#include "parse.h"
#include "lib.h"

#define MAX_REGS        1024   /*** < (CLPR_REGS_SZ - REGS), multiple of 64 ***/
#define MAXUNRESOLVED	10000
#define MAXFUNCTORS		8000
#define MAXCLAUSES		3000
#define MAXVARS			20000
#define MAXNAMES 		10000
#define MAX_HASH_P 		3073
#define MAX_HASH_L 		3073
#define MAX_VAR_OCC     64

#define DUMP_SEPARATOR  ('\001')	/* separator for dump variable strings */

#define check_codespace(i) { \
	if (CLPR_CODE_SZ - (i) <= CODE_SAFETY) CLPR_fatal("no more code space"); \
} 
#define check_goalspace(i) { \
	if ((i) + CODE_SAFETY >= MAX_GOAL_CODE) CLPR_fatal("no more goal space"); \
} 
#define is_rr(t) (bit1((t)->extra_flag)) /*** right-recursive? ***/
#define set_rr(t) (set_bit1((t)->extra_flag))
#define unset_rr(t) (unset_bit1((t)->extra_flag))
#define val_rule(t) ((t)->val == IMPL ? (t)->first->val : (t)->val)

typedef struct lablerec	{
	char *lable;
	CLPR_int val;
	CLPR_int loc;
	CLPR_int off;
	CLPR_int jump;
} LABLEREC;

typedef struct functrec {
	char arity;
	char *pname;
	CLPR_int special;
} FUNCTREC;

typedef struct rule {
	PTERM *t, *wt;
	struct rule *next;
	CLPR_int id; 
	CLPR_int startcode, endcode;
} RULE, *RULE_ptr;

#define ADDRESS    1 /*** assembly instruction arguments ***/
#define SINT       2
#define VARIABLE   3
#define REG        4
#define FUNCTOR    5
#define CONSTANT   6
#define BYTE       7
#define FLOAT      8
#define STRINGLOC  9
#define FLEXARG    10
#define CODEADDR   11

typedef struct asmrec {
	CLPR_int num;			   /* Numerical values of instructions */
	char *mnemonic;		   /* String name of instruction */
	CLPR_int nargs;			   /* Number of arguments */
	CLPR_int type[5];		   /* Types of the arguments */
} ASMREC;

typedef struct vrec {      /* var info per rule */
	struct vrec *next;
	CLPR_int val;                       
	char *str;
	CLPR_int saved_stack_loc;
	CLPR_int stack_loc;
	CLPR_int occ_top;
	struct {
		CLPR_int atom, arg, level;
	} occ[MAX_VAR_OCC + 1];	 
	CLPR_int n_live_regs;  
	CLPR_unsigned live_regs[(MAX_REGS / CLPR_WORDSIZE) + 1];
} VREC;

typedef struct hash_p_node {
	struct hash_p_node *next;
	CLPR_int name;
	CLPR_int n_rules;
	RULE *first_rule, *last_rule, *last_compiled_rule;
	CLPR_int start_hdrcode, end_hdrcode; /*** end_hdrcode == 0 for nonheaders ***/
	SPNODE *spred;
	CLPR_int flags;
	struct hash_l_node *jump_occ;
	double counter;
} HASH_P_NODE, *HASH_P_NODE_ptr;

typedef struct hash_l_node {
	char *str;
	struct hash_l_node *next;
	CLPR_int val;
	CLPR_int loc;
	CLPR_int quick_hash;
} HASH_L_NODE, *HASH_L_NODE_ptr;

typedef CLPR_int code_t;

/*** HASH_P_NODE extra bit flags */
#define H_COMPILED  (0x1)
#define H_DYNAMIC   (0x2)
#define H_PROTECTED (0x4)
#define H_SPY       (0x8)
#define H_CUT       (0x10)
#define H_ADDED     (0x20)

#endif /* CLPR_COMPILE_H */
