/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: debug.c                                              *
 ***************************************************************/

#include <string.h>
#include <stdlib.h>
#include "standard.h"
#include "config.h"
#include "parse.h"
#include "lib.h"
#include "compile.h"
#include "opcodes.h"
#include "emul.h"
#include "solver.h"
#include "main.h"

CLPR_int CLPR_inst_count[MAX_OPCODE + 1];

extern FILE *CLPR_outfile;
extern ASMREC CLPR_asmlist[];
extern CLPR_int CLPR_inst_count[];
extern NL_EQN_ptr CLPR_nl_eqn_top;

static void print_inst_freq(void);

/*---------------------------------------------------------------------------*/


CLPR_int CLPR_count_occ_in_row (CLPR_int type, CLPR_int v1, CLPR_int v2);
void CLPR_debug_error (CLPR_int n);
CLPR_int CLPR_count_id_in_eqn (EQN_ptr eqn, CLPR_int i);
void CLPR_print_eq_tableau (void);
void CLPR_print_ineq_tableau (void);
extern void CLPR_print_pre_malloc_stats (void);
void CLPR_print_emul_stats (void);
void CLPR_print_solver_stats (void);
extern void CLPR_print_spaces (CLPR_int n);
void CLPR_print_s_malloc_stats (void);
void CLPR_print_eq_stats (void);
void CLPR_print_ineq_stats (void);
void CLPR_print_nlin_stats (void);
extern CLPR_int CLPR_count_solver_malloc_space (void);
void CLPR_print_eqn (EQN_ptr eqn);
void CLPR_print_eq_occ (CLPR_int v);
void CLPR_print_ineq_occ (CLPR_int v);
void CLPR_print_nlin_occ (CLPR_int v);
void CLPR_print_nl_eqn (NL_EQN_ptr nl_eqn);
extern void CLPR_printterm (FILE *f, CLPR_int *t);
extern void CLPR_output_arith_func (CLPR_int type);
extern void CLPR_print_var2pos (FILE *f, CLPR_int *v);

void CLPR_check_eq_tableau(void)
{
CLPR_int i, j, k;
double c;
TAB_HDR_NODE_ptr hdr, hdr2;
EQN_ptr eqn;
ENTRY_ptr entry;
OCC_ptr occ;
	printf("Checking eq tableau (%d) ... ", CLPR_solver_id);
	fflush(stdout);
	for (i = 1; i <= CLPR_solver_id; i++) {
		if (!(hdr = CLPR_tableau[i])) continue;
		if (eqn = hdr->eqn) {
			entry = eqn->next;
			while (entry) { 
				j = CLPR_count_occ_in_row(EQ, entry->id, i);
				if (j < 1) CLPR_debug_error(88801);
				entry = entry->next;
			}
		}
		if (occ = hdr->eq_occ) {
			do {
				if (!(hdr2 = CLPR_tableau[occ->id])) CLPR_debug_error(88802);
				j = CLPR_count_id_in_eqn(hdr2->eqn, i);
				if (j != 1) CLPR_debug_error(88803);
				occ = occ->next;
			} while (occ);
		}
	}	
	printf("OK\n");
}

void CLPR_check_ineq_tableau(void)
{
CLPR_int i, j, k, basic_var;
double c;
TAB_HDR_NODE_ptr hdr;
EQN_ptr eqn;
EQN_PARAMS_ptr params;
ENTRY_ptr entry;
/*
	printf("Checking ineq CLPR_tableau (%d .. %d) ... ", 
		CLPR_slack_id, CLPR_SOLVER_SZ - 1);
*/
	fflush(stdout);
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--) {
		if (!(hdr = CLPR_tableau[i])) CLPR_debug_error(77701);
		if (!(eqn = hdr->eqn)) CLPR_debug_error(77702);
if (eqn->params) {
  if (eqn->constant > CLPR_EPS) CLPR_debug_error(66182);
  if (eqn->params->basic >= CLPR_slack_id && ((ENTRY_ptr) CLPR_find_entry(eqn, eqn->params->basic))->coef < -CLPR_EPS)
    CLPR_debug_error(66199);
}
		if (!(params = eqn->params)) CLPR_debug_error(77703);
		basic_var = params->basic;
		if (!(entry = eqn->next)) continue; /* CLPR_debug_error(77704); */
/*
		k = 0;
		while (entry) {
			j = CLPR_count_occ_in_row(INEQ, entry->id, i);
			if (j != 1) CLPR_debug_error(77705);
			if (entry->id == basic_var) k++;
			entry = entry->next;
		}
		if (k != 1) CLPR_debug_error(77706);
		if (!(hdr = CLPR_tableau[params->basic])) CLPR_debug_error(77707);
		if (hdr->basic_eqn != i) CLPR_debug_error(77708);
*/
		
	}
/*
	printf("OK\n");
*/
}

void CLPR_check_tableaus(void)
{
	CLPR_check_eq_tableau();
	CLPR_check_ineq_tableau();
}

CLPR_int CLPR_count_occ_in_row(CLPR_int type, CLPR_int v1, CLPR_int v2)
{
CLPR_int count;
TAB_HDR_NODE_ptr hdr;
OCC_ptr occ;
	if (!(hdr = CLPR_tableau[v1])) return -1;
	occ = (type == EQ ? hdr->eq_occ : hdr->ineq_occ);
	count = 0;
	while (occ) {
		if (occ->id == v2) count++;
		occ = occ->next;
	}
	return count;
}

CLPR_int CLPR_count_id_in_eqn(EQN_ptr eqn, CLPR_int i)
{
CLPR_int count;
ENTRY_ptr entry;
	if (!eqn) return -1;
	entry = eqn->next;
	count = 0;
	while (entry) {
		if (entry->id == i) count++;
		entry = entry->next;
	}
	return count;
}

void CLPR_debug_error(CLPR_int n)
{
	printf("Internal error: %5d\n", n);
	exit(1);
}

/*----------------------------------------------- diagnostic i/o ------------*/

void CLPR_print_tableaus(void)
{
	CLPR_print_eq_tableau();
	CLPR_print_ineq_tableau();
	if (CLPR_nlin_count) printf("... and %d non-linears\n", CLPR_nlin_count);
}

void CLPR_print_stats(void)
{
	fprintf(CLPR_outfile, "\nSTATS:\n");  
	CLPR_print_pre_malloc_stats();
	CLPR_print_emul_stats();
	CLPR_print_solver_stats();
}

void CLPR_print_e_stats(void)
{
/***
	fprintf(CLPR_outfile, "\nEMULATOR STATS:\n");  
	CLPR_print_pre_malloc_stats();
	CLPR_print_emul_stats();
***/
	printf("Spare counters: %d %d %d %d\n",
		CLPR_s_count[540], CLPR_s_count[541], CLPR_s_count[542], CLPR_s_count[543]); 
}

void CLPR_print_s_stats(void)
{
	fprintf(CLPR_outfile, "\nSOLVER STATS:\n");  
	CLPR_print_pre_malloc_stats();
	CLPR_print_solver_stats();
}

CLPR_int CLPR_unify_stats[25];
void CLPR_print_emul_stats(void)
{
	CLPR_int i;
	fprintf(CLPR_outfile, "Emulator: %d instructions\n", CLPR_s_count[299]);
	fprintf(CLPR_outfile, "\tInstruction frequencies:\n");
	print_inst_freq();
	printf("UNIFY stats:");
	for (i=0; i <= 24; i++)
		if (CLPR_unify_stats[i] > 0) printf(" (%d) = %d",i, CLPR_unify_stats[i]);
	printf("\n");
}

static void print_inst_freq(void)
{
CLPR_int i, j, count;
char *s;
	count = 0;
	for (i = 0; i <= MAX_OPCODE; i++) {
		if (!CLPR_inst_count[i]) continue;
		for (j = 0, s = (char *)NULL; j <= MAX_OPCODE; j++)
			if (i == CLPR_asmlist[j].num) {
				s = CLPR_asmlist[j].mnemonic;
				break;
			}
		if (!s) {
			fprintf(stderr, "OPCODE %d missing\n",i);
			continue;
		}
		if ((count % 3) == 0) fprintf(CLPR_outfile, "\t");
		count++;
		CLPR_print_spaces(13 - strlen(s));
		fprintf(CLPR_outfile, "%s:%6d", s, CLPR_inst_count[i]);
		if ((count % 3) == 0) fprintf(CLPR_outfile, "\n");
	}
	if ((count % 3) != 0) fprintf(CLPR_outfile, "\n");

}

void CLPR_print_solver_stats(void)
{
	CLPR_print_s_malloc_stats();
	CLPR_print_eq_stats();
	CLPR_print_ineq_stats();
	CLPR_print_nlin_stats();
}

void CLPR_print_s_malloc_stats(void)
{
fprintf(CLPR_outfile, 
	"Original Solver Malloc Space: %d blocks * %d = %d nodes (%d bytes)\n",
	CLPR_sms_blocks, SOLVER_MALLOC_NODES, CLPR_sms_blocks * SOLVER_MALLOC_NODES, 
	CLPR_sms_blocks * SOLVER_MALLOC_NODES * sizeof(SOLVER_NODE));
fprintf(CLPR_outfile, 
	"Present Solver Malloc Space: %d nodes\n", CLPR_count_solver_malloc_space());
fprintf(CLPR_outfile, 
	"\tEQN nodes: %dm, %df\n", CLPR_malloc_eqn_count, CLPR_free_eqn_count);
fprintf(CLPR_outfile, 
	"\tEQN_PARAMS: %dm, %df\n", CLPR_malloc_eqn_par_count, CLPR_free_eqn_par_count);
fprintf(CLPR_outfile, 
	"\tENTRY: %dm, %df\n", CLPR_malloc_entry_count, CLPR_free_entry_count);
fprintf(CLPR_outfile, 
	"\tOCC: %dm, %df\n", CLPR_malloc_occ_count, CLPR_free_occ_count);
fprintf(CLPR_outfile, 
	"\tNL_EQN: %dm, %df\n", CLPR_malloc_nl_eqn_count, CLPR_free_nl_eqn_count);
fprintf(CLPR_outfile, 
	"\tNL_OCC: %dm, %df\n", CLPR_malloc_nl_occ_count, CLPR_free_nl_occ_count);
fprintf(CLPR_outfile, 
	"\tNL_PARAMS1: %dm, %df\n", CLPR_malloc_nl_par1_count, CLPR_free_nl_par1_count);
fprintf(CLPR_outfile, 
	"\tNL_PARAMS2: %dm, %df\n", CLPR_malloc_nl_par2_count, CLPR_free_nl_par2_count);
fprintf(CLPR_outfile, 
	"\tHDR: %dm, %df\n", CLPR_malloc_hdr_count, CLPR_free_hdr_count);
fprintf(CLPR_outfile, 
	"\tDUMP: %dm, %df\n", CLPR_malloc_koh_count, CLPR_free_koh_count);
}

void CLPR_print_eq_stats(void)
{
	if (!(CLPR_s_count[4] + CLPR_s_count[5])) return;
	fprintf(CLPR_outfile, "EQ solver: %d calls (%d new vars, %d param vars)\n",
		CLPR_s_count[4] + CLPR_s_count[5], CLPR_s_count[4], CLPR_s_count[5]);
	fprintf(CLPR_outfile, "\tInsert_occ: %d calls\n", CLPR_s_count[14]);
	fprintf(CLPR_outfile, "\tDelete_occ: %d calls, %d extra searches\n",
		CLPR_s_count[14], CLPR_s_count[13]);
	fprintf(CLPR_outfile, "\tRestore eqn's: %d nulls, %d non-nulls\n",
		CLPR_s_count[30], CLPR_s_count[31]);
}

void CLPR_print_ineq_stats(void)
{
	if (!(CLPR_s_count[180] + CLPR_s_count[104])) return;	
	fprintf(CLPR_outfile, "INEQ_solver: %d calls (%d new rows, %d ineq_subs)\n",
		CLPR_s_count[180] + CLPR_s_count[104], CLPR_s_count[180], CLPR_s_count[104]);
	fprintf(CLPR_outfile, "\tIneq_subs: %d basics, %d non-basics\n",
		CLPR_s_count[105], CLPR_s_count[104] - CLPR_s_count[105]);
	fprintf(CLPR_outfile, "\tMain ineq loop: %d (%g pivots per constraint)\n",
		CLPR_s_count[103], 
		(double) (CLPR_s_count[103]) / (double) (CLPR_s_count[180] + CLPR_s_count[104] + 1));
	fprintf(CLPR_outfile, "\tRestore ineq's: %d nulls, %d non-nulls\n",
		CLPR_s_count[130], CLPR_s_count[131]);
	fprintf(CLPR_outfile, "\tIneq : %d success, %d fails\n",
		CLPR_s_count[122], CLPR_s_count[180]-CLPR_s_count[122]);
	fprintf(CLPR_outfile, "\tIneq subs basic: %d success, %d fails\n",
		CLPR_s_count[123], CLPR_s_count[105] - CLPR_s_count[123]);
	fprintf(CLPR_outfile, "\tMalloc params: %d backup %d gt0 %d solve_ineq\n",
		CLPR_s_count[124], CLPR_s_count[125], CLPR_s_count[126]);
	fprintf(CLPR_outfile, "\tFree params: %d backup %d gt0 %d solve_ineq\n",
		CLPR_s_count[134], CLPR_s_count[135], CLPR_s_count[136]);
}

void CLPR_print_nlin_stats(void)
{
	if (!CLPR_s_count[200]) return;
	fprintf(CLPR_outfile, "NLIN_solver: %d mults\n", CLPR_s_count[200]);
	fprintf(CLPR_outfile, "\tWakeups: %d\n", CLPR_s_count[261]);
	fprintf(CLPR_outfile, "\tNl_eqn nodes: %d orig's, %d changes\n",
		CLPR_s_count[206], CLPR_s_count[204]);
	fprintf(CLPR_outfile, "\tRestore nl_eqn's: %d\n", CLPR_s_count[230]);
}

/*---------------------------------------------------------------------------*/

void CLPR_print_hdr(TAB_HDR_NODE_ptr hdr)
{
CLPR_int i;
	if (!hdr) fprintf(CLPR_outfile, "<NULL hdr>\n");
	else {
		i = hdr->eqn->id;
		fprintf(CLPR_outfile, "HDR (%p):\n", hdr);
		fprintf(CLPR_outfile, "\teqn: "); CLPR_print_eqn(hdr->eqn);
		fprintf(CLPR_outfile, "\teq_occ: "); CLPR_print_eq_occ(i);
		fprintf(CLPR_outfile, "\tineq_occ: "); CLPR_print_ineq_occ(i);
		fprintf(CLPR_outfile, "\tnlin_occ: "); CLPR_print_nlin_occ(i);
		fprintf(CLPR_outfile, "\tbasic_eqn: %d\n", hdr->basic_eqn);
	}
}

void CLPR_print_eq_tableau(void)
{
CLPR_int i;
	fprintf(CLPR_outfile, 
		"------------------------------ Eq_tableau -----------------\n");
	for (i = CLPR_solver_id; i >= 1; i--) 
		if (CLPR_tableau[i] && CLPR_tableau[i]->eqn) CLPR_print_eqn(CLPR_tableau[i]->eqn);
	fprintf(CLPR_outfile, 
		"-----------------------------------------------------------\n");
}

void CLPR_print_ineq_tableau(void)
{
CLPR_int i;
	fprintf(CLPR_outfile, 
		"----------- Ineq_tableau ----------------------------------\n");
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--)
		if (CLPR_tableau[i] && CLPR_tableau[i]->eqn) CLPR_print_eqn(CLPR_tableau[i]->eqn);
	fprintf(CLPR_outfile, 
		"-----------------------------------------------------------\n");
}

void CLPR_print_nlins(void)
{
CLPR_int i, *t;
NL_EQN_ptr nl_eqn;
	for (nl_eqn = CLPR_nl_eqn_top; nl_eqn; nl_eqn = nl_eqn->next) {
		if (nl_eqn->type == WOKEN || nl_eqn->partner) continue;
		CLPR_print_nl_eqn(nl_eqn);
	}
}

void CLPR_print_eqn(EQN_ptr eqn)
{
ENTRY_ptr entry;

	CLPR_outfile = stderr;

	if (!eqn) {
		fprintf(CLPR_outfile, "<NULL eqn>\n");
		return;
	}
	fprintf(CLPR_outfile, "[%d]: %.2e", eqn->id, eqn->constant);
	if (eqn->params) fprintf(CLPR_outfile, " (%.2eE)", eqn->params->epsilon);
	entry = eqn->next;
	while (entry) {
		if (entry->coef > 0.0) 
			fprintf(CLPR_outfile, " +%.2e*X%d", entry->coef, entry->id);
		else fprintf(CLPR_outfile, " -%.2e*X%d", -(entry->coef), entry->id);
		if (eqn->params && eqn->params->basic == entry->id) {
			putc('^', CLPR_outfile);
		}
		entry = entry->next;
	}
	fprintf(CLPR_outfile, " (ts:%d)\n", eqn->time_stamp);
}

void CLPR_print_eq_occ(CLPR_int v)
{
OCC_ptr occ;
	if (!CLPR_tableau[v]) fprintf(CLPR_outfile, "Tab[%d] = NULL\n", v);
	else
	if (!CLPR_tableau[v]->eq_occ) fprintf(CLPR_outfile, "Tab[%d]->eq_occ = NULL\n", v);
	else {
		fprintf(CLPR_outfile, "Tableau[%d]->eq_occ: ", v);
		occ = CLPR_tableau[v]->eq_occ;
		do {
			fprintf(CLPR_outfile, "%d", occ->id);
			if (occ = occ->next) fprintf(CLPR_outfile, ", ");
		} while (occ);
		putc('\n', CLPR_outfile);
	}
}

void CLPR_print_ineq_occ(CLPR_int v)
{
OCC_ptr occ;
	if (!CLPR_tableau[v]) fprintf(CLPR_outfile, "Tab[%d] = NULL\n", v);
	else
	if (!CLPR_tableau[v]->ineq_occ) 
		fprintf(CLPR_outfile, "Tab[%d]->ineq_occ = NULL\n", v);
	else {
		fprintf(CLPR_outfile, "Tableau[%d]->ineq_occ: ", v);
		occ = CLPR_tableau[v]->ineq_occ;
		do {
			fprintf(CLPR_outfile, "%d", occ->id);
			if (occ = occ->next) fprintf(CLPR_outfile, ", ");
		} while (occ);
		putc('\n', CLPR_outfile);
	}
}

void CLPR_print_nlin_occ(CLPR_int v)
{
NL_OCC_ptr occ;
NL_EQN_ptr nl_eqn;
NL_PARAMS1_ptr params1;
	if (!CLPR_tableau[v]) fprintf(CLPR_outfile, "Tab[%d] = NULL\n", v);
	else
	if (!CLPR_tableau[v]->nlin_occ) 
			fprintf(CLPR_outfile, "Tab[%d]->nlin_occ = NULL\n", v);
	else {
		fprintf(CLPR_outfile, "Tab[%d]->nlin_occ: ", v);
		occ = CLPR_tableau[v]->nlin_occ;
		do {
			nl_eqn = occ->nl_eqn;
			params1 = nl_eqn->params1;
			fprintf(CLPR_outfile, "(ty:%d c99:%g v0:%d v1:%d v2:%d)",
				nl_eqn->type, params1->coef,
				params1->id0, params1->id1, params1->id2);
			if (occ = occ->next) fprintf(CLPR_outfile, ", ");
		} while (occ);
		putc('\n', CLPR_outfile);
	} 
}

void CLPR_print_occ_tableau(void)
{
CLPR_int i;
	fprintf(CLPR_outfile, 
		"---------------- Occ_tableau ------------------------\n");
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--) 
		if (CLPR_tableau[i]) {
			if (CLPR_tableau[i]->eq_occ) CLPR_print_eq_occ(i);
			if (CLPR_tableau[i]->ineq_occ) CLPR_print_ineq_occ(i);
			if (CLPR_tableau[i]->nlin_occ) CLPR_print_nlin_occ(i);
		}
	for (i = CLPR_solver_id; i >= 1; i--) 
		if (CLPR_tableau[i]) {
			if (CLPR_tableau[i]->eq_occ) CLPR_print_eq_occ(i);
			if (CLPR_tableau[i]->ineq_occ) CLPR_print_ineq_occ(i);
			if (CLPR_tableau[i]->nlin_occ) CLPR_print_nlin_occ(i);
		}
	fprintf(CLPR_outfile, 
		"-----------------------------------------------------\n");
}

void CLPR_print_basic_vars(void)
{
CLPR_int i;
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--) 
		if (CLPR_tableau[i] && CLPR_tableau[i]->basic_eqn)
			fprintf(CLPR_outfile, 
				"Tab[%d]->basic_eqn = %d\n", i, CLPR_tableau[i]->basic_eqn);
	for (i = CLPR_solver_id; i >= 1; i--) 
		if (CLPR_tableau[i] && CLPR_tableau[i]->basic_eqn)
			fprintf(CLPR_outfile, 
				"Tab[%d]->basic_eqn = %d\n", i, CLPR_tableau[i]->basic_eqn);
}

void CLPR_print_nl_eqn(NL_EQN_ptr nl_eqn)
{
NL_PARAMS1_ptr params1;
CLPR_int i, *t, type, v0, v1, v2;
double c, c2;
	params1 = nl_eqn->params1;
	v0 = params1->id0; v1 = params1->id1; v2 = params1->id2;
	c = params1->coef;
	if (nl_eqn->type == SOLVER_EVAL) {
CLPR_printterm(CLPR_outfile, (CLPR_int *)v0); fprintf(CLPR_outfile, " = ");
	} else fprintf(CLPR_outfile,"X%d = ", v0);	
	switch (type = nl_eqn->type) {
		case SOLVER_MULT: 
			if (!is_equal(c, 1.0)) {
				fprintf(CLPR_outfile, "%g * ", c);
			}
			fprintf(CLPR_outfile, "X%d * X%d\n", v1, v2);
			break;
		case SOLVER_ABS: 
		case SOLVER_ABS0: 
		case SOLVER_SIN:
		case SOLVER_COS:
			CLPR_output_arith_func(type);
			fprintf(CLPR_outfile, "X%d)\n", v1);
			break;
		case SOLVER_POW: 
		case SOLVER_POW0: 
		case SOLVER_POW1: 
		case SOLVER_POW2: 
			CLPR_output_arith_func(type);
			fprintf(CLPR_outfile, "X%d, X%d)\n", v1, v2);
			break;
		case SOLVER_EVAL:
			CLPR_output_arith_func(type);
			t = (CLPR_int *) CLPR_findbind((CLPR_int *)v1);
			if (CLPR_tag(*t) == TAG_PAR) fprintf(CLPR_outfile, "X%d)\n", v1);
			else {
				CLPR_print_var2pos(CLPR_outfile, t);
				fprintf(CLPR_outfile, ")\n");
			}
			break;
		default: CLPR_fatal("53743");
	}
}

