
/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: dump.c											   *
 ***************************************************************/

#include <ctype.h>

#include "standard.h"
#include "config.h"
#include "parse.h"
#include "lib.h"
#include "compile.h"
#include "emul.h"
#include "solver.h"
#include "symbols.h"
#include "main.h"

#define MAX_FOURIER_VARS (9999)
#define MAX_NLIN_VARS (9999)

#define KOHLER_BITS_PER_NODE (160) /*** cf solver.h ***/
#define	KOHLER_WORDS_PER_NODE (5)

#define MAX_FLATTEN_VAR	(200)

#define print_first_newline() { \
	if (!something_printed) { \
		fprintf(CLPR_dumpF, "\n"); \
		something_printed = TRUE; \
	} \
}
#define newline() { \
	fprintf(CLPR_dumpF, "\n"); \
}

#define sign(eqn) \
	(!is_zero(eqn->constant) ? (eqn->constant < -CLPR_EPS ? -1 : 1) : \
		(eqn->params->epsilon < -CLPR_EPS ? -1 : 1)) 
		
#define is_less(eqn1, eqn2) \
	(eqn1->constant < eqn2->constant || \
		(is_equal(eqn1->constant, eqn2->constant) && \
			eqn1->params->epsilon < eqn2->params->epsilon))

#define is_lesseq(eqn1, eqn2) \
	(eqn1->constant < eqn2->constant || \
		(is_equal(eqn1->constant, eqn2->constant) && \
			eqn1->params->epsilon <= eqn2->params->epsilon))

static CLPR_int is_dump_output;
static CLPR_int dump_mode;

static struct {
	CLPR_int index;
	CLPR_int *heap;
	char *str;
} equiv_funct_vars[MAX_DUMP_VAR];

static struct {
	CLPR_int id; 
	CLPR_int *heap;
	char *str;
} equiv_pivot_vars[MAX_DUMP_VAR];

static struct {
	CLPR_int *rep; 
	CLPR_int *bind;
	CLPR_int *heap;
	char *str;
	CLPR_int mark;
	CLPR_int cyclic;
	NL_EQN_ptr nlin_subs;
} funct_vars[MAX_DUMP_VAR];

static struct {
	CLPR_int *rep;
	CLPR_int *var, *new_var;
} flatten_vars[MAX_FLATTEN_VAR];


/***
	lin_subs(X): X = exp, X non-orig;
	nlin_subs(X): X = nlin exp, X = non-orig, X non lin subs, X non-cyclic;
	cyclic(X): X is nlin_subs, and bound to X in nlins;
	arith_looking(X): X appears in arith format, or is a lin/nlin arith subs;
	non_arith_subs(X): USED subs (X = Y if lin, X = eval(Y) if nlin);
***/

static struct {
	CLPR_int id;
	CLPR_int *heap;
	char *str;
	char lin_output;
	CLPR_int mark;
	CLPR_int printed;
	CLPR_int arith_looking; 
	EQN_ptr lin_subs; 
	NL_EQN_ptr nlin_subs;
	CLPR_int cyclic;
	CLPR_int non_arith_subs;
	CLPR_int real_class;
} pivot_vars[MAX_DUMP_VAR];

static struct { 
	CLPR_int *loc; 
	double num; 
} num_vec[MAX_PROJ_NUM];

static CLPR_int equiv_fvars_top;
static CLPR_int equiv_pvars_top;
CLPR_int CLPR_funct_vars_top;	/* export to assert(), lib.c */
CLPR_int CLPR_flatten_vars_top; 
static CLPR_int high_funct_vars_top;
static CLPR_int orig_funct_vars_top;
static CLPR_int pivot_vars_top;
static CLPR_int high_pivot_vars_top;
static CLPR_int orig_pivot_vars_top;
static CLPR_int num_vec_top;

static struct {
	CLPR_int id, pos, neg, elim;
} fourier[MAX_FOURIER_VARS + 1];
static CLPR_int f_top = 0;

static CLPR_int old_trtop, old_slack_id;
static CLPR_int overflow_warning_given;
static CLPR_int something_printed;

static CLPR_int n_origs, n_kohler_nodes;
static CLPR_int fourier_iteration;
static CLPR_int f_count1, f_count2, f_count3, f_count4, f_count5;

static CLPR_int *make_funct_var(CLPR_int *t);
static CLPR_int *make_eqn(CLPR_int *t1, CLPR_int *t2);
static CLPR_int *make_linear_form(ENTRY_ptr entry);
static CLPR_int *make_ineq_eqn(EQN_ptr eqn);
static CLPR_int *make_nl_eqn(NL_EQN_ptr nl_eqn);
static CLPR_int *make_lhs_verb_nl_eqn(NL_EQN_ptr nl_eqn);
static CLPR_int *make_nl_rhs(NL_EQN_ptr nl_eqn);
static CLPR_int *make_arith_var(CLPR_int i);
static CLPR_int *make_num(double n);
static CLPR_int *make_str(CLPR_int n, CLPR_int *t0, CLPR_int in_arith);
static CLPR_int *make_funct_var_var(CLPR_int i, CLPR_int *t);
static CLPR_int *make_term(CLPR_int *t, CLPR_int in_arith);
static CLPR_int *make_func(CLPR_int val, CLPR_int *t1, CLPR_int *t2);
static CLPR_int *make_func1(CLPR_int val, CLPR_int *t1);
static CLPR_int *make_neg(CLPR_int *t1);
static CLPR_int *make_eval_lhs(NL_EQN_ptr nl_eqn);
static CLPR_int *make_pivot_var(CLPR_int i);
static CLPR_int *collect_vars(CLPR_int *vars, char *names, CLPR_int dump1_or_2_or_3);
static CLPR_int collect_named_vars(CLPR_int *vars, CLPR_int *names);
static CLPR_int **collect_atom_vars(CLPR_int *term);
static void collect_assert_vars(CLPR_int *term);
static CLPR_int find_flatten_var(CLPR_int *t);
static void make_equiv_funct_vars(void);
static void make_equiv_pivot_vars(void);
static void make_funct_eqns(void);
static void add_flatten_eqns(void);
static void add_linear_eqns(void);
static void add_constraint(CLPR_int *e);
static void intern_numbers(void);
static void add_ineq_eqns(void);

static CLPR_int *constraint_list, *first_constraint_list;

CLPR_int CLPR_real_val;

CLPR_int *CLPR_make_cons(CLPR_int *t1, CLPR_int *t2);
double CLPR_delete_slack(EQN_ptr eqn);
KOHLER_ptr CLPR_kohler_merge(KOHLER_ptr koh1, KOHLER_ptr koh2);
INTptr CLPR_collect_dump_vars(CLPR_int *vars, char *names);

extern FUNCTREC CLPR_functors[];
extern CLPR_unsigned *CLPR_code;
extern NL_EQN_ptr CLPR_nl_eqn_top;
extern CLPR_int CLPR_trtop, *CLPR_heaptop, **CLPR_trail, CLPR_TRAIL_SZ, CLPR_MTRAIL;

extern CLPR_int *CLPR_findbind(CLPR_int *x);
extern EQN_ptr CLPR_copy_eqn(EQN_ptr eqn);
extern EQN_ptr CLPR_sum_eqns(EQN_ptr eqn1, EQN_ptr eqn2, double factor);
extern double CLPR_find_coef(EQN_ptr eqn, CLPR_int v);
extern double CLPR_delete_entry(CLPR_int type, EQN_ptr eqn, CLPR_int v);
extern double CLPR_tmp_delete_entry(EQN_ptr eqn, CLPR_int v);

/*---------------------------------------------------------------------------*/


FILE *CLPR_dumpF;


void CLPR_init_dump (void);
void CLPR_prepare_linear_eqns (void);
void CLPR_prepare_ineq_eqns (void);
void CLPR_prepare_nlin_eqns (void);
void CLPR_dump_equiv_funct_vars (void);
void CLPR_dump_funct_eqns (void);
void CLPR_dump_equiv_pivot_vars (void);
void CLPR_dump_linear_eqns (void);
void CLPR_dump_ineq_eqns (void);
void CLPR_dump_nlin_eqns (void);
void CLPR_dump_remaining_real_vars (void);
void CLPR_cleanup_dump (void);
void CLPR_add_nlin_eqns (void);
void CLPR_add_remaining_real_vars (void);
extern CLPR_int CLPR_unify (CLPR_int *t1, CLPR_int *t2);
CLPR_int CLPR_num_bits (CLPR_unsigned x);
extern void CLPR_restore_null_ineq (CLPR_int *i);
void CLPR_pop_dump_trail (CLPR_int i);
void CLPR_collect_deep_vars (CLPR_int *term);
void CLPR_collect_nlin_vars (void);
CLPR_int CLPR_is_funct_var (CLPR_int *t);
void CLPR_add_equiv_funct_var (CLPR_int v, CLPR_int *h, char *s);
void CLPR_add_funct_var (CLPR_int *t1, CLPR_int *t2, CLPR_int *h, char *s);
CLPR_int CLPR_is_pivot_var (CLPR_int v);
void CLPR_add_equiv_pivot_var (CLPR_int v, CLPR_int *h, char *s);
void CLPR_add_pivot_var (CLPR_int v, CLPR_int *h, char *s);
extern CLPR_int CLPR_is_ground (CLPR_int v, double *c_ptr);
void CLPR_get_eval_vars (CLPR_int *t);
extern void CLPR_printterm (FILE *f, CLPR_int *t);
void CLPR_push_dump_trail (CLPR_int type, CLPR_int v, EQN_ptr eqn);
extern void CLPR_add_eqns (CLPR_int type, EQN_ptr eqn1, EQN_ptr eqn2, double factor);
extern void CLPR_add_simple_entry (CLPR_int type, EQN_ptr eqn, CLPR_int v, double c);
extern void CLPR_tmp_undelete_entry (EQN_ptr eqn);
CLPR_int CLPR_find_slack (EQN *eqn, double *c_ptr);
extern void CLPR_delete_eqn (EQN_ptr eqn, CLPR_int type);
extern void CLPR_mult_eqn (CLPR_int type, EQN_ptr eqn, double factor);
void CLPR_inc_f_neg (CLPR_int id);
void CLPR_inc_f_pos (CLPR_int id);
void CLPR_kohler_setup (void);
void CLPR_elim_quasi (void);
void CLPR_fourier_motzkin (void);
void CLPR_normalise_ineq (void);
void CLPR_free_kohler_nodes (KOHLER_ptr koh);
CLPR_int CLPR_check_cyclic (CLPR_int v0, CLPR_int v1);
CLPR_int CLPR_check_eval_cyclic (CLPR_int *v0, CLPR_int *v1);
CLPR_int CLPR_quasi_similar (EQN_ptr eqn1, EQN_ptr eqn2);
CLPR_int CLPR_next_fourier_id (void);
CLPR_int CLPR_kohler_merge_count (KOHLER_ptr koh1, KOHLER_ptr koh2);
void CLPR_dec_f_neg (CLPR_int id);
void CLPR_dec_f_pos (CLPR_int id);
extern void CLPR_delete_occ (CLPR_int type, CLPR_int i, CLPR_int v);
extern CLPR_int CLPR_free_eqn_nodes (EQN_ptr eqn);
extern void CLPR_insert_occ (CLPR_int type, CLPR_int i, CLPR_int v);
CLPR_int CLPR_f_metric (CLPR_int i);
void CLPR_dump_funct_var_var (CLPR_int i, CLPR_int *t);
void CLPR_output_funct_var (CLPR_int *t);
void CLPR_output_term (CLPR_int *t);
void CLPR_output_var_str (char *s);
void CLPR_output_number (double c);
void CLPR_output_arith_var (CLPR_int i, CLPR_int brackets);
void CLPR_mark_eqn_arith_looking (EQN_ptr eqn);
void CLPR_output_verbatim_arith_var (CLPR_int i);
void CLPR_output_linear_form (ENTRY_ptr entry);
void CLPR_mark_ineq_arith_looking (EQN_ptr eqn);
void CLPR_output_ineq_eqn (EQN_ptr eqn);
void CLPR_output_nl_eqn (NL_EQN_ptr nl_eqn);
void CLPR_mark_nlin_arith_looking (NL_EQN_ptr nl_eqn);
CLPR_int CLPR_nl_has_been_subs (CLPR_int x);
void CLPR_output_lhs_verb_nl_eqn (NL_EQN_ptr nl_eqn);
CLPR_int CLPR_check_non_arith_subs (EQN_ptr eqn);
void CLPR_output_nl_rhs (NL_EQN_ptr nl_eqn);
extern void CLPR_print_var2pos (FILE *f, CLPR_int *v);
void CLPR_output_eval_lhs (NL_EQN_ptr nl_eqn);
void CLPR_output_printed_term (CLPR_int *t, CLPR_int priority);
extern CLPR_int CLPR_get_op_priority (char *name, CLPR_int narg, CLPR_int *left, CLPR_int *right);
extern void CLPR_restore_null_eqn (CLPR_int *i);
extern void CLPR_restore_eqn (CLPR_int *i);
extern void CLPR_restore_ineq (CLPR_int *i);
extern void CLPR_print_eqn (EQN_ptr eqn);
void CLPR_special_print_eqn (EQN_ptr eqn);
extern CLPR_int CLPR_makenum (CLPR_int v, double c);

static void expand (EQN_ptr eqn1, EQN_ptr eqn2, double factor);
static void contract (EQN_ptr eqn);

void CLPR_dump1(FILE *f, CLPR_int *vars, char *names)
{
	is_dump_output = TRUE;
	CLPR_dumpF = f;
	CLPR_init_dump();
	collect_vars(vars, names, 1); /*** conservative estimate of pivot vars ***/
/***
CLPR_print_collected_vars();
CLPR_print_tableaus();
CLPR_print_nlins();
***/
	CLPR_prepare_linear_eqns();
/***
CLPR_print_eq_tableau();
***/
	CLPR_prepare_ineq_eqns();
	CLPR_prepare_nlin_eqns();
	CLPR_dump_equiv_funct_vars();
	CLPR_dump_funct_eqns();
	CLPR_dump_equiv_pivot_vars();
	CLPR_dump_linear_eqns();
	CLPR_dump_ineq_eqns();
	CLPR_dump_nlin_eqns();
	CLPR_dump_remaining_real_vars();
	CLPR_cleanup_dump();
}

CLPR_int CLPR_dump2(FILE *f, CLPR_int *vars, CLPR_int *names)
{
	vars = CLPR_findbind(vars);
	names = CLPR_findbind(names);
	if (!is_cons(vars) && !is_nil(vars))
		CLPR_runtime_error("bad 1st argument to dump/2");
	if (!is_cons(names) && !is_nil(names))
		CLPR_runtime_error("bad 2nd argument to dump/2");
	is_dump_output = TRUE;
	CLPR_dumpF = f;
	CLPR_init_dump();
	if (!collect_vars(vars, (char *)names, 2)) {
	  CLPR_cleanup_dump();
	  return FALSE; 
	}
	CLPR_prepare_linear_eqns();
	CLPR_prepare_ineq_eqns();
	CLPR_prepare_nlin_eqns();
	CLPR_dump_equiv_funct_vars();
	CLPR_dump_funct_eqns();
	CLPR_dump_equiv_pivot_vars();
	CLPR_dump_linear_eqns();
	CLPR_dump_ineq_eqns();
	CLPR_dump_nlin_eqns();
	CLPR_dump_remaining_real_vars();
	CLPR_cleanup_dump();
	return TRUE;
}

CLPR_int CLPR_dump3(CLPR_int *vars, CLPR_int *user_vars, CLPR_int *into)
{
CLPR_int i;
CLPR_int *new_proj_vars;
CLPR_int *tmp;
	is_dump_output = FALSE;
	dump_mode = TRUE;
	CLPR_dumpF = stdout;
	into = CLPR_findbind(into);
	vars = CLPR_findbind(vars);
	user_vars = CLPR_findbind(user_vars);
	if (!is_unbound_var(*into) && !is_daemon(into) &&
			!is_cons(into) && !is_nil(into))
		CLPR_runtime_error("bad 2nd argument to dump/3");
	if (!is_unbound_var(*user_vars) && !is_daemon(into) &&
			 !is_cons(user_vars) && !is_nil(user_vars))
		CLPR_runtime_error("bad 3rd argument to dump/3");
	first_constraint_list = CLPR_heaptop++;
	constraint_list = first_constraint_list;
	*first_constraint_list = addtag(CLPR_TAG_NIL, EMPTYLIST);
	CLPR_init_dump();
	new_proj_vars = collect_vars(vars, NULL, 3);
	CLPR_prepare_linear_eqns();
	CLPR_prepare_ineq_eqns();
	CLPR_prepare_nlin_eqns();
	make_equiv_funct_vars();
	make_funct_eqns();
	make_equiv_pivot_vars();
	add_linear_eqns();
	add_ineq_eqns();
	CLPR_add_nlin_eqns();
	CLPR_add_remaining_real_vars();
	add_flatten_eqns();
	*constraint_list = addtag(CLPR_TAG_NIL, EMPTYLIST);
	tmp = first_constraint_list;
	CLPR_cleanup_dump(); /*** hereafter no undo'ing ***/

	if (!CLPR_unify((CLPR_int *)&tmp, into)) return FALSE;
	intern_numbers();
	if (!CLPR_unify((CLPR_int *)&new_proj_vars, user_vars)) return FALSE; 
	return TRUE;
}

CLPR_int *CLPR_get_rule_constraints(CLPR_int *term) /*** for assert/1 only ***/
          
{
CLPR_int i;
static CLPR_int *tmp;
	is_dump_output = FALSE;
	dump_mode = FALSE;	
	term = CLPR_findbind(term);
	first_constraint_list = CLPR_heaptop++;
	constraint_list = first_constraint_list;
	*first_constraint_list = addtag(CLPR_TAG_NIL, EMPTYLIST);
	CLPR_init_dump();
	collect_assert_vars(term);
	CLPR_prepare_linear_eqns();
	CLPR_prepare_ineq_eqns();
	CLPR_prepare_nlin_eqns();
	make_equiv_funct_vars();
	make_funct_eqns();
	make_equiv_pivot_vars();
	add_linear_eqns();
	add_ineq_eqns();
	CLPR_add_nlin_eqns();
	CLPR_add_remaining_real_vars();
	add_flatten_eqns();
	*constraint_list = addtag(CLPR_TAG_NIL, EMPTYLIST);

	/* NO cleanup at this point */
	/* CLPR_cleanup_dump();*/ /*** hereafter no undo'ing ***/ 
	intern_numbers();
	return first_constraint_list;
}

static char kohler_cnt_bits[256];
void CLPR_init_kohler(void)
{
CLPR_int i;
	for (i = 0; i < 256; i++) kohler_cnt_bits[i] = CLPR_num_bits(i);
}

CLPR_int CLPR_num_bits(CLPR_unsigned x)
{
CLPR_int cnt = 0;
	while (x) {
		if (x & 1) cnt++;
		x = x >> 1;
	}
	return cnt;
}

void CLPR_init_dump(void)
{
CLPR_int i;
	something_printed = FALSE;
	old_trtop = CLPR_trtop;
	CLPR_flatten_vars_top = equiv_fvars_top = equiv_pvars_top = 
		CLPR_funct_vars_top = orig_pivot_vars_top = 
			pivot_vars_top = num_vec_top = 0;
	old_slack_id = CLPR_slack_id;
	for (i = 1; i <= f_top; i++) 
		fourier[i].id = fourier[i].pos = fourier[i].neg = fourier[i].elim = 0; 
	f_top = 0;
}

void CLPR_cleanup_dump(void)
{
CLPR_int i, j;
TAB_HDR_NODE_ptr hdr;
	for (i = CLPR_slack_id; i < old_slack_id; i++) CLPR_restore_null_ineq((CLPR_int *)i);
	CLPR_slack_id = old_slack_id;
	i = CLPR_trtop;
	while (--i >= old_trtop) CLPR_pop_dump_trail(i);
	CLPR_trtop = old_trtop;
	for (i = 1; i <= pivot_vars_top; i++) {
		hdr = CLPR_tableau[j = pivot_vars[i].id];
		if (!(hdr->eqn || hdr->eq_occ || hdr->ineq_occ || hdr->nlin_occ)) {
			free_HDR(hdr);
			CLPR_tableau[j] = NULL;
		} else hdr->next = NULL;
	}
}

/*--------------------------------------------- collect certain vars --------*/

static CLPR_int *collect_vars(CLPR_int *vars, char *names, CLPR_int dump1_or_2_or_3)
{
CLPR_int i;
CLPR_int *new_vars;
	if (dump1_or_2_or_3 == 2) 
		new_vars = (CLPR_int *) collect_named_vars(vars, (CLPR_int *)names);
	else new_vars = CLPR_collect_dump_vars(vars, names);
	orig_funct_vars_top = CLPR_funct_vars_top;
	orig_pivot_vars_top = pivot_vars_top;
	for (i = 1; i <= CLPR_funct_vars_top; i++) CLPR_collect_deep_vars(funct_vars[i].rep);
	/* CLPR_exp_collect_nlin_vars();*/ /* TEST only */
	CLPR_collect_nlin_vars();
	return new_vars;
}

static void collect_assert_vars(CLPR_int *term)
{
	CLPR_collect_deep_vars(term);
	orig_funct_vars_top = CLPR_funct_vars_top;
	orig_pivot_vars_top = pivot_vars_top;
	CLPR_collect_nlin_vars();
}

INTptr CLPR_collect_dump_vars(CLPR_int *vars, char *names)
{
CLPR_int i, j, *listp, *rep, *orig;
CLPR_int *first, *prev_list;
char *namep, *varname;
	namep = names;
	first = (INTptr) NULL;
	prev_list = (INTptr) &first;
	for (listp = CLPR_findbind(vars); CLPR_tag(*listp) != CLPR_TAG_NIL; 
		listp = CLPR_findbind(listp)) {
		varname = namep;
		if (is_dump_output) for (; *namep++ != DUMP_SEPARATOR; ) ;
		if (CLPR_tag(*listp) == CLPR_TAG_CONS) {
			if (!is_dump_output) {
				*prev_list = addtag(CLPR_TAG_CONS, CLPR_heaptop);
				make_unbound_var(CLPR_heaptop);
			}
			orig = (CLPR_int *) pmask(*listp);
			rep = CLPR_findbind(orig);
			if (CLPR_tag(*rep) != TAG_PAR) { 
				if (i = CLPR_is_funct_var(rep)) {
					CLPR_add_equiv_funct_var(
						i, funct_vars[i].heap, funct_vars[i].str);
					funct_vars[i].heap = CLPR_heaptop;
					funct_vars[i].str = varname;
				} else 
					CLPR_add_funct_var(orig, rep, CLPR_heaptop, varname);
			} else {
				if (i = CLPR_is_pivot_var(j = mask(*rep))) {
					CLPR_add_equiv_pivot_var(
						j, pivot_vars[i].heap, pivot_vars[i].str);
					pivot_vars[i].heap = CLPR_heaptop;
					pivot_vars[i].str = varname;
				} else {
					CLPR_add_pivot_var(mask(*rep), CLPR_heaptop, varname);
				}
			}
			if (!is_dump_output) {
				prev_list = ++CLPR_heaptop;
				CLPR_heaptop++;
			}
			listp = (CLPR_int *) pmask(*listp) + 1;
		} else 
			CLPR_fatal("dump requires a first-argument list");
	}
	if (!is_dump_output) *prev_list = addtag(CLPR_TAG_NIL, EMPTYLIST);
	return first;
}

static CLPR_int **collect_atom_vars(CLPR_int *term)
{
CLPR_int i, j, *rep, *orig;
CLPR_int *first, *prev_list;
CLPR_int n, k, **vars;

	n = arity_of(mask(*term));
	vars = (CLPR_int **) CLPR_heaptop;
	CLPR_heaptop += n;
	for (k=1; k <= n; k++) {
		make_unbound_var(CLPR_heaptop);
		vars[k-1] = CLPR_heaptop;
		orig = term + k;
		rep = CLPR_findbind(orig);
		if (is_unbound_var(*rep) || !is_param(rep)) { 
			if (i = CLPR_is_funct_var(rep)) {
				CLPR_add_equiv_funct_var(
					i, funct_vars[i].heap, "");
				funct_vars[i].heap = CLPR_heaptop;
				funct_vars[i].str = "";
			} else 	CLPR_add_funct_var(orig, rep, CLPR_heaptop, "");
		} else {
			if (i = CLPR_is_pivot_var(j = mask(*rep))) {
				CLPR_add_equiv_pivot_var(
					j, pivot_vars[i].heap, "");
				pivot_vars[i].heap = CLPR_heaptop;
				pivot_vars[i].str = "";
			} else CLPR_add_pivot_var(mask(*rep), CLPR_heaptop, "");
		}
		CLPR_heaptop++;
	}
	return vars;
}

CLPR_int collect_named_vars(CLPR_int *vars, CLPR_int *names)
{
CLPR_int i, j, *listp, *rep, *orig;
char *varname;
CLPR_int *varterm;
	for (listp = CLPR_findbind(vars); CLPR_tag(*listp) != CLPR_TAG_NIL; 
		listp = CLPR_findbind(listp)) {
		if (is_cons(listp)) {
			if (!is_cons(names)) return FALSE;
			orig = (CLPR_int *) pmask(*listp);
			rep = CLPR_findbind(orig);
			varterm = CLPR_findbind((CLPR_int *) pmask(*names));
			if (!is_con(varterm)) return FALSE;
			varname = string_of(mask(*varterm));
			if (CLPR_tag(*rep) != TAG_PAR) { 
				if (i = CLPR_is_funct_var(rep)) {
					CLPR_add_equiv_funct_var(
						i, funct_vars[i].heap, funct_vars[i].str);
					funct_vars[i].heap = CLPR_heaptop;
					funct_vars[i].str = varname;
				} else 
					CLPR_add_funct_var(orig, rep, CLPR_heaptop, varname);
			} else {
				if (i = CLPR_is_pivot_var(j = mask(*rep))) {
					CLPR_add_equiv_pivot_var(
						j, pivot_vars[i].heap, pivot_vars[i].str);
					pivot_vars[i].heap = CLPR_heaptop;
					pivot_vars[i].str = varname;
				} else {
					CLPR_add_pivot_var(mask(*rep), CLPR_heaptop, varname);
				}
			}
			listp = (CLPR_int *) pmask(*listp) + 1;
			names = CLPR_findbind((CLPR_int *) pmask(*names) + 1);
		} else return FALSE;
	}
	if (is_nil(names)) return TRUE;
	else return FALSE;
}

void CLPR_collect_deep_vars(CLPR_int *term)
{
CLPR_int i, n;
char *s;
double c;
EQN_ptr eqn;
	CLPR_deref(term);
	if (is_unbound_var(*term) || is_daemon(term)) {
		if (CLPR_is_funct_var(term)) return;
		if (is_dump_output) CLPR_add_funct_var(term, term, NULL, NULL);
		else {
			make_unbound_var(CLPR_heaptop);
			CLPR_add_funct_var(term, term, CLPR_heaptop++, NULL);
		}
	} else	
		switch ((CLPR_unsigned) CLPR_tag(*term)) {
		case TAG_CON:
		case CLPR_TAG_NIL:
			return;
		case TAG_PAR:
			i = mask(*term);
			if (CLPR_is_ground(i, &c) || CLPR_is_pivot_var(i)) return;
			if (is_dump_output) CLPR_add_pivot_var(mask(*term), NULL, NULL);
			else {
				make_unbound_var(CLPR_heaptop);
				CLPR_add_pivot_var(mask(*term), CLPR_heaptop++, NULL);
			}
			break;
		case CLPR_TAG_CONS:
			term = (CLPR_int *) pmask(*term);
			CLPR_collect_deep_vars(term);
			CLPR_collect_deep_vars(++term);
			return;
		case TAG_STR:
			n = arity_of(*term);
			for (i = 1; i <= n; i++) CLPR_collect_deep_vars(++term);
			return;
		default:
			CLPR_fatal("34541");
	}
}

void CLPR_collect_nlin_vars(void) /*** all nlin vars are pivot vars ***/
{
CLPR_int v0, v1, v2, *t0, *t1, *h;
NL_EQN_ptr nl_eqn;
	for (nl_eqn = CLPR_nl_eqn_top; nl_eqn; nl_eqn = nl_eqn->next) {  
		if (nl_eqn->type == WOKEN || nl_eqn->partner) continue;
		if (nl_eqn->type == SOLVER_EVAL) {
			t0 =  (CLPR_int *) nl_eqn->params1->id0;
			t1 =  (CLPR_int *) nl_eqn->params1->id1;
			CLPR_get_eval_vars(t0);
			CLPR_deref(t1);
			if (!CLPR_is_funct_var(t1)) {
				if (is_dump_output) CLPR_add_funct_var(t1, t1, NULL, NULL);
				else {
					make_unbound_var(CLPR_heaptop);
					CLPR_add_funct_var(t1, t1, CLPR_heaptop++, NULL);
				}
			}
		} else {
			if ((v0 = nl_eqn->params1->id0) && !CLPR_is_pivot_var(v0)) {
				if (!is_dump_output) {
					h = CLPR_heaptop++;
					make_unbound_var(h);
				} else h = NULL;					
				CLPR_add_pivot_var(v0, h, NULL);
			}
			if ((v1 = nl_eqn->params1->id1) && !CLPR_is_pivot_var(v1)) {
				if (!is_dump_output) {
					h = CLPR_heaptop++;
					make_unbound_var(h);
				} else h = NULL;					
				CLPR_add_pivot_var(v1, h, NULL);
			}
			if ((v2 = nl_eqn->params1->id2) && !CLPR_is_pivot_var(v2)) {
				if (!is_dump_output) {
					h = CLPR_heaptop++;
					make_unbound_var(h);
				} else h = NULL;					
				CLPR_add_pivot_var(v2, h, NULL);
			}
		}			
	}
}

void CLPR_exp_collect_nlin_vars(void) /*** all nlin vars are pivot vars ***/
{
CLPR_int v0, v1, v2, *t0, *t1, *h;
NL_EQN_ptr nl_eqn;
	for (nl_eqn = CLPR_nl_eqn_top; nl_eqn; nl_eqn = nl_eqn->next) {  
		if (nl_eqn->type == WOKEN || nl_eqn->partner) continue;
		if (nl_eqn->type == SOLVER_EVAL) {
			t0 =  (CLPR_int *) nl_eqn->params1->id0;
			t1 =  (CLPR_int *) nl_eqn->params1->id1;
			CLPR_get_eval_vars(t0);
			CLPR_deref(t1);
			if (!CLPR_is_funct_var(t1)) {
				if (is_dump_output) CLPR_add_funct_var(t1, t1, NULL, NULL);
				else {
					make_unbound_var(CLPR_heaptop);
					CLPR_add_funct_var(t1, t1, CLPR_heaptop++, NULL);
				}
			}
		} else {
			if ((v0 = nl_eqn->params1->id0) && !CLPR_is_pivot_var(v0)) {
				if (!is_dump_output) {
					h = CLPR_heaptop++;
					make_unbound_var(h);
				} else h = NULL;					
				CLPR_add_pivot_var(v0, h, NULL);
			}
		}
	}
	for (nl_eqn = CLPR_nl_eqn_top; nl_eqn; nl_eqn = nl_eqn->next) {  
		if (nl_eqn->type == SOLVER_EVAL) continue;
			if ((v1 = nl_eqn->params1->id1) && !CLPR_is_pivot_var(v1)) {
				if (!is_dump_output) {
					h = CLPR_heaptop++;
					make_unbound_var(h);
				} else h = NULL;					
				CLPR_add_pivot_var(v1, h, NULL);
			}
			if ((v2 = nl_eqn->params1->id2) && !CLPR_is_pivot_var(v2)) {
				if (!is_dump_output) {
					h = CLPR_heaptop++;
					make_unbound_var(h);
				} else h = NULL;					
				CLPR_add_pivot_var(v2, h, NULL);
			}
	}
}

void CLPR_get_eval_vars(CLPR_int *t)
{
CLPR_int i, n;
double c;
	CLPR_deref(t);
	if (is_unbound_var(*t)) {
		if (!CLPR_is_funct_var(t)) {
			if (is_dump_output) CLPR_add_funct_var(t, t, NULL, NULL);
			else {
				make_unbound_var(CLPR_heaptop);
				CLPR_add_funct_var(t, t, CLPR_heaptop++, NULL);
			}
		}
	} else {
		switch ((CLPR_unsigned) CLPR_tag(*t)) {
			case TAG_CON:
			case CLPR_TAG_NIL:
			case TAG_DAEMON: return;
			case CLPR_TAG_CONS:
				t = (CLPR_int *) pmask(*t);
				CLPR_get_eval_vars(t);
				CLPR_get_eval_vars(t + 1);
				return;
			case TAG_STR:
				n = arity_of(*t);
				for (i = 1; i <= n; i++) CLPR_get_eval_vars(t + i);
				return;
			case TAG_PAR:
				i = mask(*t);
				if (CLPR_is_ground(i, &c) || CLPR_is_pivot_var(i)) return;
				if (is_dump_output) CLPR_add_pivot_var(i, NULL, NULL);
				else {
					make_unbound_var(CLPR_heaptop);
					CLPR_add_pivot_var(i, CLPR_heaptop++, NULL);
				}
				break;
			default: CLPR_fatal("34545");
		}
	}
}

void CLPR_add_funct_var(CLPR_int *t1, CLPR_int *t2, CLPR_int *h, char *s) /*** no duplicate check ***/
                 
        
{
	if (CLPR_funct_vars_top >= MAX_DUMP_VAR - 1) 
		CLPR_fatal("dump overflow (funct vars)");
	funct_vars[++CLPR_funct_vars_top].bind = t1;
	funct_vars[CLPR_funct_vars_top].rep = t2;
	funct_vars[CLPR_funct_vars_top].heap = h;
	funct_vars[CLPR_funct_vars_top].str = s;
	funct_vars[CLPR_funct_vars_top].mark = 0;
}

static CLPR_int find_flatten_var(CLPR_int *t)
{
CLPR_int i;
double dummy;
	t = CLPR_findbind(t);
	if (is_cons(t) || (is_func(t) && !is_val_hat(mask(*t)))
			|| is_con(t) || is_nil(t) || is_num(t, &dummy)) /* TODO */
		return 0;
	else {	/* ^+, var, daemon */
		for (i = 1; i <= CLPR_flatten_vars_top; i++) 
			if (t == flatten_vars[i].rep) return i;
		if (++CLPR_flatten_vars_top >= MAX_FLATTEN_VAR) CLPR_fatal("55712");
		if (is_unbound_var(*t) || is_daemon(t)) 
			flatten_vars[CLPR_flatten_vars_top].var = make_funct_var(t);
		else if (is_param(t))
			flatten_vars[CLPR_flatten_vars_top].var = make_pivot_var(mask(*t));
		else {
			flatten_vars[CLPR_flatten_vars_top].var = CLPR_heaptop;
			make_unbound_var(CLPR_heaptop++);
		}
		flatten_vars[CLPR_flatten_vars_top].rep = t;
		make_unbound_var(CLPR_heaptop);
		flatten_vars[CLPR_flatten_vars_top].new_var = CLPR_heaptop++;
		return CLPR_flatten_vars_top;
	}
}

CLPR_int *CLPR_flatten_var_term(CLPR_int *t)
{
CLPR_int i;
	if (!(i = find_flatten_var(t))) return t;
	else return flatten_vars[i].new_var;
}

void CLPR_print_flatten_vars(void)
{
CLPR_int i;
	for (i= 1; i <= CLPR_flatten_vars_top; i++) {
		printf("t = "); CLPR_printterm(stdout, flatten_vars[i].rep);
		printf(" var = %p",flatten_vars[i].var); 
/*		printf(" var = "); CLPR_printterm(stdout, flatten_vars[i].var);  */
		printf(" new_var = "); CLPR_printterm(stdout, flatten_vars[i].new_var);
		printf("\n"); 
	}
}

void CLPR_add_pivot_var(CLPR_int v, CLPR_int *h, char *s) /*** duplicate check ***/
          
        
{
CLPR_int i;
TAB_HDR_NODE_ptr hdr;
	if (!(hdr = CLPR_tableau[v])) {
		malloc_HDR(hdr);
		CLPR_tableau[v] = hdr;
	} else if (hdr->next) {
		if (pivot_vars[(CLPR_int) hdr->next].id != v) {
			CLPR_fatal("66621");
		} 
		return;
	}
	if (pivot_vars_top >= MAX_DUMP_VAR - 1) 
		CLPR_fatal("dump overflow (pivot vars)");
	pivot_vars[++pivot_vars_top].id = v;
	pivot_vars[pivot_vars_top].heap = h;
	pivot_vars[pivot_vars_top].str = s;
	pivot_vars[pivot_vars_top].mark = 0;
	pivot_vars[pivot_vars_top].printed = FALSE;
	pivot_vars[pivot_vars_top].arith_looking = FALSE;
	pivot_vars[pivot_vars_top].non_arith_subs = FALSE;
	pivot_vars[pivot_vars_top].real_class = 0;
	hdr->next = (TAB_HDR_NODE_ptr) pivot_vars_top;
}

void CLPR_add_equiv_funct_var(CLPR_int v, CLPR_int *h, char *s) /*** no duplicate check ***/
          
        
{
	if (equiv_fvars_top >= MAX_DUMP_VAR - 1) 
		CLPR_fatal("dump overflow (equivalent funct vars)");
	equiv_funct_vars[++equiv_fvars_top].index = v;
	equiv_funct_vars[equiv_fvars_top].heap = h;
	equiv_funct_vars[equiv_fvars_top].str = s;
}

void CLPR_add_equiv_pivot_var(CLPR_int v, CLPR_int *h, char *s) /*** no duplicate check ***/
          
        
{
	if (equiv_pvars_top >= MAX_DUMP_VAR - 1) 
		CLPR_fatal("dump overflow (equivalent pivot vars)");
	equiv_pivot_vars[++equiv_pvars_top].id = v;
	equiv_pivot_vars[equiv_pvars_top].heap = h;
	equiv_pivot_vars[equiv_pvars_top].str = s;
}

/*---------------------------------------------------------------------------*/

void CLPR_prepare_linear_eqns(void)
{ 
CLPR_int i, j, k, v, v0, v1, min_priority;
EQN_ptr eqn0, eqn1, eqn2;
ENTRY_ptr entry, entry2, found;
OCC_ptr occ, occ99;
double c, c1, c2;
TAB_HDR_NODE_ptr hdr0, hdr2;

	for (i = 1; i <= pivot_vars_top; i++) { /*** decreasing priority order ***/
		pivot_vars[i].lin_output = FALSE;
		pivot_vars[i].lin_subs = NULL;
		v0 = pivot_vars[i].id;
		if (!(hdr0 = CLPR_tableau[v0]) || !(eqn0 = hdr0->eqn)) continue;

		/*** find lowest pivot var or nonpivot var in rhs(v0) ***/
		entry = eqn0->next;
		min_priority = 0;
		for (found = NULL; entry; entry = entry->next) {
			j = CLPR_is_pivot_var(entry->id);
			if (!j) {
				min_priority = pivot_vars_top + 1;
				found = entry;
				break;
			} else if (j > i && j > min_priority) {
				min_priority = j;
				found = entry;
			}
		}
		if (!found) {
			pivot_vars[i].lin_output = (i <= orig_pivot_vars_top);
			if (i > orig_pivot_vars_top) pivot_vars[i].lin_subs = eqn0;
			continue;
		}
		pivot_vars[i].lin_output = FALSE;
		pivot_vars[i].lin_subs = NULL;
		v = found->id;
		c = CLPR_tmp_delete_entry(eqn0, v);
		hdr0->eqn = NULL;

		/*** subs in eq tableau ***/
		occ = (CLPR_tableau[v = found->id] ? CLPR_tableau[v]->eq_occ : NULL); 
		while (occ) {
			occ99 = occ->next;
			v1 = occ->id;
			if ((k = CLPR_is_pivot_var(v1)) && k > i) {
				eqn1 = CLPR_tableau[v1]->eqn;
				if (eqn1->time_stamp) {
					CLPR_push_dump_trail(BTK_EQN, v1, eqn1);
					CLPR_tableau[v1]->eqn = eqn1 = CLPR_copy_eqn(eqn1);
					eqn1->time_stamp = 0;
				}
				c1 = CLPR_delete_entry(EQ, eqn1, v);
				CLPR_add_eqns(EQ, eqn1, eqn0, -(c1 / c));
				CLPR_add_simple_entry(EQ, eqn1, v0, (c1 / c));
			}
			occ = occ99;
		}

		/*** subs in ineq tableau ***/
		occ = (CLPR_tableau[v] ? CLPR_tableau[v]->ineq_occ : NULL); 
		while (occ) {
			occ99 = occ->next;
			v1 = occ->id;
			eqn1 = CLPR_tableau[v1]->eqn;
			if (eqn1->time_stamp) {
				CLPR_push_dump_trail(BTK_INEQ, v1, eqn1);
				CLPR_tableau[v1]->eqn = eqn1 = CLPR_copy_eqn(eqn1);
				eqn1->time_stamp = 0;
			}
			c1 = CLPR_delete_entry(INEQ, eqn1, v);	
			CLPR_add_eqns(INEQ, eqn1, eqn0, -(c1 / c));
			CLPR_add_simple_entry(INEQ, eqn1, v0, (c1 / c));
			occ = occ99;
		}

		if (min_priority > pivot_vars_top) { /*** nonpivot var ***/
			CLPR_tmp_undelete_entry(eqn0);
			hdr0->eqn = eqn0;
			continue;
		}

		/*** lower priority dump var ***/
		CLPR_push_dump_trail(BTK_NULL_EQN, v, NULL); 
		malloc_EQN(eqn2);
		if (min_priority > orig_pivot_vars_top) 
			pivot_vars[min_priority].lin_subs = eqn2;
		eqn2->next = NULL;
		eqn2->time_stamp = 0;
		eqn2->id = v;
		eqn2->constant = 0.0;
		eqn2->params = NULL;
		entry2 = eqn0->next;
		CLPR_add_eqns(EQ, eqn2, eqn0, -(1.0 / c));
		CLPR_add_simple_entry(EQ, eqn2, v0, (1.0 / c));
		if (!(hdr2 = CLPR_tableau[v])) {
			malloc_HDR(hdr2);
			CLPR_tableau[v] = hdr2;
		}
		hdr2->eqn = eqn2;
		CLPR_tmp_undelete_entry(eqn0);
		hdr0->eqn = eqn0;
	}
}

void CLPR_prepare_ineq_eqns(void)
{
CLPR_int i, j;
EQN_ptr eqn, eqn_i, eqn_j;
ENTRY_ptr entry;
OCC_ptr occ, occ99;
double c, c_i, c_j, dummy;
CLPR_int s_i;
KOHLER_ptr koh;
	for (i = old_slack_id; i < CLPR_SOLVER_SZ; i++) {
		CLPR_push_dump_trail(BTK_INEQ, i, CLPR_tableau[i]->eqn);
		CLPR_tableau[i]->eqn = CLPR_copy_eqn(CLPR_tableau[i]->eqn);
	}
	for (i = old_slack_id; i < CLPR_SOLVER_SZ; i++) { /*** one slack per row ***/
		eqn_i = CLPR_tableau[i]->eqn;
		if (!(s_i = CLPR_find_slack(eqn_i, &c_i))) {
		  CLPR_delete_eqn(eqn_i, INEQ); 
		  continue;
		}
		occ = CLPR_tableau[s_i]->ineq_occ;
		if (!occ->next) continue;
		CLPR_tmp_delete_entry(eqn_i, s_i);
		for (; occ; occ = occ99) {
			occ99 = occ->next;
			if ((j = occ->id) == i) continue; 
			eqn_j = CLPR_tableau[j]->eqn;
			c_j = CLPR_delete_entry(INEQ, eqn_j, s_i);
			CLPR_add_eqns(INEQ, eqn_j, eqn_i, -(c_j / c_i));
		}
		CLPR_tmp_undelete_entry(eqn_i);
		CLPR_mult_eqn(INEQ, eqn_i, 1/(c_i));
	}

	/*** standardize: exp <= 0 ***/
	for (i = CLPR_SOLVER_SZ - 1; i >= old_slack_id; i--) {
		eqn = CLPR_tableau[i]->eqn;
		/*** may have row without slack **/
		if (!eqn->next) continue;
		c = CLPR_delete_slack(eqn);
		if (c < -CLPR_EPS) CLPR_mult_eqn(INEQ, eqn, -1.0);
		for (entry = eqn->next; entry; entry = entry->next) {
			if ((j = entry->id) >= CLPR_slack_id || CLPR_is_pivot_var(j)) continue;
			if (entry->coef < -CLPR_EPS) CLPR_inc_f_neg(j);
			else CLPR_inc_f_pos(j);	
		}
	}
	CLPR_kohler_setup();
	CLPR_elim_quasi();
	if (f_top) CLPR_fourier_motzkin();
	CLPR_elim_quasi();
	CLPR_normalise_ineq();
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--) 
		if (koh = CLPR_tableau[i]->eqn->params->kohler) {
			CLPR_free_kohler_nodes(koh);
			CLPR_tableau[i]->eqn->params->kohler = NULL;
		}
}

void CLPR_prepare_nlin_eqns(void) /*** set up nlin subs vars ***/
{
CLPR_int i;
NL_EQN *nl_eqn;
CLPR_int *t;
	for (i = 1; i <= pivot_vars_top; i++) {
		pivot_vars[i].nlin_subs = NULL;
		pivot_vars[i].cyclic = FALSE;
		pivot_vars[i].mark = 0;
	}
	for (i = 1; i <= CLPR_funct_vars_top; i++) {
		funct_vars[i].nlin_subs = NULL;
		funct_vars[i].cyclic = FALSE;
		funct_vars[i].mark = 0;
	}
	for (nl_eqn = CLPR_nl_eqn_top; nl_eqn; nl_eqn = nl_eqn->next) {
		if (nl_eqn->type == WOKEN || nl_eqn->partner) continue;
		if (nl_eqn->type != SOLVER_EVAL) 
			i = CLPR_is_pivot_var(nl_eqn->params1->id0);
		else {
			t = CLPR_findbind((CLPR_int *)nl_eqn->params1->id0);
			if (CLPR_tag(*t) == TAG_PAR) i = CLPR_is_pivot_var(mask(*t));
			else {
				if (!is_unbound_var(*t) && !is_daemon(t)) continue;
				if (!(i = CLPR_is_funct_var(t))) CLPR_fatal("66112");
				if (!funct_vars[i].nlin_subs && orig_funct_vars_top < i)
					funct_vars[i].nlin_subs = nl_eqn;
				continue;
			}
		}
		if (!pivot_vars[i].lin_subs && orig_pivot_vars_top < i) 
			if (!pivot_vars[i].nlin_subs) 
			  pivot_vars[i].nlin_subs = nl_eqn;
	}
	for (i = 1; i <= pivot_vars_top; i++) {
		if (pivot_vars[i].nlin_subs && 
		    CLPR_check_cyclic(pivot_vars[i].id, pivot_vars[i].id)) 
				pivot_vars[i].cyclic = TRUE;
	}
	for (i = 1; i <= CLPR_funct_vars_top; i++) {
		if (funct_vars[i].nlin_subs && 
		    CLPR_check_eval_cyclic(funct_vars[i].rep, funct_vars[i].rep))
				funct_vars[i].cyclic = TRUE;
	}
}

CLPR_int CLPR_check_cyclic(CLPR_int v0, CLPR_int v1) /*** is v0 "bound" to v1 ? ***/
           
{
CLPR_int i, w1, w2, *t;
EQN_ptr eqn;
ENTRY_ptr entry;
NL_EQN_ptr nl_eqn;
	if (!(i = CLPR_is_pivot_var(v0))) return FALSE;
	else if (pivot_vars[i].mark == v1) return FALSE;
	else if (eqn = pivot_vars[i].lin_subs) {
		for (entry = eqn->next; entry; entry = entry->next) {
			if (entry->id == v1) return TRUE;
			else {
				pivot_vars[i].mark = v1;
				if (CLPR_check_cyclic(entry->id, v1)) return TRUE;
			}
		}
		return FALSE;
	} else if (nl_eqn = pivot_vars[i].nlin_subs) {
		if (nl_eqn->type == SOLVER_EVAL) {
			t = (CLPR_int *) nl_eqn->params1->id1;
			CLPR_deref(t);
			return CLPR_check_eval_cyclic(t, (CLPR_int *)v1);
		}
		w1 = nl_eqn->params1->id1;
		if (w1 == v1) return TRUE;
		w2 = nl_eqn->params1->id2;
		if (w2 == v1) return TRUE;
		pivot_vars[i].mark = v1;
		if (CLPR_check_cyclic(w1, v1)) return TRUE;
		if (CLPR_check_cyclic(w2, v1)) return TRUE;
		return FALSE;
	} else return FALSE;
}

CLPR_int CLPR_check_eval_cyclic(CLPR_int *v0, CLPR_int *v1) /*** is v0 "bound" to v1 ? ***/
             
{
CLPR_int i, *t;
EQN_ptr eqn;
ENTRY_ptr entry;
NL_EQN_ptr nl_eqn;

	if (!(i = CLPR_is_funct_var(v0))) return FALSE;
	else if (funct_vars[i].mark == (CLPR_int) v1) return FALSE;
	if (nl_eqn = funct_vars[i].nlin_subs) {
		t = (CLPR_int *) nl_eqn->params1->id1;
		CLPR_deref(t);
		if (t == v1) return TRUE;
		funct_vars[i].mark = (CLPR_int) v1;
		if (CLPR_check_eval_cyclic(t, v1)) return TRUE;
		return FALSE;
	} else return FALSE;
}

void CLPR_elim_quasi(void)
{
CLPR_int i, j;
EQN_ptr eqn_i, eqn_j;
OCC_ptr occ, occ99;
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--) {
		eqn_i = CLPR_tableau[i]->eqn;
		if (eqn_i->time_stamp == -1) continue; 
		if (!eqn_i->next) continue;
		for (occ = CLPR_tableau[eqn_i->next->id]->ineq_occ; occ; occ = occ99) {
			occ99 = occ->next;
			eqn_j = CLPR_tableau[occ->id]->eqn;
			if (i == occ->id || eqn_j->time_stamp == -1) continue;
			switch (CLPR_quasi_similar(eqn_i, eqn_j)) {
				case 0: break;
				case 1: contract(eqn_j); break;
				case 2: contract(eqn_i); goto NEXT;
			}
		}
		NEXT: ;
	}
}

void CLPR_normalise_ineq(void)
{
CLPR_int i, j;
EQN_ptr eqn;
ENTRY_ptr entry;
double coef;
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--) {
		eqn = CLPR_tableau[i]->eqn;
		if (eqn->time_stamp == -1) continue; 
		if (!(entry = eqn->next)) continue;
		if (is_equal(fabs(entry->coef), 1.0)) continue;
		coef = fabs(entry->coef);
		eqn->constant = eqn->constant/coef;
		eqn->params->epsilon = eqn->params->epsilon/coef;
		for (; entry; entry = entry->next)
			entry->coef = entry->coef/coef;
	}
}

void CLPR_fourier_motzkin(void)
{
TAB_HDR_NODE_ptr hdr;
EQN_ptr eqn, eqn1, eqn2;
ENTRY_ptr entry;
OCC_ptr occ, occ2, occ88, occ99, occ_pos, occ_neg;
OCC first_pos_occ, first_neg_occ;
KOHLER_ptr koh1, koh2;
CLPR_int i, j, k, elimvar;
double c, c1, c2;

/***
	printf("Fourier: %d ineqs, %d elim vars\n", CLPR_count_all_ineq(), f_top);
***/
	for (i = 1; i <= f_top; i++) { 
		fourier_iteration++;
		k = CLPR_next_fourier_id();
		elimvar = fourier[k].id;
/***
printf("%d: elimvar X%d (%d +ve, %d -ve, metric %d), total %d ineqs\n", 
	i, elimvar, fourier[k].pos, fourier[k].neg, CLPR_f_metric(k), CLPR_count_all_ineq());
f_count1 = f_count2 = f_count3 = f_count4 = 0;
***/
		fourier[k].elim = TRUE;
		if (fourier[k].pos == 0 || fourier[k].neg == 0) {
			fourier_iteration--;
			if (hdr = CLPR_tableau[elimvar]) {
				for (occ = hdr->ineq_occ; occ; occ = occ99) {
					occ99 = occ->next;
					contract(CLPR_tableau[occ->id]->eqn);
				}
			}
			continue;
		}
		occ_pos = &first_pos_occ;
		occ_neg = &first_neg_occ;
		for (occ = CLPR_tableau[elimvar]->ineq_occ; occ; occ = occ99) {
			occ99 = occ->next;
			eqn = CLPR_tableau[occ->id]->eqn;
			eqn->time_stamp = -1;
			c = CLPR_delete_entry(INEQ, eqn, elimvar);
			malloc_OCC(occ2);
			occ2->next = NULL;
			occ2->id = eqn->id;
			occ2->coef = c;
			if (c > CLPR_EPS) {
				occ_pos->next = occ2;
				occ_pos = occ2; 
			} else {
				occ_neg->next = occ2;
				occ_neg = occ2;
			}
		}
		occ_pos = first_pos_occ.next;
		occ_neg = first_neg_occ.next;

		/*** expand! ***/
		for (occ = occ_pos; occ; occ = occ->next) {
			eqn1 = CLPR_tableau[occ->id]->eqn;
			koh1 = eqn1->params->kohler;
			for (occ2 = occ_neg; occ2; occ2 = occ2->next) {
				eqn2 = CLPR_tableau[occ2->id]->eqn;
				koh2 = eqn2->params->kohler;
				f_count1++;
				/* HACK */
				if (CLPR_kohler_merge_count(koh1, koh2) > fourier_iteration + 1) {
#ifdef STATS
					f_count2++;
					CLPR_fourier_stats();
#endif
				} else
					expand(eqn1, eqn2, -(occ->coef / occ2->coef));
			}
		}
		for (occ = occ_pos; occ; occ = occ99) {
			occ99 = occ->next;
			contract(CLPR_tableau[occ->id]->eqn);
			free_OCC(occ);
		}
		for (occ = occ_neg; occ; occ = occ99) {
			occ99 = occ->next;
			contract(CLPR_tableau[occ->id]->eqn);
			free_OCC(occ);
		}
	}
}

void CLPR_fourier_stats(void)
{
if (f_count1 && !(f_count1 % 1000)) {
printf("Expand attempt %d (Kohler'd %d, Expanded %d, Quasi'd %d k2ed %d)\n",
f_count1, f_count2, f_count3, f_count4, f_count5);
}
}

void CLPR_kohler_setup(void) /*** Node 1, bits 0 - 159, words 0 - 4, etc ***/
{
CLPR_int i, j, eqn_number, node, word;
EQN_ptr eqn;
KOHLER_NODE first;
KOHLER_ptr koh0, koh;
	n_origs = CLPR_SOLVER_SZ - old_slack_id; /*** numbered 0 .. n ***/
	i = n_origs / KOHLER_BITS_PER_NODE;
	n_kohler_nodes = (n_origs % KOHLER_BITS_PER_NODE ? i + 1 : i);
	fourier_iteration = 0;
	for (i = CLPR_SOLVER_SZ - 1; i >= old_slack_id; i--) {
		eqn_number = CLPR_SOLVER_SZ - i - 1;
		eqn = CLPR_tableau[i]->eqn;
		koh0 = &first;
		for (j = 1; j <= n_kohler_nodes; j++) { /*** setup kohler chain ***/
			malloc_KOHLER(koh);
			koh0->next = koh;
			koh0 = koh;
		}
		eqn->params->kohler = first.next;
		node = (eqn_number / KOHLER_BITS_PER_NODE) + 1;
		word = (eqn_number % KOHLER_BITS_PER_NODE) / (sizeof(CLPR_int) * 8);
		koh->bits[word] |= 
			(CLPR_unsigned) 1 << (eqn_number % (sizeof(CLPR_int) * 8));
	}
}

KOHLER_ptr CLPR_kohler_merge(KOHLER_ptr koh1, KOHLER_ptr koh2)
{
CLPR_int i, j;
KOHLER_NODE first;
KOHLER_ptr koh0, koh;
	koh0 = &first;
	for (j = 1; j <= n_kohler_nodes; j++) { 
		malloc_KOHLER(koh);
		koh0->next = koh;
		koh0 = koh;
	}
	koh0 = first.next;
	do {
		for (i = 0; i < KOHLER_WORDS_PER_NODE; i++)
			koh0->bits[i] = koh1->bits[i] | koh2->bits[i];
		koh0 = koh0->next;
		koh1 = koh1->next;
		koh2 = koh2->next;
	} while (koh0);
	return first.next;
}

CLPR_int CLPR_kohler_count(KOHLER_ptr koh)
{
CLPR_int i, j, count;
CLPR_unsigned m;

	count = 0;
	do {
		for (i = 0; i < KOHLER_WORDS_PER_NODE; i++) {
			m = koh->bits[i];
			for (j = 1; j <= 4; j++, m = m >> 8)
				count += kohler_cnt_bits[m & 255];
		}
		koh = koh->next; 
	} while (koh);
	return count;
}

CLPR_int CLPR_kohler_subset(EQN_ptr eqn1, EQN_ptr eqn2)
{
register CLPR_int i, j, k1,k2,k3;
CLPR_int sub1,sub2;
KOHLER_ptr koh1, koh2;

    sub1 = sub2 = 1;
    koh1 = eqn1->params->kohler;
    koh2 = eqn2->params->kohler;
    do {
        for (i = 0; i < KOHLER_WORDS_PER_NODE; i++) {
            if (sub1 || sub2) {
                k1 = koh1->bits[i];
                k2 = koh2->bits[i];
                k3 = k1 & k2;
                if (k1 != k3) sub1 = 0;
                if (k2 != k3) sub2 = 0;
            }
            if (!sub1 && !sub2) return 0;
        }
        koh1 = koh1->next;
        koh2 = koh2->next;
    } while (koh1 && koh2);
    return (sub1 ? 1 : 2);
}

CLPR_int CLPR_kohler_merge_count(KOHLER_ptr koh1, KOHLER_ptr koh2)
{
CLPR_int i, j, count;
CLPR_unsigned m;
	count = 0;
	do {
		for (i = 0; i < KOHLER_WORDS_PER_NODE; i++) {
			m = (koh1->bits[i] | koh2->bits[i]);
			for (j = 1; j <= 4; j++, m = m >> 8)
				count += kohler_cnt_bits[m & 255];
		}
		koh1 = koh1->next; 
		koh2 = koh2->next; 
	} while (koh1);
	return count;
}

void contract(EQN_ptr eqn) /*** do not free EQN's ***/
            
{
CLPR_int i, j;
ENTRY_ptr entry, entry99;
EQN_ptr eqn_i;

	CLPR_free_kohler_nodes(eqn->params->kohler);
	i = eqn->id;
	for (entry = eqn->next; entry; entry = entry99) {
		entry99 = entry->next;
		j = entry->id;
		if (entry->coef < -CLPR_EPS) CLPR_dec_f_neg(j);
		else CLPR_dec_f_pos(j);
		CLPR_delete_occ(INEQ, i, j);
		free_ENTRY(entry);
	}
	eqn_i = CLPR_tableau[i]->eqn;
	eqn_i->next = NULL;
	eqn_i->constant = 0.0;
	eqn_i->time_stamp = -1;
	eqn_i->params->basic = 0;
	eqn_i->params->epsilon = 0.0;
	eqn_i->params->kohler = NULL;
}

void expand(EQN_ptr eqn1, EQN_ptr eqn2, double factor) /*** new eqn = eqn1 + eqn2 * factor ***/
                   
              
{
CLPR_int i, which;
EQN_ptr eqn, eqn_i;
TAB_HDR_NODE_ptr hdr;
ENTRY_ptr entry;
KOHLER_ptr koh;
OCC_ptr occ;

	eqn = CLPR_copy_eqn(eqn1);
	eqn->id = 0;
	eqn->params->kohler = 
		CLPR_kohler_merge(eqn1->params->kohler, eqn2->params->kohler); 
    
	/*** subset test, to use or not? 
    for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--) {
        eqn_i = CLPR_tableau[i]->eqn;
        if (eqn_i->time_stamp == -1) continue;
        if (!eqn_i->next) continue;
        if (CLPR_kohler_subset(eqn, eqn_i) == 2) {
            f_count5++;
            CLPR_free_eqn_nodes(eqn);
            return;
        }
    }
	***/

	CLPR_add_eqns(INEQ, eqn, eqn2, factor);
	if (!eqn->next) {
		CLPR_free_kohler_nodes(eqn->params->kohler);
		CLPR_free_eqn_nodes(eqn);
		return; /*** eqn->constant <= 0 ***/
	}
	eqn->time_stamp = 0;

	for (occ = CLPR_tableau[eqn->next->id]->ineq_occ; occ; occ = occ->next) {
		eqn_i = CLPR_tableau[occ->id]->eqn;
		if (eqn_i->time_stamp == -1) continue;
		if (CLPR_quasi_similar(eqn, eqn_i) == 2) {
			f_count4++;
			CLPR_free_kohler_nodes(eqn->params->kohler);
			CLPR_free_eqn_nodes(eqn);
			return;
		}
	}
	f_count3++;
	malloc_HDR(hdr);
	hdr->eqn = eqn;
	eqn->id = --CLPR_slack_id;
	if (CLPR_slack_id <= CLPR_solver_id)
		 CLPR_fatal("Solver tableau overflow (fourier-motzkin)");
	CLPR_tableau[CLPR_slack_id] = hdr;
	for (entry = eqn->next; entry; entry = entry->next) {
		i = entry->id;
		if (entry->coef < -CLPR_EPS) CLPR_inc_f_neg(i);
		else CLPR_inc_f_pos(i);
		CLPR_insert_occ(INEQ, CLPR_slack_id, i);
	}
}

void CLPR_inc_f_pos(CLPR_int id)
{
CLPR_int i;
	if (CLPR_is_pivot_var(id)) return;
	for (i = 1; i <= f_top; i++) if (fourier[i].id == id) break;
	if (i > f_top) {
		fourier[i].id = id;
		f_top++;
	}
	fourier[i].pos++;
}

void CLPR_inc_f_neg(CLPR_int id)
{
CLPR_int i;
	if (CLPR_is_pivot_var(id)) return;
	for (i = 1; i <= f_top; i++) if (fourier[i].id == id) break;
	if (i > f_top) {
		fourier[i].id = id;
		f_top++;
	}
	fourier[i].neg++;
}

void CLPR_dec_f_pos(CLPR_int id)
{
CLPR_int i;
	if (CLPR_is_pivot_var(id)) return;
	for (i = 1; i <= f_top; i++) if (fourier[i].id == id) break;
	if (i > f_top) CLPR_fatal("78932");
	fourier[i].pos--;
}

void CLPR_dec_f_neg(CLPR_int id)
{
CLPR_int i;
	if (CLPR_is_pivot_var(id)) return;
	for (i = 1; i <= f_top; i++) if (fourier[i].id == id) break;
	if (i > f_top) CLPR_fatal("78932");
	fourier[i].neg--;
}

CLPR_int CLPR_next_fourier_id(void)
{
CLPR_int i, least;
	least = 0;
	for (i = 1; i <= f_top; i++) {
		if (fourier[i].elim) continue; 
		if (CLPR_f_metric(i) < CLPR_f_metric(least)) least = i;
	}
	return least;
}

CLPR_int CLPR_f_metric(CLPR_int i)
{
	if (!i) return MAXPOSINT;
	if (0 < i && i <= f_top) return fourier[i].pos * fourier[i].neg;
	CLPR_fatal("23929");
}

CLPR_int CLPR_quasi_similar(EQN_ptr eqn1, EQN_ptr eqn2) /*** if TRUE, returns the stronger (1, 2) ***/
                   
{
ENTRY_ptr entry1, entry2 = (ENTRY_ptr) NULL;
double c;
	if ((entry1 = eqn1->next) && (entry2 = eqn2->next)) {
		c = entry1->coef / entry2->coef;
		if (c < -CLPR_EPS) return 0;
	}
	while (entry1 && entry2) {
		if (entry1->id == entry2->id &&
		    is_equal(c, entry1->coef / entry2->coef)) {
			entry1 = entry1->next;
			entry2 = entry2->next;
		} else return 0;
	}
	if (entry1 || entry2) return 0;
	if (eqn1->constant > c * eqn2->constant ||
		(is_equal(eqn1->constant, c * eqn2->constant) && 
			eqn1->params->epsilon >= c * eqn2->params->epsilon)) return 1;
	return 2; /*** strictly stronger ***/
}

/*------------------------------------------------------- DUMP! -------------*/

void CLPR_dump_funct_eqns(void)
{
CLPR_int i, *t, *t2;
	for (i = 1; i <= CLPR_funct_vars_top; i++) {
		t = funct_vars[i].bind;
		t2 = CLPR_findbind(t);
		if (is_unbound_var(*t2) || is_daemon(t2)) {
			CLPR_dump_funct_var_var(i, t);
			continue;
		}
		print_first_newline();
		CLPR_output_funct_var(t);
		fprintf(CLPR_dumpF, " = ");
		CLPR_output_term(t2);
		fprintf(CLPR_dumpF, "\n");
	}
}

void CLPR_dump_funct_var_var(CLPR_int i, CLPR_int *t) /*** t is bind, unbound ***/
          
{
CLPR_int j, *t2;
	t2 = CLPR_findbind(t);
	for (j = CLPR_funct_vars_top; j > i; j--) {
		if (funct_vars[j].rep == t2) {
			print_first_newline();
			CLPR_output_funct_var(t);
			fprintf(CLPR_dumpF, " = ");
			CLPR_output_var_str(funct_vars[j].str);
			fprintf(CLPR_dumpF, "\n");
			return;
		}
	}
}

void CLPR_dump_equiv_funct_vars(void)
{
CLPR_int i;
	if (is_dump_output && equiv_fvars_top) print_first_newline();
	for (i = 1; i <= equiv_fvars_top; i++) {
		CLPR_output_var_str(equiv_funct_vars[i].str);
		fprintf(CLPR_dumpF, " = ");
		CLPR_output_var_str(funct_vars[equiv_funct_vars[i].index].str);
		fprintf(CLPR_dumpF, "\n");
	}
}

void CLPR_dump_equiv_pivot_vars(void)
{
CLPR_int i, v;	
double c;
	if (is_dump_output && equiv_pvars_top) print_first_newline();
	for (i = 1; i <= equiv_pvars_top; i++) {
		CLPR_output_var_str(equiv_pivot_vars[i].str);
		fprintf(CLPR_dumpF, " = ");
		v = equiv_pivot_vars[i].id;
		if (CLPR_is_ground(v, &c)) CLPR_output_number(c);
		else CLPR_output_arith_var(v, FALSE);
		fprintf(CLPR_dumpF, "\n");
	}
}

void CLPR_dump_linear_eqns(void)
{
EQN_ptr eqn;
CLPR_int i;
double c;
	for (i = 1; i <= pivot_vars_top; i++) { 
		eqn = CLPR_tableau[pivot_vars[i].id]->eqn;
		if (pivot_vars[i].lin_output || pivot_vars[i].lin_subs) /* real eqn */
			CLPR_mark_eqn_arith_looking(eqn);
		if (!pivot_vars[i].lin_output) continue;
		c = eqn->constant;
		print_first_newline();
		CLPR_output_verbatim_arith_var(pivot_vars[i].id);
		fprintf(CLPR_dumpF, " = ");
		if (!eqn->next) {
			CLPR_output_number(c);
			newline();
		} else {
			CLPR_output_linear_form(eqn->next);
			if (c > CLPR_EPS) {
				fprintf(CLPR_dumpF, " + ");
				CLPR_output_number(c);
			} else if (c < -CLPR_EPS) {
				fprintf(CLPR_dumpF, " - ");
				CLPR_output_number(-c);
			}
			newline();
		}
	}
}

void CLPR_dump_ineq_eqns(void)
{
CLPR_int i;
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--) {
		if (CLPR_tableau[i]->eqn->time_stamp == -1) continue;
		CLPR_mark_ineq_arith_looking(CLPR_tableau[i]->eqn);
		print_first_newline();
		CLPR_output_ineq_eqn(CLPR_tableau[i]->eqn);
	}
}

void CLPR_dump_nlin_eqns(void)
{
CLPR_int i, *t, x;
NL_EQN_ptr nl_eqn;
	for (nl_eqn = CLPR_nl_eqn_top; nl_eqn; nl_eqn = nl_eqn->next) {
		if (nl_eqn->type == WOKEN || nl_eqn->partner) continue;
		if (nl_eqn->type == SOLVER_EVAL) {
			t = CLPR_findbind((CLPR_int *)nl_eqn->params1->id0);
			if (is_param(t)) {
				i = CLPR_is_pivot_var(mask(*t));
				if (pivot_vars[i].nlin_subs == nl_eqn) continue;
				CLPR_output_nl_eqn(nl_eqn);
			} else {
				if ((i = CLPR_is_funct_var(t)) && 
					(funct_vars[i].nlin_subs == nl_eqn)) continue;
				/*
				if (is_unbound_var(*t)) {
					i = CLPR_is_funct_var(t);
					if (funct_vars[i].nlin_subs == nl_eqn) continue;
				}
				*/
				CLPR_output_nl_eqn(nl_eqn);
			}
			continue;
		} else {
			CLPR_mark_nlin_arith_looking(nl_eqn);
			i = CLPR_is_pivot_var(x = nl_eqn->params1->id0);
			if (pivot_vars[i].nlin_subs == nl_eqn) {
				if (is_power_nlin(nl_eqn) && !CLPR_nl_has_been_subs(x)) 
					CLPR_output_lhs_verb_nl_eqn(nl_eqn);
			} else CLPR_output_nl_eqn(nl_eqn);
		}
	}
	for (i = 1; i <= pivot_vars_top; i++) {
		if (pivot_vars[i].cyclic) CLPR_output_nl_eqn(pivot_vars[i].nlin_subs);
	}
	for (i = 1; i <= CLPR_funct_vars_top; i++) {
		if (funct_vars[i].cyclic) CLPR_output_nl_eqn(pivot_vars[i].nlin_subs);
	}
}

void CLPR_dump_remaining_real_vars(void)
{
CLPR_int i;
	for (i = 1; i <= orig_pivot_vars_top; i++) {
		if (!pivot_vars[i].arith_looking && !pivot_vars[i].real_class) {
			print_first_newline();
			fprintf(CLPR_dumpF, "real(");
			CLPR_output_verbatim_arith_var(pivot_vars[i].id);
			fprintf(CLPR_dumpF, ")\n");
		}
	}
	for (i = orig_pivot_vars_top + 1; i <= pivot_vars_top; i++) {
		if (!pivot_vars[i].arith_looking && !pivot_vars[i].real_class &&
		   (pivot_vars[i].printed || pivot_vars[i].non_arith_subs)) {
			print_first_newline();
			fprintf(CLPR_dumpF, "real(");
			CLPR_output_arith_var(pivot_vars[i].id, FALSE);
			fprintf(CLPR_dumpF, ")\n");
		}
	}
}

/*------------------------------------------------------- print -------------*/

void CLPR_output_var_str(char *s)
{
	while (*s && *s != DUMP_SEPARATOR) putc(*s++, CLPR_dumpF);
}

void CLPR_output_verbatim_arith_var(CLPR_int i)
{
CLPR_int j;
char *s;
	if (j = CLPR_is_pivot_var(i)) {
		pivot_vars[j].printed = TRUE;
		if (s = pivot_vars[j].str) CLPR_output_var_str(s);
		else fprintf(CLPR_dumpF, "_t%d", pivot_vars[j].id);
	} else {
		if (i < old_slack_id) fprintf(CLPR_dumpF, "_t%d", i);
		else fprintf(CLPR_dumpF, "_S%d", i);
	}
}

void CLPR_output_arith_var(CLPR_int i, CLPR_int brackets)
{
char *s;
CLPR_int j, k;
double c;
EQN_ptr eqn;
NL_EQN_ptr nl_eqn;
	if (CLPR_is_ground(i, &c)) CLPR_output_number(c);
	else if (j = CLPR_is_pivot_var(i)) {
		if (pivot_vars[j].lin_subs) {
			eqn = CLPR_tableau[i]->eqn;
			if (CLPR_check_non_arith_subs(eqn)) pivot_vars[j].non_arith_subs = TRUE;
			c = eqn->constant;
			if (eqn->next) {
				brackets = brackets && (eqn->next->next || !is_zero(c));
				if (brackets) fprintf(CLPR_dumpF, "(");
				CLPR_output_linear_form(eqn->next);
				if (c > CLPR_EPS) { 
					fprintf(CLPR_dumpF, " + ");
					CLPR_output_number(c);
				} else if (c < -CLPR_EPS) {
					fprintf(CLPR_dumpF, " - ");
					CLPR_output_number(-c);
				}
				if (brackets) fprintf(CLPR_dumpF, ")");
			} else CLPR_output_number(c);
		} else if (pivot_vars[j].cyclic) CLPR_output_verbatim_arith_var(i);
		else if (nl_eqn = pivot_vars[j].nlin_subs) {
			pivot_vars[j].non_arith_subs = TRUE;
			CLPR_output_nl_rhs(nl_eqn);
		} else CLPR_output_verbatim_arith_var(i);
	} else CLPR_output_verbatim_arith_var(i);
}

CLPR_int CLPR_check_non_arith_subs(EQN_ptr eqn) /*** x -> t where t is non-arith looking ***/
            
{
CLPR_int i;
NL_EQN_ptr nl_eqn;
	/*** x = c*y ? ***/
	if (!eqn->next || eqn->next->next || !is_zero(eqn->constant)) return FALSE;
	/*** x = 1.0*y ? ***/
	if (!is_equal(eqn->next->coef, 1.0)) return FALSE;	
	i = CLPR_is_pivot_var(eqn->next->id);
	nl_eqn = pivot_vars[i].nlin_subs;
	if (!nl_eqn || nl_eqn->type == SOLVER_EVAL) return TRUE;
	return FALSE;
}

void CLPR_output_arith_func(CLPR_int type)
{
	switch (type) {
	case SOLVER_ABS: 
	case SOLVER_ABS0: 
		fprintf(CLPR_dumpF, "abs(");
		break;
	case SOLVER_POW:
	case SOLVER_POW0:
	case SOLVER_POW1:
	case SOLVER_POW2:
	case SOLVER_POW02:
		fprintf(CLPR_dumpF, "pow(");
		break;
	case SOLVER_SIN:
	case SOLVER_SIN0:
		fprintf(CLPR_dumpF, "sin(");
		break;
	case SOLVER_COS:
	case SOLVER_COS0:
		fprintf(CLPR_dumpF, "cos(");
		break;
	case SOLVER_MIN:
	case SOLVER_MIN0:
	case SOLVER_MIN1:
	case SOLVER_MIN2:
		fprintf(CLPR_dumpF, "min(");
		break;
	case SOLVER_MAX:
	case SOLVER_MAX0:
	case SOLVER_MAX1:
	case SOLVER_MAX2:
		fprintf(CLPR_dumpF, "max(");
		break;
	case SOLVER_EVAL:
		fprintf(CLPR_dumpF, "eval(");
		break;
	case SOLVER_ARCSIN:
		fprintf(CLPR_dumpF, "arcsin(");
		break;
	case SOLVER_ARCCOS:
		fprintf(CLPR_dumpF, "arccos(");
		break;
	default: CLPR_fatal("82313");
	}
}

void CLPR_output_funct_var(CLPR_int *t)
{
CLPR_int i;
char *s;
	if (i = CLPR_is_funct_var(t)) {
		if (funct_vars[i].nlin_subs) {
			CLPR_output_nl_rhs(funct_vars[i].nlin_subs);
			return;
		}
		s = funct_vars[i].str;
		if (s) CLPR_output_var_str(s);
		else CLPR_print_var2pos(CLPR_dumpF, funct_vars[i].rep);
	} else {
		CLPR_print_var2pos(CLPR_dumpF, t);
	}
}

void CLPR_output_linear_eqn(EQN_ptr eqn)
{
CLPR_int i, j;
double c;
	CLPR_output_verbatim_arith_var(eqn->id);
	fprintf(CLPR_dumpF, " = ");
	c = eqn->constant;
	if (!eqn->next) {
		CLPR_output_number(c);
		newline();
	} else {
		CLPR_output_linear_form(eqn->next);
		if (c > CLPR_EPS) {
			fprintf(CLPR_dumpF, " + ");
			CLPR_output_number(c);
		} else if (c < -CLPR_EPS) {
			fprintf(CLPR_dumpF, " - ");
			CLPR_output_number(-c);
		}
		newline();
	}
}		

void CLPR_output_ineq_eqn(EQN_ptr eqn)
{
CLPR_int all_positive, first_operand, constant_printed;
double c0, e0, c;
ENTRY_ptr entry;
	if (!eqn->next) return;
	c0 = -(eqn->constant);
	e0 = -(eqn->params->epsilon); /*** exp <= c0 + e0 ***/
	if (e0 >= CLPR_EPS) CLPR_fatal("10077");
	all_positive = first_operand = TRUE;
	for (entry = eqn->next; entry; entry = entry->next) {
		if ((c = entry->coef) > CLPR_EPS) {
			if (!first_operand) fprintf(CLPR_dumpF, " + ");
			first_operand = FALSE;
			if (!is_equal(fabs(c), 1.0)) {
				CLPR_output_number(c);
				fprintf(CLPR_dumpF, "*");
			}
			CLPR_output_arith_var(entry->id, FALSE);
		} else all_positive = FALSE;
	}
	constant_printed = FALSE;
	if (!all_positive) {
		if (first_operand) {
			CLPR_output_number(-c0);
			constant_printed = TRUE;
		} else if (c0 < -CLPR_EPS) {
			fprintf(CLPR_dumpF, " + ");
			CLPR_output_number(-c0);
			constant_printed = TRUE;
		}
	}
	if (e0) fprintf(CLPR_dumpF, " < "); else fprintf(CLPR_dumpF, " <= ");
	first_operand = TRUE;			
	for (entry = eqn->next; entry; entry = entry->next) {
		if ((c = entry->coef) < -CLPR_EPS) {
			if (!first_operand) fprintf(CLPR_dumpF, " + ");
			first_operand = FALSE;
			if (!is_equal(fabs(c), 1.0)) {
				CLPR_output_number(-c);
				fprintf(CLPR_dumpF, "*");
			}
			CLPR_output_arith_var(entry->id, FALSE);
		}
	}
	if (!constant_printed) {
		if (first_operand) CLPR_output_number(c0);
		else if (!is_zero(c0)) {
			fprintf(CLPR_dumpF, " + ");
			CLPR_output_number(c0);
		}
	}
	newline();
}

void CLPR_output_nl_eqn(NL_EQN_ptr nl_eqn)
{
	print_first_newline();
	if (nl_eqn->type == SOLVER_EVAL) CLPR_output_eval_lhs(nl_eqn);
	else CLPR_output_arith_var(nl_eqn->params1->id0, FALSE);
	fprintf(CLPR_dumpF, " = ");
	CLPR_output_nl_rhs(nl_eqn);
	fprintf(CLPR_dumpF, "\n");
}

void CLPR_output_lhs_verb_nl_eqn(NL_EQN_ptr nl_eqn)
{
	print_first_newline();
	CLPR_output_verbatim_arith_var(nl_eqn->params1->id0);
	fprintf(CLPR_dumpF, " = ");
	CLPR_output_nl_rhs(nl_eqn);
	fprintf(CLPR_dumpF, "\n");
}

void CLPR_output_linear_form(ENTRY_ptr entry) /*** c1 * e1 + ... cn * en, no constant ***/
                
{
double c;
char first_operand;
	first_operand = TRUE;
	while (entry) {
		c = entry->coef;
		if (!first_operand) {
			if (c > CLPR_EPS) fprintf(CLPR_dumpF, " + ");
			else if (c < -CLPR_EPS) fprintf(CLPR_dumpF, " - ");
		} else 
			if (c < -CLPR_EPS) fprintf(CLPR_dumpF, "-");
		if (!is_equal(fabs(c), 1.0)) {
			if (c > CLPR_EPS) {
				CLPR_output_number(c);
				fprintf(CLPR_dumpF, "*");
			} else {
				CLPR_output_number(-c);
				fprintf(CLPR_dumpF, "*");
			}
		}
		CLPR_output_arith_var(entry->id, FALSE);
		entry = entry->next;
		first_operand = FALSE;
	}
}

void CLPR_output_cons(CLPR_int *term)
{
	CLPR_output_term((CLPR_int *) pmask(*term));
	term = CLPR_findbind((CLPR_int *) pmask(*term) + 1);
	if (CLPR_tag(*term) == CLPR_TAG_CONS) {
		fprintf(CLPR_dumpF, ", ");
		CLPR_output_cons(term);
	} else 
		if (CLPR_tag(*term) == CLPR_TAG_NIL) return;
		else {
			fprintf(CLPR_dumpF, " | ");
			CLPR_output_term(term);
		}
}

void CLPR_output_term(CLPR_int *t)
{
	CLPR_output_printed_term(t, MAX_OP_PRECEDENCE+1);
}

void CLPR_output_printed_term(CLPR_int *t, CLPR_int priority)
{
CLPR_int i, j, n;
char *s;
double c;
EQN_ptr eqn;
CLPR_int current, left, right;
	t = CLPR_findbind(t);
	if (is_unbound_var(*t) || is_daemon(t)) CLPR_output_funct_var(t);		
	else {
		switch ((CLPR_unsigned) CLPR_tag(*t)) {
		case TAG_CON:
			fprintf(CLPR_dumpF, "%s", string_of(*t));
			break;
		case CLPR_TAG_NIL:
			fprintf(CLPR_dumpF, "[]");
			break;
		case TAG_NUM:
			CLPR_output_number(*((double *) pmask(*t)));
			break;
		case TAG_PAR:
			i = mask(*t);
			/* FIX: expand to include cases from CLPR_output_arith_var */
			CLPR_output_arith_var(i, FALSE);
			break;
		case CLPR_TAG_CONS:
			fprintf(CLPR_dumpF, "[");
			CLPR_output_cons(t);
			fprintf(CLPR_dumpF, "]");
			break;
		case TAG_STR:
			n = arity_of(*t);
			s = string_of(*t);
			current = CLPR_get_op_priority(s, n, &left, &right);
			switch (opkind_of(*t)) {
				case NONE:
					if (!isalpha(*s)) fprintf(CLPR_dumpF, "'%s'(", s);
					else fprintf(CLPR_dumpF, "%s(", s);
					for (i = 1; i <= n; i++) {
						CLPR_output_term(++t);
						if (i != n) fprintf(CLPR_dumpF, ", ");
					}
					fprintf(CLPR_dumpF, ")");
					break;
				case PREFIX:
					if (current > priority) fprintf(CLPR_dumpF, "(");
					fprintf(CLPR_dumpF, "%s ", s);
					CLPR_output_printed_term(++t, right);
					if (current > priority) fprintf(CLPR_dumpF, ")");
					break;
				case POSTFIX:
					if (current > priority) fprintf(CLPR_dumpF, "(");
					CLPR_output_printed_term(++t, left);
					fprintf(CLPR_dumpF, " %s", s);
					if (current > priority) fprintf(CLPR_dumpF, ")");
					break;
				case BINARY:
					if (current > priority) fprintf(CLPR_dumpF, "(");
					CLPR_output_printed_term(++t, left);
					fprintf(CLPR_dumpF, " %s ", s);
					CLPR_output_printed_term(++t, right);
					if (current > priority) fprintf(CLPR_dumpF, ")");
					break;
				default:
					CLPR_fatal("illegal opkind in CLPR_output_term");
			}
			break;
		default:
			CLPR_fatal("34563");
		}
	}
}

void CLPR_output_number(double c)
{
	if (is_zero(c)) fprintf(CLPR_dumpF, "0");
	else fprintf(CLPR_dumpF, "%g", c);
}

void CLPR_output_nl_rhs(NL_EQN_ptr nl_eqn)
{
NL_PARAMS1_ptr params1;
CLPR_int i, *t, type, v0, v1, v2;
double c, c2;
	params1 = nl_eqn->params1;
	v0 = params1->id0; v1 = params1->id1; v2 = params1->id2;
	c = params1->coef;
	switch (type = nl_eqn->type) {
		case SOLVER_MULT: 
			if (!is_equal(c, 1.0)) {
				CLPR_output_number(c);
				fprintf(CLPR_dumpF, " * ");
			}
			CLPR_output_arith_var(v1, TRUE); 
			fprintf(CLPR_dumpF, " * "); 
			CLPR_output_arith_var(v2, TRUE);
			break;
		case SOLVER_ABS: 
		case SOLVER_ABS0: 
		case SOLVER_SIN:
		case SOLVER_SIN0:
		case SOLVER_COS:
		case SOLVER_COS0:
		case SOLVER_ARCSIN:
		case SOLVER_ARCCOS:			
			CLPR_output_arith_func(type);
			CLPR_output_arith_var(v1, FALSE); 
			fprintf(CLPR_dumpF, ")");
			break;
		case SOLVER_POW: 
		case SOLVER_POW0: 
		case SOLVER_POW1: 
		case SOLVER_POW2: 
		case SOLVER_POW02:
		case SOLVER_MIN:
		case SOLVER_MIN1:
		case SOLVER_MIN2:
		case SOLVER_MAX:
		case SOLVER_MAX1:
		case SOLVER_MAX2:
			CLPR_output_arith_func(type);
			CLPR_output_arith_var(v1, FALSE); 
			fprintf(CLPR_dumpF, ", ");
			CLPR_output_arith_var(v2, FALSE); 
			fprintf(CLPR_dumpF, ")");
			break;
		case SOLVER_EVAL:
			CLPR_output_arith_func(type);
			t = CLPR_findbind((CLPR_int *)v1);
			if ((i= CLPR_is_funct_var(t)) && funct_vars[i].nlin_subs)
				CLPR_output_nl_rhs(funct_vars[i].nlin_subs);
			else CLPR_output_funct_var(t);
			fprintf(CLPR_dumpF, ")");
			break;
		default: CLPR_fatal("23743");
	}
}

void CLPR_output_eval_lhs(NL_EQN_ptr nl_eqn)
{
	CLPR_int v0, *t;
	v0 = nl_eqn->params1->id0;
	t = CLPR_findbind((CLPR_int *)v0);
	if (CLPR_tag(*t) == TAG_PAR) {
		CLPR_output_arith_var(mask(*t), FALSE);
	} else {
		CLPR_output_term(t);
	}
}

void CLPR_mark_eqn_arith_looking(EQN_ptr eqn)
{
CLPR_int i, j, a, b;
ENTRY_ptr entry;
	i = CLPR_is_pivot_var(eqn->id);
	if (!eqn->next || 
		eqn->next->next || 
		!is_zero(eqn->constant) ||
		!is_equal(eqn->next->coef, 1.0)) {
		if (i) pivot_vars[i].arith_looking = TRUE;
		for (entry = eqn->next; entry; entry = entry->next) 
			if (j = CLPR_is_pivot_var(entry->id))
				pivot_vars[j].arith_looking = TRUE;
		return;
	}
	/*** x = y ***/
	j = CLPR_is_pivot_var(eqn->next->id);
	if (i && j) {
		if (pivot_vars[i].arith_looking) 
			pivot_vars[j].arith_looking = TRUE;
		else if (pivot_vars[j].arith_looking) 
			pivot_vars[i].arith_looking = TRUE;
		for (a = i; pivot_vars[a].real_class; a = pivot_vars[a].real_class) ;
		for (b = j; pivot_vars[b].real_class; b = pivot_vars[b].real_class) ;
		pivot_vars[a].real_class = b;
	}
}

void CLPR_mark_ineq_arith_looking(EQN_ptr eqn)
{
CLPR_int i;
ENTRY_ptr entry;
	for (entry = eqn->next; entry; entry = entry->next) 
		if (i = CLPR_is_pivot_var(entry->id)) pivot_vars[i].arith_looking = TRUE;
}

void CLPR_mark_nlin_arith_looking(NL_EQN_ptr nl_eqn)
{
CLPR_int i;
	if (nl_eqn->type == SOLVER_EVAL) return;
	if (i = CLPR_is_pivot_var(nl_eqn->params1->id0)) 
		pivot_vars[i].arith_looking = TRUE;
	if (i = CLPR_is_pivot_var(nl_eqn->params1->id1)) 
		pivot_vars[i].arith_looking = TRUE;
	if (i = CLPR_is_pivot_var(nl_eqn->params1->id2)) 
		pivot_vars[i].arith_looking = TRUE;
}

/*--------------------------------------------------- Aux stuff -------------*/

CLPR_int CLPR_find_slack(EQN *eqn, double *c_ptr)
{
ENTRY_ptr entry;
	entry = eqn->next;
	while (entry) {
		if (entry->id >= old_slack_id) {
			*c_ptr = entry->coef;
			return entry->id;
		}
		entry = entry->next;
	}
	return 0;
}

double CLPR_delete_slack(EQN_ptr eqn)
{
ENTRY first;
register ENTRY_ptr prev_entry, entry;
double c;
	first.next = entry = eqn->next;
	prev_entry = &first;
	while (entry && entry->id < old_slack_id) {
		prev_entry = entry;
		entry = entry->next;
	}
	c = entry->coef;
	prev_entry->next = entry->next;
 	free_ENTRY(entry);
	eqn->next = first.next;
	if (eqn->id) CLPR_delete_occ(INEQ, eqn->id, entry->id);
	return c;
}

void CLPR_push_dump_trail(CLPR_int type, CLPR_int v, EQN_ptr eqn)
{
	if (CLPR_trtop >= CLPR_MTRAIL) {
		printf("CLPR_trtop %d %d\n", CLPR_trtop, CLPR_MTRAIL);
		CLPR_fatal("Dump trail overflow");
	}
	if (eqn) push_trail(type, eqn);
	else push_trail(type, v);
}

void CLPR_pop_dump_trail(CLPR_int i)
{
CLPR_int j;
	j = (CLPR_int) CLPR_trail[i];
	switch (tagnum(j)) {
		case NUM_BTK_NULL_EQN: CLPR_restore_null_eqn((CLPR_int *)mask(j)); break;
		case NUM_BTK_EQN: CLPR_restore_eqn((CLPR_int *)pmask(j)); break;
		case NUM_BTK_NULL_INEQ: CLPR_restore_null_ineq((CLPR_int *)mask(j)); break;
		case NUM_BTK_INEQ: CLPR_restore_ineq((CLPR_int *)pmask(j)); break;
		default: printf("pop dump %x, tag %x\n", j, tagnum(j));
			CLPR_fatal("92345");
	}
}

void CLPR_print_dump_trail(void)
{
CLPR_int i,j;
	i = CLPR_trtop;
	while (--i >= old_trtop) {
		j = (CLPR_int) CLPR_trail[i];
		printf("CLPR_trail[%d]: ", i); 
		switch (tagnum(j)) {
			case NUM_BTK_NULL_EQN: 
				printf("<null eqn>\n"); break;
			case NUM_BTK_EQN: 
				CLPR_print_eqn((EQN_ptr) pmask(j));
				break;
			case NUM_BTK_NULL_INEQ: 
				printf("<null ineq>\n"); break;
			case NUM_BTK_INEQ: 
				CLPR_print_eqn((EQN_ptr) pmask(j));
				break;
			default: CLPR_fatal("92311");
		}
	}
}

CLPR_int CLPR_is_pivot_var(CLPR_int v) /*	pivot var uses next in header nodes */
      
{
TAB_HDR_NODE_ptr hdr;
	if (!(hdr = CLPR_tableau[v])) return 0;
	else return (CLPR_int) hdr->next;
}

CLPR_int CLPR_is_funct_var(CLPR_int *t)
{
CLPR_int i;
	funct_vars[0].bind = funct_vars[0].rep = t;
	i = CLPR_funct_vars_top;	
	while (funct_vars[i].bind != t && funct_vars[i].rep != t) i--;
	return i;
}

CLPR_int CLPR_heap_to_pivot_var(CLPR_int *t)
{
CLPR_int i;
	pivot_vars[0].heap = t;
	i = pivot_vars_top;	
	while (pivot_vars[i].heap != t) i--;
	return i;
}

CLPR_int CLPR_heap_to_free_var(CLPR_int *t)
{
CLPR_int i;
	funct_vars[0].heap = t;
	i = CLPR_funct_vars_top;	
	while (funct_vars[i].heap != t) i--;
	if (!i) {
		flatten_vars[0].new_var = t;
		for (i = CLPR_flatten_vars_top; ; i--) {
			if (flatten_vars[i].new_var == t) break;
			else if (flatten_vars[i].var == t) {
				i = CLPR_flatten_vars_top + i;
				break;
			}
		}
		if (i) i += CLPR_funct_vars_top;
	}
	return i;
}

void CLPR_free_kohler_nodes(KOHLER_ptr koh)
{
KOHLER_ptr koh99;
	while (koh) {
		koh99 = koh->next;
		free_KOHLER(koh);
		koh = koh99;
	}
}

void CLPR_print_collected_vars(void)
{
CLPR_int i;
char *s;
	printf("Orig: pivots %d functs %d, Tops: pivots %d functs = %d\n",
		orig_pivot_vars_top, orig_funct_vars_top, 
		pivot_vars_top, CLPR_funct_vars_top);
	for (i = 1; i <= CLPR_funct_vars_top; i++) {
		printf("funct_vars[%d] = b:%p, r:%p ", 
			i, funct_vars[i].bind, funct_vars[i].rep);
			CLPR_print_var2pos(stdout, funct_vars[i].rep);
/*
		if (funct_vars[i].heap) {
			printf(" h = ");
			CLPR_print_var2pos(stdout, funct_vars[i].heap);
		}
*/
		if (s = funct_vars[i].str) {
			printf(" (");
			CLPR_output_var_str(s);
			printf(")\n");
		} else printf(" (no name)\n");
	}
	for (i = 1; i <= pivot_vars_top; i++) {
		printf("pivot_vars[%d] = id:%d", i, pivot_vars[i].id);
/*
		if (pivot_vars[i].heap) {
			printf(" h = ");
			CLPR_print_var2pos(stdout, pivot_vars[i].heap);
		}
*/
		if (s = pivot_vars[i].str) {
			printf(" (");
			CLPR_output_var_str(s);
			printf(")\n");
		} else printf(" (no name)\n");
	}
}

void CLPR_print_fourier(void)
{
CLPR_int i;
	for (i = 1; i <= f_top; i++) {
		if (fourier[i].elim) continue;
		CLPR_output_arith_var(fourier[i].id, FALSE);
		printf(": +ve %d -ve %d\n", fourier[i].pos, fourier[i].neg);
	}
}


/* ------------------------------------------------------------------------ */

void CLPR_special_print_tableau(CLPR_int arg)
{
CLPR_int i;
	printf(
		"%%---------- Ineq_tableau (%d) -----------------------------\n", arg);
	printf("p%d(X19) :- \n", arg);
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--)
		if (CLPR_tableau[i] && CLPR_tableau[i]->eqn) {
			CLPR_special_print_eqn(CLPR_tableau[i]->eqn);
			printf(",\n");
		}
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--) {
		printf("X%d >= 0", i);
		if (i > CLPR_slack_id) printf(",\n"); else printf(".\n");
	}
	printf(
		"%%----------------------------------------------------------\n");
}

void CLPR_special_print_eqn(EQN_ptr eqn)
{
ENTRY_ptr entry;
CLPR_int first;
	if (!eqn) CLPR_fatal("8618");
	if (!is_zero(eqn->params->epsilon)) CLPR_fatal("8619");
	first = TRUE;
	entry = eqn->next;
	if (!entry) return;
	while (entry) {
		if (entry->coef > 0.0) {
			if (first) printf( " %g*X%d", entry->coef, entry->id);
			else printf( " +%g*X%d", entry->coef, entry->id);
		} else 
			printf(" -%g*X%d", -(entry->coef), entry->id);
		entry = entry->next;
		first = FALSE;
	}
	printf( " = %g", -eqn->constant);
}

CLPR_int CLPR_count_all_ineq(void)
{
CLPR_int i, count;
	count = 0;
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--)
		if (CLPR_tableau[i]->eqn->time_stamp != -1) count++;
	return count;
}

/*--------------------------------------------- dump/3 ----------------------*/

static void make_equiv_funct_vars(void)
{
CLPR_int i;
	for (i = 1; i <= equiv_fvars_top; i++) {
		add_constraint(make_eqn(equiv_funct_vars[i].heap, 
			funct_vars[equiv_funct_vars[i].index].heap));				
	}
}

static void make_equiv_pivot_vars(void)
{
CLPR_int i, j;	
	for (i = 1; i <= equiv_pvars_top; i++) {
		j = CLPR_is_pivot_var(equiv_pivot_vars[i].id);
		add_constraint(make_eqn(equiv_pivot_vars[i].heap, 
			pivot_vars[i].heap));	
	}
}

static void make_funct_eqns(void)
{
CLPR_int i, *t, *t2;
CLPR_int *tmp;
	for (i = 1; i <= CLPR_funct_vars_top; i++) {
		t = funct_vars[i].bind;
		t2 = CLPR_findbind(t);
		if (is_unbound_var(*t2) || is_daemon(t2)) {
			tmp = make_funct_var_var(i, funct_vars[i].bind);
			if (tmp) add_constraint(tmp);
		} else 	
			add_constraint(make_eqn(make_funct_var(t), make_term(t2, FALSE)));
	}
}

static void add_flatten_eqns(void)
{
CLPR_int i;
CLPR_int *t;
	for (i = 1; i <= CLPR_flatten_vars_top; i++) {
		t = flatten_vars[i].rep;
		if (!is_unbound_var(*t) && !is_daemon(t)) {
			add_constraint(t = make_eqn(flatten_vars[i].var, 
				make_term(flatten_vars[i].rep, FALSE)));
		}
	}
	for (i = 1; i <= CLPR_flatten_vars_top; i++) {
		 add_constraint(make_eqn(flatten_vars[i].var, 
			make_func1(HAT_QUOTEVAL, flatten_vars[i].new_var)));
	}
}

static void add_linear_eqns(void)
{
EQN_ptr eqn;
CLPR_int i, *t1, *t2;
double c;
	for (i = 1; i <= pivot_vars_top; i++) {
		eqn = CLPR_tableau[pivot_vars[i].id]->eqn;
		if (pivot_vars[i].lin_output || pivot_vars[i].lin_subs)
			CLPR_mark_eqn_arith_looking(eqn);
		if (!pivot_vars[i].lin_output) continue;
		if (!(t1 = pivot_vars[i].heap)) CLPR_fatal("38455");
		c = eqn->constant;
		if (!eqn->next) t2 = make_num(c);
		else {
			t2 = make_linear_form(eqn->next);
			if (c > CLPR_EPS) t2 = make_func(HAT_PLUSVAL, t2, make_num(c));
			else if (c < -CLPR_EPS) t2 = make_func(HAT_MINUSVAL, t2, make_num(-c));
		}
		add_constraint(make_eqn(t1, t2));
	}
}

static CLPR_int *make_linear_form(ENTRY_ptr entry)
{
double c;
char first_operand;
CLPR_int *t, *t1;
	first_operand = TRUE;
	while (entry) {
		c = entry->coef;
		if (!is_equal(fabs(c), 1.0)) {
			if (c > CLPR_EPS) t1 = make_func(HAT_MULTVAL,
								 make_num(c), make_arith_var(entry->id));
			else t1 = make_func(HAT_MULTVAL,
								make_num(-c), make_arith_var(entry->id));
		} else t1 = make_arith_var(entry->id);
		if (!first_operand) {
			if (c > CLPR_EPS) t = make_func(HAT_PLUSVAL, t, t1);
			else if (c < -CLPR_EPS) t = make_func(HAT_MINUSVAL, t, t1);
		} else {
			if (c < -CLPR_EPS) t = make_neg(t1); 
			else t = t1;
		}
		entry = entry->next;
		first_operand = FALSE;
	}
	return t;
}

static CLPR_int *make_func(CLPR_int val, CLPR_int *t1, CLPR_int *t2)
{
CLPR_int *t;
	t = CLPR_heaptop;
	*(CLPR_heaptop++) = addtag(TAG_STR, val);
	make_bound_var(CLPR_heaptop++, t1);
	make_bound_var(CLPR_heaptop++, t2);
	return t;
}

static CLPR_int *make_func1(CLPR_int val, CLPR_int *t1)
{
CLPR_int *t;
	t = CLPR_heaptop;
	*(CLPR_heaptop++) = addtag(TAG_STR, val);
	make_bound_var(CLPR_heaptop++, t1);
	return t;
}

static CLPR_int *make_neg(CLPR_int *t1)
{
CLPR_int *t;
        t = CLPR_heaptop;
        *(CLPR_heaptop++) = addtag(TAG_STR,HAT_U_MINUSVAL);
        make_bound_var(CLPR_heaptop++,t1);
        return t;
}

static CLPR_int *make_funct_var_var(CLPR_int i, CLPR_int *t) /*** t is bind, unbound ***/
          
{
CLPR_int j, *t2;
	t2 = CLPR_findbind(t);
	for (j = CLPR_funct_vars_top; j > i; j--) {
		if (funct_vars[j].rep == t2) {
			return make_eqn(make_funct_var(t), funct_vars[j].heap);
		}
	}
	return NULL;
}

static void add_constraint(CLPR_int *e)
{
	if (!e) return;
	if (dump_mode) {
		*constraint_list = addtag(CLPR_TAG_CONS, CLPR_heaptop);
		make_bound_var(CLPR_heaptop, e);
		constraint_list = CLPR_heaptop + 1;
		CLPR_heaptop += 2;
	} else {
		*constraint_list = (CLPR_int) CLPR_heaptop;
		make_bound_var(CLPR_heaptop, addtag(TAG_STR, COMMA));
		make_bound_var(CLPR_heaptop + 1, e);
		constraint_list = CLPR_heaptop + 2;
		CLPR_heaptop += 3;
	}
}

static CLPR_int *make_num(double n)
{
CLPR_int *t;
	if (num_vec_top >= MAX_PROJ_NUM)
		CLPR_runtime_error("too many constants in dump/3\n");
	if (is_zero(n)) n = 0.0;
	num_vec[++num_vec_top].loc = CLPR_heaptop;
	num_vec[num_vec_top].num = n;
	*CLPR_heaptop = addtag(TAG_MISC, NULL);
	t = CLPR_heaptop++;
	return t;
}

static void intern_numbers(void)
{
CLPR_int i;
	for (i = 1; i <= num_vec_top; i++) {
		CLPR_makenum(++CLPR_solver_id, num_vec[i].num);
		*(num_vec[i].loc) = addtag(TAG_PAR, CLPR_solver_id);
	}
}

static CLPR_int *make_funct_var(CLPR_int *t)
{
CLPR_int i;
NL_EQN_ptr nl;
	if (i = CLPR_is_funct_var(t)) {
		if ((nl = funct_vars[i].nlin_subs)) return make_nl_rhs(nl);
		else return funct_vars[i].heap;
	} else {
		if (CLPR_funct_vars_top >= MAX_DUMP_VAR - 1) {
			CLPR_runtime_error("dump overflow (make numbers)");
		}
		funct_vars[++CLPR_funct_vars_top].bind = t;
		funct_vars[CLPR_funct_vars_top].rep = CLPR_findbind(t);
		funct_vars[CLPR_funct_vars_top].str = NULL;
        make_unbound_var(CLPR_heaptop);
		funct_vars[CLPR_funct_vars_top].heap = CLPR_heaptop++;
        return funct_vars[CLPR_funct_vars_top].heap;
	}
}

static CLPR_int *make_pivot_var(CLPR_int i)
{
CLPR_int j;
	if (j = CLPR_is_pivot_var(i)) {
		pivot_vars[j].printed = TRUE;
		return pivot_vars[j].heap;
	}
	CLPR_fatal("66184");
}

static CLPR_int *make_eqn(CLPR_int *t1, CLPR_int *t2)
{
CLPR_int *t;
	t = CLPR_heaptop;
	*(CLPR_heaptop++) = addtag(TAG_STR, EQVAL);
	make_bound_var(CLPR_heaptop++, t1);
	make_bound_var(CLPR_heaptop++, t2);
	return t;
}

CLPR_int *CLPR_make_cons(CLPR_int *t1, CLPR_int *t2)
{
CLPR_int *t;
	t = CLPR_heaptop;
	CLPR_heaptop = CLPR_heaptop + 3;
	*t = addtag(CLPR_TAG_CONS, t + 1);
	make_bound_var(t + 1, t1);
	make_bound_var(t + 2, t2);
	return t;
}

/*
 * ^+(t1, t2): if ti is var, daemon or ^+, 
 *  flatten using new var Xi, then make new eqn ti = ^quote(Xi).
 */

static CLPR_int *make_str(CLPR_int n, CLPR_int *t0, CLPR_int in_arith)
{
CLPR_int i, *t, *t1;
	t = CLPR_heaptop;
	CLPR_heaptop = CLPR_heaptop + n + 1;
	*t = *t0;
	if (is_val_hat(mask(*t0))) {
		t1 = t;
		t = make_func1(HAT_QUOTEVAL, t);
		for (i = 1; i <= n; i++)
			make_bound_var(t1 + i, CLPR_flatten_var_term(t0 + i));
	} else {
		for (i = 1; i <= n; i++) 
			make_bound_var((t + i), make_term(t0 + i, in_arith));
	}
	return t;
}

static CLPR_int *make_term(CLPR_int *t, CLPR_int in_arith)
{
CLPR_int *t1, *t2, n;
char *s;
double c;
EQN_ptr eqn;
CLPR_int *t0;
	t = CLPR_findbind(t);
	if (is_unbound_var(*t) || is_daemon(t)) return make_funct_var(t);	
	else {
		switch ((CLPR_unsigned) CLPR_tag(*t)) {
		case TAG_CON:
		case CLPR_TAG_NIL:
		case TAG_NUM: return t;
		case TAG_PAR: return make_arith_var(mask(*t));
		case CLPR_TAG_CONS:
			t1 = make_term(head_of(t), in_arith);
			t2 = make_term(tail_of(t), in_arith);
			return CLPR_make_cons(t1, t2);
		case TAG_STR:
			return make_str(arity_of(*t), t, in_arith);
		default:
			CLPR_fatal("34872");
		}
	}
}

static CLPR_int *make_arith_var(CLPR_int i)
{
CLPR_int *t, j;
double c;
EQN_ptr eqn;
NL_EQN_ptr nl_eqn;

	if (CLPR_is_ground(i, &c)) return make_num(c);
	if (j = CLPR_is_pivot_var(i)) {
		if (pivot_vars[j].lin_subs) {
			eqn = CLPR_tableau[i]->eqn;
			if (CLPR_check_non_arith_subs(eqn)) pivot_vars[j].non_arith_subs = TRUE;
			c = eqn->constant;
			if (eqn->next) {
				t = make_linear_form(eqn->next);
				if (c > CLPR_EPS) return make_func(HAT_PLUSVAL, t, make_num(c));
				else if (c < -CLPR_EPS) 
					return make_func(HAT_MINUSVAL, t, make_num(-c));
				else return t;
			} else return make_num(c);
		} else if (pivot_vars[j].cyclic) return make_pivot_var(i);
		else if (nl_eqn = pivot_vars[j].nlin_subs) {
			pivot_vars[j].non_arith_subs = TRUE;
			return make_nl_rhs(nl_eqn);
		} else return make_pivot_var(i);
	} else return make_pivot_var(i);
}

static void add_ineq_eqns(void)
{
CLPR_int i;
	for (i = CLPR_SOLVER_SZ - 1; i >= CLPR_slack_id; i--) {
		if (CLPR_tableau[i]->eqn->time_stamp == -1) continue;
		CLPR_mark_ineq_arith_looking(CLPR_tableau[i]->eqn);
		add_constraint(make_ineq_eqn(CLPR_tableau[i]->eqn));
	}
}

static CLPR_int *make_ineq_eqn(EQN_ptr eqn)
{
double c0, e0, c;
CLPR_int t, *t1, *t2;
CLPR_int all_positive, first_operand, constant_printed;
ENTRY_ptr entry;
	if (!eqn->next) return (CLPR_int *) NULL;
	c0 = -(eqn->constant);
	e0 = -(eqn->params->epsilon); /*** exp <= c0 + e0 ***/
	if (e0 >= CLPR_EPS) CLPR_fatal("10078");
	t1 = make_linear_form(eqn->next);
	t2 = make_num(c0);
	return (is_zero(e0) ? make_func(LEVAL, t1, t2) : make_func(LTVAL, t1, t2));
}

void CLPR_add_nlin_eqns(void)
{
CLPR_int i, *t, x;
NL_EQN_ptr nl_eqn;

	for (nl_eqn = CLPR_nl_eqn_top; nl_eqn; nl_eqn = nl_eqn->next) {
		if (nl_eqn->type == WOKEN || nl_eqn->partner) continue;
		if (nl_eqn->type == SOLVER_EVAL) {
			t = CLPR_findbind((CLPR_int *)nl_eqn->params1->id0);
			if (is_param(t)) {
				i = CLPR_is_pivot_var(mask(*t));
				if (pivot_vars[i].nlin_subs == nl_eqn) continue;
				add_constraint(make_nl_eqn(nl_eqn));
			} else {
				if ((i = CLPR_is_funct_var(t)) && 
					(funct_vars[i].nlin_subs == nl_eqn)) continue;
				/*
				if (is_unbound_var(*t)) {
					i = CLPR_is_funct_var(t);
					if (funct_vars[i].nlin_subs) continue;
				}
				*/
				add_constraint(make_nl_eqn(nl_eqn));
			}
		} else {
			CLPR_mark_nlin_arith_looking(nl_eqn);
			x = nl_eqn->params1->id0;
			if (!(i = CLPR_is_pivot_var(x))) CLPR_fatal("66895");
			if (pivot_vars[i].nlin_subs == nl_eqn) {
				if (is_power_nlin(nl_eqn) && !CLPR_nl_has_been_subs(x))
					add_constraint(make_lhs_verb_nl_eqn(nl_eqn));
			} else add_constraint(make_nl_eqn(nl_eqn));
		}
	}
	for (i = 1; i <= pivot_vars_top; i++) {
		if (pivot_vars[i].cyclic) 
			add_constraint(make_nl_eqn(pivot_vars[i].nlin_subs));
	}
	for (i = 1; i <= CLPR_funct_vars_top; i++) {
		if (funct_vars[i].cyclic) 
			add_constraint(make_nl_eqn(pivot_vars[i].nlin_subs));
	}	

}

void CLPR_add_remaining_real_vars(void)
{
CLPR_int i;
	for (i = 1; i <= orig_pivot_vars_top; i++) {
		if (!pivot_vars[i].arith_looking && !pivot_vars[i].real_class) 
			add_constraint(make_func1(CLPR_real_val, 
				make_arith_var(pivot_vars[i].id)));
	}
	for (i = orig_pivot_vars_top + 1; i <= pivot_vars_top; i++) {
		if (!pivot_vars[i].arith_looking && !pivot_vars[i].real_class &&
		   (pivot_vars[i].printed || pivot_vars[i].non_arith_subs))
			add_constraint(make_func1(CLPR_real_val, 
				make_arith_var(pivot_vars[i].id)));
	}
}

static CLPR_int *make_nl_eqn(NL_EQN_ptr nl_eqn)
{
CLPR_int *left;
	if (nl_eqn->type == SOLVER_EVAL) 
		left = make_eval_lhs(nl_eqn);
	else left = make_arith_var(nl_eqn->params1->id0);
	return make_eqn(left, make_nl_rhs(nl_eqn));
}

static CLPR_int *make_lhs_verb_nl_eqn(NL_EQN_ptr nl_eqn)
{
CLPR_int *left;
	left = make_pivot_var(nl_eqn->params1->id0);
	return make_eqn(left, make_nl_rhs(nl_eqn));
}

static CLPR_int *make_nl_rhs(NL_EQN_ptr nl_eqn)
{
CLPR_int *right, *t;
NL_PARAMS1_ptr params1;
CLPR_int i, type, v0, v1, v2;
double c, c2;
	params1 = nl_eqn->params1;
	v0 = params1->id0; v1 = params1->id1; v2 = params1->id2;
	c = params1->coef;
	switch (type = nl_eqn->type) {
		case SOLVER_MULT: 
			t = make_func(HAT_MULTVAL, make_arith_var(v1),
					 make_arith_var(v2));
			if (!is_equal(c, 1.0)) 
				t =  make_func(HAT_MULTVAL, make_num(c), t);
			return t;
		case SOLVER_ABS: 
		case SOLVER_ABS0: return make_func1(HAT_ABSVAL, make_arith_var(v1));
		case SOLVER_SIN:  return make_func1(HAT_SINVAL, make_arith_var(v1));
		case SOLVER_COS:  return make_func1(HAT_COSVAL, make_arith_var(v1));
		case SOLVER_ARCSIN:  
			return make_func1(HAT_ARCSINVAL, make_arith_var(v1));
		case SOLVER_ARCCOS:  
			return make_func1(HAT_ARCCOSVAL, make_arith_var(v1));
		case SOLVER_POW: 
		case SOLVER_POW0: 
		case SOLVER_POW1:
		case SOLVER_POW2:
		case SOLVER_POW02:
			return make_func(HAT_POWVAL, make_arith_var(v1),
						make_arith_var(v2));
		case SOLVER_MIN:
		case SOLVER_MIN0:
		case SOLVER_MIN1:
		case SOLVER_MIN2:
			return make_func(HAT_MINVAL, make_arith_var(v1),
						make_arith_var(v2));
		case SOLVER_MAX:
		case SOLVER_MAX0:
		case SOLVER_MAX1:
		case SOLVER_MAX2:
			return make_func(HAT_MAXVAL, make_arith_var(v1),
						make_arith_var(v2));
		case SOLVER_EVAL:
			t = CLPR_findbind((CLPR_int *)v1);
			if ((i = CLPR_is_funct_var(t)) && funct_vars[i].nlin_subs)
				right = make_nl_rhs(funct_vars[i].nlin_subs);
			else right = make_funct_var(t);
			return make_func1(HAT_EVALVAL, right);
		default: CLPR_fatal("35812");
	}
}

static CLPR_int *make_eval_lhs(NL_EQN_ptr nl_eqn)
{
CLPR_int v0, *t;
	v0 = nl_eqn->params1->id0;
	t = CLPR_findbind((CLPR_int *)v0);
	return (is_param(t) ? make_arith_var(mask(*t)) : make_term(t, FALSE));
}

CLPR_int CLPR_nl_has_been_subs(CLPR_int x) /*** conservative check for x being subs'd ***/
      
{
TAB_HDR_NODE_ptr hdr;
OCC_ptr occ1;
NL_OCC_ptr occ2;
CLPR_int j;
	if (!(hdr = CLPR_tableau[x])) return FALSE;
	for (occ1 = hdr->eq_occ; occ1; occ1 = occ1->next) {
		j = CLPR_is_pivot_var(occ1->id);
		if (pivot_vars[j].lin_output) return TRUE;
	}
	if (hdr->ineq_occ) return TRUE;
	for (occ2 = hdr->nlin_occ; occ2; occ2 = occ2->next) {
		j = CLPR_is_pivot_var(occ2->nl_eqn->params1->id0);
		if (j && !pivot_vars[j].nlin_subs) return TRUE;
	}
	return FALSE;
}

