/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: emit.c                                               *
 ***************************************************************/

#include "standard.h"
#include "config.h"
#include "parse.h"
#include "symbols.h"
#include "lib.h"
#include "compile.h"
#include "opcodes.h"
#include "emul.h"

#define clear_last()			last_opcode=BADINST
#define set_last(op)		{ last_opcode=op; last_opaddr = CLPR_pcode; }
#define reset_last(op)		last_opcode=op;
#define match_last(op)		(opt && last_opcode == op)
#define gen_word(op)		CLPR_code[CLPR_pcode++]=(CLPR_unsigned) op
#define gen_lastword(op, off)	CLPR_code[last_opaddr+off]=(CLPR_unsigned) op
#define lastword_at(off)	(CLPR_code[last_opaddr + off])

static CLPR_int last_opcode=BADINST, last_opaddr;

static CLPR_int opt = 1;

CLPR_int CLPR_has_been_saved;
CLPR_int CLPR_save_nvars;
CLPR_int CLPR_must_do_save;
double *CLPR_floatp;
double CLPR_float_array[2];
#define floatstore (CLPR_float_array[0])

extern CLPR_int CLPR_pcode;

extern VREC *CLPR_notevar(CLPR_int val);
extern CLPR_int CLPR_find_loc_for_lable(CLPR_int val, char *s);
extern CLPR_int CLPR_punresolved;
extern CLPR_unsigned *CLPR_code;
extern LABLEREC CLPR_unresolved[];
extern CLPR_int CLPR_s_count[];
extern FUNCTREC CLPR_functors[];
extern SPNODE *CLPR_spred[];
extern HASH_L_NODE_ptr CLPR_p_jump_occ(CLPR_int name);

/*---------------------------------------------------------------------------*/


void CLPR_emit_realsave (void);
CLPR_int CLPR_fold_multi2 (CLPR_int oldop, CLPR_int newop, CLPR_int arg1, CLPR_int arg2);
CLPR_int CLPR_fold_multi1 (CLPR_int oldop, CLPR_int newop, CLPR_int arg1);
void CLPR_emit_addr (char *s, CLPR_int val);
extern void CLPR_assign_p_jump_occ (CLPR_int name, HASH_L_NODE_ptr occ);
extern void CLPR_add_unres_lable (char *s, CLPR_int val, CLPR_int pc, CLPR_int off);
extern CLPR_int CLPR_val_of_cur_rule (void);
extern void CLPR_copy_unres_lable (CLPR_int from_pc, CLPR_int pc);
extern CLPR_int CLPR_p_cut (CLPR_int name);
extern CLPR_int CLPR_hashbuiltina (char *s, CLPR_int n);

void CLPR_init_emit(void)
{
	clear_last();
}

/*--------------------------------------------------- get -------------------*/

void CLPR_emit_gettnum(double f, CLPR_int R)
{
	clear_last();
	floatstore = f;
	CLPR_floatp = &floatstore;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GETTNUM;
	CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
	CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}

void CLPR_emit_getpnum(double f, CLPR_int R)
{
	clear_last();
	CLPR_emit_realsave();
	floatstore = f;
	CLPR_floatp = &floatstore;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GETPNUM;
	CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
	CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}

void CLPR_emit_getnil(CLPR_int R)
{
	clear_last();
	if (R == 1) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET1NIL;
	else if (R == 2) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET2NIL;
	else if (R == 3) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET3NIL;
	else {	
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GETNIL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
	}
}		

void CLPR_emit_getcon(CLPR_int val, CLPR_int R)
{
	clear_last();
	if (R == 1) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET1CON;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
	} else if (R == 2) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET2CON;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
	} else if (R == 3) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET3CON;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
	} else {	
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GETCON;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
	}
}		

void CLPR_emit_getstr(CLPR_int val, CLPR_int R)
{
	clear_last();
	if (R == 1) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET1STR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
	} else if (R == 2) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET2STR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
	} else if (R == 3) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET3STR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
	} else {	
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GETSTR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
	}
}		

void CLPR_emit_getlist(CLPR_int R)
{
	clear_last();
	if (R == 1) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET1LIST;
	else if (R == 2) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET2LIST;
	else if (R == 3) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET3LIST;
	else {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GETLIST;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
	}
}		

void CLPR_emit_gettvar(CLPR_int R1, CLPR_int R2)
{
	if (R1 == 1) {
		set_last(GET1TVAR);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET1TVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else if (R1 == 2) {
		set_last(GET1TVAR);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET2TVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else if (R1 == 3) {
		set_last(GET1TVAR);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET3TVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else {
		set_last(GETTVAR);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GETTVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R1;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	}
}		

void CLPR_emit_gettval(CLPR_int R1, CLPR_int R2)
{
	clear_last();
	if (R1 == 1) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET1TVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else if (R1 == 2) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET2TVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else if (R1 == 3) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET3TVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GETTVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R1;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	}
}		

void CLPR_emit_getpvar(CLPR_int S1, CLPR_int R2)
{
	CLPR_emit_realsave();
	if (S1 == 1) {
		set_last(GET1PVAR);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET1PVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else if (S1 == 2) {
		set_last(GET2PVAR);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET2PVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else if (S1 == 3) {
		set_last(GET3PVAR);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET3PVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else {
		set_last(GETPVAR);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GETPVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) S1;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	}
}		

void CLPR_emit_getpval(CLPR_int S1, CLPR_int R2)
{
	clear_last();
	CLPR_emit_realsave();
	if (S1 == 1) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET1PVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else if (S1 == 2) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET2PVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else if (S1 == 3) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GET3PVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	} else {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) GETPVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) S1;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	}
}		

/*--------------------------------------------------- uni -------------------*/

void CLPR_emit_univoid(CLPR_int n)
{
	if (match_last(UNIVOID)) {
		gen_lastword(UNIVOIDN, 0);
		reset_last(UNIVOIDN);
		gen_word(1 + n);
	} else if (match_last(UNIVOIDN)) {
		gen_lastword(lastword_at(1)+n, 1);
	} else {
		if (n == 1) {
			set_last(UNIVOID);
			gen_word(UNIVOID);
		} else {
			set_last(UNIVOIDN);
			gen_word(UNIVOIDN);
			gen_word(n);
		}
	}
}		

void CLPR_emit_uninum(double f)
{
	clear_last();
	floatstore = f;
	CLPR_floatp = &floatstore;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNINUM;
	CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
	CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
}

void CLPR_emit_uninil(void)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNINIL;
}		

void CLPR_emit_unilist(void)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNILIST;
}		

void CLPR_emit_unicon(CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNICON;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
}		

void CLPR_emit_unistr(CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNISTR;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
}		

void CLPR_emit_uni2ppvar(CLPR_int R1, CLPR_int R2)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = UNI2PPVAR;
	CLPR_code[CLPR_pcode++] = R1;
	CLPR_code[CLPR_pcode++] = R2;
}

void CLPR_emit_uni2ptvar(CLPR_int R1, CLPR_int R2)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = UNI2PTVAR;
	CLPR_code[CLPR_pcode++] = R1;
	CLPR_code[CLPR_pcode++] = R2;
}

void CLPR_emit_uni2tpvar(CLPR_int R1, CLPR_int R2)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = UNI2TPVAR;
	CLPR_code[CLPR_pcode++] = R1;
	CLPR_code[CLPR_pcode++] = R2;
}

void CLPR_emit_uni2ttvar(CLPR_int R1, CLPR_int R2)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = UNI2TTVAR;
	CLPR_code[CLPR_pcode++] = R1;
	CLPR_code[CLPR_pcode++] = R2;
}

void CLPR_emit_uni2tvvar(CLPR_int R)
{
	if (opt) {
		set_last(UNI2TVVAR);
		gen_word(UNI2TVVAR);
		gen_word(R);
	} else {
		gen_word(UNITVAR); gen_word(R);
		gen_word(UNIVOID);
		clear_last();
	}
}

void CLPR_emit_uni2vtvar(CLPR_int R)
{
	if (opt) {
		set_last(UNI2VTVAR);
		gen_word(UNI2VTVAR);
		gen_word(R);
	} else {
		gen_word(UNIVOID);
		gen_word(UNITVAR); gen_word(R);
		clear_last();
	}
}

void CLPR_emit_uni2pvvar(CLPR_int S)
{
	CLPR_emit_realsave();
	if (opt) {
		set_last(UNI2PVVAR);
		gen_word(UNI2PVVAR);
		gen_word(S);
	} else {
		gen_word(UNIPVAR); gen_word(S);
		gen_word(UNIVOID);
		clear_last();
	}
}

void CLPR_emit_uni2vpvar(CLPR_int S)
{
	CLPR_emit_realsave();
	if (opt) {
		set_last(UNI2VPVAR);
		gen_word(UNI2VPVAR);
		gen_word(S);
	} else {
		gen_word(UNIVOID);
		gen_word(UNIPVAR); gen_word(S);
		clear_last();
	}
}

void CLPR_emit_unitvar(CLPR_int R)
{
	clear_last();
	if (R == 1) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNIT1VAR;
	else if (R == 2) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNIT2VAR;
	else if (R == 3) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNIT3VAR;
	else {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNITVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
	}
}		

void CLPR_emit_unipvar(CLPR_int R)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNIPVAR;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}		

void CLPR_emit_unitval(CLPR_int R)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNITVAL;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}		

void CLPR_emit_unipval(CLPR_int R)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) UNIPVAL;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}		

/*--------------------------------------------------- put -------------------*/

void CLPR_emit_puttnum(double f, CLPR_int R)
{
	clear_last();
	floatstore = f;
	CLPR_floatp = &floatstore;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PUTTNUM;
	CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
	CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}

void CLPR_emit_putpnum(double f, CLPR_int R)
{
	clear_last();
	CLPR_emit_realsave();
	floatstore = f;
	CLPR_floatp = &floatstore;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PUTPNUM;
	CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
	CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}

void CLPR_emit_putnil(CLPR_int reg)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PUTNIL;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) reg;
}

void CLPR_emit_putcon(CLPR_int val, CLPR_int reg)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PUTCON;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) reg;
}

void CLPR_emit_putlist(CLPR_int reg)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PUTLIST;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) reg;
}

void CLPR_emit_putpval(CLPR_int S1, CLPR_int R2)
{
	if (!CLPR_fold_multi2(PUTPVAL, PUTPVALN, S1, R2)) {
		CLPR_emit_realsave();
		set_last(PUTPVAL);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PUTPVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) S1;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	}
}

void CLPR_emit_putuval(CLPR_int S1, CLPR_int R2)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PUTUVAL;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) S1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
}

void CLPR_emit_putpvar(CLPR_int S1, CLPR_int R2)
{
	if (!CLPR_fold_multi2(PUTPVAR, PUTPVARN, S1, R2)) {
		CLPR_emit_realsave();
		set_last(PUTPVAR);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PUTPVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) S1;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R2;
	}
}

void CLPR_emit_puttvar(CLPR_int R)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PUTTVAR;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}

void CLPR_emit_putstr(CLPR_int val, CLPR_int R)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PUTSTR;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}

/*--------------------------------------------------- bld -------------------*/

void CLPR_emit_bldvoid(void)
{
	if (match_last(BLDVOID)) {
		gen_lastword(BLDVOIDN, 0);
		reset_last(BLDVOIDN);
		gen_word(2);
	} else if (match_last(BLDVOIDN)) {
		gen_lastword(lastword_at(1)+1, 1);
	} else {
		set_last(BLDVOID);
		gen_word(BLDVOID);
	}
}		

void CLPR_emit_bldnum(double f)
{
	clear_last();
	floatstore = f;
	CLPR_floatp = &floatstore;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) BLDNUM;
	CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
	CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
}

void CLPR_emit_bldnil(void)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) BLDNIL;
}

void CLPR_emit_bldlist(void)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) BLDLIST;
}		

void CLPR_emit_bldcon(CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) BLDCON;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
}

void CLPR_emit_bldstr(CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) BLDSTR;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) val;
}

void CLPR_emit_bldtvar(CLPR_int R)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) BLDTVAR;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}

void CLPR_emit_bldtval(CLPR_int R)
{
	if (!CLPR_fold_multi1(BLDTVAL, BLDTVALN, R)) {
		set_last(BLDTVAL);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) BLDTVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
	}
}

void CLPR_emit_bldpval(CLPR_int R)
{
	if (!CLPR_fold_multi1(BLDPVAL, BLDPVALN, R)) {
		CLPR_emit_realsave();
		set_last(BLDPVAL);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) BLDPVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
	}
}

void CLPR_emit_bldpvar(CLPR_int R)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) BLDPVAR;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}

/*--------------------------------------------------- switch ----------------*/

void CLPR_emit_switchnil(char *s, CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SWITCHONNIL;
	CLPR_emit_addr(s, val);
}

void CLPR_emit_switchdot(char *s, CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SWITCHONDOT;
	CLPR_emit_addr(s, val);
}

void CLPR_emit_switchcon(char *s, CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SWITCHONCON;
	CLPR_emit_addr(s, val);
}

void CLPR_emit_switch2con(char *s, CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SWITCH2ONCON;
	CLPR_emit_addr(s, val);
}

void CLPR_emit_switchfunc(char *s, CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SWITCHONFUNC;
	CLPR_emit_addr(s, val);
}

void CLPR_emit_switch2func(char *s, CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SWITCH2ONFUNC;
	CLPR_emit_addr(s, val);
}

void CLPR_emit_switchreal(char *s, CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SWITCHONREAL;
	CLPR_emit_addr(s, val);
}

void CLPR_emit_switch2real(char *s, CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SWITCH2ONREAL;
	CLPR_emit_addr(s, val);
}

void CLPR_emit_switchtype(char *s1, char *s2, char *s3, char *s4, char *s5, CLPR_int val1, CLPR_int val2, CLPR_int val3, CLPR_int val4, CLPR_int val5)
{
	clear_last();
	if (val1 == FAILVAL && val2 == FAILVAL && val5 == FAILVAL) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SWITCHONCONFUNC;
		CLPR_emit_addr(s3, val3);
		CLPR_emit_addr(s4, val4);
	} else {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SWITCHONTYPE;
		CLPR_emit_addr(s1, val1);
		CLPR_emit_addr(s2, val2);
		CLPR_emit_addr(s3, val3);
		CLPR_emit_addr(s4, val4);
		CLPR_emit_addr(s5, val5);
	}
}

/*------------------------------------------------------ arith --------------*/

void CLPR_emit_reginit(CLPR_int R)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) REGINIT;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
}

void CLPR_emit_initpf(double c)
{
	clear_last();
	if (is_zero(c)) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) INIT0PF;
	else {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) INITPF;
		floatstore = c;
		CLPR_floatp = &floatstore;
		CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
		CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
	}
}

void CLPR_emit_addpfvar(double c, CLPR_int S)
{
	clear_last();
	CLPR_emit_realsave();
	if (is_equal(c, 1.0)) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) ADD1PFVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) S;
	} else {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) ADDPFVAR;
		floatstore = c;
		CLPR_floatp = &floatstore;
		CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
		CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) S;
	}
}

void CLPR_emit_addpfval(double c, CLPR_int S)
{
	clear_last();
	CLPR_emit_realsave();
	if (is_equal(c, 1.0)) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) ADD1PFVAL;
		CLPR_code[CLPR_pcode++] =  (CLPR_unsigned) S;
	} else {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) ADDPFVAL;
		floatstore = c;
		CLPR_floatp = &floatstore;
		CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
		CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) S;
	}
}

void CLPR_emit_addtpfvar(double c, CLPR_int R)
{
	clear_last();
	if (is_equal(c, 1.0)) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) ADDT1PFVAR;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
	} else {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) ADDTPFVAR;
		floatstore = c;
		CLPR_floatp = &floatstore;
		CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
		CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
	}
}

void CLPR_emit_addtpfval(double c, CLPR_int R)
{
	clear_last();
	if (is_equal(c, 1.0)) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) ADDT1PFVAL;
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
	} else {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) ADDTPFVAL;
		floatstore = c;
		CLPR_floatp = &floatstore;
		CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
		CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) R;
	}
}

void CLPR_emit_solve_pf0(void)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SOLVE_PF0;
}

void CLPR_emit_solve_ge0(void)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SOLVE_GE0;
}

void CLPR_emit_solve_gt0(void)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SOLVE_GT0;
}

void CLPR_emit_mult(CLPR_unsigned opc, double c, CLPR_int i1, CLPR_int i2, CLPR_int i3)
{
	clear_last();
	CLPR_emit_realsave();
	if (is_equal(c, 1.0)) {
		CLPR_code[CLPR_pcode++] = SOLVE_MULT1;
		CLPR_code[CLPR_pcode++] = opc;
	} else {
		CLPR_code[CLPR_pcode++] = SOLVE_MULT;
		CLPR_code[CLPR_pcode++] = opc;
		floatstore = c;
		CLPR_floatp = &floatstore;
		CLPR_code[CLPR_pcode++] = *((CLPR_unsigned *) CLPR_floatp);
		CLPR_code[CLPR_pcode++] = *(((CLPR_unsigned *) CLPR_floatp) + 1);
	}
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i2;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i3;
}

void CLPR_emit_min(CLPR_unsigned opc, CLPR_int i1, CLPR_int i2, CLPR_int i3)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = SOLVE_MIN;
	CLPR_code[CLPR_pcode++] = opc;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i2;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i3;
}

void CLPR_emit_max(CLPR_unsigned opc, CLPR_int i1, CLPR_int i2, CLPR_int i3)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = SOLVE_MAX;
	CLPR_code[CLPR_pcode++] = opc;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i2;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i3;
}

void CLPR_emit_pow(CLPR_unsigned opc, CLPR_int i1, CLPR_int i2, CLPR_int i3)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = SOLVE_POW;
	CLPR_code[CLPR_pcode++] = opc;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i2;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i3;
}

void CLPR_emit_abs(CLPR_unsigned opc, CLPR_int i1, CLPR_int i2)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = SOLVE_ABS;
	CLPR_code[CLPR_pcode++] = opc;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i2;
}

void CLPR_emit_sin(CLPR_unsigned opc, CLPR_int i1, CLPR_int i2)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = SOLVE_SIN;
	CLPR_code[CLPR_pcode++] = opc;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i2;
}

void CLPR_emit_cos(CLPR_unsigned opc, CLPR_int i1, CLPR_int i2)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = SOLVE_COS;
	CLPR_code[CLPR_pcode++] = opc;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i2;
}

void CLPR_emit_arcsin(CLPR_unsigned opc, CLPR_int i1, CLPR_int i2)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = SOLVE_ARCSIN;
	CLPR_code[CLPR_pcode++] = opc;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i2;
}

void CLPR_emit_arccos(CLPR_unsigned opc, CLPR_int i1, CLPR_int i2)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = SOLVE_ARCCOS;
	CLPR_code[CLPR_pcode++] = opc;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i2;
}

/*------------------------------------------------------ misc ---------------*/

void CLPR_emit_inst(CLPR_int w)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) w;
}

void CLPR_emit_word(CLPR_int w)
{
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) w;
}

void CLPR_emit_doubleword(CLPR_int w1, CLPR_int w2)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) w1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) w2;
}

void CLPR_emit_crestore(void)
{
	if (!(CLPR_must_do_save && !CLPR_has_been_saved)) {
		clear_last();
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) CRESTORE;
	}
}

void CLPR_emit_restore(void)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) RESTORE;
}


void CLPR_emit_proceed(void)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PROCEED;
}

void CLPR_emit_callp(char *s, CLPR_int val)
{
HASH_L_NODE_ptr occ;
	clear_last();
	CLPR_emit_realsave();
/*
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) CALLP;
	CLPR_emit_addr(s, val);
*/
	gen_word(CALLI);
	occ = (HASH_L_NODE_ptr) CLPR_emalloc(sizeof(HASH_L_NODE));
	occ->val = val;
	occ->loc = CLPR_pcode;
	occ->next = CLPR_p_jump_occ(val);
	CLPR_assign_p_jump_occ(val, occ);
	gen_word(val);
}

void CLPR_emit_addr(char *s, CLPR_int val)
{
	if (!s) CLPR_code[CLPR_pcode++] = 0;
	else CLPR_add_unres_lable(s, val, CLPR_pcode++, 0);
}

void CLPR_emit_jump(char *s, CLPR_int val)
{
HASH_L_NODE_ptr occ;
CLPR_int entry_pc;

	if (opt && (val == CLPR_val_of_cur_rule()) && 
	   ((entry_pc = CLPR_find_loc_for_lable(val, s)) != -1)) {
		switch(CLPR_code[entry_pc]) {
		case SWITCHONTYPE:
			gen_word(ENTERSWTYPE);	
			CLPR_copy_unres_lable(entry_pc+1, CLPR_pcode++);
			CLPR_copy_unres_lable(entry_pc+2, CLPR_pcode++);
			CLPR_copy_unres_lable(entry_pc+3, CLPR_pcode++);
			CLPR_copy_unres_lable(entry_pc+4, CLPR_pcode++);
			CLPR_copy_unres_lable(entry_pc+5, CLPR_pcode++);
			CLPR_add_unres_lable(s, val, CLPR_pcode++, 6);
			break;
		case SWITCHONCON:
			gen_word(ENTERSWCON);	
			CLPR_add_unres_lable(s, val, CLPR_pcode++, 0);
			break;
		case SWITCHONFUNC:
			gen_word(ENTERSWFUNC);	
			CLPR_add_unres_lable(s, val, CLPR_pcode++, 0);
			break;
		case SWITCHONCONFUNC:
			gen_word(ENTERSWCONFUNC);	
			CLPR_copy_unres_lable(entry_pc+1, CLPR_pcode++);
			CLPR_copy_unres_lable(entry_pc+2, CLPR_pcode++);
			CLPR_add_unres_lable(s, val, CLPR_pcode++, 3);
			break;
		case SWITCHONREAL:
			gen_word(ENTERSWREAL);	
			CLPR_add_unres_lable(s, val, CLPR_pcode++, 0);
			break;
		case TRY:
			goto NORMAL_ENTRY;  /* CHECK */
/***
			gen_word(ENTERTRYELSE);	
			CLPR_copy_unres_lable(entry_pc+1, CLPR_pcode++);
			gen_word(CLPR_code[entry_pc+2]);
			CLPR_add_unres_lable(s, val, CLPR_pcode++, 3);
			break;
***/
		default:
			goto NORMAL_ENTRY;
		}
		clear_last();
		return;
	}
NORMAL_ENTRY:
	clear_last();
/*
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) JUMP;
	CLPR_emit_addr(s, val);
*/
	gen_word(JUMPI);
	occ = (HASH_L_NODE_ptr) CLPR_emalloc(sizeof(HASH_L_NODE));
	occ->val = val;
	occ->loc = CLPR_pcode;
	occ->next = CLPR_p_jump_occ(val);
	CLPR_assign_p_jump_occ(val, occ);
	gen_word(val);
}

void CLPR_emit_try(char *s, CLPR_int val, CLPR_int n)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) TRY;
	CLPR_add_unres_lable(s, val, CLPR_pcode++, 0);
	if (CLPR_p_cut(val)) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) n+1;
	else CLPR_code[CLPR_pcode++] = (CLPR_unsigned) n;
}

void CLPR_emit_retry(char *s, CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) RETRY;
	CLPR_add_unres_lable(s, val, CLPR_pcode++, 0);
}

void CLPR_emit_trust(char *s, CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) TRUST;
	CLPR_add_unres_lable(s, val, CLPR_pcode++, 0);
}

void CLPR_emit_dump(char *str, CLPR_int *CLPR_regs)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) DUMP;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) CLPR_hashbuiltina(str, 0);
	CLPR_emit_word(CLPR_regs[0]);
}

void CLPR_emit_printf(CLPR_int *CLPR_regs)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) PRINTF;
/*
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) CLPR_hashbuiltina(str, 0);
*/
	CLPR_emit_word(CLPR_regs[0]);
	CLPR_emit_word(CLPR_regs[1]);
}

void CLPR_emit_read(char *str, CLPR_int num, CLPR_int *CLPR_regs)
{
VREC *v;
CLPR_int tmp;
	clear_last();
	CLPR_emit_word(READREG);
	CLPR_emit_word(CLPR_regs[0]);
/*
	v = CLPR_notevar(num);
	if (v->stack_loc) {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) READ;
		CLPR_code[CLPR_pcode++] = v->stack_loc;
	} else {
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) READREG;
		CLPR_make_live_reg((tmp = CLPR_getreg()), v);
		CLPR_code[CLPR_pcode++] = tmp;
	}
*/
}

void CLPR_emit_print(CLPR_int *CLPR_regs)
{
	clear_last();
	CLPR_emit_word(PRINT);
	CLPR_emit_word(CLPR_regs[0]);
}

void CLPR_emit_retract(CLPR_int *CLPR_regs)
{
	clear_last();
	CLPR_emit_word(RETRACT);
	CLPR_emit_word(CLPR_regs[0]);
}

void CLPR_emit_roole(CLPR_int *CLPR_regs)
{
	clear_last();
	CLPR_emit_word(ROOLE);
	CLPR_emit_word(CLPR_regs[0]);
	CLPR_emit_word(CLPR_regs[1]);
}

void CLPR_emit_jumpx(CLPR_int *CLPR_regs)
{
	clear_last();
	CLPR_emit_word(JUMPX);
	CLPR_emit_word(CLPR_regs[0]);
}

void CLPR_emit_callx(CLPR_int *CLPR_regs)
{
	clear_last();
	CLPR_emit_word(CALLX);
	CLPR_emit_word(CLPR_regs[0]);
}

void CLPR_emit_spcall(CLPR_int val, CLPR_int *CLPR_regs)
{
	CLPR_int i;

	clear_last();
	CLPR_emit_word(SPCALL);
	CLPR_emit_word(val);
	for (i=0; i < CLPR_spred[val]->arity; i++) CLPR_emit_word(CLPR_regs[i]);
}

void CLPR_emit_halt(void)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) HALT;
}

void CLPR_emit_maybesave(CLPR_int nvars)
{
	CLPR_must_do_save = TRUE;
	CLPR_has_been_saved = FALSE;
	CLPR_save_nvars = nvars;
}

void CLPR_emit_realsave(void)
{
	if (CLPR_must_do_save && !CLPR_has_been_saved) {
		clear_last();
		CLPR_code[CLPR_pcode++] = (CLPR_unsigned) SAVE;
		CLPR_code[CLPR_pcode++] = CLPR_save_nvars;
		CLPR_must_do_save = FALSE;
		CLPR_has_been_saved = TRUE;
	}
}

void CLPR_emit_savecurcp(CLPR_int R)
{
	clear_last();
	CLPR_emit_word(SAVECURCP);
	CLPR_emit_word(R);
}

void CLPR_emit_saveprevcp(CLPR_int R)
{
	clear_last();
	CLPR_emit_word(SAVEPREVCP);
	CLPR_emit_word(R);
}

void CLPR_emit_savecut(CLPR_int R, CLPR_int Y)
{
	clear_last();
	CLPR_emit_word(SAVECUT);
	CLPR_emit_word(R);
	CLPR_emit_word(Y);
}

void CLPR_emit_fastcut(CLPR_int R)
{
	clear_last();
	CLPR_emit_word(FASTCUT);
	CLPR_emit_word(R);
}

void CLPR_emit_docut(CLPR_int Y)
{
	clear_last();
	CLPR_emit_word(DOCUT);
	CLPR_emit_word(Y);
}

void CLPR_emit_dbg_entry(PTERM *t, CLPR_int num)
{
	clear_last();
	CLPR_emit_inst(DBGENTRY);
	CLPR_emit_word(t->val);
	CLPR_emit_word(num);
}

CLPR_int CLPR_fold_multi1(CLPR_int oldop, CLPR_int newop, CLPR_int arg1)
{
	CLPR_int old1;

	if (match_last(oldop)) {
		old1 = lastword_at(1);
		reset_last(newop);
		gen_lastword(newop, 0);
		gen_lastword(2, 1);
		gen_word(old1);
		gen_word(arg1);
	} else if (match_last(newop)) {
		gen_lastword(lastword_at(1) + 1, 1);
		gen_word(arg1);
	} else return FALSE;
	return TRUE;
}

CLPR_int CLPR_fold_multi2(CLPR_int oldop, CLPR_int newop, CLPR_int arg1, CLPR_int arg2)
{
	CLPR_int old1, old2;

	if (match_last(oldop)) {
		old1 = lastword_at(1);
		old2 = lastword_at(2);
		reset_last(newop);
		gen_lastword(newop, 0);
		gen_lastword(2, 1);
		gen_lastword(old1, 2);
		gen_word(old2);
		gen_word(arg1);
		gen_word(arg2);
	} else if (match_last(newop)) {
		gen_lastword(lastword_at(1) + 1, 1);
		gen_word(arg1);
		gen_word(arg2);
	} else return FALSE;
	return TRUE;
}

void CLPR_emit_eval(CLPR_unsigned opc, CLPR_int i1, CLPR_int i2)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = SOLVE_EVAL;
	CLPR_code[CLPR_pcode++] = opc;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i1;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) i2;
}

void CLPR_emit_special_savecut(CLPR_int r)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) DIRSAVECUT;
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) r;
}

void CLPR_emit_dynamic_try(CLPR_int val)
{
	clear_last();
	if (arity_of(val) > 0) CLPR_code[CLPR_pcode++] = (CLPR_unsigned) DYNAMIC_TRY;
	else CLPR_code[CLPR_pcode++] = (CLPR_unsigned) DYNAMIC_0_TRY;
	CLPR_emit_word(arity_of(val));
	CLPR_emit_word(val);
}

void CLPR_emit_dyn_call(CLPR_int val)
{
	clear_last();
	CLPR_emit_realsave();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) DYN_CALL;
	CLPR_emit_word(val);
}

void CLPR_emit_dyn_jump(CLPR_int val)
{
	clear_last();
	CLPR_code[CLPR_pcode++] = (CLPR_unsigned) DYN_JUMP;
	CLPR_emit_word(val);
}

