/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: emul.c	                                           *
 ***************************************************************/

#if defined(SYS5) || defined(OS2V2)
#include <sys/types.h>
#include <sys/times.h>
#endif
#if defined(BSD) && !defined(OS2V2)
#include <sys/time.h>
#include <sys/resource.h>
#endif

#include "standard.h"
#include "config.h"
#include "parse.h"
#include "symbols.h"
#include "lib.h"
#include "compile.h"
#include "opcodes.h"
#include "emul.h"
#include "solver.h"
#include "main.h"

#define MIN_HEAP_SZ   256
#define MIN_LSTACK_SZ 256
#define MIN_TRAIL_SZ  256

#define MAX_SAVE_CONTEXT 1024

#define CHECK_WAKEUP 
#define CHECK_OVERFLOW

#ifdef CHECK_OVERFLOW
#define CheckHeap()  \
	if (CLPR_heaptop >= maxheap) { \
		fprintf(stderr, "Heapsize = %d\n", CLPR_HEAP_SZ); \
		CLPR_fatal("Heap overflow"); \
	} 
#define CheckStack()  \
	if (CLPR_stacktop >= maxlstack) { \
		fprintf(stderr, "Stacksize = %d\n", CLPR_LSTACK_SZ); \
		CLPR_fatal("Stack overflow"); \
	} 
#define CheckCP()  \
	if (CLPR_stacktop >= maxlstack) { \
		fprintf(stderr, "Stacksize = %d\n", CLPR_LSTACK_SZ); \
		CLPR_fatal("Stack overflow"); \
	} 
#define CheckTrail() \
	if (CLPR_trtop >= maxtrail) { \
		fprintf(stderr, "Trailsize = %d\n", CLPR_TRAIL_SZ); \
		CLPR_fatal("Trail overflow"); \
	}
#else
#define CheckHeap()
#define CheckStack()
#define CheckCP()
#define CheckTrail()
#endif

struct {
	CLPR_int *retreg, *curar, *stacktop, *heaptop, trtop,
		*lastcp, solver_id, slack_id, nlin_count, 
		*runable, have_runable;
	NL_EQN_ptr nl_eqn_top;
} CLPR_save[MAX_SAVE_CONTEXT];

CLPR_int *CLPR_findbind(CLPR_int *x);
CLPR_int *CLPR_val_to_pc(CLPR_int val);
RULE_ptr CLPR_quick_find(RULE_ptr r, CLPR_int val, double fp_val, RULE_ptr *prev_rule, CLPR_int quoted), CLPR_p_first_rule(CLPR_int name);
CLPR_int *CLPR_run_dynamic(CLPR_int *pc, CLPR_int first, CLPR_int can_index);
CLPR_int *CLPR_pterm_to_heap(PTERM *t);

extern CLPR_int *CLPR_pterm_to_quoted_heap(PTERM *t, CLPR_int use_collected_vars);
extern FILE *CLPR_outfile;
extern code_t *CLPR_alloc_mem(CLPR_int n);
extern void CLPR_free_nlin_nodes(NL_EQN_ptr nl_eqn);

/*-------------------------------------------------------- globals ----------*/

CLPR_int *CLPR_stack;	             /* Register File / Heap / Stack */
CLPR_int **CLPR_trail;	            /* Trail pointers */

CLPR_int CLPR_REGS_SZ = DEF_REGS_SZ;
CLPR_int CLPR_HEAP_START = DEF_REGS_SZ;
CLPR_int CLPR_LSTACK_SZ = DEF_LSTACK_SZ;
CLPR_int CLPR_HEAP_SZ = DEF_HEAP_SZ;
CLPR_int CLPR_TRAIL_SZ = DEF_TRAIL_SZ;
CLPR_int CLPR_MTRAIL = DEF_TRAIL_SZ + TRAIL_SAFETY;
CLPR_int CLPR_MSTACK = 
	DEF_REGS_SZ + DEF_HEAP_SZ + HEAP_SAFETY + DEF_LSTACK_SZ + STACK_SAFETY;
CLPR_int CLPR_LSTACK = DEF_REGS_SZ + DEF_HEAP_SZ + HEAP_SAFETY;

CLPR_int *CLPR_runable;
bool CLPR_have_runable = FALSE;

CLPR_int *CLPR_regs;	      /* Pointer to register file */
CLPR_int *CLPR_heap;	      /* Pointer to heap */
CLPR_int *CLPR_safeheap;      /* Pointer to heap equivalent of top choicepoint */
CLPR_int *CLPR_lstack;	      /* Pointer to stack */
static CLPR_int *maxlstack;
static CLPR_int *maxheap;
static CLPR_int maxtrail;

CLPR_int *CLPR_retreg;	      /* Return Register */
CLPR_int *CLPR_curar;	      /* Current activation record */
CLPR_int *CLPR_stacktop;      /* Current top of CLPR_stack */
CLPR_int *CLPR_heaptop;	      /* Current top of CLPR_heap */
CLPR_int CLPR_trtop;	      /* Current top of CLPR_trail */
CLPR_int *CLPR_lastcp;	      /* Latest choice point on CLPR_stack */
static CLPR_int *s;	      /* Unification read/write mode registers */

CLPR_int CLPR_stamp;
CLPR_int *CLPR_first_lastcp;

double CLPR_EPS = DEFAULT_EPS;
CLPR_int CLPR_pcode = 1;
CLPR_int CLPR_epcode;
CLPR_int CLPR_savep;
char *CLPR_ts;
double CLPR_tf_array[2], *CLPR_tfp;
#define tf (CLPR_tf_array[0])
double CLPR_stored_constant;
CLPR_int CLPR_tag_pair[MAX_TAGS][MAX_TAGS]; 
CLPR_int CLPR_swt_offsets[MAX_TAGS + 1];

static CLPR_int true_atom;

#if defined(SYS5) || defined(OS2V2)
double aix_utime;
struct tms time1, time2;
#endif
#if defined(BSD) && !defined(OS2V2)
struct rusage usg;
struct timeval time1, time2;
#endif

extern float CLPR_int_to_float(CLPR_int i);
extern PTERM *CLPR_lookup_name(char *sym, CLPR_int quick_cmp);

extern NL_EQN_ptr CLPR_nl_eqn_top;
extern CLPR_int CLPR_inst_count[];
extern FILE *CLPR_cur_stream, *CLPR_error_stream;
extern SPNODE *CLPR_spred[];
extern SPNODE CLPR_sptab[];
extern FUNCTREC CLPR_functors[];
extern CLPR_unsigned *CLPR_code;
extern CLPR_int CLPR_startcode[], CLPR_endcode[];
extern FILE *CLPR_infile;
extern FILE *CLPR_outfile;
extern FILE *CLPR_see_input;
extern CLPR_int CLPR_see_input_t;
extern CLPR_int CLPR_TRACE, CLPR_CLAM_TRACE;

/*---------------------------------------------------------------------------*/


extern CLPR_int CLPR_sum_str (register char *s);
extern void CLPR_storename (char *s, CLPR_int val, CLPR_int narg);
extern char *CLPR_maketag (CLPR_int n, char *extra, char *str, CLPR_int arity);
extern void CLPR_assign_p_protected (CLPR_int name, CLPR_int flag);
extern void CLPR_init_solver_goal (void);
extern CLPR_int CLPR_free_eqn_nodes (EQN_ptr eqn);
extern CLPR_int CLPR_free_occ_nodes (OCC_ptr occ);
extern CLPR_int CLPR_free_nl_occ_nodes (NL_OCC_ptr occ);
void CLPR_solver_backtrack (CLPR_int i);
extern CLPR_int CLPR_wake_runable (void);
extern CLPR_int CLPR_p_dynamic (CLPR_int name);
extern void CLPR_make_runable (CLPR_int *t);
extern CLPR_int CLPR_is_ground (CLPR_int v, double *c_ptr);
CLPR_int CLPR_unify (CLPR_int *t1, CLPR_int *t2);
extern CLPR_int CLPR_s_evar (double c, CLPR_int v);
extern void CLPR_init_eqn (void);
extern CLPR_int CLPR_s_econ (double c);
extern CLPR_int CLPR_s_gt0 (double c);
extern CLPR_int CLPR_s_ge0 (double c);
extern CLPR_int CLPR_makenum (CLPR_int v, double c);
extern CLPR_int CLPR_solve_eqn_1 (CLPR_int v, double c);
extern CLPR_int CLPR_solve_eqn_2 (CLPR_int v0, CLPR_int v1, double c);
extern CLPR_int CLPR_s_mult (double c, CLPR_int v0, CLPR_int v1, CLPR_int v2);
extern CLPR_int CLPR_s_min (CLPR_int v0, CLPR_int v1, CLPR_int v2);
extern CLPR_int CLPR_s_max (CLPR_int v0, CLPR_int v1, CLPR_int v2);
extern CLPR_int CLPR_s_pow (CLPR_int v0, CLPR_int v1, CLPR_int v2);
extern CLPR_int CLPR_s_abs (CLPR_int v0, CLPR_int v1);
extern CLPR_int CLPR_s_eval (CLPR_int *t1, CLPR_int *t2);
extern CLPR_int CLPR_s_sin (CLPR_int v0, CLPR_int v1);
extern CLPR_int CLPR_s_cos (CLPR_int v0, CLPR_int v1);
extern CLPR_int CLPR_s_arcsin (CLPR_int v0, CLPR_int v1);
extern CLPR_int CLPR_s_arccos (CLPR_int v0, CLPR_int v1);
CLPR_int CLPR_retract_rule (CLPR_int *t, CLPR_int *pc, CLPR_int first);
CLPR_int CLPR_match_rule (CLPR_int *head, CLPR_int *body, CLPR_int *pc, CLPR_int first);
CLPR_int CLPR_emul_sys_pred (CLPR_int *pc);
extern CLPR_int CLPR_do_read (FILE *f, CLPR_int *t);
extern void CLPR_printterm (FILE *f, CLPR_int *t);
extern void CLPR_warn_err (char *s, ...);
extern CLPR_int CLPR_do_printf (FILE *F, CLPR_int *t, char *fmt, CLPR_int flush_flag);
extern CLPR_int CLPR_do_dump (FILE *f, CLPR_int *vars, char *names);
void CLPR_emul_dbg_entry (CLPR_int *pc);
CLPR_int CLPR_emul_halt (void);
extern void CLPR_restore_null_eqn (CLPR_int *i);
extern void CLPR_restore_eqn (CLPR_int *i);
extern void CLPR_restore_null_ineq (CLPR_int *i);
extern void CLPR_restore_ineq (CLPR_int *i);
extern void CLPR_restore_nl_eqn (CLPR_int *i);
extern CLPR_int CLPR_daemons_remaining (void);
extern void CLPR_readline (char *s, CLPR_int n);
extern CLPR_int CLPR_p_spy (CLPR_int name);
extern CLPR_int CLPR_p_compiled (CLPR_int name);
extern CLPR_int CLPR_p_start_hdrcode (CLPR_int name);
extern CLPR_int CLPR_p_protected (CLPR_int name);
CLPR_int CLPR_get_index_val (CLPR_int *t, double *c);
void CLPR_push_cp (void);
extern void CLPR_assign_p_last_rule (CLPR_int name, RULE_ptr rule);
extern void CLPR_assign_p_first_rule (CLPR_int name, RULE_ptr rule);
extern void CLPR_assign_p_compiled (CLPR_int name, CLPR_int flag);
extern void CLPR_assign_p_start_hdrcode (CLPR_int name, CLPR_int i);
extern void CLPR_assign_p_end_hdrcode (CLPR_int name, CLPR_int i);
extern void CLPR_free_rule (RULE_ptr r);
void CLPR_untrail_to (CLPR_int new_trtop);
extern CLPR_int CLPR_get_quoted_func (CLPR_int v);

void CLPR_init_engine(void)
{
PTERM *t;

	if (CLPR_HEAP_SZ < MIN_HEAP_SZ) CLPR_fatal("heap size too small");
	if (CLPR_LSTACK_SZ < MIN_LSTACK_SZ) CLPR_fatal("local stack size too small");
	if (CLPR_TRAIL_SZ < MIN_TRAIL_SZ) CLPR_fatal("trail size too small");

	CLPR_HEAP_START = CLPR_REGS_SZ;
	CLPR_MSTACK = CLPR_REGS_SZ + CLPR_HEAP_SZ + HEAP_SAFETY + CLPR_LSTACK_SZ + STACK_SAFETY;
	CLPR_LSTACK = CLPR_HEAP_SZ + HEAP_SAFETY + CLPR_REGS_SZ;
	CLPR_MTRAIL = CLPR_TRAIL_SZ + TRAIL_SAFETY;

	CLPR_stack = (CLPR_int *) CLPR_alloc_mem(CLPR_MSTACK);
	if (!CLPR_stack) CLPR_fatal("not enough memory to initialize clpr (stacks)");
	CLPR_trail = (CLPR_int **) CLPR_alloc_mem(CLPR_TRAIL_SZ);
	if (!CLPR_trail) CLPR_fatal("not enough memory to initialize clpr (trail)");

	CLPR_pcode = FAIL_ADDR;
	t = CLPR_lookup_name("$fail", CLPR_sum_str("$fail"));
	CLPR_storename(CLPR_maketag(0, "", "$fail", 0), t->val, 0);
	CLPR_code[FAIL_ADDR] = FFAIL;
	CLPR_assign_p_protected(t->val, TRUE);

	CLPR_pcode = GFAIL_ADDR;
	t = CLPR_lookup_name("$fail", CLPR_sum_str("$gfail"));
	CLPR_storename(CLPR_maketag(0, "", "$gfail", 0), t->val, 0);
	CLPR_code[GFAIL_ADDR] = GFAIL;
	CLPR_assign_p_protected(t->val, TRUE);

	CLPR_swt_offsets[tagnum(CLPR_TAG_NIL)] =  1;
	CLPR_swt_offsets[tagnum(CLPR_TAG_CONS)] = 2;
	CLPR_swt_offsets[tagnum(TAG_CON)] =  3;
	CLPR_swt_offsets[tagnum(TAG_STR)] =  4;
	CLPR_swt_offsets[tagnum(TAG_PAR)] =  5;

	true_atom = addtag(TAG_CON, TRUEVAL);
}

void CLPR_init_unify(void)
{
	CLPR_tag_pair[0][0] = 21;
	CLPR_tag_pair[0][tagnum(TAG_CON)] = 1;
	CLPR_tag_pair[0][tagnum(CLPR_TAG_NIL)] = 2;
	CLPR_tag_pair[0][tagnum(TAG_PAR)] = 3;
	CLPR_tag_pair[0][tagnum(TAG_NUM)] = 4;
	CLPR_tag_pair[0][tagnum(CLPR_TAG_CONS)] = 5;
	CLPR_tag_pair[0][tagnum(TAG_STR)] = 6;
	CLPR_tag_pair[0][tagnum(TAG_DAEMON)] = 23;
	CLPR_tag_pair[tagnum(TAG_CON)][0] = 7;
	CLPR_tag_pair[tagnum(CLPR_TAG_NIL)][0] = 8;
	CLPR_tag_pair[tagnum(TAG_PAR)][0] = 9;
	CLPR_tag_pair[tagnum(TAG_NUM)][0] = 10;
	CLPR_tag_pair[tagnum(CLPR_TAG_CONS)][0] = 11;
	CLPR_tag_pair[tagnum(TAG_STR)][0] = 12;
	CLPR_tag_pair[tagnum(TAG_DAEMON)][0] = 22;
	CLPR_tag_pair[tagnum(TAG_DAEMON)][tagnum(TAG_CON)] = 22;
	CLPR_tag_pair[tagnum(TAG_DAEMON)][tagnum(CLPR_TAG_NIL)] = 22;
	CLPR_tag_pair[tagnum(TAG_DAEMON)][tagnum(TAG_PAR)] = 22;
	CLPR_tag_pair[tagnum(TAG_DAEMON)][tagnum(TAG_NUM)] = 22;
	CLPR_tag_pair[tagnum(TAG_DAEMON)][tagnum(CLPR_TAG_CONS)] = 22;
	CLPR_tag_pair[tagnum(TAG_DAEMON)][tagnum(TAG_STR)] = 22;
	CLPR_tag_pair[tagnum(TAG_CON)][tagnum(TAG_DAEMON)] = 23;
	CLPR_tag_pair[tagnum(CLPR_TAG_NIL)][tagnum(TAG_DAEMON)] = 23;
	CLPR_tag_pair[tagnum(TAG_PAR)][tagnum(TAG_DAEMON)] = 23;
	CLPR_tag_pair[tagnum(TAG_NUM)][tagnum(TAG_DAEMON)] = 23;
	CLPR_tag_pair[tagnum(CLPR_TAG_CONS)][tagnum(TAG_DAEMON)] = 23;
	CLPR_tag_pair[tagnum(TAG_STR)][tagnum(TAG_DAEMON)] = 23;
	CLPR_tag_pair[tagnum(TAG_PAR)][tagnum(TAG_PAR)] = 13;
	CLPR_tag_pair[tagnum(TAG_CON)][tagnum(TAG_CON)] = 14;
	CLPR_tag_pair[tagnum(CLPR_TAG_NIL)][tagnum(CLPR_TAG_NIL)] = 15;
	CLPR_tag_pair[tagnum(TAG_STR)][tagnum(TAG_STR)] = 16;
	CLPR_tag_pair[tagnum(CLPR_TAG_CONS)][tagnum(CLPR_TAG_CONS)] = 17;
	CLPR_tag_pair[tagnum(TAG_NUM)][tagnum(TAG_NUM)] = 18;
	CLPR_tag_pair[tagnum(TAG_NUM)][tagnum(TAG_PAR)] = 19; 
	CLPR_tag_pair[tagnum(TAG_PAR)][tagnum(TAG_NUM)] = 20;
	CLPR_tag_pair[tagnum(TAG_DAEMON)][tagnum(TAG_DAEMON)] = 24;
}

void CLPR_init_rungoal(CLPR_int interactive)
{
CLPR_int *tmp1;
	tmp1 = CLPR_stacktop = CLPR_stacktop + CPFRAME_HEIGHT;
	tmp1[NEXTEP] = (CLPR_int) (CLPR_code + GFAIL_ADDR);
	tmp1[PREVCP] = (CLPR_int) CLPR_lastcp;
	tmp1[OHEAPTOP] = (CLPR_int) CLPR_heaptop;
	tmp1[OTRTOP] = CLPR_trtop;
	tmp1[CPRETPT] = (CLPR_int) CLPR_retreg;
	tmp1[OCURAR] = (CLPR_int) CLPR_curar;
	tmp1[NUMPARS] = 0;
	tmp1[SOLVER_ID] = CLPR_solver_id;
	tmp1[SLACK_ID] = CLPR_slack_id;
	CLPR_lastcp = CLPR_stacktop;
	CLPR_safeheap = CLPR_heaptop;
	CLPR_stamp++;
	if (interactive) CLPR_first_lastcp = CLPR_lastcp;
}

/*---------------------------------------------------------------------------*/

void CLPR_save_context(void)
{
CLPR_int i;
	if (CLPR_savep >= MAX_SAVE_CONTEXT - 1) CLPR_fatal("too many recursive rungoals");
	CLPR_save[CLPR_savep].retreg = CLPR_retreg;
	CLPR_save[CLPR_savep].curar = CLPR_curar;
	CLPR_save[CLPR_savep].stacktop = CLPR_stacktop;
	CLPR_save[CLPR_savep].heaptop = CLPR_heaptop;
	CLPR_save[CLPR_savep].trtop = CLPR_trtop;
	CLPR_save[CLPR_savep].lastcp = CLPR_lastcp;
	CLPR_save[CLPR_savep].solver_id = CLPR_solver_id;
	CLPR_save[CLPR_savep].slack_id = CLPR_slack_id;
	CLPR_save[CLPR_savep].nlin_count = CLPR_nlin_count;
	CLPR_save[CLPR_savep].nl_eqn_top = CLPR_nl_eqn_top;
	CLPR_save[CLPR_savep].runable = CLPR_runable;
	CLPR_save[CLPR_savep].have_runable = CLPR_have_runable;
	CLPR_savep++;
	CLPR_init_solver_goal();
}

void CLPR_restore_context(void)
{
register CLPR_int i, j;
TAB_HDR_NODE_ptr hdr;
	CLPR_savep--;
	for (i = CLPR_trtop - 1; i >= CLPR_save[CLPR_savep].trtop; i--) {
		j = (CLPR_int) CLPR_trail[i];
		switch (tagnum(j)) {
				case NUM_BTK_INEQ:
#ifdef STATS
	CLPR_s_count[134]++;
#endif	
				case NUM_BTK_EQN:
					CLPR_free_eqn_nodes((EQN_ptr)pmask(j));
					break;
				case NUM_BTK_NLIN:
					CLPR_free_nlin_nodes((NL_EQN_ptr)pmask(j));
					break;
		}
	}
	for (i = CLPR_save[CLPR_savep].solver_id + 1; i <= CLPR_solver_id; i++) {
		if (hdr = CLPR_tableau[i]) {
				CLPR_free_eqn_nodes(hdr->eqn);
				CLPR_free_occ_nodes(hdr->eq_occ);
				CLPR_free_occ_nodes(hdr->ineq_occ);
				CLPR_free_nl_occ_nodes(hdr->nlin_occ);
				free_HDR(hdr);
				CLPR_tableau[i] = NULL;
		}
	}
	for (i = CLPR_save[CLPR_savep].slack_id - 1; i >= CLPR_slack_id; i--) {
		if (hdr = CLPR_tableau[i]) {
#ifdef STATS
	if (hdr->eqn) CLPR_s_count[135]++;
	if (hdr->next) CLPR_s_count[136]++;
#endif
				CLPR_free_eqn_nodes(hdr->eqn);
				CLPR_free_occ_nodes(hdr->eq_occ);
				CLPR_free_occ_nodes(hdr->ineq_occ);
				CLPR_free_nl_occ_nodes(hdr->nlin_occ);
				CLPR_free_eqn_nodes((EQN_ptr) hdr->next);
				free_HDR(hdr);
				CLPR_tableau[i] = NULL;
		}
	}
	CLPR_retreg = CLPR_save[CLPR_savep].retreg;
	CLPR_curar = CLPR_save[CLPR_savep].curar;
	CLPR_stacktop = CLPR_save[CLPR_savep].stacktop;
	CLPR_heaptop = CLPR_save[CLPR_savep].heaptop;
	CLPR_trtop = CLPR_save[CLPR_savep].trtop;
	CLPR_lastcp = CLPR_save[CLPR_savep].lastcp;
	CLPR_solver_id = CLPR_save[CLPR_savep].solver_id;
	CLPR_slack_id = CLPR_save[CLPR_savep].slack_id;
	CLPR_nlin_count = CLPR_save[CLPR_savep].nlin_count;
	CLPR_nl_eqn_top = CLPR_save[CLPR_savep].nl_eqn_top;
	CLPR_runable = CLPR_save[CLPR_savep].runable;
	CLPR_have_runable = CLPR_save[CLPR_savep].have_runable;
#ifdef STATS
	if (CLPR_savep == 0) CLPR_check_solver_malloc_space();
#endif
}


void CLPR_int_clean(void)
{
	while (CLPR_savep) CLPR_restore_context();
}

void CLPR_first_engine(void)
{
	CLPR_heaptop = s = CLPR_heap = CLPR_stack + CLPR_HEAP_START;
	CLPR_stacktop = CLPR_curar = CLPR_lstack = CLPR_stack + CLPR_LSTACK;
	maxlstack = CLPR_stack + CLPR_MSTACK - STACK_SAFETY;
	maxheap = CLPR_stack + CLPR_LSTACK - HEAP_SAFETY;
	maxtrail = var2pos((CLPR_trail + CLPR_TRAIL_SZ - TRAIL_SAFETY), CLPR_trail);
	CLPR_regs = CLPR_stack + REGS;
	CLPR_safeheap = CLPR_heap;
	CLPR_stamp = 1;
	CLPR_lastcp = NULL;
	CLPR_savep = 0;
	CLPR_tfp = &tf;
#if defined(SYS5) || defined(OS2V2)
	times(&time1);
#endif
#if defined(BSD) && !defined(OS2V2)
	getrusage(RUSAGE_SELF,&usg);
	time1 = usg.ru_utime;
#endif
}

/********************************** emulate! *********************************/
CLPR_int CLPR_rungoal(CLPR_int ipc, CLPR_int interactive)
/*****************************************************************************/
        
                
{
register CLPR_int *pc;
register CLPR_int *Regs;
register CLPR_int *tmp1, *tmp2;
register CLPR_int mode_read, *s;
register CLPR_int ti;
CLPR_int ti1, ti2;
CLPR_int *tmp3, *tmp4;
CLPR_int *Regs1, *Regs2, *Regs3;
double c;
float f;
CLPR_int *cp_for_cut;
CLPR_int first_dyn_call = FALSE;

	CLPR_runable = CLPR_heaptop++;
	*CLPR_runable = addtag(CLPR_TAG_NIL, EMPTYLIST);
	pc = ptrpc(ipc);
	Regs = CLPR_regs;
	Regs1 = CLPR_regs + 1;
	Regs2 = CLPR_regs + 2;
	Regs3 = CLPR_regs + 3;
	CLPR_init_rungoal(interactive);
	goto LOOP;

/***************/
/***/ FAIL: /***/
/***************/

#ifdef CTRACE
	if (CLPR_CLAM_TRACE) fprintf(CLPR_outfile, "FAILING\n");
#endif
	CLPR_runable = CLPR_heaptop++;
		*CLPR_runable = addtag(CLPR_TAG_NIL, EMPTYLIST);
	ti = (CLPR_int) mask(*(CLPR_lastcp + OTRTOP));
	for (ti1 = CLPR_trtop - 1; ti1 >= ti; ti1--) {
		ti2 = (CLPR_int) CLPR_trail[ti1];
		if (CLPR_tag(ti2) == BTK_BIND) make_unbound_var((CLPR_int *) pmask(ti2));
#ifdef CYCLIC_BIND
		else if (CLPR_tag(ti2) == BTK_TEST) 
		  make_bound_var((CLPR_int *) pmask(ti2), **((CLPR_int **) pmask(ti2)));
#endif
#ifdef DESTRUCTIVE_UPDATE
		else if (CLPR_tag(ti2) == BTK_TEST2) 
			make_bound_var((CLPR_int *) pmask(ti2), CLPR_trail[--ti1]);
#endif
		else CLPR_solver_backtrack(ti2);
	}

	ti2 = mask(*(CLPR_lastcp + NUMPARS));
	tmp1 = CLPR_lastcp + ti2;
	tmp2 = Regs + ti2;
	while (tmp2 > Regs) *tmp2-- = *tmp1--;
	pc = (CLPR_int *) pmask(*tmp1);
	tmp1 -= 2;
	CLPR_heaptop = (CLPR_int *) pmask(*tmp1--);
	CLPR_trtop = (CLPR_int) mask(*tmp1--); 
	CLPR_retreg = (CLPR_int *) pmask(*tmp1--); 
	CLPR_curar = (CLPR_int *) pmask(*tmp1); 
	tmp1 -= 2; 
	CLPR_solver_id = (CLPR_int) mask(*tmp1--); 
	CLPR_slack_id = (CLPR_int) mask(*tmp1--); 
	CLPR_stamp--;

/***************/
/***/ LOOP: /***/ 
/***************/
#ifdef STATS
	CLPR_s_count[299]++; 
	CLPR_inst_count[*pc]++;
#endif

#ifdef CTRACE
	if (CLPR_CLAM_TRACE) CLPR_clam_trace(intpc(pc));
#endif

	/************/
	switch (*pc) {
	/************/

	caseop(BADINST): CLPR_fatal("Zero opcode");
	caseop(NOOP):
		pc++;
		goto LOOP;
	caseop(CALLP):
#ifdef CHECK_WAKEUP
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
#endif
		CheckHeap();
		CheckTrail();
		CLPR_retreg = pc + 2;
		pc = (CLPR_int *) *(pc + 1);
		goto LOOP;
	caseop(CALLI):
#ifdef CHECK_WAKEUP
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
#endif
		CheckHeap();
		CheckTrail();
		CLPR_retreg = pc + 2;
		tmp1 = pc;
		if (pc = CLPR_val_to_pc(pc[1])) {
			tmp1[0] = CALLP;
			tmp1[1] = (CLPR_int) pc;
			goto LOOP;
		} else {
			CLPR_warning("%s/%d not defined, failing",
				 string_of(tmp1[1]), arity_of(tmp1[1]));
			goto FAIL;
		}
	caseop(JUMP):
#ifdef CHECK_WAKEUP
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
#endif
		CheckHeap();
		CheckTrail();
		pc = (CLPR_int *) *(pc + 1);
		goto LOOP;
	caseop(JUMPI):
#ifdef CHECK_WAKEUP
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
#endif
		CheckHeap();
		CheckTrail();
		tmp1 = pc;
		if (pc = CLPR_val_to_pc(pc[1])) {
			tmp1[0] = JUMP;
			tmp1[1] = (CLPR_int) pc;
			goto LOOP;
		} else {
			CLPR_warning("%s/%d not defined, failing",
				 string_of(tmp1[1]), arity_of(tmp1[1]));
			goto FAIL;
		}
	caseop(PROCEED):
		pc = CLPR_retreg;
		goto LOOP;
	caseop(CALLX):
		CLPR_retreg = pc + 1;
	caseop(JUMPX):
#ifdef CHECK_WAKEUP
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
#endif
		CheckHeap();
		CheckTrail();
		tmp1 = tmp2 = Regs1;
		CLPR_deref(tmp1);
		ti = mask(*tmp1);
		if (CLPR_tag(*tmp1) == TAG_CON) {
			if (!(pc = CLPR_val_to_pc(ti))) {
				CLPR_warning("call/1: %s not defined, failing", string_of(ti));
				goto FAIL;
			}
		} else if (CLPR_tag(*tmp1) == TAG_STR) {
			if (!(pc = CLPR_val_to_pc(ti))) {
				CLPR_warning("call/1: %s not defined, failing", string_of(ti));
				goto FAIL;
			}
			ti1 = arity_of(ti);
			do {
				tmp1++;
				*tmp2++ = (CLPR_int) tmp1;
			} while (--ti1);
		} else if (CLPR_tag(*tmp1) == CLPR_TAG_CONS) {
			if (!(pc = CLPR_val_to_pc(DOT))) {
				CLPR_warning("call/1: `.' not defined, failing", 0);
				goto FAIL;
			}
			Regs[1] = (CLPR_int) pmask(*tmp1);
			Regs[2] = (CLPR_int) (((CLPR_int *) pmask(*tmp1)) + 1);
		} else if (CLPR_tag(*tmp1) == CLPR_TAG_NIL) {
			if (!(pc = CLPR_val_to_pc(EMPTYLIST))) {
				CLPR_warning("call/1: [] not defined, failing", 0);
				goto FAIL;
			}
		} else CLPR_runtime_error("bad argument to call/1");
		if (CLPR_p_dynamic(ti)) first_dyn_call = TRUE;
		goto LOOP;
	caseop(UNI2TTVAR):
		if (mode_read) {
				make_bound_var(Regs + *(pc + 1), s++);
				make_bound_var(Regs + *(pc + 2), s++);
				pc += 3;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_bound_var(Regs + *(pc + 1), s++);
				make_unbound_var(s);
				make_bound_var(Regs + *(pc + 2), s++);
				CLPR_heaptop = s;
				pc += 3;
				goto LOOP;
		}
	caseop(UNI2PTVAR):
		if (mode_read) {
				make_bound_var(CLPR_curar + *(pc + 1), s++);
				make_bound_var(Regs + *(pc + 2), s++);
				pc += 3;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_bound_var(CLPR_curar + *(pc + 1), s++);
				make_unbound_var(s);
				make_bound_var(Regs + *(pc + 2), s++);
				CLPR_heaptop = s;
				pc += 3;
				goto LOOP;
		}
	caseop(UNI2TPVAR):
		if (mode_read) {
				make_bound_var(Regs + *(pc + 1), s++);
				make_bound_var(CLPR_curar + *(pc + 2), s++);
				pc += 3;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_bound_var(Regs + *(pc + 1), s++);
				make_unbound_var(s);
				make_bound_var(CLPR_curar + *(pc + 2), s++);
				CLPR_heaptop = s;
				pc += 3;
				goto LOOP;
		}
	caseop(UNI2PPVAR):
		if (mode_read) {
				make_bound_var(CLPR_curar + *(pc + 1), s++);
				make_bound_var(CLPR_curar + *(pc + 2), s++);
				pc += 3;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_bound_var(CLPR_curar + *(pc + 1), s++);
				make_unbound_var(s);
				make_bound_var(CLPR_curar + *(pc + 2), s++);
				CLPR_heaptop = s;
				pc += 3;
				goto LOOP;
		}
	case (unsigned char) UNI2TVVAR:
		if (mode_read) {
				make_bound_var(Regs + *(pc + 1), s);
				s += 2;
				pc += 2;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_bound_var(Regs + *(pc + 1), s);
				make_unbound_var(s+1);
				s += 2;
				CLPR_heaptop = s;
				pc += 2;
				goto LOOP;
		}
	case (unsigned char) UNI2VTVAR:
		if (mode_read) {
				make_bound_var(Regs + *(pc + 1), s+1);
				s += 2;
				pc += 2;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_unbound_var(s+1);
				make_bound_var(Regs + *(pc + 1), s+1);
				s += 2;
				CLPR_heaptop = s;
				pc += 2;
				goto LOOP;
		}
	case (unsigned char) UNI2PVVAR:
		if (mode_read) {
				make_bound_var(CLPR_curar + *(pc + 1), s);
				s += 2;
				pc += 2;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_bound_var(CLPR_curar + *(pc + 1), s);
				make_unbound_var(s+1);
				s += 2;
				CLPR_heaptop = s;
				pc += 2;
				goto LOOP;
		}
	case (unsigned char) UNI2VPVAR:
		if (mode_read) {
				make_bound_var(CLPR_curar + *(pc + 1), s+1);
				s += 2;
				pc += 2;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_unbound_var(s+1);
				make_bound_var(CLPR_curar + *(pc + 1), s+1);
				s += 2;
				CLPR_heaptop = s;
				pc += 2;
				goto LOOP;
		}
	caseop(UNILIST):
		if (mode_read) {
				tmp1 = s;
				CLPR_deref(tmp1);
				if (CLPR_tag(*tmp1) == CLPR_TAG_CONS) {
					s = (CLPR_int *) pmask(*tmp1);
					pc++;
					goto LOOP;
				} else if (is_unbound_var(*tmp1)) {
					*tmp1 = addtag(CLPR_TAG_CONS, CLPR_heaptop);
					s = CLPR_heaptop;
					mode_read = FALSE;
					push_check_trail(BTK_BIND, tmp1);
					pc++;
					goto LOOP;
				} else if (is_daemon(tmp1)) {
					make_bound_var(tmp1, CLPR_heaptop);
					*CLPR_heaptop = addtag(CLPR_TAG_CONS, CLPR_heaptop + 1);
					CLPR_heaptop++;
					s = CLPR_heaptop;
					mode_read = FALSE;
					push_check_trail(BTK_DBIND, tmp1);
					CLPR_make_runable(tmp1 + 1);
					pc++;
					goto LOOP;
				} else goto FAIL;
		} else {
				CLPR_heaptop++;
				*s = addtag(CLPR_TAG_CONS, CLPR_heaptop);
				s = CLPR_heaptop;
				pc++;
				goto LOOP;
		}
	caseop(UNISTR):
		if (mode_read) {
				tmp1 = s;
				CLPR_deref(tmp1);
				if (*tmp1 == addtag1(TAG_STR, *(pc + 1))) {
					s = tmp1 + 1;
					pc += 2;
					goto LOOP;
				} else if (is_unbound_var(*tmp1)) {
					make_bound_var(tmp1, (CLPR_int) CLPR_heaptop);
					*CLPR_heaptop = addtag1(TAG_STR, *(pc + 1));
					s = ++CLPR_heaptop;
					mode_read = FALSE;
					push_check_trail(BTK_BIND, tmp1);
					pc += 2;
					goto LOOP;
				} else if (is_daemon(tmp1)) {
					CLPR_make_runable(tmp1 + 1);
					make_bound_var(tmp1, (CLPR_int) CLPR_heaptop);
					*CLPR_heaptop = addtag1(TAG_STR, *(pc + 1));
					s = ++CLPR_heaptop;
					mode_read = FALSE;
					push_check_trail(BTK_DBIND, tmp1);
					pc += 2;
					goto LOOP;
				} else goto FAIL;
		} else {
				make_bound_var(s, (CLPR_int) (s + 1));
				*(s + 1)  = addtag1(TAG_STR, *(pc + 1));
				s += 2;
				CLPR_heaptop = s;
				pc += 2;
				goto LOOP;
		}
	caseop(ENTERSWTYPE):
#ifdef CHECK_WAKEUP
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
#endif
		CheckHeap();
		CheckTrail();	/* Fall thru */
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		if (CLPR_tag(ti = *tmp1) == CLPR_TAG_CONS) {
				s = (CLPR_int *) pmask(ti);
				mode_read = TRUE;
				pc = (CLPR_int *) *(pc + 2);
				goto LOOP;
		} else if (is_unbound_var(ti) || is_daemon(tmp1)) {
				pc = (CLPR_int *) pc[6];
				goto LOOP;
		} else {
				mode_read = TRUE;
				pc = (CLPR_int *) *(pc + CLPR_swt_offsets[tagnum(ti)]);
				goto LOOP;
		}
	caseop(SWITCHONTYPE):
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		if (CLPR_tag(ti = *tmp1) == CLPR_TAG_CONS) {
				s = (CLPR_int *) pmask(ti);
				mode_read = TRUE;
				pc = (CLPR_int *) *(pc + 2);
				goto LOOP;
		} else if (is_unbound_var(ti) || is_daemon(tmp1)) {
				pc += 6;
				goto LOOP;
		} else {
				mode_read = TRUE;
				pc = (CLPR_int *) *(pc + CLPR_swt_offsets[tagnum(ti)]);
				goto LOOP;
		}
	caseop(ENTERSWCONFUNC):
#ifdef CHECK_WAKEUP
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
#endif
		CheckHeap();
		CheckTrail();	/* Fall thru */
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		if (CLPR_tag(ti = *tmp1) == TAG_STR) {
				pc = (CLPR_int *) *(pc + 2);
				mode_read = TRUE;
				goto LOOP;
		} else if (CLPR_tag(ti) == TAG_CON) {
				pc = (CLPR_int *) *(pc + 1);
				mode_read = TRUE;
				goto LOOP;
		} else {
				pc = (CLPR_int *) pc[3];
				goto LOOP;
		}
	caseop(SWITCHONCONFUNC):
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		if (CLPR_tag(ti = *tmp1) == TAG_STR) {
				pc = (CLPR_int *) *(pc + 2);
				mode_read = TRUE;
				goto LOOP;
		} else if (CLPR_tag(ti) == TAG_CON) {
				pc = (CLPR_int *) *(pc + 1);
				mode_read = TRUE;
				goto LOOP;
		} else {
				pc += 3;
				goto LOOP;
		}
	caseop(SWITCHONNIL):
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == CLPR_TAG_NIL || is_unbound_var(*tmp1)
				 || is_daemon(tmp1)) {
				pc += 2;
				goto LOOP;
		} 
		pc = (CLPR_int *) *(pc + 1);
		goto LOOP;
	caseop(SWITCHONDOT):
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == CLPR_TAG_CONS || is_unbound_var(*tmp1)
				 || is_daemon(tmp1)) {
				pc += 2;
				goto LOOP;
		} 
		pc = (CLPR_int *) *(pc + 1);
		goto LOOP;
	caseop(ENTERSWCON):
#ifdef CHECK_WAKEUP
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
#endif
		CheckHeap();
		CheckTrail();
		pc = (INSTptr) pc[1];	/* Fall thru */
	caseop(SWITCHONCON):
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		ti = *tmp1;
		tmp2 = (CLPR_int *) *(pc + 1);
		if (CLPR_tag(ti) == TAG_CON) {
				tmp1 = pc + 2;
				do {
					if (*tmp1 == ti) { pc = (CLPR_int *) *(tmp1 + 1); goto LOOP; }
					tmp1 += 2;
				} while (tmp1 != tmp2);
				pc = (CLPR_int *) *(tmp1 + 1);
				goto LOOP;
		}
		if (is_unbound_var(ti) || (CLPR_tag(ti) == TAG_DAEMON)) {
				pc = tmp2 + 2;
				goto LOOP;
		}
		pc = (CLPR_int *) *(tmp2 + 1);
		goto LOOP;
	caseop(SWITCH2ONCON): /*** needs ti from swtype, swcfunc ***/
		tmp1 = pc + 2;
		tmp2 = (CLPR_int *) *(pc + 1);
		do {
				if (*tmp1 == ti) { pc = (CLPR_int *) *(tmp1 + 1); goto LOOP; }
				tmp1 += 2;
		} while (tmp1 != tmp2);				
		pc = (CLPR_int *) *(tmp2 + 1);
		goto LOOP;
	caseop(ENTERSWFUNC):
#ifdef CHECK_WAKEUP
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
#endif
		CheckHeap();
		CheckTrail();
		pc = (INSTptr) pc[1];	/* Fall thru */
	caseop(SWITCHONFUNC):
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		ti = *tmp1;
		tmp2 = (CLPR_int *) *(pc + 1);
		if (CLPR_tag(ti) == TAG_STR) {
				tmp3 = pc + 2;
				do {
					if (*tmp3 == ti) { 
						pc = (CLPR_int *) *(tmp3 + 1); 
						s = tmp1 + 1;
						mode_read = TRUE;
						goto LOOP; 
					}
					tmp3 += 2;
				} while (tmp3 != tmp2);
				pc = (CLPR_int *) *(tmp2 + 1);
				goto LOOP;
		}
		if (is_unbound_var(ti) || (CLPR_tag(ti) == TAG_DAEMON)) {
				pc = tmp2 + 2;
				goto LOOP;
		}
		pc = (CLPR_int *) *(tmp2 + 1);
		goto LOOP;
	caseop(SWITCH2ONFUNC): /*** needs tmp1 from swtype,swcfunc ***/
		ti = *tmp1;
		tmp3 = pc + 2;
		tmp2 = (CLPR_int *) *(pc + 1);
		do {
				if (*tmp3 == ti) { 
					pc = (CLPR_int *) *(tmp3 + 1); 
					s = tmp1 + 1;
					goto LOOP; 
				}
				tmp3 += 2;
		} while (tmp3 != tmp2);				
		pc = (CLPR_int *) *(tmp2 + 1);
		goto LOOP;
	caseop(ENTERSWREAL):
#ifdef CHECK_WAKEUP
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
#endif
		CheckHeap();
		CheckTrail();
		pc = (INSTptr) pc[1];	/* Fall thru */
	caseop(SWITCHONREAL):
	caseop(SWITCH2ONREAL):
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		ti = *tmp1;
		tmp2 = (CLPR_int *) *(pc + 1);
		if (CLPR_tag(ti) == TAG_PAR) {
				if (CLPR_is_ground(mask(ti), &c)) {
					f = (float) c;
					tmp1 = pc + 2;
					do {
						if (is_equal(CLPR_int_to_float(*tmp1), f)) {
							pc = (CLPR_int *) *(tmp1 + 1);
							goto LOOP;
						}
						tmp1 += 2;
					} while (tmp1 != tmp2);
					pc = (CLPR_int *) *(tmp1 + 1);
					goto LOOP;
				}
				pc = tmp2 + 2;
				goto LOOP;
		} else if (is_unbound_var(ti) || (CLPR_tag(ti) == TAG_DAEMON)) {
				pc = tmp2 + 2;
				goto LOOP;
		} else {
				pc = (CLPR_int *) *(tmp2 + 1);
				goto LOOP;
		}
	caseop(GETPVAR):
		tmp1 = CLPR_curar + *(pc + 1);
		*tmp1 = Regs[*(pc + 2)];
		pc += 3;
		goto LOOP;
	caseop(GETPVAL):
		if (CLPR_unify(Regs + *(pc + 2), CLPR_curar + *(pc + 1))) {
				pc += 3;
				goto LOOP;
		} else goto FAIL;
	caseop(GETTVAR):
		Regs[*(pc + 1)] = Regs[*(pc + 2)];
		pc += 3;
		goto LOOP;
	caseop(GETTVAL):
		if (CLPR_unify(Regs + *(pc + 1), Regs + *(pc + 2))) {
				pc += 3;
				goto LOOP;
		} else goto FAIL;
	caseop(GETSTR):
		tmp1 = (CLPR_int *) (Regs + *(pc + 2));
		CLPR_deref(tmp1);
		if (*tmp1 == addtag1(TAG_STR, *(pc + 1))) {
				mode_read = TRUE;
				s = tmp1 + 1;
				pc += 3;
				goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
				mode_read = FALSE;
				make_bound_var(tmp1, (CLPR_int) CLPR_heaptop);
				push_check_trail(BTK_BIND, tmp1);
				*CLPR_heaptop = addtag1(TAG_STR, *(pc + 1));
				s = ++CLPR_heaptop;
				pc += 3;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				CLPR_make_runable(tmp1 + 1);
				mode_read = FALSE;
				make_bound_var(tmp1, (CLPR_int) CLPR_heaptop);
				push_check_trail(BTK_DBIND, tmp1);
				*CLPR_heaptop = addtag1(TAG_STR, *(pc + 1));
				s = ++CLPR_heaptop;
				pc += 3;
				goto LOOP;
		} else goto FAIL;
	caseop(GETLIST):
		tmp1 = (CLPR_int *) (Regs + *(pc + 1));
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == CLPR_TAG_CONS) {
				s = (CLPR_int *) pmask((CLPR_int) (*tmp1));
				mode_read = TRUE;
				pc += 2;
				goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
				*tmp1 = addtag(CLPR_TAG_CONS, CLPR_heaptop);
				s = CLPR_heaptop;
				push_check_trail(BTK_BIND, tmp1);
				mode_read = FALSE;
				pc += 2;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				CLPR_make_runable(tmp1 + 1);
				make_bound_var(tmp1, CLPR_heaptop);
				*CLPR_heaptop = addtag(CLPR_TAG_CONS, CLPR_heaptop + 1);
				CLPR_heaptop++;
				s = CLPR_heaptop;
				push_check_trail(BTK_DBIND, tmp1);
				mode_read = FALSE;
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GETCON):
		ti = *(pc + 1);
		tmp1 = (CLPR_int *) (Regs + *(pc + 2));
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == TAG_CON) {
			if (mask(*tmp1) == ti) {
				pc += 3;
				goto LOOP;
			} else goto FAIL;
		} else if (is_unbound_var(*tmp1)) {
			*tmp1 = addtag(TAG_CON, ti);
			push_check_trail(BTK_BIND, tmp1);
			pc += 3;
			goto LOOP;
		} else if (is_daemon(tmp1)) {
			make_bound_var(tmp1, CLPR_heaptop);
			*CLPR_heaptop = addtag(TAG_CON, ti);
			CLPR_heaptop++;
			push_check_trail(BTK_DBIND, tmp1);
			CLPR_make_runable(tmp1 + 1);
			pc += 3;
			goto LOOP;
		} else goto FAIL;
	caseop(GETNIL):
		tmp1 = (CLPR_int *) (Regs + *(pc + 1));
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == CLPR_TAG_NIL) {
			pc += 2;
			goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
			*tmp1 = addtag(CLPR_TAG_NIL, EMPTYLIST);
			push_check_trail(BTK_BIND, tmp1);
			pc += 2;
			goto LOOP;
		} else if (is_daemon(tmp1)) {
			make_bound_var(tmp1, CLPR_heaptop);
			*CLPR_heaptop = addtag(CLPR_TAG_NIL, EMPTYLIST);
			CLPR_heaptop++;
			push_check_trail(BTK_DBIND, tmp1);
			CLPR_make_runable(tmp1 + 1);
			pc += 2;
			goto LOOP;
		} else goto FAIL;
	caseop(GET1PVAR):	          /*.................. get1 ...*/ 
		*(CLPR_curar + 1) = Regs[*(pc + 1)];
		pc += 2;
		goto LOOP;
	caseop(GET1PVAL):
		if (CLPR_unify(Regs + *(pc + 1), CLPR_curar + 1)) {
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET1TVAR):
		*Regs1 = Regs[*(pc + 1)];
		pc += 2;
		goto LOOP;
	caseop(GET1TVAL):
		if (CLPR_unify(Regs1, Regs + *(pc + 1))) {
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET1STR):
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		if (*tmp1 == addtag1(TAG_STR, *(pc + 1))) {
				mode_read = TRUE;
				s = tmp1 + 1;
				pc += 2;
				goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
				mode_read = FALSE;
				make_bound_var(tmp1, (CLPR_int) CLPR_heaptop);
				push_check_trail(BTK_BIND, tmp1);
				*CLPR_heaptop = addtag1(TAG_STR, *(pc + 1));
				s = ++CLPR_heaptop;
				pc += 2;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				CLPR_make_runable(tmp1 + 1);
				mode_read = FALSE;
				make_bound_var(tmp1, (CLPR_int) CLPR_heaptop);
				push_check_trail(BTK_DBIND, tmp1);
				*CLPR_heaptop = addtag1(TAG_STR, *(pc + 1));
				s = ++CLPR_heaptop;
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET1LIST):
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == CLPR_TAG_CONS) {
				s = (CLPR_int *) pmask((CLPR_int) (*tmp1));
				mode_read = TRUE;
				pc++;
				goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
				*tmp1 = addtag(CLPR_TAG_CONS, CLPR_heaptop);
				s = CLPR_heaptop;
				push_check_trail(BTK_BIND, tmp1);
				mode_read = FALSE;
				pc++;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				CLPR_make_runable(tmp1 + 1);
				make_bound_var(tmp1, CLPR_heaptop);
				*CLPR_heaptop = addtag(CLPR_TAG_CONS, CLPR_heaptop + 1);
				CLPR_heaptop++;
				s = CLPR_heaptop;
				push_check_trail(BTK_DBIND, tmp1);
				mode_read = FALSE;
				pc++;
				goto LOOP;
		} else goto FAIL;
	caseop(GET1CON):
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == TAG_CON) {
				if (mask(*tmp1) == *(pc + 1)) {
					pc += 2;
					goto LOOP;
				} else goto FAIL;
		} else if (is_unbound_var(*tmp1)) {
				*tmp1 = addtag(TAG_CON, *(pc + 1));
				push_check_trail(BTK_BIND, tmp1);
				pc += 2;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				make_bound_var(tmp1, CLPR_heaptop);
				*CLPR_heaptop = addtag(TAG_CON, *(pc + 1));
				CLPR_heaptop++;
				push_check_trail(BTK_DBIND, tmp1);
				CLPR_make_runable(tmp1 + 1);
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET1NIL):
		tmp1 = Regs1;
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == CLPR_TAG_NIL) {
				pc++;
				goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
				*tmp1 = addtag(CLPR_TAG_NIL, EMPTYLIST);
				push_check_trail(BTK_BIND, tmp1);
				pc++;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				make_bound_var(tmp1, CLPR_heaptop);
				*CLPR_heaptop = addtag(CLPR_TAG_NIL, EMPTYLIST);
				CLPR_heaptop++;
				push_check_trail(BTK_DBIND, tmp1);
				CLPR_make_runable(tmp1 + 1);
				pc++;
				goto LOOP;
		} else goto FAIL;
	caseop(GET2PVAR):	          /*.................. get2 ...*/ 
		*(CLPR_curar + 2) = Regs[*(pc + 1)];
		pc += 2;
		goto LOOP;
	caseop(GET2PVAL):
		if (CLPR_unify(Regs + *(pc + 1), CLPR_curar + 2)) {
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET2TVAR):
		*Regs2 = Regs[*(pc + 1)];
		pc += 2;
		goto LOOP;
	caseop(GET2TVAL):
		if (CLPR_unify(Regs2, Regs + *(pc + 1))) {
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET2STR):
		tmp1 = Regs2;
		CLPR_deref(tmp1);
		if (*tmp1 == addtag1(TAG_STR, *(pc + 1))) {
				mode_read = TRUE;
				s = tmp1 + 1;
				pc += 2;
				goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
				mode_read = FALSE;
				make_bound_var(tmp1, (CLPR_int) CLPR_heaptop);
				push_check_trail(BTK_BIND, tmp1);
				*CLPR_heaptop = addtag1(TAG_STR, *(pc + 1));
				s = ++CLPR_heaptop;
				pc += 2;
				goto LOOP;
		 } else if (is_daemon(tmp1)) {
				CLPR_make_runable(tmp1 + 1);
				mode_read = FALSE;
				make_bound_var(tmp1, (CLPR_int) CLPR_heaptop);
				push_check_trail(BTK_DBIND, tmp1);
				*CLPR_heaptop = addtag1(TAG_STR, *(pc + 1));
				s = ++CLPR_heaptop;
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET2LIST):
		tmp1 = Regs2;
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == CLPR_TAG_CONS) {
				s = (CLPR_int *) pmask((CLPR_int) (*tmp1));
				mode_read = TRUE;
				pc++;
				goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
				*tmp1 = addtag(CLPR_TAG_CONS, CLPR_heaptop);
				s = CLPR_heaptop;
				push_check_trail(BTK_BIND, tmp1);
				mode_read = FALSE;
				pc++;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				CLPR_make_runable(tmp1 + 1);
				make_bound_var(tmp1, CLPR_heaptop);
				*CLPR_heaptop = addtag(CLPR_TAG_CONS, CLPR_heaptop + 1);
				CLPR_heaptop++;
				s = CLPR_heaptop;
				push_check_trail(BTK_DBIND, tmp1);
				mode_read = FALSE;
				pc++;
				goto LOOP;
		} else goto FAIL;
	caseop(GET2CON):
		tmp1 = Regs2;
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == TAG_CON) {
				if (mask(*tmp1) == *(pc + 1)) {
					pc += 2;
					goto LOOP;
				} else goto FAIL;
		} else if (is_unbound_var(*tmp1)) {
				*tmp1 = addtag(TAG_CON, *(pc + 1));
				push_check_trail(BTK_BIND, tmp1);
				pc += 2;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				make_bound_var(tmp1, CLPR_heaptop);
				*CLPR_heaptop = addtag(TAG_CON, *(pc + 1));
				CLPR_heaptop++;
				push_check_trail(BTK_DBIND, tmp1);
				CLPR_make_runable(tmp1 + 1);
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET2NIL):
		tmp1 = Regs2;
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == CLPR_TAG_NIL) {
				pc++;
				goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
				*tmp1 = addtag(CLPR_TAG_NIL, EMPTYLIST);
				push_check_trail(BTK_BIND, tmp1);
				pc++;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				make_bound_var(tmp1, CLPR_heaptop);
				*CLPR_heaptop = addtag(CLPR_TAG_NIL, EMPTYLIST);
				CLPR_heaptop++;
				push_check_trail(BTK_DBIND, tmp1);
				CLPR_make_runable(tmp1 + 1);
				pc++;
				goto LOOP;
		} else goto FAIL;
	caseop(GET3PVAR):	          /*.................. get3 ...*/ 
		*(CLPR_curar + 3) = Regs[*(pc + 1)];
		pc += 2;
		goto LOOP;
	caseop(GET3PVAL):
		if (CLPR_unify(Regs + *(pc + 1), CLPR_curar + 3)) {
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET3TVAR):
		*Regs3 = Regs[*(pc + 1)];
		pc += 2;
		goto LOOP;
	caseop(GET3TVAL):
		if (CLPR_unify(Regs3, Regs + *(pc + 1))) {
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET3STR):
		tmp1 = Regs3;
		CLPR_deref(tmp1);
		if (*tmp1 == addtag1(TAG_STR, *(pc + 1))) {
				mode_read = TRUE;
				s = tmp1 + 1;
				pc += 2;
				goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
				mode_read = FALSE;
				make_bound_var(tmp1, (CLPR_int) CLPR_heaptop);
				push_check_trail(BTK_BIND, tmp1);
				*CLPR_heaptop = addtag1(TAG_STR, *(pc + 1));
				s = ++CLPR_heaptop;
				pc += 2;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				CLPR_make_runable(tmp1 + 1);
				mode_read = FALSE;
				make_bound_var(tmp1, (CLPR_int) CLPR_heaptop);
				push_check_trail(BTK_DBIND, tmp1);
				*CLPR_heaptop = addtag1(TAG_STR, *(pc + 1));
				s = ++CLPR_heaptop;
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET3LIST):
		tmp1 = Regs3;
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == CLPR_TAG_CONS) {
				s = (CLPR_int *) pmask((CLPR_int) (*tmp1));
				mode_read = TRUE;
				pc++;
				goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
				*tmp1 = addtag(CLPR_TAG_CONS, CLPR_heaptop);
				s = CLPR_heaptop;
				push_check_trail(BTK_BIND, tmp1);
				mode_read = FALSE;
				pc++;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				CLPR_make_runable(tmp1 + 1);
				make_bound_var(tmp1, CLPR_heaptop);
				*CLPR_heaptop = addtag(CLPR_TAG_CONS, CLPR_heaptop + 1);
				CLPR_heaptop++;
				s = CLPR_heaptop;
				push_check_trail(BTK_DBIND, tmp1);
				mode_read = FALSE;
				pc++;
				goto LOOP;
		} else goto FAIL;
	caseop(GET3CON):
		tmp1 = Regs3;
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == TAG_CON) {
				if (mask(*tmp1) == *(pc + 1)) {
					pc += 2;
					goto LOOP;
				} else goto FAIL;
		} else if (is_unbound_var(*tmp1)) {
				*tmp1 = addtag(TAG_CON, *(pc + 1));
				push_check_trail(BTK_BIND, tmp1);
				pc += 2;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				make_bound_var(tmp1, CLPR_heaptop);
				*CLPR_heaptop = addtag(TAG_CON, *(pc + 1));
				CLPR_heaptop++;
				push_check_trail(BTK_DBIND, tmp1);
				CLPR_make_runable(tmp1 + 1);
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(GET3NIL):
		tmp1 = Regs3;
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == CLPR_TAG_NIL) {
				pc++;
				goto LOOP;
		} else if (is_unbound_var(*tmp1)) {
				*tmp1 = addtag(CLPR_TAG_NIL, EMPTYLIST);
				push_check_trail(BTK_BIND, tmp1);
				pc++;
				goto LOOP;
		} else if (is_daemon(tmp1)) {
				make_bound_var(tmp1, CLPR_heaptop);
				*CLPR_heaptop = addtag(CLPR_TAG_NIL, EMPTYLIST);
				CLPR_heaptop++;
				push_check_trail(BTK_DBIND, tmp1);
				CLPR_make_runable(tmp1 + 1);
				pc++;
				goto LOOP;
		} else goto FAIL;
	caseop(UNITVAR):
		if (mode_read) {
				make_bound_var(Regs + *(pc + 1), s++);
				pc += 2;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_bound_var(Regs + *(pc + 1), s++);
				CLPR_heaptop++;
				pc += 2;
				goto LOOP;
		}
	caseop(UNIT1VAR):
		if (mode_read) {
				make_bound_var(Regs1, s++);
				pc++;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_bound_var(Regs1, s++);
				CLPR_heaptop++;
				pc++;
				goto LOOP;
		}
	caseop(UNIT2VAR):
		if (mode_read) {
				make_bound_var(Regs2, s++);
				pc++;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_bound_var(Regs2, s++);
				CLPR_heaptop++;
				pc++;
				goto LOOP;
		}
	caseop(UNIT3VAR):
		if (mode_read) {
				make_bound_var(Regs3, s++);
				pc++;
				goto LOOP;
		} else {
				make_unbound_var(s);
				make_bound_var(Regs3, s++);
				CLPR_heaptop++;
				pc++;
				goto LOOP;
		}
	caseop(UNIVOID):
		if (mode_read) {
				s++;
				pc++;
				goto LOOP;
		} else {
				make_unbound_var(s);
				CLPR_heaptop = ++s;
				pc++;
				goto LOOP;
		}
	caseop(UNIVOIDN):
		if (mode_read) {
				s += pc[1];
				pc += 2;
				goto LOOP;
		} else {
				tmp1 = s;
				tmp2 = s + pc[1];
				do {
					make_unbound_var(tmp1++);
				} while (tmp1 != tmp2);
				CLPR_heaptop = s = tmp2;
				pc += 2;
				goto LOOP;
		}
	caseop(UNITVAL):
		if (!mode_read) {
				tmp1 = (CLPR_int *) (Regs + *(pc + 1));
				CLPR_deref(tmp1);
				if (is_unbound_var(*tmp1)) {
					if (tmp1 < s) make_bound_var(s, tmp1);
					else {
						make_bound_var(tmp1, s);
						make_unbound_var(s);
						push_check_trail(BTK_BIND, tmp1);
					}
				} else if ((CLPR_tag(*tmp1) == TAG_STR) || is_daemon(tmp1))
					 make_bound_var(s, tmp1);
				else *s = *tmp1;
				s = ++CLPR_heaptop;
				pc += 2;
				goto LOOP;
		} else if (CLPR_unify(Regs + *(pc + 1), s)) {
				s++;
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(UNIPVAR):
		if (!mode_read) {
				make_unbound_var(s);
				make_bound_var(CLPR_curar + *(pc + 1), s);
				s = ++CLPR_heaptop;
				pc += 2;
				goto LOOP;
		} else {
				tmp1 = CLPR_curar + *(pc + 1);
				tmp2 = s;
				CLPR_deref(tmp2);
				if (is_unbound_var(*tmp2) || is_daemon(tmp2)
						 || CLPR_tag(*tmp2) == TAG_STR)
					make_bound_var(tmp1, tmp2);
				else *tmp1 = *tmp2;
				push_check_trail(BTK_BIND, tmp1);
				s++;
				pc += 2;
				goto LOOP;
		}
	caseop(UNIPVAL):
		if (!mode_read) {
				tmp1 = CLPR_curar + *(pc + 1);
				CLPR_deref(tmp1);
				if (is_unbound_var(*tmp1)) {
					make_unbound_var(s);
					make_bound_var(tmp1, s);
					push_check_trail(BTK_BIND, tmp1);
				} else if ((CLPR_tag(*tmp1) == TAG_STR) || is_daemon(tmp1))
					make_bound_var(s, tmp1);
				else *s = *tmp1;
				s = ++CLPR_heaptop;
				pc += 2;
				goto LOOP;
		} else if (CLPR_unify(CLPR_curar + *(pc + 1), s)) {
				s++;
				pc += 2;
				goto LOOP;
		} else goto FAIL;
	caseop(UNICON):
		if (!mode_read) {
				*s = addtag1(TAG_CON, *(pc + 1));
				s = ++CLPR_heaptop;
				pc += 2;
				goto LOOP;
		} else {
				tmp1 = s;
				CLPR_deref(tmp1);
				if (is_unbound_var(*tmp1)) {
					*tmp1 = addtag(TAG_CON, *(pc + 1));
					push_check_trail(BTK_BIND, tmp1);
							} else if (is_daemon(tmp1)) {
								    make_bound_var(tmp1, CLPR_heaptop);
					*CLPR_heaptop = addtag(TAG_CON, *(pc + 1));
								    CLPR_heaptop++;
					push_check_trail(BTK_DBIND, tmp1);
								    CLPR_make_runable(tmp1 + 1);
				} else if (*tmp1 != addtag1(TAG_CON, *(pc + 1))) goto FAIL;
				s++;
				pc += 2;
				goto LOOP;
		}
	caseop(UNINIL):
		if (!mode_read) {
				*s = addtag(CLPR_TAG_NIL, EMPTYLIST);
				s = ++CLPR_heaptop;
				pc = pc + 1;
				goto LOOP;
		} else {
				tmp1 = s;
				CLPR_deref(tmp1);
				if (is_unbound_var(*tmp1)) {
					*tmp1 = addtag(CLPR_TAG_NIL, EMPTYLIST);
					push_check_trail(BTK_BIND, tmp1);
				} else if (is_daemon(tmp1)) {
					make_bound_var(tmp1, CLPR_heaptop);
					*CLPR_heaptop = addtag(CLPR_TAG_NIL, EMPTYLIST);
					CLPR_heaptop++;
					push_check_trail(BTK_DBIND, tmp1);
					CLPR_make_runable(tmp1 + 1);
				} else if (*tmp1 != addtag(CLPR_TAG_NIL, EMPTYLIST)) goto FAIL;
					s++;
					pc++;
					goto LOOP;
		}
	caseop(PUTPVAR):
		tmp1 = (CLPR_int *) (CLPR_curar + *(pc + 1));
		make_unbound_var(tmp1);
		make_bound_var(Regs + *(pc + 2), tmp1);
		pc += 3;
		goto LOOP;
	caseop(PUTPVARN):
		ti = pc[1];
		tmp2 = pc + 2;
		do {
			tmp1 = (CLPR_int *) (CLPR_curar + *tmp2);
			make_unbound_var(tmp1);
			make_bound_var(Regs + tmp2[1], tmp1);
			tmp2 += 2;
		} while (--ti > 0);
		pc = tmp2;
		goto LOOP;
	caseop(PUTPVAL):
		tmp1 = CLPR_curar + *(pc + 1);
		if (is_unbound_var(*tmp1) || is_daemon(tmp1)) {
				make_bound_var(Regs + *(pc + 2), tmp1);
				pc += 3;
				goto LOOP;
		} else {
				make_bound_var(Regs + *(pc + 2), *tmp1);
				pc += 3;
				goto LOOP;
		}
	caseop(PUTPVALN):
		ti = pc[1];
		tmp2 = pc + 2;
		do {
			tmp1 = CLPR_curar + *tmp2;
			if (is_unbound_var(*tmp1) || is_daemon(tmp1)) 
				make_bound_var(Regs + tmp2[1], tmp1);
			else make_bound_var(Regs + tmp2[1], *tmp1);
			tmp2 += 2;
		} while (--ti > 0);
		pc = tmp2;
		goto LOOP;
	caseop(PUTUVAL):
		ti1 = *(pc + 2);
		tmp1 = CLPR_curar + *(pc + 1);
		CLPR_deref(tmp1);
		if (is_unbound_var(*tmp1)) {
			if (tmp1 >= CLPR_curar) {
				make_unbound_var(CLPR_heaptop);
				make_bound_var(Regs + *(pc + 2), CLPR_heaptop);
				make_bound_var(tmp1, CLPR_heaptop++);
				push_check_trail(BTK_BIND, tmp1);
			} else make_bound_var(Regs + *(pc + 2), tmp1);
			pc = pc + 3;
			goto LOOP;
		} else if (is_daemon(tmp1)) {
			make_bound_var(Regs + *(pc + 2), tmp1);
			pc = pc + 3;
			goto LOOP;
		} else {
			if (is_func(tmp1)) make_bound_var(Regs + *(pc + 2), tmp1);
			else make_bound_var(Regs + *(pc + 2), *tmp1);
			pc = pc + 3;
			goto LOOP;
		}
	caseop(PUTTVAR):
		make_bound_var(Regs + *(pc + 1), CLPR_heaptop);
		make_unbound_var(CLPR_heaptop); 
		CLPR_heaptop++;
		pc += 2;
		goto LOOP;
	caseop(PUTTVAL):
		Regs[*(pc + 2)] = Regs[*(pc + 1)];
		pc = pc + 3;
		goto LOOP;
	caseop(PUTSTR):
		*CLPR_heaptop = addtag(TAG_STR, *(pc + 1));
		make_bound_var(Regs + *(pc + 2), CLPR_heaptop);
		s = ++CLPR_heaptop;
		pc = pc + 3;
		goto LOOP;
	caseop(PUTLIST):
		Regs[*(pc + 1)] = addtag(CLPR_TAG_CONS, CLPR_heaptop);
		s = CLPR_heaptop;
		pc += 2;
		goto LOOP;
	caseop(PUTCON):
		Regs[*(pc + 2)] = addtag(TAG_CON, *(pc + 1));
		pc = pc + 3;
		goto LOOP;
	caseop(PUTNIL):
		Regs[*(pc + 1)] = addtag(CLPR_TAG_NIL, EMPTYLIST);
		pc = pc + 2;
		goto LOOP;
	caseop(BLDPVAR):
		make_unbound_var(s);
		make_bound_var(CLPR_curar + *(pc + 1), s);
		s++;
		CLPR_heaptop++;
		pc = pc + 2;
		goto LOOP;
	caseop(BLDPVAL):
		tmp1 = (CLPR_int *) (CLPR_curar + *(pc + 1));
		CLPR_deref(tmp1);
		if (is_unbound_var(*tmp1)) {
			if (below_heap(tmp1)) make_bound_var(s, tmp1);
			else {
				make_unbound_var(s);
				make_bound_var(tmp1, s);
				push_check_trail(BTK_BIND, tmp1);
			}
		} else if ((CLPR_tag(*tmp1) == TAG_STR) || is_daemon(tmp1))
			make_bound_var(s, tmp1);
		else make_bound_var(s, *tmp1);
		CLPR_heaptop = ++s;
		pc += 2;
		goto LOOP;
	caseop(BLDPVALN):
		ti = pc[1];
		tmp1 = pc + 1;
		s--;
		for (; ti > 0; ti--) {
			tmp2 = (CLPR_int *) (CLPR_curar + tmp1[ti]);
			CLPR_deref(tmp2);
			if (is_unbound_var(*tmp2)) {
				if (below_heap(tmp2)) make_bound_var(s+ti, tmp2);
				else {
					make_unbound_var(s+ti);
					make_bound_var(tmp2, s+ti);
					push_check_trail(BTK_BIND, tmp2);
				}
			} else if ((CLPR_tag(*tmp2) == TAG_STR) || is_daemon(tmp2))
				make_bound_var(s+ti, tmp2);
			else make_bound_var(s+ti, *tmp2);
		}
		CLPR_heaptop = s = s + pc[1] + 1;
		pc = pc + pc[1] + 2;			
		goto LOOP;
	caseop(BLDTVAR):
		make_unbound_var(s);
		make_bound_var(Regs + *(pc + 1), s++);
		CLPR_heaptop++;
		pc += 2;
		goto LOOP;
	caseop(BLDVOID):
		make_unbound_var(s);
		CLPR_heaptop = ++s;
		pc++;
		goto LOOP;
	caseop(BLDVOIDN):
		ti1 = pc[1];
		ti = ti1;
		s--;
		do {
			make_unbound_var(s + ti);
		} while (--ti > 0);
		CLPR_heaptop = s = s + ti1 + 1;
		pc += 2;
		goto LOOP;
	caseop(BLDTVAL):
		tmp1 = (CLPR_int *) (Regs + *(pc + 1));
		CLPR_deref(tmp1);
		if (is_unbound_var(*tmp1)) {
			make_unbound_var(s);
			make_bound_var(tmp1, s);
			push_check_trail(BTK_BIND, tmp1);
		} else if ((CLPR_tag(*tmp1) == TAG_STR) || is_daemon(tmp1)) 
			make_bound_var(s, tmp1);
		else *s = *tmp1;
		s++;
		CLPR_heaptop++;
		pc += 2;
		goto LOOP;
	caseop(BLDTVALN):
		ti = pc[1];
		tmp1 = pc + 1;
		s--;
		for (; ti > 0; ti--) {
			tmp2 = (CLPR_int *) (Regs + tmp1[ti]);
			CLPR_deref(tmp2);
			if (is_unbound_var(*tmp2)) {
				if (below_heap(tmp2)) make_bound_var(s+ti, tmp2);
				else {
					make_unbound_var(s+ti);
					make_bound_var(tmp2, s+ti);
					push_check_trail(BTK_BIND, tmp2);
				}
			} else if ((CLPR_tag(*tmp2) == TAG_STR) || is_daemon(tmp2))
				make_bound_var(s+ti, tmp2);
			else make_bound_var(s+ti, *tmp2);
		}
		CLPR_heaptop = s = s + pc[1] + 1;
		pc = pc + pc[1] + 2;			
		goto LOOP;
	caseop(BLDCON):
		*(s++) = addtag1(TAG_CON, *(pc + 1));
		CLPR_heaptop++;
		pc += 2;
		goto LOOP;
	caseop(BLDNIL):
		*s = addtag(CLPR_TAG_NIL, EMPTYLIST);
		CLPR_heaptop = ++s;
		pc++;
		goto LOOP;
	caseop(BLDLIST):
		*s = addtag(CLPR_TAG_CONS, s + 1);
		CLPR_heaptop = ++s;
		pc++;
		goto LOOP;
	caseop(BLDSTR):
		*s = (CLPR_int) (s + 1);
		*(s + 1) = addtag1(TAG_STR, *(pc + 1));
		CLPR_heaptop = s = s + 2;
		pc += 2;
		goto LOOP;		
	caseop(SAVE):
		CheckStack();
		tmp1 = CLPR_curar; 
		CLPR_curar = CLPR_stacktop + 2;
		CLPR_stacktop = CLPR_curar + *(pc + 1);
		*(CLPR_curar + FATHER) = (CLPR_int) tmp1;
		*(CLPR_curar + RETPT) = (CLPR_int) CLPR_retreg;
		pc += 2;
		goto LOOP;
	caseop(RESTORE):
		CLPR_retreg = (CLPR_int *) pmask(*(CLPR_curar + RETPT));
		CLPR_curar = (CLPR_int *) pmask(*(CLPR_curar + FATHER));
		pc++;
		goto LOOP;
	caseop(CRESTORE):
		if (CLPR_curar > CLPR_lastcp) CLPR_stacktop = CLPR_curar - 2;
		CLPR_retreg = (CLPR_int *) pmask(*(CLPR_curar + RETPT));
		CLPR_curar = (CLPR_int *) pmask(*(CLPR_curar + FATHER));
		pc++;
		goto LOOP;
	caseop(ENTERTRYELSE):
#ifdef CHECK_WAKEUP
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
#endif
		CheckHeap();	/* Fall-thru */
	caseop(TRY):
		CheckStack();
		tmp1 = CLPR_stacktop;
		*(++tmp1) = CLPR_slack_id;
		*(++tmp1) = CLPR_solver_id;
		*(++tmp1) = ti = *(pc + 2);
		*(++tmp1) = (CLPR_int) CLPR_curar;
		*(++tmp1) = (CLPR_int) CLPR_retreg;
		*(++tmp1) = CLPR_trtop;
		*(++tmp1) = (CLPR_int) CLPR_heaptop;
		*(++tmp1) = (CLPR_int) CLPR_lastcp;
		*(++tmp1) = (CLPR_int) (pc + 3);
		CLPR_lastcp = tmp1;
		CLPR_safeheap = CLPR_heaptop;
		tmp2 = Regs;
		CLPR_stacktop = tmp1 + ti;
		while (tmp1 < CLPR_stacktop) *(++tmp1) = *(++tmp2);
		pc = (CLPR_int *) *(pc + 1);
		CLPR_stamp++;
		goto LOOP;
	caseop(RETRY):
		*(CLPR_lastcp + NEXTEP) = (CLPR_int) (pc + 2);
		CLPR_stacktop = CLPR_lastcp + mask(*(CLPR_lastcp + NUMPARS));
		pc = (CLPR_int *) *(pc + 1);
		CLPR_stamp++;
		goto LOOP;
	caseop(TRUST):
		CLPR_stacktop = CLPR_lastcp - CPFRAME_HEIGHT;
		CLPR_safeheap = (CLPR_int *) pmask(*(CLPR_lastcp + OHEAPTOP));
		CLPR_lastcp = (CLPR_int *) pmask(*(CLPR_lastcp + PREVCP));
		pc = (CLPR_int *) *(pc + 1);
		goto LOOP;
	caseop(FASTCUT):
		CLPR_lastcp = (CLPR_int *) pmask(Regs[pc[1]]);
		pc += 2;
		goto LOOP;
	caseop(SAVECURCP):
		Regs[pc[1]] = (CLPR_int) CLPR_lastcp;
		pc += 2;
		goto LOOP;
	caseop(SAVEPREVCP):
		Regs[pc[1]] = CLPR_lastcp[PREVCP];
		pc += 2;
		goto LOOP;
	caseop(SAVECUT):
		*(CLPR_curar + pc[2]) = (CLPR_int) Regs[pc[1]];
		pc += 3;
		goto LOOP;
	caseop(DIRSAVECUT):
		*(CLPR_curar + *(pc + 1)) = (CLPR_int) CLPR_lastcp; 
		pc += 2;
		goto LOOP;
	caseop(DOCUT):
		CLPR_lastcp = (CLPR_int *) pmask(*(CLPR_curar + *(pc + 1)));
		pc += 2;
		goto LOOP;
	caseop(REGINIT):
		make_unbound_var(Regs + *(pc + 1));
		pc += 2;
		goto LOOP;
	case (CLPR_unsigned) ADD1PFVAR: 
		tmp1 = CLPR_curar + *(pc + 1);
		*tmp1 = addtag(TAG_PAR, ++CLPR_solver_id);
		push_check_trail(BTK_BIND, tmp1);
		if (!CLPR_s_evar(1.0, mask(*tmp1))) goto FAIL;
		pc += 2;
		goto LOOP;
	caseop(ADDPFVAR):
		*((CLPR_int *) CLPR_tfp) = *(pc + 1);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 2);
		tmp1 = CLPR_curar + *(pc + 3);
		*tmp1 = addtag(TAG_PAR, ++CLPR_solver_id);
		push_check_trail(BTK_BIND, tmp1);
		if (!CLPR_s_evar(tf, mask(*tmp1))) goto FAIL;
		pc += 4;
		goto LOOP;
	caseop(ADD1PFVAL):
		tmp1 = CLPR_curar + *(pc + 1);
		CLPR_deref(tmp1);
		if (is_unbound_var(*tmp1)) {
			*tmp1 = addtag(TAG_PAR, ++CLPR_solver_id);
			push_check_trail(BTK_BIND, tmp1);
		} else if (is_daemon(tmp1)) {
			make_bound_var(tmp1, CLPR_heaptop);
			*CLPR_heaptop = addtag(TAG_PAR, ++CLPR_solver_id);
			CLPR_heaptop++;
			push_check_trail(BTK_DBIND, tmp1);
			CLPR_make_runable(tmp1 + 1);
			tmp1 = (CLPR_int *) pmask(*tmp1);
		} else if (!is_param(tmp1)) goto FAIL; /* type clash */
		if (!CLPR_s_evar(1.0, mask(*tmp1))) goto FAIL;
		pc += 2;
		goto LOOP;
	caseop(ADDPFVAL):
		*((CLPR_int *) CLPR_tfp) = *(pc + 1);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 2);
		tmp1 = CLPR_curar + *(pc + 3);
		CLPR_deref(tmp1);
		if (is_unbound_var(*tmp1)) {
			*tmp1 = addtag(TAG_PAR, ++CLPR_solver_id);
			push_check_trail(BTK_BIND, tmp1);
		} else if (is_daemon(tmp1)) {
			make_bound_var(tmp1, CLPR_heaptop);
			*CLPR_heaptop = addtag(TAG_PAR, ++CLPR_solver_id);
			CLPR_heaptop++;
			push_check_trail(BTK_DBIND, tmp1);
			CLPR_make_runable(tmp1 + 1);
			tmp1 = (CLPR_int *) pmask(*tmp1);
		} else if (!is_param(tmp1)) goto FAIL; /* type clash */
		if (!CLPR_s_evar(tf, mask(*tmp1))) goto FAIL;
		pc += 4;
		goto LOOP;
	caseop(ADDT1PFVAR):
		tmp1 = Regs + *(pc + 1);
		*tmp1 = addtag(TAG_PAR, ++CLPR_solver_id);
		if (!CLPR_s_evar(1.0, mask(*tmp1))) goto FAIL;
		pc += 2;
		goto LOOP;
	caseop(ADDTPFVAR):
		*((CLPR_int *) CLPR_tfp) = *(pc + 1);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 2);
		tmp1 = Regs + *(pc + 3);
		*tmp1 = addtag(TAG_PAR, ++CLPR_solver_id);
		if (!CLPR_s_evar(tf, mask(*tmp1))) goto FAIL;
		pc += 4;
		goto LOOP;
	caseop(ADDT1PFVAL):
		tmp1 = Regs + *(pc + 1);
		CLPR_deref(tmp1);
		if (is_unbound_var(*tmp1)) {
			*tmp1 = addtag(TAG_PAR, ++CLPR_solver_id);
			push_check_trail(BTK_BIND, tmp1);
		} else if (is_daemon(tmp1)) {
			make_bound_var(tmp1, CLPR_heaptop);
			*CLPR_heaptop = addtag(TAG_PAR, ++CLPR_solver_id);
			CLPR_heaptop++;
			push_check_trail(BTK_DBIND, tmp1);
			CLPR_make_runable(tmp1 + 1);
			tmp1 = (CLPR_int *) pmask(*tmp1);
		} else if (!is_param(tmp1)) goto FAIL; /* type clash */
		if (!CLPR_s_evar(1.0, mask(*tmp1))) goto FAIL;
		pc += 2;
		goto LOOP;
	caseop(ADDTPFVAL):
		*((CLPR_int *) CLPR_tfp) = *(pc + 1);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 2);
		tmp1 = Regs + *(pc + 3);
		CLPR_deref(tmp1);
		if (is_unbound_var(*tmp1)) {
			*tmp1 = addtag(TAG_PAR, ++CLPR_solver_id);
			push_check_trail(BTK_BIND, tmp1);
		} else if (is_daemon(tmp1)) {
			make_bound_var(tmp1, CLPR_heaptop);
			*CLPR_heaptop = addtag(TAG_PAR, ++CLPR_solver_id);
			CLPR_heaptop++;
			push_check_trail(BTK_DBIND, tmp1);
			CLPR_make_runable(tmp1 + 1);
			tmp1 = (CLPR_int *) pmask(*tmp1);
		} else if (!is_param(tmp1)) goto FAIL; /* type clash */
		if (!CLPR_s_evar(tf, mask(*tmp1))) goto FAIL;
		pc += 4;
		goto LOOP;
	caseop(INITPF):
		*((CLPR_int *) CLPR_tfp) = *(pc + 1);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 2);
		CLPR_stored_constant = -tf;
		pc += 3;
		CLPR_init_eqn();
		goto LOOP;
	caseop(INIT0PF):
		CLPR_stored_constant = 0.0;
		pc = pc + 1;
		CLPR_init_eqn();
		goto LOOP;
	caseop(SOLVE_PF0):
		if (!CLPR_s_econ(-1.0 * CLPR_stored_constant)) goto FAIL;
		pc = pc + 1;
		goto LOOP;
	caseop(SOLVE_GT0):
		if (!CLPR_s_gt0(CLPR_stored_constant)) goto FAIL;		
		pc = pc + 1;
		goto LOOP;
	caseop(SOLVE_GE0):
		if (!CLPR_s_ge0(CLPR_stored_constant)) goto FAIL;
		pc = pc + 1;
		goto LOOP;
	caseop(PUTPNUM):
		*((CLPR_int *) CLPR_tfp) = *(pc + 1);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 2);
		CLPR_makenum(++CLPR_solver_id, tf);
		CLPR_curar[*(pc + 3)] = addtag(TAG_PAR, CLPR_solver_id); 
		pc += 4;
		goto LOOP;
	caseop(PUTTNUM):
		*((CLPR_int *) CLPR_tfp) = *(pc + 1);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 2);
		CLPR_makenum(++CLPR_solver_id, tf);
		Regs[*(pc + 3)] = addtag(TAG_PAR, CLPR_solver_id); 
		pc += 4;
		goto LOOP;
	caseop(GETPNUM):
		*((CLPR_int *) CLPR_tfp) = *(pc + 1);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 2);
		tmp1 = CLPR_curar + *(pc + 3);
		CLPR_deref(tmp1);
		if (is_unbound_var(*tmp1)) {
			CLPR_makenum(++CLPR_solver_id, tf);
			*tmp1 = addtag(TAG_PAR, CLPR_solver_id);
			push_check_trail(BTK_BIND, tmp1);
			pc += 4;
			goto LOOP;
		} else if (is_daemon(tmp1)) {
			CLPR_makenum(++CLPR_solver_id, tf);
			make_bound_var(tmp1, CLPR_heaptop);
			*CLPR_heaptop = addtag(TAG_PAR, CLPR_solver_id);
			CLPR_heaptop++;
			push_check_trail(BTK_DBIND, tmp1);
			CLPR_make_runable(tmp1 + 1);
			pc += 4;
			goto LOOP;
		} else {
			if (CLPR_tag(*tmp1) == TAG_PAR) 
				if (CLPR_is_ground(mask(*tmp1), &c)) 
					if (is_equal(c, tf)) {
						pc += 4;
						goto LOOP;
					} else goto FAIL;
				else if (CLPR_solve_eqn_1(mask(*tmp1), tf)) {
					pc += 4;
					goto LOOP;
				} else goto FAIL;
			else goto FAIL;
		}
	caseop(GETTNUM):
		*((CLPR_int *) CLPR_tfp) = *(pc + 1);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 2);
		tmp1 = Regs + *(pc + 3);
		CLPR_deref(tmp1);
		if (is_unbound_var(*tmp1)) {
			CLPR_makenum(++CLPR_solver_id, tf);
			*tmp1 = addtag(TAG_PAR, CLPR_solver_id);
			push_check_trail(BTK_BIND, tmp1);
			pc += 4;
			goto LOOP;
		} else if (is_daemon(tmp1)) {
			CLPR_makenum(++CLPR_solver_id, tf);
			make_bound_var(tmp1, CLPR_heaptop);
			*CLPR_heaptop = addtag(TAG_PAR, CLPR_solver_id);
			CLPR_heaptop++;
			push_check_trail(BTK_DBIND, tmp1);
			CLPR_make_runable(tmp1 + 1);
			pc += 4;
			goto LOOP;
		} else {
			if (CLPR_tag(*tmp1) == TAG_PAR) 
				if (CLPR_is_ground(mask(*tmp1), &c)) 
					if (is_equal(c, tf)) {
						pc += 4;
						goto LOOP;
					} else goto FAIL;
				else if (CLPR_solve_eqn_1(mask(*tmp1), tf)) {
					pc += 4;
					goto LOOP;
				} else goto FAIL;
			else goto FAIL;
		}
	caseop(BLDNUM):
		*((CLPR_int *) CLPR_tfp) = *(pc + 1);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 2);
		CLPR_makenum(++CLPR_solver_id, tf);
		*(s++) = addtag(TAG_PAR, CLPR_solver_id);
		CLPR_heaptop++;
		pc += 3;
		goto LOOP;
	caseop(UNINUM):
		*((CLPR_int *) CLPR_tfp) = *(pc + 1);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 2);
		CLPR_makenum(++CLPR_solver_id, tf);
		if (!mode_read) {
			*s = addtag(TAG_PAR, CLPR_solver_id);
			s++;
			CLPR_heaptop++;
			pc += 3;
			goto LOOP;
		} else {
			tmp1 = s;
			CLPR_deref(tmp1);
			if (is_unbound_var(*tmp1)) {
				*tmp1 = addtag(TAG_PAR, CLPR_solver_id);
				push_check_trail(BTK_BIND, tmp1);
			} else if (is_daemon(tmp1)) {
			    make_bound_var(tmp1, CLPR_heaptop);
				*CLPR_heaptop = addtag(TAG_PAR, CLPR_solver_id);
			    CLPR_heaptop++;
				push_check_trail(BTK_DBIND, tmp1);
			    CLPR_make_runable(tmp1 + 1);
			} else {
				if (CLPR_tag(*tmp1) != TAG_PAR || 
					!CLPR_solve_eqn_2(mask(*tmp1), CLPR_solver_id, 1.0)) 
					goto FAIL;
			}
			s++;
			pc += 3;
			goto LOOP;
		}
	caseop(SOLVE_MULT):
		ti = *(pc + 1);
		*((CLPR_int *) CLPR_tfp) = *(pc + 2);
		*(((CLPR_int *) CLPR_tfp) + 1) = *(pc + 3);
		tmp1 = (bit6(ti) ? CLPR_curar + *(pc + 4) : Regs + *(pc + 4));
		if (bit5(ti)) check_tag_par(tmp1) else make_tag_par(tmp1);
		tmp2 = (bit4(ti) ? CLPR_curar + *(pc + 5) : Regs + *(pc + 5));
		if (bit3(ti)) check_tag_par(tmp2) else make_tag_par(tmp2);
		tmp3 = (bit2(ti) ? CLPR_curar + *(pc + 6) : Regs + *(pc + 6));
		if (bit1(ti)) check_tag_par(tmp3) else make_tag_par(tmp3);
		if (!CLPR_s_mult(tf, mask(*tmp1), mask(*tmp2), mask(*tmp3))) goto FAIL;
		pc += 7;
		goto LOOP;
	caseop(SOLVE_MULT1):
		ti = *(pc + 1);
		tmp1 = (bit6(ti) ? CLPR_curar + *(pc + 2) : Regs + *(pc + 2));
		if (bit5(ti)) check_tag_par(tmp1) else make_tag_par(tmp1);
		tmp2 = (bit4(ti) ? CLPR_curar + *(pc + 3) : Regs + *(pc + 3));
		if (bit3(ti)) check_tag_par(tmp2) else make_tag_par(tmp2);
		tmp3 = (bit2(ti) ? CLPR_curar + *(pc + 4) : Regs + *(pc + 4));
		if (bit1(ti)) check_tag_par(tmp3) else make_tag_par(tmp3);
		if (!CLPR_s_mult(1.0, mask(*tmp1), mask(*tmp2), mask(*tmp3))) goto FAIL;
		pc += 5;
		goto LOOP;
	caseop(SOLVE_MIN):
		ti = *(pc + 1);
		tmp1 = (bit6(ti) ? CLPR_curar + *(pc + 2) : Regs + *(pc + 2));
		if (bit5(ti)) check_tag_par(tmp1) else make_tag_par(tmp1);
		tmp2 = (bit4(ti) ? CLPR_curar + *(pc + 3) : Regs + *(pc + 3));
		if (bit3(ti)) check_tag_par(tmp2) else make_tag_par(tmp2);
		tmp3 = (bit2(ti) ? CLPR_curar + *(pc + 4) : Regs + *(pc + 4));
		if (bit1(ti)) check_tag_par(tmp3) else make_tag_par(tmp3);
		if (!CLPR_s_min(mask(*tmp1), mask(*tmp2), mask(*tmp3))) goto FAIL;
		pc += 5;
		goto LOOP;
	caseop(SOLVE_MAX):
		ti = *(pc + 1);
		tmp1 = (bit6(ti) ? CLPR_curar + *(pc + 2) : Regs + *(pc + 2));
		if (bit5(ti)) check_tag_par(tmp1) else make_tag_par(tmp1);
		tmp2 = (bit4(ti) ? CLPR_curar + *(pc + 3) : Regs + *(pc + 3));
		if (bit3(ti)) check_tag_par(tmp2) else make_tag_par(tmp2);
		tmp3 = (bit2(ti) ? CLPR_curar + *(pc + 4) : Regs + *(pc + 4));
		if (bit1(ti)) check_tag_par(tmp3) else make_tag_par(tmp3);
		if (!CLPR_s_max(mask(*tmp1), mask(*tmp2), mask(*tmp3))) goto FAIL;
		pc += 5;
		goto LOOP;
	caseop(SOLVE_POW):
		ti = *(pc + 1);
		tmp1 = (bit6(ti) ? CLPR_curar + *(pc + 2) : Regs + *(pc + 2));
		if (bit5(ti)) check_tag_par(tmp1) else make_tag_par(tmp1);
		tmp2 = (bit4(ti) ? CLPR_curar + *(pc + 3) : Regs + *(pc + 3));
		if (bit3(ti)) check_tag_par(tmp2) else make_tag_par(tmp2);
		tmp3 = (bit2(ti) ? CLPR_curar + *(pc + 4) : Regs + *(pc + 4));
		if (bit1(ti)) check_tag_par(tmp3) else make_tag_par(tmp3);
		if (!CLPR_s_pow(mask(*tmp1), mask(*tmp2), mask(*tmp3))) goto FAIL;
		pc += 5;
		goto LOOP;
	caseop(SOLVE_ABS):
		ti = *(pc + 1);
		tmp1 = (bit4(ti) ? CLPR_curar + *(pc + 2) : Regs + *(pc + 2));
		if (bit3(ti)) check_tag_par(tmp1) else make_tag_par(tmp1);
		tmp2 = (bit2(ti) ? CLPR_curar + *(pc + 3) : Regs + *(pc + 3));
		if (bit1(ti)) check_tag_par(tmp2) else make_tag_par(tmp2);
		if (!CLPR_s_abs(mask(*tmp1), mask(*tmp2))) goto FAIL;
		pc += 4;
		goto LOOP;
	caseop(SOLVE_EVAL):
		ti = *(pc + 1);
		tmp1 = (bit4(ti) ? CLPR_curar + *(pc + 2) : Regs + *(pc + 2));
		if (!bit3(ti)) { 
			make_unbound_var(CLPR_heaptop);
			 make_bound_var(tmp1,(CLPR_heaptop++));
		 }
		tmp2 = (bit2(ti) ? CLPR_curar + *(pc + 3) : Regs + *(pc + 3));
		if (!bit1(ti)) { 
			make_unbound_var(CLPR_heaptop); 
			make_bound_var(tmp2,(CLPR_heaptop++));
		 }
		if (!CLPR_s_eval(tmp1, tmp2)) goto FAIL;
		pc += 4;
		goto LOOP;
	caseop(SOLVE_SIN):
		ti = *(pc + 1);
		tmp1 = (bit4(ti) ? CLPR_curar + *(pc + 2) : Regs + *(pc + 2));
		if (bit3(ti)) check_tag_par(tmp1) else make_tag_par(tmp1);
		tmp2 = (bit2(ti) ? CLPR_curar + *(pc + 3) : Regs + *(pc + 3));
		if (bit1(ti)) check_tag_par(tmp2) else make_tag_par(tmp2);
		if (!CLPR_s_sin(mask(*tmp1), mask(*tmp2))) goto FAIL;
		pc += 4;
		goto LOOP;
	caseop(SOLVE_COS):
		ti = *(pc + 1);
		tmp1 = (bit4(ti) ? CLPR_curar + *(pc + 2) : Regs + *(pc + 2));
		if (bit3(ti)) check_tag_par(tmp1) else make_tag_par(tmp1);
		tmp2 = (bit2(ti) ? CLPR_curar + *(pc + 3) : Regs + *(pc + 3));
		if (bit1(ti)) check_tag_par(tmp2) else make_tag_par(tmp2);
		if (!CLPR_s_cos(mask(*tmp1), mask(*tmp2))) goto FAIL;
		pc += 4;
		goto LOOP;
	caseop(SOLVE_ARCSIN):
		ti = *(pc + 1);
		tmp1 = (bit4(ti) ? CLPR_curar + *(pc + 2) : Regs + *(pc + 2));
		if (bit3(ti)) check_tag_par(tmp1) else make_tag_par(tmp1);
		tmp2 = (bit2(ti) ? CLPR_curar + *(pc + 3) : Regs + *(pc + 3));
		if (bit1(ti)) check_tag_par(tmp2) else make_tag_par(tmp2);
		if (!CLPR_s_arcsin(mask(*tmp1), mask(*tmp2))) goto FAIL;
		pc += 4;
		goto LOOP;
	caseop(SOLVE_ARCCOS):
		ti = *(pc + 1);
		tmp1 = (bit4(ti) ? CLPR_curar + *(pc + 2) : Regs + *(pc + 2));
		if (bit3(ti)) check_tag_par(tmp1) else make_tag_par(tmp1);
		tmp2 = (bit2(ti) ? CLPR_curar + *(pc + 3) : Regs + *(pc + 3));
		if (bit1(ti)) check_tag_par(tmp2) else make_tag_par(tmp2);
		if (!CLPR_s_arccos(mask(*tmp1), mask(*tmp2))) goto FAIL;
		pc += 4;
		goto LOOP;
	caseop(RETRACT):
		tmp1 = Regs + pc[1];
		CLPR_deref(tmp1);
		if (below_heap(tmp1) && !is_func(tmp1)) {
			*CLPR_heaptop = *tmp1;
			tmp1 = CLPR_heaptop++;
		}
		if (CLPR_retract_rule(tmp1, pc, TRUE)) {
			pc += 3;
			goto LOOP;
		} else goto FAIL;
	caseop(RETRACT_AGAIN):
		if (CLPR_retract_rule(0, pc, FALSE)) {
			pc++;
			goto LOOP;
		} else goto FAIL;
	caseop(ROOLE):
		tmp1 = Regs + pc[1];
		CLPR_deref(tmp1);
		if (below_heap(tmp1) && !is_func(tmp1)) {
			*CLPR_heaptop = *tmp1;
			tmp1 = CLPR_heaptop++;
		}
		tmp2 = Regs + pc[2];
		CLPR_deref(tmp2);
		if (below_heap(tmp2) && !is_func(tmp2)) {
			*CLPR_heaptop = *tmp2;
			tmp2 = CLPR_heaptop++;
		}
		if (CLPR_match_rule(tmp1, tmp2, pc, TRUE)) {
			pc += 4;
			goto LOOP;
		} else goto FAIL;
	caseop(ROOLE_AGAIN):
		if (CLPR_match_rule(0, 0, pc, FALSE)) {
			pc++;
			goto LOOP;
		} else goto FAIL;
	caseop(SPCALL):
		ti = CLPR_spred[pc[1]]->arity;
		if (CLPR_emul_sys_pred(pc)) {
			pc += 2 + ti;
			goto LOOP;
		} else goto FAIL;
	caseop(READ):
		if (!CLPR_do_read(CLPR_see_input, CLPR_curar + *(pc + 1))) goto FAIL;
		pc += 2;	
		goto LOOP;
	caseop(READREG):
		if (!CLPR_do_read(CLPR_see_input, Regs + *(pc + 1))) goto FAIL;
		pc += 2;	
		goto LOOP;
	caseop(PRINT):
		CLPR_printterm(CLPR_outfile, &Regs[pc[1]]);
		pc += 2;
		goto LOOP;
	caseop(PRINTF):
		tmp1 = CLPR_findbind(&Regs[pc[1]]);
		if (!is_con(tmp1)) {
			CLPR_warn_err("incorrect format argument for printf/2");
			goto FAIL;
		}
		CLPR_do_printf(CLPR_outfile, &Regs[pc[2]], string_of(mask(*tmp1)), TRUE);
		pc += 3;
		goto LOOP;
	caseop(DUMP):
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
		CLPR_do_dump(CLPR_outfile, &Regs[pc[2]], string_of(mask(pc[1])));
		pc += 3;
		goto LOOP;
	caseop(DBGENTRY):
		CLPR_emul_dbg_entry(pc);
		pc += 3;
		goto LOOP;		
	caseop(FFAIL):
		goto FAIL;
	caseop(GFAIL):
		if (interactive) fprintf(CLPR_outfile, "\n*** No\n");
		return FALSE;
	caseop(HALT):
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
		if (!interactive) return TRUE;
		if (CLPR_emul_halt()) goto END_RUNGOAL;
		goto FAIL;
	caseop(JZF):
		ti = *(pc + 1);
		if (*((double *)(Regs + ti)) == 0.0) pc = (CLPR_int *) *(pc + 2);
		else pc = pc + 3;
		goto LOOP;
	caseop(JGF):
		ti = *(pc + 1);
		if (*((double *)(Regs + ti)) > 0.0)	pc = (CLPR_int *) *(pc + 2);
		else pc = pc + 3;
		goto LOOP;
	caseop(LITF):
		ti = *(pc + 3);
		*(Regs + ti) = *(pc + 1);
		*(Regs + ti + 1) = *(pc + 2);
		pc = pc + 4;
		goto LOOP;
	caseop(SUBF):
		*((double *)(Regs + *(pc + 3))) = 
		*((double *)(Regs + *(pc + 1))) - 
		*((double *)(Regs + *(pc + 2)));
		pc = pc + 4;
		goto LOOP;
	caseop(ADDF):
		*((double *)(Regs + *(pc + 3))) = 
		*((double *)(Regs + *(pc + 1))) +
		*((double *)(Regs + *(pc + 2)));
		pc = pc + 4;
		goto LOOP;
	caseop(MULF):
		*((double *)(Regs + *(pc + 3))) = 
		*((double *)(Regs + *(pc + 1))) *
		*((double *)(Regs + *(pc + 2)));
		pc = pc + 4;
		goto LOOP;
	caseop(DIVF):
		*((double *)(Regs + *(pc + 3))) = 
		*((double *)(Regs + *(pc + 1))) /
		*((double *)(Regs + *(pc + 2)));
		pc = pc + 4;
		goto LOOP;
	caseop(MVF):
		ti = *(pc + 1);
		ti1 = *(pc + 2);
		*(Regs + ti1++) = *(Regs + ti++);
		*(Regs + ti1) = *(Regs + ti);
		pc = pc + 3;
		goto LOOP;
	caseop(STF):
		ti = *(pc + 1);
		tmp1 = CLPR_curar + *(pc + 2);
		*(tmp1++) = TAG_NUM; /* for garbage collector's benefit */
		*(tmp1++) = *(Regs + (ti++));
		*tmp1 = *(Regs + ti);
		pc = pc + 3;
		goto LOOP;
	caseop(LDF):
		ti = *(pc + 2);
		tmp1 = CLPR_curar + *(pc + 1) + 1;
		*(Regs + (ti++)) = *(tmp1++);
		*(Regs + ti) = *tmp1;
		pc = pc + 3;
		goto LOOP;
	caseop(GETF):
		tmp1 = Regs + *(pc + 1);
		tmp2 = Regs + *(pc + 2);
		CLPR_deref(tmp1);
		if (CLPR_tag(*tmp1) == TAG_PAR && CLPR_is_ground(mask(*tmp1), &c)) {
			tf = c;
			*tmp2 = *((CLPR_int *) CLPR_tfp);
			 *(tmp2 + 1) = *(((CLPR_int *) CLPR_tfp) + 1);
		} else CLPR_runtime_error("GETF violated");
		pc = pc + 3;
		goto LOOP;
	caseop(PUTF):
		tmp1 = Regs + *(pc + 1);
		tmp2 = Regs + *(pc + 2);
		*((CLPR_int *) CLPR_tfp) = *tmp1;
		*(((CLPR_int *) CLPR_tfp) + 1) = *(tmp1 + 1);
		CLPR_makenum(++CLPR_solver_id, tf);
		*tmp2 = addtag(TAG_PAR, CLPR_solver_id); 
		pc = pc + 3;
		goto LOOP;
	caseop(DYNAMIC_TRY):
		if ((pc = CLPR_run_dynamic(pc, first_dyn_call, TRUE))) {
			first_dyn_call = FALSE;
			goto LOOP;
		}
		first_dyn_call = FALSE;
		goto FAIL;
	caseop(DYNAMIC_0_TRY):
		if ((pc = CLPR_run_dynamic(pc, first_dyn_call, FALSE))) {
			first_dyn_call = FALSE;
			goto LOOP;
		}
		first_dyn_call = FALSE;
		goto FAIL;
	caseop(DYN_CALL):
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
		first_dyn_call = TRUE;
		if (CLPR_heap >= maxheap || CLPR_trtop >= maxtrail) 
			CLPR_fatal("heap/trail overflow");
		CLPR_retreg = pc + 2;
		if (pc = (CLPR_int *) CLPR_val_to_pc(ti = *(pc + 1))) {
			Regs[arity_of(ti)+1] = (CLPR_int) CLPR_lastcp;
			goto LOOP;
		}
		first_dyn_call = FALSE;
		goto FAIL;
	caseop(DYN_JUMP):
		if (CLPR_have_runable && !CLPR_wake_runable()) goto FAIL;
		first_dyn_call = TRUE;
		if (CLPR_heap >= maxheap || CLPR_trtop >= maxtrail) 
			CLPR_fatal("heap/trail overflow");
		if (pc = (CLPR_int *) CLPR_val_to_pc(ti = *(pc + 1))) {
			Regs[arity_of(ti)+1] = (CLPR_int) CLPR_lastcp;
			goto LOOP;
		}
		first_dyn_call = FALSE;
		goto FAIL;
	default:
	   	fprintf(CLPR_error_stream, "(CLPR_code[%d] = %d)\n", intpc(pc), *pc);
		CLPR_fatal("Illegal Instruction");
	}

	/**********/
	END_RUNGOAL: return TRUE;
	/**********/
}

/*---------------------------------------------------------------------------*/

void CLPR_solver_backtrack(CLPR_int i)
{
	switch (tagnum(i)) {
		case NUM_BTK_NULL_EQN: CLPR_restore_null_eqn((CLPR_int *)mask(i)); break;
		case NUM_BTK_EQN: CLPR_restore_eqn((CLPR_int *)pmask(i)); break;
		case NUM_BTK_NULL_INEQ: CLPR_restore_null_ineq((CLPR_int *)mask(i)); break;
		case NUM_BTK_INEQ: CLPR_restore_ineq((CLPR_int *)pmask(i)); break;
		case NUM_BTK_NLIN: CLPR_restore_nl_eqn((CLPR_int *)pmask(i)); break;
		case NUM_BTK_DBIND: *((CLPR_int *)pmask(i)) = TAG_DAEMON; break;
		default: fprintf(stderr, "-- %d --\n", tagnum(i)); CLPR_fatal("79980");
	}
}

CLPR_int CLPR_emul_halt(void)
{
char req[MAX_MESSAGE];
	if (CLPR_lastcp != CLPR_first_lastcp) {
		if (!CLPR_nlin_count && !CLPR_daemons_remaining()) printf("\n*** Retry? ");
		else printf("\n*** (Maybe) Retry? ");
		fflush(stdout);
/*
		fgets(req, 99, CLPR_cur_stream);
		fgets(req, 99, CLPR_see_input);
		CLPR_readline(req, 99);
*/
		CLPR_readline(req, 99);
		if (*req == ';' || *req == 'y' || *req == 'Y') return FALSE;
	} else {
		if (!CLPR_nlin_count && !CLPR_daemons_remaining()) printf("\n*** Yes\n");
		else printf("\n*** Maybe\n");
	}
	return TRUE;
}

CLPR_int CLPR_emul_sys_pred(CLPR_int *pc)
{
CLPR_int ti, (*builtin)(), flag;

	ti = *(pc + 1); 
	builtin = CLPR_spred[ti]->func;
	flag = TRUE;
	switch (CLPR_spred[ti]->arity) {
		case 0:  flag = ((*builtin)()); break;
		case 1:  flag = ((*builtin)(&CLPR_regs[pc[2]])); break;
		case 2:  flag = ((*builtin)(&CLPR_regs[pc[2]], &CLPR_regs[pc[3]])); break;
		case 3:  flag = ((*builtin)(&CLPR_regs[pc[2]], &CLPR_regs[pc[3]], &CLPR_regs[pc[4]]));
				break;
		case 4:  flag = ((*builtin)(&CLPR_regs[pc[2]], &CLPR_regs[pc[3]],
						 &CLPR_regs[pc[4]], &CLPR_regs[pc[5]])); 
				break;
		default: CLPR_fatal("runtime: arity of system predicate too high");
	}
	return flag;
}

void CLPR_emul_dbg_entry(CLPR_int *pc)
{
CLPR_int val, rno, i, n;
	if (CLPR_TRACE) {
		val = *(pc + 1);
		if (!CLPR_p_spy(val)) return;
		rno = *(pc + 2);
		n = arity_of(val);
		if (n > 0) {
				fprintf(CLPR_outfile, "TRY(%d): %s(", rno, string_of(val));
				for (i = 1; i <= n; i++) {
					CLPR_printterm(CLPR_outfile, &CLPR_regs[i]);
					if (i < n) fprintf(CLPR_outfile,", ");				
				}
				fprintf(CLPR_outfile, ")\n");
		} else fprintf(CLPR_outfile, "TRY(%d): %s\n", rno, string_of(val));
	}
}


/*---------------------------------------------------------- unify ----------*/

#define MAX_UNIFY_LEVELS 4096
#define UNIFY_SAFETY 999

CLPR_int CLPR_unify(CLPR_int *t1, CLPR_int *t2)
{
register CLPR_int tp1, tp2;
CLPR_int *CLPR_stack[MAX_UNIFY_LEVELS], top, i, a, *tmp;
double c1, c2;

	top = 0;
	loop {
/*
		CLPR_deref(t1);
		CLPR_deref(t2);
*/
DEREF1:	tp1 = *t1;
		if (tp1 > 0) {
				t1 = (CLPR_int *) tp1;
				goto DEREF1;
		}
DEREF2:	tp2 = *t2;
		if (tp2 > 0) {
				t2 = (CLPR_int *) tp2;
				goto DEREF2;
		}
		switch (CLPR_tag_pair[(CLPR_int) tagnum(tp1)][(CLPR_int) tagnum(tp2)]) {
/*
		switch (CLPR_tag_pair[(CLPR_int) tagnum(*t1)][(CLPR_int) tagnum(*t2)]) {
*/
		case 1: /* 0, TAG_CON  */
		case 2: /* 0, CLPR_TAG_NIL  */
		case 3: /* 0, TAG_PAR  */
		case 4: /* 0, TAG_NUM  */
		case 5: /* 0, CLPR_TAG_CONS */
				*t1 = *t2;
				if (t1 >= CLPR_heap && !safe(t1)) push_trail(BTK_BIND, t1);
				break;
		case 6: /* 0, TAG_STR */
				make_bound_var(t1, t2);
				if (t1 >= CLPR_heap && !safe(t1)) push_trail(BTK_BIND, t1);
				break;
		case 7:  /* TAG_CON, 0  */
		case 8:  /* CLPR_TAG_NIL, 0  */
		case 9:  /* TAG_PAR, 0  */
		case 10: /* TAG_NUM, 0  */
		case 11: /* CLPR_TAG_CONS, 0 */
				*t2 = *t1;
				if (t2 >= CLPR_heap && !safe(t2)) push_trail(BTK_BIND, t2);
				break;
		case 12: /* TAG_STR, 0  */
				make_bound_var(t2, t1);
				if (t2 >= CLPR_heap && !safe(t2)) push_trail(BTK_BIND, t2);
				break;
		case 13: /* TAG_PAR, TAG_PAR */
				if (CLPR_solve_eqn_2(mask(*t1), mask(*t2), 1.0)) break;
				return FALSE;
		case 14: /* TAG_CON, TAG_CON */
				if (*t1 == *t2) break;
				return FALSE;
		case 15: /* CLPR_TAG_NIL, CLPR_TAG_NIL */
				break;
		case 16: /* TAG_STR, TAG_STR */
				if (t1 == t2) break;
				if ((a = mask(*t1)) == mask(*t2)) {
#ifdef CYCLIC_BIND
				  push_trail(BTK_TEST, t1);
				  *t1 = t2;
#endif
					if (top >= MAX_UNIFY_LEVELS - UNIFY_SAFETY) 
						CLPR_fatal("unify stack blown");
					tmp = t1 + CLPR_functors[a].arity;
					do {
						CLPR_stack[++top] = ++t1;
						CLPR_stack[++top] = ++t2;
					} while (t1 < tmp);
					break;
				}
				return FALSE;
		case 17: /* CLPR_TAG_CONS, CLPR_TAG_CONS */
				if (*t1 != *t2) {
					if (top >= MAX_UNIFY_LEVELS - 2) CLPR_fatal("unify stack blown");
					t1 = (CLPR_int *) pmask(*t1);
					t2 = (CLPR_int *) pmask(*t2);
					CLPR_stack[++top] = t1++;
					CLPR_stack[++top] = t2++;
					CLPR_stack[++top] = t1;
					CLPR_stack[++top] = t2;
				}
				break;
		case 18: /* TAG_NUM, TAG_NUM */
				c1 = *((double *)(pmask(*t1)));
				c2 = *((double *)(pmask(*t2)));
				if (is_equal(c1, c2)) break;
				return FALSE;
		case 19: /* TAG_NUM, TAG_PAR */
				if (CLPR_solve_eqn_1((CLPR_int)t2, *((double *) (pmask(*t1))))) break;
				return FALSE;
		case 20: /* TAG_PAR, TAG_NUM */
				if (CLPR_solve_eqn_1((CLPR_int)t1, *((double *) (pmask(*t2))))) break;
				return FALSE;
		case 21:
				if (t1 > t2) {
					make_bound_var(t1, t2);
					if (!safe(t1)) push_trail(BTK_BIND, t1);
				} else if (t1 != t2) {
					make_bound_var(t2, t1);
					if (!safe(t2)) push_trail(BTK_BIND, t2);
				}
				break;
		case 22: /* TAG_DAEMON, (*) */
				if (is_unbound_var(*t2)) {
					make_bound_var(t2, t1);
					if (!safe(t2)) push_trail(BTK_BIND, t2);
				} else {				  
					make_bound_var(t1, t2);
					CLPR_make_runable(t1 + 1);
					if (!safe(t1)) push_trail(BTK_DBIND, t1);
				}
				break;
		case 23: /* (*), TAG_DAEMON */
				if (is_unbound_var(*t1)) {
					make_bound_var(t1, t2);
					if (!safe(t1)) push_trail(BTK_BIND, t1);
				} else {				  
					make_bound_var(t2, t1);
					CLPR_make_runable(t2 + 1);
					if (!safe(t2)) push_trail(BTK_DBIND, t2);
				}
				break;
		case 24: /* TAG_DAEMON, TAG_DAEMON */
				make_bound_var(t2, t1);
				CLPR_make_runable(t2 + 1);
				if (!safe(t2)) push_trail(BTK_DBIND, t2);
				break;
		default:
				return FALSE;
		}
		if (!top) return TRUE;
		t2 = CLPR_stack[top--];
		t1 = CLPR_stack[top--];
	} 
}

/*---------------------------------------------------------- aux stuff ------*/

CLPR_int *CLPR_findbind(CLPR_int *x)
{
	CLPR_deref(x);
	return x;
}

CLPR_int *CLPR_val_to_pc(CLPR_int val)
{
	if (CLPR_p_compiled(val)) return ptrpc(CLPR_p_start_hdrcode(val));
	else return NULL;
}

CLPR_int CLPR_retract_rule(CLPR_int *t, CLPR_int *pc, CLPR_int first)
{
CLPR_int val, *t1, *t2;
CLPR_int *tmp1;
RULE_ptr r, prev_rule;
CLPR_int old_solver_id, old_trtop, *old_heaptop;
CLPR_int has_args, index_val;
double fp_val;

/*
	if (first) t = CLPR_findbind(t);
	else t = (CLPR_int *) CLPR_findbind(CLPR_lastcp[1]);
*/
	if (!first) t = (CLPR_int *) CLPR_lastcp[1];
	if (!is_func(t) && !is_con(t)) CLPR_runtime_error("bad argument to retract/1");
	if ((val = mask(*t)) == IMPL) {
		t1 = CLPR_findbind(t + 1);
		val = mask(*t1);
	} else t1 = t;
	if (CLPR_p_protected(val)) 
		CLPR_runtime_error("retract failed: %s/%d is protected",
				string_of(val), arity_of(val));
	if (!CLPR_p_dynamic(val))
		CLPR_runtime_error("retract failed: %s/%d is not dynamic",
				string_of(val), arity_of(val));
	if (!CLPR_p_compiled(val)) {
		if (first) return FALSE;
		goto CLEANUP_FAIL;
	}
	prev_rule = NULL;
	if (is_func(t1) && ((index_val = CLPR_get_index_val(t1 + 1, &fp_val)) != 0)) {
		has_args = TRUE;
		r = CLPR_quick_find(CLPR_p_first_rule(val), index_val, fp_val, &prev_rule, TRUE);
	} else {
		has_args = FALSE;
		r = CLPR_p_first_rule(val);
	}
	if (first) {
		CLPR_push_cp();
		CLPR_lastcp[NEXTEP] = (CLPR_int) (pc + 2);
		CLPR_lastcp[1] = (CLPR_int) t;
		CLPR_stacktop = CLPR_lastcp + 1;
	} else CLPR_stamp++;
	old_solver_id = CLPR_solver_id;
	old_heaptop = CLPR_heaptop;
	old_trtop = CLPR_trtop;
	while (r) {
		if (r->endcode == -1) goto MORE;
		t2 = CLPR_pterm_to_heap(r->t);
		if (CLPR_unify(t, (CLPR_int *)&t2)) {
			if (CLPR_have_runable && !CLPR_wake_runable()) goto UNDO;
			if (prev_rule) {
				prev_rule->next = r->next;
				if (!r->next) CLPR_assign_p_last_rule(val, prev_rule); 
			} else {
				if (r->next) CLPR_assign_p_first_rule(val, r->next);
				else {
					CLPR_assign_p_first_rule(val, NULL);
					CLPR_assign_p_last_rule(val, NULL);
					CLPR_assign_p_compiled(val, FALSE);
#ifdef FREE_CODE_SPACE
					CLPR_free_hdrcode(val);
#else
                    CLPR_assign_p_start_hdrcode(val, 0);
                    CLPR_assign_p_end_hdrcode(val, 0);
#endif
				}
			}
			CLPR_free_rule(r);
			return TRUE;
		} else {
UNDO:
			CLPR_untrail_to(old_trtop);
			CLPR_trtop = old_trtop;
			CLPR_solver_id = old_solver_id;
			CLPR_heaptop = old_heaptop;
		}
MORE:
		prev_rule = r;
		if (has_args) 
			r = CLPR_quick_find(r->next, index_val, fp_val, &prev_rule, TRUE);
		else r = r->next;
	}
	CLEANUP_FAIL:
	CLPR_stacktop = CLPR_lastcp - CPFRAME_HEIGHT;
	CLPR_lastcp = (CLPR_int *) pmask(*(CLPR_lastcp + PREVCP));
	CLPR_safeheap = (CLPR_int *) pmask(*(CLPR_lastcp + OHEAPTOP));
	return FALSE;
}

CLPR_int CLPR_match_rule(CLPR_int *head, CLPR_int *body, CLPR_int *pc, CLPR_int first)
{
CLPR_int val, old_solver_id, old_trtop, *old_heaptop;
CLPR_int *head_hterm, *body_hterm; 
PTERM *head_pterm, *body_pterm;
RULE_ptr r, old_r, dummy;
CLPR_int current_id, has_args, index_val;
double fp_val;
CLPR_int *t1;
/*
	if (first) {
		head = CLPR_findbind(head);
		body = CLPR_findbind(body);
	} else {
		head = (CLPR_int *) CLPR_lastcp[1];
		body = (CLPR_int *) CLPR_lastcp[2];
	}
*/
	if (!first) {
		head = (CLPR_int *) CLPR_lastcp[1];
		body = (CLPR_int *) CLPR_lastcp[2];
	}
	if (!is_func(head) && !is_con(head)) 
		CLPR_runtime_error("bad first argument to rule/2");
	if (is_param(body)) CLPR_runtime_error("bad second argument to rule/2");
	
	val = mask(*head);
	if (!CLPR_p_dynamic(val))
		CLPR_runtime_error("rule failed: %s/%d is not dynamic",
				string_of(val), arity_of(val));
	if (!CLPR_p_compiled(val)) {
		if (first) return FALSE;
		goto CLEANUP_FAIL;
	}
	r = CLPR_p_first_rule(val);
	if (first) {
		CLPR_push_cp();
		CLPR_lastcp[NEXTEP] = (CLPR_int) (pc + 3);
		CLPR_lastcp[1] = (CLPR_int) head;
		CLPR_lastcp[2] = (CLPR_int) body;
		/*** set below  CLPR_lastcp[3,4] ***/
		CLPR_stacktop = CLPR_lastcp + 4;
	} else {
		current_id = *(CLPR_lastcp + 3);
		old_r = (RULE_ptr) *(CLPR_lastcp + 4);
		CLPR_stamp++;
		if (old_r->id != current_id)
				for (; r && r->id <= current_id; r = r->next);
		else r = old_r->next;
	}
	if (r && is_func(head) && 
				((index_val = CLPR_get_index_val(head+1, &fp_val)) != 0)) {
		has_args = TRUE;
		r = CLPR_quick_find(r, index_val, fp_val, &dummy, TRUE);
	} else has_args = FALSE;

	old_solver_id = CLPR_solver_id;
	old_heaptop = CLPR_heaptop;
	old_trtop = CLPR_trtop;

	while (r) {
		if (r->endcode == -1) goto MORE;
		head_pterm = head_of_rule(r->wt);
		body_pterm = body_of_rule(r->wt);
		head_hterm = CLPR_pterm_to_quoted_heap(head_pterm, FALSE);
		if (CLPR_unify(head, (CLPR_int *)&head_hterm)) {
			if (body_pterm) {
				body_hterm = CLPR_pterm_to_quoted_heap(body_pterm, TRUE);
			} else body_hterm = &true_atom;
			if (!CLPR_unify(body, (CLPR_int *)&body_hterm)) goto UNDO;
			if (CLPR_have_runable && !CLPR_wake_runable()) goto UNDO;
			*(CLPR_lastcp + 3) = r->id;
			*(CLPR_lastcp + 4) = (CLPR_int) r;
			return TRUE;
		} else {
UNDO:
			CLPR_untrail_to(old_trtop);
			CLPR_trtop = old_trtop;
			CLPR_solver_id = old_solver_id;
			CLPR_heaptop = old_heaptop;
		}
MORE:
		if (has_args) r = CLPR_quick_find(r->next, index_val, fp_val, &dummy, TRUE);
		else r = r->next;
	}
	CLEANUP_FAIL:
	CLPR_stacktop = CLPR_lastcp - CPFRAME_HEIGHT;
	CLPR_lastcp = (CLPR_int *) pmask(*(CLPR_lastcp + PREVCP));
	CLPR_safeheap = (CLPR_int *) pmask(*(CLPR_lastcp + OHEAPTOP));
	return FALSE;
}

void CLPR_untrail_to(CLPR_int new_trtop)
{
CLPR_int i, t;
	for (i = CLPR_trtop - 1; i >= new_trtop; i--) {
		t = (CLPR_int) CLPR_trail[i];
		if (CLPR_tag(t) == BTK_BIND) make_unbound_var((CLPR_int *) pmask(t));
		else CLPR_solver_backtrack(t);
	}
}

void CLPR_push_cp(void)
{
CLPR_int *tmp;
	if (CLPR_stacktop < maxlstack && CLPR_trtop < maxtrail) {
		tmp = CLPR_stacktop;
		*(++tmp) = CLPR_slack_id;
		*(++tmp) = CLPR_solver_id;
		*(++tmp) = 0;
		*(++tmp) = (CLPR_int) CLPR_curar;
		*(++tmp) = (CLPR_int) CLPR_retreg;
		*(++tmp) = CLPR_trtop;
		*(++tmp) = (CLPR_int) CLPR_heaptop;
		*(++tmp) = (CLPR_int) CLPR_lastcp;
		*(++tmp) = (CLPR_int) FAIL_ADDR;
		CLPR_stacktop = CLPR_lastcp = tmp;
		CLPR_safeheap = CLPR_heaptop;
		CLPR_stamp++;
	} else if (CLPR_stacktop >= maxlstack) CLPR_fatal("stack overflow");
	else CLPR_fatal("trail overflow");
}


CLPR_int *CLPR_run_dynamic(CLPR_int *pc, CLPR_int first, CLPR_int can_index)
{
RULE_ptr r, dummy;
CLPR_int id, n, i;
CLPR_int index_val;
PTERM *first_arg;
double fp_val;

	if (first) { /*** pseudo-indexing ***/
		r = CLPR_p_first_rule(pc[2]);
		if (can_index && 
					((index_val = CLPR_get_index_val(&CLPR_regs[1], &fp_val)) != 0))
				r = CLPR_quick_find(r, index_val, fp_val, &dummy, FALSE);
		if (!r) return NULL;
		CLPR_push_cp();
		CLPR_lastcp[NEXTEP] = (CLPR_int) pc;
		CLPR_lastcp[NUMPARS] = n = *(pc + 1);
		for (i = 1; i <= n; i++) CLPR_lastcp[i] = CLPR_regs[i];
		CLPR_lastcp[n + 1] = r->id;
		CLPR_lastcp[n + 2] = (CLPR_int) r;
		CLPR_stacktop = CLPR_lastcp + n + 2;
		pc = (CLPR_int *) (CLPR_code + r->startcode);
		return pc;
	} else {
		id = CLPR_lastcp[CLPR_lastcp[NUMPARS] + 1];
		r = (RULE_ptr) CLPR_lastcp[CLPR_lastcp[NUMPARS] + 2];
		if (r->id == id) {
				r = r->next;
		} else {
			for (r = CLPR_p_first_rule(*(pc + 2)); r && r->id <= id; r = r->next);
		}
		if (r && can_index && 
					((index_val = CLPR_get_index_val(&CLPR_regs[1], &fp_val)) != 0))
				r = CLPR_quick_find(r, index_val, fp_val, &dummy, FALSE);
		if (!r) goto CLEANUP_FAIL;
		CLPR_lastcp[CLPR_lastcp[NUMPARS] + 1] = r->id;
		CLPR_lastcp[CLPR_lastcp[NUMPARS] + 2] = (CLPR_int) r;
		CLPR_stamp++;
		pc = (CLPR_int *) (CLPR_code + r->startcode);
		return pc;
	}
	CLEANUP_FAIL:
	CLPR_stacktop = CLPR_lastcp - CPFRAME_HEIGHT;
	CLPR_lastcp = (CLPR_int *) pmask(*(CLPR_lastcp + PREVCP));
	CLPR_safeheap = (CLPR_int *) pmask(*(CLPR_lastcp + OHEAPTOP));
	return NULL;
}

RULE_ptr CLPR_quick_find(RULE_ptr r, CLPR_int val, double fp_val, RULE_ptr *prev_rule, CLPR_int quoted)
{
PTERM *first_arg;
RULE_ptr prev;
CLPR_int val1;
	prev = *prev_rule;
	if (val > 0) {
		for (; r; r = r->next) {
			first_arg = head_of_rule(r->t)->first;
			val1 = (quoted && is_arith_term(first_arg) ?
				CLPR_get_quoted_func(first_arg->val) : first_arg->val);
			if ((is_p_functor(first_arg) && val1 == val) || 
				 is_p_var(first_arg)) {
					*prev_rule = prev;
					return r;
			}
			prev = r;
		}
	} else if (val < 0) {
		for (r = r; r; r = r->next) {
			first_arg = head_of_rule(r->t)->first;
			if ((is_p_real(first_arg) && is_equal(first_arg->rval, fp_val)) ||
				is_p_var(first_arg)) {
					*prev_rule = prev;
					return r;
			}
			prev = r;
		}
	}
	return r;
}

CLPR_int CLPR_get_index_val(CLPR_int *t, double *c)
{
CLPR_int tg;
	t = CLPR_findbind(t);
	tg = CLPR_tag(*t);
	if (tg == CLPR_TAG_CONS) return DOT;
	else if (tg == TAG_CON || tg == TAG_STR) {
		return mask(*t);
	} else if (tg == TAG_PAR && CLPR_is_ground(mask(*t), c)) return -1;
	else return 0;
}

CLPR_int CLPR_make_a_con(CLPR_int *t, CLPR_int val)
{
	t = CLPR_findbind(t);
	if (is_unbound_var(*t)) {
		*t = addtag(TAG_CON, val);
		push_check_trail(BTK_BIND, t);
		return TRUE;
	} else if (is_daemon(t)) {
		*t = addtag(TAG_CON, val);
		CLPR_make_runable(t + 1);
		push_check_trail(BTK_DBIND, t);
		if (CLPR_have_runable && !CLPR_wake_runable()) return FALSE;
		else return TRUE;
	} else if (is_con(t))
		return mask(*t) == val;
	else return FALSE;
}

CLPR_int CLPR_make_a_nil(CLPR_int *t)
{
	t = CLPR_findbind(t);
	if (is_unbound_var(*t)) {
		*t = addtag(CLPR_TAG_NIL, EMPTYLIST);
		push_check_trail(BTK_BIND, t);
		return TRUE;
	} else if (is_daemon(t)) {
		*t = addtag(CLPR_TAG_NIL, EMPTYLIST);
		CLPR_make_runable(t + 1);
		push_check_trail(BTK_DBIND, t);
		if (CLPR_have_runable && !CLPR_wake_runable()) return FALSE;
		else return TRUE;
	} else if (is_nil(t)) return TRUE;
	else return FALSE;
}

CLPR_int CLPR_make_a_num(CLPR_int *t, double val)
{
double c;
	t = CLPR_findbind(t);
	if (is_unbound_var(*t)) {
		CLPR_makenum(++CLPR_solver_id, val);
		*t = addtag(TAG_PAR, CLPR_solver_id);
		push_check_trail(BTK_BIND, t);
		return TRUE;
	} else if (is_daemon(t)) {
		CLPR_makenum(++CLPR_solver_id, val);
		*t = addtag(TAG_PAR, CLPR_solver_id);
		CLPR_make_runable(t + 1);
		push_check_trail(BTK_DBIND, t);
		if (CLPR_have_runable && !CLPR_wake_runable()) return FALSE;
		else return TRUE;
	} else if (is_param(t)) {
		if (CLPR_is_ground(mask(*t), &c)) {
			if (is_equal(c, val)) return TRUE;
			else return FALSE;
		} else if (CLPR_solve_eqn_1(mask(*t), val)) return TRUE;
		else return FALSE;
	} else return FALSE;
}

CLPR_int CLPR_make_a_new_fun(CLPR_int *t, CLPR_int val, CLPR_int arity)
{
CLPR_int *h, i, CLPR_eflag=FALSE;
	t = CLPR_findbind(t);
	if (is_func(t)) return (val == mask(*t));
	if (is_daemon(t)) CLPR_eflag = TRUE;
	else if (!is_unbound_var(*t)) return FALSE;
	h = CLPR_heaptop;
	CLPR_heaptop += arity + 1;
	make_bound_var(t, (CLPR_int) h);
	make_bound_var(h, addtag(TAG_STR, val));
	for (i=0; i < arity; i++)
		make_unbound_var(&h[i+1]);
	if (CLPR_eflag) {
		CLPR_make_runable(t + 1);
		push_check_trail(BTK_DBIND, t);
	} else push_check_trail(BTK_BIND, t);
	return TRUE;
}

CLPR_int CLPR_make_a_new_list(CLPR_int *t)
{
CLPR_int *h, CLPR_eflag=FALSE;
	t = CLPR_findbind(t);
	if (is_daemon(t)) CLPR_eflag = TRUE;
	else if (!is_unbound_var(*t)) return FALSE;
	h = CLPR_heaptop;
	CLPR_heaptop += 2;
	make_bound_var(t, addtag(CLPR_TAG_CONS, h));	
	make_unbound_var(h);
	make_unbound_var(h+1);
	if (CLPR_eflag) {
		CLPR_make_runable(t + 1);
		push_check_trail(BTK_DBIND, t);
	} else push_check_trail(BTK_BIND, t);
	return TRUE;
}

void CLPR_push_destructive_trail(CLPR_int *t, CLPR_int val)
{
    if (t < CLPR_safeheap || (t > CLPR_lstack && t < CLPR_lastcp)) { 
		CLPR_trail[CLPR_trtop++] = (CLPR_int *) val;
        CLPR_trail[CLPR_trtop++] = (CLPR_int *) addtag(BTK_TEST2, (CLPR_int *) t); 
	}
}

	
