/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: emul.h                                               *
 ***************************************************************/

#include "standard.h"

#define CPFRAME_HEIGHT     9        /* Number of entries below */
#define NEXTEP             0        /* Next environment pointer */
#define PREVCP            -1        /* Previous choice point on stack */
#define OHEAPTOP          -2        /* Previous heap top */
#define OTRTOP            -3        /* Previous trail top */
#define CPRETPT           -4        /* Return Point */
#define OCURAR            -5        /* Previous current activation record */
#define NUMPARS           -6        /* Number of parameters */
#define SOLVER_ID         -7        /* Max solver id */
#define SLACK_ID          -8        /* Max solver slack id */

#define FRAME_HEIGHT       2
#define FATHER             0
#define RETPT             -1

#define MAX_TAGS       (16)
#define NONTAG_BITS	(CLPR_WORDSIZE - 4)
#define def_tag(x)	(((CLPR_int)(x)) << NONTAG_BITS)

#define TAG_CON        def_tag(0x8)
#define TAG_STR        def_tag(0x9)
#define TAG_NUM        def_tag(0xa)
#define TAG_PAR        def_tag(0xb)
#define TAG_MISC       def_tag(0xc)
#define CLPR_TAG_CONS  def_tag(0xd)
#define CLPR_TAG_NIL   def_tag(0xe)
#define TAG_DAEMON     def_tag(0xf)
#define TAG_UNBOUND    def_tag(0)

#define BTK_BIND        def_tag(0x0) /* trail tags */
#define BTK_NULL_EQN    def_tag(0x1)
#define BTK_EQN         def_tag(0x2)
#define BTK_INEQ        def_tag(0x3)
#define BTK_NULL_INEQ   def_tag(0x4)
#define BTK_NLIN        def_tag(0x5)
#define BTK_DBIND       def_tag(0x6)
#define BTK_TEST	def_tag(0x7)
#define BTK_TEST2	def_tag(0x8)
#define BTK_TEST3	def_tag(0x9)

#define NUM_BTK_BIND        (0)
#define NUM_BTK_NULL_EQN    (1)
#define NUM_BTK_EQN         (2)
#define NUM_BTK_INEQ        (3)
#define NUM_BTK_NULL_INEQ   (4)
#define NUM_BTK_NLIN        (5)
#define NUM_BTK_DBIND       (6)

#define DBG_MASK_ON         (1)
#define DBG_MASK_SPY        (2)

#define IO_READ             (1)
#define IO_WRITE            (2)

#define GOAL_START          (5)
#define FAIL_ADDR           (0)
#define GFAIL_ADDR          (1)

/*---------------------------------------------------------------------------*/

#define bit1(i) ((i) & (CLPR_int)0x01)
#define bit2(i) ((i) & (CLPR_int)0x02)
#define bit3(i) ((i) & (CLPR_int)0x04)
#define bit4(i) ((i) & (CLPR_int)0x08)
#define bit5(i) ((i) & (CLPR_int)0x10)
#define bit6(i) ((i) & (CLPR_int)0x20)

#define set_bit1(i) (i) |= (CLPR_int)0x01
#define set_bit2(i) (i) |= (CLPR_int)0x02
#define set_bit3(i) (i) |= (CLPR_int)0x04
#define set_bit4(i) (i) |= (CLPR_int)0x08
#define set_bit5(i) (i) |= (CLPR_int)0x10
#define set_bit6(i) (i) |= (CLPR_int)0x20

#define unset_bit1(i) (i) &= ~(CLPR_int)0x01
#define unset_bit2(i) (i) &= ~(CLPR_int)0x02
#define unset_bit3(i) (i) &= ~(CLPR_int)0x04
#define unset_bit4(i) (i) &= ~(CLPR_int)0x08
#define unset_bit5(i) (i) &= ~(CLPR_int)0x10
#define unset_bit6(i) (i) &= ~(CLPR_int)0x20

/*---------------------------------------------------------------------------*/

#define safe(x) (((x) > CLPR_lastcp) || ((x) < CLPR_lstack && (x) >= CLPR_safeheap))
#define CLPR_tag(i) (((CLPR_int) (i)) & def_tag(0xF))

#ifdef AIX        /* problem with cc with unsigned shifts ! */
#define tagnum(i) (((i) >> NONTAG_BITS) & 0xF)
#else
#define tagnum(i) ( ((CLPR_unsigned)(i)) >> NONTAG_BITS)
#endif
#define mask(i) ((i) & (def_tag(0x1) - 1))
#define addtag(t, i) ((CLPR_int) ((t) | (((CLPR_int)(i)) & (def_tag(0x1) - 1))))
#define addtag1(t, i) ((CLPR_int) (((CLPR_int) (t)) | (i)))
#define add_misctag(i) (i)

/*---------------------------------------------------------------------------*/

#ifdef IBMRT
#ifdef AIX
#define pmask(i) ((((CLPR_int) (i)) & 0x0FFFFFFF) | 0x20000000)
#else
#define pmask(i) ((((CLPR_int) (i)) & 0x0FFFFFFF) | 0x10000000)
#endif
#else
#ifdef HP835
#define pmask(i) ((((CLPR_int) (i)) & 0x0FFFFFFF) | 0x40000000)
#else
#ifdef MIPS
#define pmask(i) ((((CLPR_int) (i)) & 0x0FFFFFFF) | 0x10000000)
#else
#ifdef __alpha__
#define pmask(i) ((((CLPR_int) (i)) & 0x0FFFFFFFFFFFFFFF) | 0x0000000000000000)
#else
#define pmask(i) (((CLPR_int) (i)) & 0x0FFFFFFF)
#endif
#endif
#endif
#endif

#define intptr(p) ((CLPR_unsigned) p)
#define MASK4 def_tag(0xf)

/*---------------------------------------------------------------------------*/

#define is_con(t) (CLPR_tag(*(t)) == TAG_CON)
#define is_nil(t) (CLPR_tag(*(t)) == CLPR_TAG_NIL)
#define is_func(t) (CLPR_tag(*(t)) == TAG_STR)
#define is_cons(t) (CLPR_tag(*(t)) == CLPR_TAG_CONS)
#define is_param(t) (CLPR_tag(*(t)) == TAG_PAR)
#define is_num(t, v) ((CLPR_tag(*(t)) == TAG_PAR) && CLPR_is_ground(mask(*(t)), v))

#define is_bound_var(i) ((i) > 0)
#define is_unbound_var(i) (!(i))
#define is_daemon(i) (CLPR_tag(*(i)) == TAG_DAEMON)
#define is_pointer(i) ((CLPR_int)(i) >= 0)
#define is_data(i) ((i) < 0)
#define make_unbound_var(i) (*(i) = (CLPR_int) NULL)
#define make_bound_var(i, j) (*(i) = (CLPR_int) (j))

#define is_stack_ptr(p) (p >= CLPR_lstack)
#define is_heap_ptr(p) (p < CLPR_heaptop && p >= CLPR_heap)
#define below_heap(p) (p < CLPR_heaptop)

#define arity_of(t)		(CLPR_functors[mask(t)].arity)
#define string_of(t)    (CLPR_functors[mask(t)].pname)
#define opkind_of(t)    (CLPR_functors[mask(t)].special)
#define intpc(p) ((((CLPR_int) (p)) - ((CLPR_int) CLPR_code)) / sizeof(CLPR_int))
#define ptrpc(i) ((CLPR_int *) (CLPR_code + (i)))
#define var2pos(p, base) ((((CLPR_int) (p)) - ((CLPR_int) base)) / sizeof(CLPR_int))

#define head_of(t) ((CLPR_int *) (pmask(*(t))))
#define tail_of(t) (((CLPR_int *) pmask(*(t))) + 1)

/*---------------------------------------------------------------------------*/

#define CLPR_deref(p) while (*(p) > 0) p = (CLPR_int *) *(p);

#define push_trail(btk_type, addr) \
    CLPR_trail[CLPR_trtop++] = (CLPR_int *) addtag(btk_type, (CLPR_int *) (addr)) 

#define push_check_trail(btk_type, addr) \
    if ((addr) < CLPR_safeheap || ((addr) > CLPR_lstack && (addr) < CLPR_lastcp)) { \
        CLPR_trail[CLPR_trtop++] = (CLPR_int *) addtag(btk_type, (CLPR_int *) (addr)); \
    }

#define check_tag_par(tmp) { \
	CLPR_deref(tmp); \
	if (is_unbound_var(*(tmp))) { \
		*(tmp) = addtag(TAG_PAR, ++CLPR_solver_id); \
		push_check_trail(BTK_BIND, (tmp)); \
	} else if (is_daemon(tmp)) { \
		make_bound_var(tmp, CLPR_heaptop); \
		*CLPR_heaptop++ = addtag(TAG_PAR, ++CLPR_solver_id); \
		push_check_trail(BTK_DBIND, tmp); \
		CLPR_make_runable(tmp + 1); \
		tmp = (CLPR_int *) pmask(*tmp); \
	} else if (!is_param(tmp)) goto FAIL; \
}

#define make_tag_par(tmp) { \
	*(tmp) = addtag(TAG_PAR, ++CLPR_solver_id); \
	push_check_trail(BTK_BIND, (tmp)); \
}

#define caseop(x) case (CLPR_int) x

/*---------------------------------------------------------------------------*/

extern void CLPR_push_destructive_trail (CLPR_int *t, CLPR_int val);

