/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: index.c                                              *
 ***************************************************************/

#include <string.h>
#include "standard.h"
#include "config.h"
#include "parse.h"
#include "symbols.h"
#include "lib.h"
#include "compile.h"
#include "opcodes.h"
#include "emul.h"
#include "main.h"

#define MAX_INDEX_CHAIN 256 /*** do not increase ***/
#define MAX_INDEX_CLAUSES 256
#define MAX_INDEX_TYPES N_TOKEN_TYPES

#define FASTLABEL	(1)
#define NOFASTLABEL	(0)

#define TYPELABEL	(1)
#define CHAINLABEL	(0)

#define FAIL_PROC "E$0_$fail/0"

#define type_chain(i)   (CLPR_index_chains[i].type)
#define val_chain(i)    (CLPR_index_chains[i].val)
#define length_chain(i) (CLPR_index_chains[i].tail[0])
#define rule_in_chain(c,n) (CLPR_index_chains[c].tail[n])

struct chain {
	CLPR_int val;
	CLPR_int type;
	CLPR_int tail[MAX_INDEX_CHAIN + 1];
} CLPR_index_chains[MAX_INDEX_CLAUSES + 2];

static CLPR_int n_rules;
static char *p_name;
static CLPR_int p_arity;
static CLPR_int p_val;

static CLPR_int first_rule[MAX_INDEX_TYPES + 1];
static CLPR_int freq_rule[MAX_INDEX_TYPES + 1];
static CLPR_int card_type[MAX_INDEX_TYPES + 1];
static CLPR_int n_atoms, n_funcs, n_reals;

static char *make_itag(CLPR_int type, CLPR_int val, CLPR_int block);
static char *make_rtag(CLPR_int rule_number);
static char *make_ttag(CLPR_int type);

char *CLPR_make_ftag(char *name, CLPR_int arity, CLPR_int rule_number);

extern FUNCTREC CLPR_functors[];
extern CLPR_int CLPR_CODEGEN_DEBUG;

extern char *CLPR_maketag(CLPR_int n, char *extra, char *str, CLPR_int arity);
extern float CLPR_int_to_float(CLPR_int i);
extern char *CLPR_char_malloc(CLPR_int n);
 
/*---------------------------------------------------------------------------*/


extern void CLPR_storename (char *s, CLPR_int val, CLPR_int narg);
extern CLPR_int CLPR_p_cut (CLPR_int name);
extern void CLPR_emit_savecurcp (CLPR_int R);
void CLPR_gen_all_chain (CLPR_int n_rules);
void CLPR_init_index (void);
void CLPR_make_index_chains (RULE *r);
void CLPR_gen_general_chain (void);
void CLPR_choose_a_switch (void);
extern void CLPR_emit_try (char *s, CLPR_int val, CLPR_int n);
extern void CLPR_emit_retry (char *s, CLPR_int val);
extern void CLPR_emit_trust (char *s, CLPR_int val);
void CLPR_add_to_all_chains (CLPR_int rule_number);
void CLPR_add_to_chain (CLPR_int type, CLPR_int val, CLPR_int rule_number);
extern CLPR_int CLPR_float_to_int (float f);
extern void CLPR_show_pterm (PTERM *t);
void CLPR_gen_switch1type (CLPR_int type);
void CLPR_gen_switchtype (void);
extern void CLPR_emit_switchdot (char *s, CLPR_int val);
extern void CLPR_emit_switchnil (char *s, CLPR_int val);
void CLPR_gen_default_chain (void);
extern void CLPR_emit_switchcon (char *s, CLPR_int val);
extern void CLPR_emit_switchfunc (char *s, CLPR_int val);
extern void CLPR_emit_switchreal (char *s, CLPR_int val);
void CLPR_gen_jump_table (CLPR_int type);
void CLPR_gen_index_chain (CLPR_int type, CLPR_int val);
extern void CLPR_emit_switchtype (char *s1, char *s2, char *s3, char *s4, char *s5, CLPR_int val1, CLPR_int val2, CLPR_int val3, CLPR_int val4, CLPR_int val5);
void CLPR_gen_secondary_switch1type (CLPR_int type);
extern void CLPR_emit_switch2con (char *s, CLPR_int val);
extern void CLPR_emit_switch2func (char *s, CLPR_int val);
extern void CLPR_emit_switch2real (char *s, CLPR_int val);
extern void CLPR_emit_word (CLPR_int w);
extern void CLPR_emit_addr (char *s, CLPR_int val);

void CLPR_gen_index_code(RULE *r, char *str, CLPR_int a, CLPR_int nclauses)
{
CLPR_int i, j, narg;
char *s;
	s = CLPR_maketag(0, "", str, a);
	if (r->wt->val == IMPL) {
		p_val = r->wt->first->val;
		narg = r->wt->first->narg;
		CLPR_storename(s, p_val, r->wt->first->narg);
	} else {
		p_val = r->wt->val;
		narg = r->wt->narg;
		CLPR_storename(s, p_val, r->wt->narg);
	}
	if (CLPR_p_cut(p_val)) CLPR_emit_savecurcp(narg+1);
	p_name = str;
	p_arity = a;
	n_rules = nclauses;
	if (!a || nclauses > MAX_INDEX_CLAUSES) {
		CLPR_gen_all_chain(nclauses);
		return;
	}
	CLPR_init_index();
	CLPR_make_index_chains(r);
	if (freq_rule[0] == n_rules) CLPR_gen_general_chain();
	else CLPR_choose_a_switch();
	return;
}

void CLPR_gen_all_chain(CLPR_int n_rules)
{
	CLPR_int i;

	CLPR_emit_try(make_rtag(1), p_val, p_arity);
	for (i = 2; i < n_rules; i++) CLPR_emit_retry(make_rtag(i), p_val);
	CLPR_emit_trust(make_rtag(i), p_val);
}

void CLPR_make_index_chains(RULE *r)
{
CLPR_int i;
PTERM *arg1;
	for (i = 1; r ; r = r->next, i++) {
		if (r->wt->val == IMPL) arg1 = r->wt->first->first;
		else arg1 = r->wt->first;
		if (is_p_var(arg1) || (is_arith_term(arg1) && !is_p_real(arg1))) {
			if (!first_rule[0]) first_rule[0] = i;
			freq_rule[0]++;
			CLPR_add_to_all_chains(i);
		} else if (is_p_emptylist(arg1)) {
			if (!first_rule[EMPTYLIST]) first_rule[EMPTYLIST] = i;	
			freq_rule[EMPTYLIST]++;
			CLPR_add_to_chain(EMPTYLIST, arg1->val, i);
		} else if (is_p_list(arg1)) {
			if (!first_rule[DOT]) first_rule[DOT] = i;	
			freq_rule[DOT]++;
			CLPR_add_to_chain(DOT, arg1->val, i);
		} else if (is_p_constant(arg1)) {
			if (!first_rule[ATOM]) first_rule[ATOM] = i;	
			freq_rule[ATOM]++;
			CLPR_add_to_chain(ATOM, arg1->val, i);
		} else if (is_p_func(arg1)) {
			if (!first_rule[FUNC]) first_rule[FUNC] = i;	
			freq_rule[FUNC]++;
			CLPR_add_to_chain(FUNC, arg1->val, i);
		} else if (is_p_real(arg1)) {
			if (!first_rule[REAL]) first_rule[REAL] = i;	
			freq_rule[REAL]++;
			CLPR_add_to_chain(REAL, CLPR_float_to_int((float) arg1->rval), i);
		} else {
			CLPR_show_pterm(arg1);
			CLPR_fatal("52070");
		}
	}
}

void CLPR_choose_a_switch(void)
{
CLPR_int i, count, type;
	count = 0;
	for (i = 1; i <= MAX_INDEX_TYPES; i++) 
		if (freq_rule[i]) {
			count++;
			type = i;
		}
	if (count == 1) CLPR_gen_switch1type(type);
	else CLPR_gen_switchtype();
}

void CLPR_gen_switch1type(CLPR_int type) /*** 1 type ***/
         
{
CLPR_int i, r0, l0, val;
char *s;
	val = p_val;
	if (type == EMPTYLIST || type == DOT) {
		r0 = first_rule[0];
		l0 = length_chain(0);
		if (l0 == 0) {
			s = FAIL_PROC; 
			val = FAILVAL;
		}
		else if (l0 == 1) s = make_rtag(r0);
		else s = make_itag(0, 0, TYPELABEL);
		if (type == DOT) CLPR_emit_switchdot(s, val);
		else CLPR_emit_switchnil(s, val);
		CLPR_gen_general_chain();
		if (l0 >= 2) {
			CLPR_storename(s, val, p_arity);
			CLPR_gen_default_chain();
		}
		return;
	}
#ifdef NEWINDEX
	if (type != REAL && card_type[type] == 1 && 
			(freq_rule[type] == length_chain(first_rule[type]) ||
			rule_in_chain(0,1) == rule_in_chain(first_rule[type],1))) {
		CLPR_gen_single_chain(NOFASTLABEL, first_rule[type]);
		return;
	}
#endif
	switch (type) {
		case ATOM: CLPR_emit_switchcon(make_ttag(ATOM), p_val); break;
		case FUNC: CLPR_emit_switchfunc(make_ttag(FUNC), p_val); break;
		case REAL: CLPR_emit_switchreal(make_ttag(REAL), p_val);
	}
	CLPR_gen_jump_table(type);
	CLPR_gen_general_chain();
	for (i = 1; i <= n_rules; i++) 
		if (type_chain(i) == type && length_chain(i) >= 2) {
			val = CLPR_index_chains[i].val;
			CLPR_storename(make_itag(type, val, CHAINLABEL), p_val, p_arity);
			CLPR_gen_index_chain(type, val);
		}
	if (length_chain(0) >= 2) {
		CLPR_storename(make_itag(0, 0, TYPELABEL), p_val, p_arity);
		CLPR_gen_default_chain();
	}
}

void CLPR_gen_switchtype(void) /*** at least two types ***/
{
char *s[MAX_INDEX_TYPES + 1];
CLPR_int val[MAX_INDEX_TYPES + 1];
CLPR_int has_chain[MAX_INDEX_TYPES + 1];
CLPR_int i;
	for (i = 1; i <= MAX_INDEX_TYPES; i++) {
		val[i] = p_val;
		has_chain[i] = FALSE;
	}
	switch (length_chain(0)) {
	case 0:
		for (i = 1; i <= MAX_INDEX_TYPES; i++) {
			if (freq_rule[i] == 0) {
				s[i] = FAIL_PROC;
				val[i] = FAILVAL;
			} else if (freq_rule[i] == 1) {
				if (i == EMPTYLIST || i == DOT)
					s[i] = CLPR_make_ftag(p_name, p_arity, first_rule[i]);
				else s[i] = make_rtag(first_rule[i]);
			} else {
				s[i] = make_itag(i, 0, TYPELABEL);
				has_chain[i] = TRUE;
			}
		}
		break;
	case 1:
		for (i = 1; i <= MAX_INDEX_TYPES; i++) {
			if (freq_rule[i] == 0) s[i] = make_rtag(first_rule[0]);
 			else {
				s[i] = make_itag(i, 0, TYPELABEL);
				has_chain[i] = TRUE;
			}
		}
		break;
	default:
		for (i = 1; i <= MAX_INDEX_TYPES; i++) {
			if (freq_rule[i] == 0) s[i] = make_itag(0, 0, TYPELABEL);
			else {
				s[i] = make_itag(i, 0, TYPELABEL); 
				has_chain[i] = TRUE;
			}
		}
	}
	CLPR_emit_switchtype(s[EMPTYLIST],   s[DOT],   s[ATOM],   s[FUNC],   s[REAL],
	                val[EMPTYLIST], val[DOT], val[ATOM], val[FUNC], val[REAL]);
	CLPR_gen_general_chain();
	for (i = 1; i <= MAX_INDEX_TYPES; i++)
		if (has_chain[i]) { /*** (s[i] && *(s[i]) == 'L') { ***/
			CLPR_storename(s[i], p_val, p_arity);
			CLPR_gen_secondary_switch1type(i);
		}
	if (length_chain(0) >= 2) {
		CLPR_storename(make_itag(0, 0, TYPELABEL), p_val, p_arity);
		CLPR_gen_default_chain();
	}
}

void CLPR_gen_secondary_switch1type(CLPR_int type)
{
CLPR_int i, n, val;
char *s1, *s2;
	if (type == EMPTYLIST || type == DOT) {
		CLPR_gen_index_chain(type, 0);
		return;
	}
#ifdef NEWINDEX
	if (type != REAL && card_type[type] == 1) {
		CLPR_gen_single_chain(NOFASTLABEL, first_rule[type]);
		return;
	}
#endif
	switch (type) {
		case ATOM: CLPR_emit_switch2con(make_ttag(ATOM), p_val); break;
		case FUNC: CLPR_emit_switch2func(make_ttag(FUNC), p_val); break;
		case REAL: CLPR_emit_switch2real(make_ttag(REAL), p_val);
	}
	CLPR_gen_jump_table(type);
	for (i = 1; i <= n_rules; i++) 
		if (type_chain(i) == type && length_chain(i) >= 2) {
			val = CLPR_index_chains[i].val;
			CLPR_storename(make_itag(type, val, CHAINLABEL), p_val, p_arity);
			CLPR_gen_index_chain(type, val);
		}
}

void CLPR_gen_general_chain(void)
{
CLPR_int i;
	CLPR_emit_try(make_rtag(1), p_val, p_arity);
	for (i = 2; i < n_rules; i++) CLPR_emit_retry(make_rtag(i), p_val);
	CLPR_emit_trust(make_rtag(i), p_val);
}

void CLPR_gen_jump_table(CLPR_int type) /*** (# val's) + 1 doublewords ***/
         
{
CLPR_int i, val, arity;
	for (i = 1; i <= n_rules; i++) {
		if (type_chain(i) != type) continue;
		val = CLPR_index_chains[i].val;
		switch (type) {
			case ATOM: CLPR_emit_word(TAG_CON | val); break;
			case FUNC: CLPR_emit_word(TAG_STR | val); break;
			case REAL: CLPR_emit_word(val);
		}
		if (length_chain(i) == 1) 
			CLPR_emit_addr(CLPR_make_ftag(p_name, p_arity, i), p_val);
		else CLPR_emit_addr(make_itag(type, val, CHAINLABEL), p_val);
	}
	CLPR_storename(make_ttag(type), p_val, p_arity);
	CLPR_emit_word(0);
	switch (length_chain(0)) {
		case 0:  CLPR_emit_addr(FAIL_PROC, FAILVAL); break;
		case 1:  CLPR_emit_addr(make_rtag(first_rule[0]), p_val); break;
		default: CLPR_emit_addr(make_itag(0, 0, TYPELABEL), p_val);
	}
}

void CLPR_gen_index_chain(CLPR_int type, CLPR_int val)
{
CLPR_int i, r;
	r = first_rule[type];
	if (type != EMPTYLIST && type != DOT)
		while (CLPR_index_chains[r].val != val) r++;
	if (first_rule[type] == CLPR_index_chains[r].tail[1]) 
		CLPR_emit_try(CLPR_make_ftag(p_name, p_arity, CLPR_index_chains[r].tail[1]), 
			p_val, p_arity);
	else
		CLPR_emit_try(make_rtag(CLPR_index_chains[r].tail[1]), p_val, p_arity);
	for (i = 2; i < length_chain(r); i++) 
		CLPR_emit_retry(make_rtag(CLPR_index_chains[r].tail[i]), p_val);
	CLPR_emit_trust(make_rtag(CLPR_index_chains[r].tail[i]), p_val);
}

void CLPR_gen_single_chain(CLPR_int fast, CLPR_int start)
{
CLPR_int i, r;
	r = start;
	if (fast == FASTLABEL && r == CLPR_index_chains[r].tail[1]) 
		CLPR_emit_try(CLPR_make_ftag(p_name, p_arity, CLPR_index_chains[r].tail[1]), 
			p_val, p_arity);
	else
		CLPR_emit_try(make_rtag(CLPR_index_chains[r].tail[1]), p_val, p_arity);
	for (i = 2; i < length_chain(r); i++) 
		CLPR_emit_retry(make_rtag(CLPR_index_chains[r].tail[i]), p_val);
	CLPR_emit_trust(make_rtag(CLPR_index_chains[r].tail[i]), p_val);
}

void CLPR_gen_default_chain(void)
{
CLPR_int i;
	CLPR_emit_try(make_rtag(CLPR_index_chains[0].tail[1]), p_val, p_arity);
	for (i = 2; i < length_chain(0); i++)
		CLPR_emit_retry(make_rtag(CLPR_index_chains[0].tail[i]), p_val);
	CLPR_emit_trust(make_rtag(CLPR_index_chains[0].tail[i]), p_val);
}

/*----------------------------------------------------------- aux stuff -----*/

void CLPR_init_index(void)
{
CLPR_int i;
	n_atoms = n_funcs = n_reals = 0;
	for (i = 0; i <= MAX_INDEX_TYPES; i++) 
		card_type[i] = first_rule[i] = freq_rule[i] = 0;
	for (i = 0; i <= MAX_INDEX_CLAUSES; i++) 
		type_chain(i) = val_chain(i) = length_chain(i) = 0;
}

void CLPR_add_to_chain(CLPR_int type, CLPR_int val, CLPR_int rule_number)
{
CLPR_int i, j;
	if (type == EMPTYLIST || type == DOT) {
		if (!(i = first_rule[type])) {
			i = rule_number;
			type_chain(i) = type;
		}
		val_chain(i) = 0; 
		j = length_chain(i);
		CLPR_index_chains[i].tail[++j] = rule_number;
		length_chain(i) = j;
		return;
	}				
	for (i = 1; i <= rule_number; i++) {
		if (type_chain(i) != type || val_chain(i) != val) continue;
		j = length_chain(i);
		CLPR_index_chains[i].tail[++j] = rule_number;
		length_chain(i) = j;
		return;
	}
	type_chain(rule_number) = type;
	val_chain(rule_number) = val;
	i = length_chain(rule_number);
	CLPR_index_chains[rule_number].tail[++i] = rule_number;
	length_chain(rule_number) = i;
	card_type[type]++;
	if (type == ATOM) n_atoms++;
	else if (type == FUNC) n_funcs++;
	else if (type == REAL) n_reals++;
}

void CLPR_add_to_all_chains(CLPR_int rule_number)
{
CLPR_int i, j;
	for (i = 0; i <= n_rules; i++) {
		j = length_chain(i);
		CLPR_index_chains[i].tail[++j] = rule_number;
		length_chain(i) = j;
	}
}

void CLPR_print_chains(void)
{
CLPR_int i, j, r, count, last;
	count = 0;
	for (i = 1; i <= MAX_INDEX_TYPES; i++) {
		if (!(r = first_rule[i])) continue;
		printf("Type %d: ", i);
		if (type_chain(r) == REAL) 
			printf("%f: ", CLPR_int_to_float(val_chain(r)));
		else 
			printf("%s: ", string_of(CLPR_index_chains[r].val));
		for (j = 1; j <= length_chain(r); j++) 
			printf(" %d", CLPR_index_chains[r].tail[j]);
		putchar('\n');
	}
	printf("def: ");
	i = length_chain(0);
	if (i) for (j = 1; j <= i; j++) printf(" %d", CLPR_index_chains[0].tail[j]);
	else printf(" $fail");
	putchar('\n');
}

static char *make_itag(CLPR_int type, CLPR_int val, CLPR_int block)
{
char buf[256], *s;

	switch (type) {
	case EMPTYLIST:
		sprintf(buf, "L_%s/%d_nil", p_name, p_arity);
		break;
	case DOT:
		sprintf(buf, "L_%s/%d_list", p_name, p_arity);
		break;
	case ATOM:
	case FUNC:
		if (block)
			sprintf(buf, "L_%s/%d_%d", p_name, p_arity, type);
		else
			sprintf(buf, "L_%s/%d_%d:%d", p_name, p_arity, type, val);
		break;
	case REAL:
		if (block)
			sprintf(buf, "L_%s/%d_%d", p_name, p_arity, type);
		else
			sprintf(buf, "L_%s/%d_%d:$%d", p_name, p_arity, type, val);
		break;
	case 0:
		if (val == 0 && block)
			sprintf(buf, "L_%s/%d_def", p_name, p_arity);
		else CLPR_fatal("make_itag 13871");
		break;
	default:
		CLPR_fatal("make_itag 13871");
	}
	s = CLPR_char_malloc(strlen(buf) + 1);
	strcpy(s, buf);
	return s;
}
 
static char *make_rtag(CLPR_int rule_number)
{
char buf[256], *s;
	sprintf(buf, "E$%d_%s/%d", rule_number, p_name, p_arity);
	s = CLPR_char_malloc(strlen(buf) + 1);
	strcpy(s, buf);
	return s;
}

char *CLPR_make_ftag(char *name, CLPR_int arity, CLPR_int rule_number)
{
char buf[256], *s;
	if (CLPR_CODEGEN_DEBUG) sprintf(buf, "E$%d_%s/%d", rule_number, name, arity);
	else sprintf(buf, "F$%d_%s/%d", rule_number, name, arity);
	s = CLPR_char_malloc(strlen(buf) + 1);
	strcpy(s, buf);
	return s;
}
 
static char *make_ttag(CLPR_int type) /*** only for last jump table entry ***/
         
{
char buf[256], *s;
	switch (type) {
		case EMPTYLIST: s = "nil"; break;
		case DOT: s = "list"; break;
		case ATOM: s = "con"; break;
		case FUNC: s = "func"; break;
		case REAL: s = "real"; break;
		default: CLPR_fatal("94711");
	}
	sprintf(buf, "T_%s/%d_%s", p_name, p_arity, s);
	s = CLPR_char_malloc(strlen(buf) + 1);
	strcpy(s, buf);
	return s;
}
 
