/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: init.c                                               *
 ***************************************************************/

#include <stdarg.h>
#include <stdlib.h>
#include "standard.h"
#include "config.h"
#include "parse.h"
#include "lib.h"
#include "symbols.h"
#include "opcodes.h"
#include "emul.h"
#include "solver.h"
#include "table.h"
#include "main.h"

char *CLPR_CLP_SUFFIX;

PTERM *CLPR_null_list;	/* setup some predefined stuff for parser */
struct HASHNODE *CLPR_hnode_emptylist;
char *CLPR_cons_str;		/* dito */
SPNODE *CLPR_spred[MAX_SPRED];

struct {
	char *name;
	CLPR_int narg;
	CLPR_int val;
} CLPR_def_p_func[] = {
	{".", 2, DOT},
	{"|", 2, 4},
	{",", 2, COMMA},
	{":-", 2, IMPL},
	{"[]", 0, EMPTYLIST},
	{"?-", 1, QUERY},
	{"<stdin>", 0, STDINVAL},
	{"<stdout>", 0, STDOUTVAL},
	{"end", 0, END},
	{"user", 0, USER},
        {"::-", 1, COMMAND},
	{":-", 1, DIRECTIVE},
	{";", 2, DISJ},
	{"+", 2, PLUSVAL},
	{"-", 1, U_MINUSVAL},
	{"-", 2, MINUSVAL},
	{"*", 2, MULTVAL},
	{"/", 2, DIVVAL},
	{"sin", 1, SINVAL},
	{"cos", 1, COSVAL},
	{"abs", 1, ABSVAL},
	{"min", 2, MINVAL},
	{"max", 2, MAXVAL},
	{"pow", 2, POWVAL},
	{"eval", 1, EVALVAL},
	{"quote", 1, QUOTEVAL},
	{"arcsin", 1, ARCSINVAL},
	{"arccos", 1, ARCCOSVAL},
/***********************************/
	{"^+", 2, HAT_PLUSVAL},
	{"^-", 1, HAT_U_MINUSVAL},
	{"^-", 2, HAT_MINUSVAL},
	{"^*", 2, HAT_MULTVAL},
	{"^/", 2, HAT_DIVVAL},
	{"^sin", 1, HAT_SINVAL},
	{"^cos", 1, HAT_COSVAL},
	{"^abs", 1, HAT_ABSVAL},
	{"^min", 2, HAT_MINVAL},
	{"^max", 2, HAT_MAXVAL},
	{"^pow", 2, HAT_POWVAL},
	{"^eval", 1, HAT_EVALVAL},
	{"^quote", 1, HAT_QUOTEVAL},
	{"^arcsin", 1, HAT_ARCSINVAL},
	{"^arccos", 1, HAT_ARCCOSVAL},
/***********************************/
	{"=", 2, EQVAL},
	{"<", 2, LTVAL},
	{">", 2, GTVAL},
	{"<=", 2, LEVAL},
	{">=", 2, GEVAL},
	{"dump", 1, DUMPVAL},
	{"retract", 1, RETRACTVAL},
	{"rule", 2, RULEVAL},
	{"printf", 2, PRINTFVAL},
	{"print", 1, PRINTVAL},
	{"write", 1, WRITEVAL},
	{"$call", 1, CALLXVAL},
	{"read", 1, READVAL},
	{"$fail", 0, FAILVAL},
	{"$gfail", 0, GFAILVAL},
	{"!", 0, CUTVAL},
	{"true", 0, TRUEVAL},
	{(char *) 0, 0, 0}
};

extern char *CLPR_CLPRLIB;
extern struct HASHNODE *CLPR_tok_hnode;
extern PTERM *CLPR_lookup_name(char *sym, CLPR_int quick_cmp);
extern SPNODE CLPR_sptab[];
extern CLPR_int CLPR_real_val;
extern FILE *CLPR_see_input;

/*---------------------------------------------------------------------------*/

extern CLPR_int CLPR_hashbuiltina (char *s, CLPR_int n);
extern void CLPR_assign_p_protected (CLPR_int name, CLPR_int flag);
void CLPR_init_malloc (void);
extern void CLPR_init_compiler (void);
extern void CLPR_init_solver (void);
void CLPR_init_name (void);
extern void CLPR_init_hash_p_table (void);
extern void CLPR_init_hash_l_table (void);
extern void CLPR_init_unify (void);
extern char CLPR_readchar (void);
extern void CLPR_init_engine (void);
extern void CLPR_init_kohler (void);
extern CLPR_int CLPR_sum_str (register char *s);
extern PTERM *CLPR_func_narg_pterm (PTERM *t, CLPR_int narg);
extern CLPR_int CLPR_do_op (char *sym, char *style, CLPR_int pri);
extern void CLPR_setconst (char *sym, double num);
void CLPR_init_fatal (char *s, ...);
void CLPR_test_pmask (void);
extern void CLPR_check_solver_sizes (void);

void CLPR_addspred(char *s, CLPR_int n, CLPR_int (*f) (/* ??? */), SPNODE *pp, CLPR_int val)
{
	CLPR_int i;

	CLPR_add_func(s, n, val);
	i = CLPR_hashbuiltina(s,n);
	if (i != val) CLPR_fatal("92018");
	CLPR_assign_p_protected(i, TRUE);
	CLPR_spred[i] = pp;
}

void CLPR_init_all(void)
{
	CLPR_see_input = stdin;
	CLPR_init_malloc();
	CLPR_init_compiler();
	CLPR_init_solver();
	/* Shifted to main.c */
	/**** CLPR_init_scanner(); ****/
	CLPR_init_name();
	CLPR_init_hash_p_table();
	CLPR_init_hash_l_table();
	CLPR_init_unify();
	CLPR_readchar();
	CLPR_init_engine();
	CLPR_init_kohler();
	CLPR_real_val = CLPR_hashbuiltina("real", 1);
	if ((CLPR_CLP_SUFFIX = (char *) getenv("CLPRSUFFIX")) == (char *) NULL) {
#ifdef DEF_CLP_SUFFIX
		CLPR_CLP_SUFFIX = DEF_CLP_SUFFIX;
#else
		CLPR_CLP_SUFFIX = "";
#endif
	}
}

void CLPR_init_name(void) /*** predefined names for parser ***/
{
PTERM *t;
CLPR_int i, lib_val;

	for (i = 0; CLPR_def_p_func[i].name; i++) {
		CLPR_add_func(CLPR_def_p_func[i].name, 
		         CLPR_def_p_func[i].narg, 
                 CLPR_def_p_func[i].val);
	}
#ifdef LIBPATH
#	ifdef OS2V2 
	/* for some reason the quotes are being deleted from the makefile define */
	/* when using GNU make for OS/2 */
	if (!(CLPR_CLPRLIB = (char *) getenv("CLPR_CLPRLIB"))) CLPR_CLPRLIB = "LIBPATH";
#	else
	if (!(CLPR_CLPRLIB = (char *) getenv("CLPR_CLPRLIB"))) CLPR_CLPRLIB = LIBPATH;
#	endif
#else
	if (!(CLPR_CLPRLIB = (char *) getenv("CLPR_CLPRLIB"))) CLPR_CLPRLIB = ".";
#endif
	CLPR_add_func(CLPR_CLPRLIB, 0, FLIBDIR);

	CLPR_lookup_name("[]", CLPR_sum_str("[]"));
	CLPR_hnode_emptylist = CLPR_tok_hnode;
	CLPR_null_list = CLPR_lookup_name("[]", CLPR_sum_str("[]"));
	CLPR_func_narg_pterm(CLPR_null_list, 0);
	t = CLPR_lookup_name(".", CLPR_sum_str("."));
	CLPR_cons_str = t->str;

	CLPR_do_op(".",   "xfy", 51);
	CLPR_do_op(",",  "xfy", 252);
	CLPR_do_op(":-",  "xfx", 255);
	CLPR_do_op("-",   "fy",  21);
	CLPR_do_op(":-",  "fx",  255);
	CLPR_do_op("?-",  "fx",  255);
	CLPR_do_op("::-", "fx",  255);
	CLPR_do_op("^+", "yfx",  31); 
	CLPR_do_op("^-", "yfx",  31); 
	CLPR_do_op("^*", "yfx",  21); 
	CLPR_do_op("^/", "yfx",  21); 

	CLPR_setconst("#c",  2.99792458e+8);
	CLPR_setconst("#e",  E_VAL);
	CLPR_setconst("#g", 9.80665);
	CLPR_setconst("#h",  6.626176e-34);
	CLPR_setconst("#p",  PI_VAL); 
	CLPR_setconst("#p_2",  PI_2_VAL); 
	CLPR_setconst("#p_4",  PI_4_VAL); 
	CLPR_setconst("#sqrt2",  SQRT2_VAL); 
	CLPR_setconst("#sqrt1_2",  SQRT1_2_VAL); 
	CLPR_setconst("#ec",  1.6021892e-19);
	CLPR_setconst("#cm2in",  0.393701);
	CLPR_setconst("#km2mile",  0.62137);
	CLPR_setconst("#gm2oz",  0.03527);
	CLPR_setconst("#kg2lb",  2.20462);
	CLPR_setconst("#l2gal",  0.21998);
	CLPR_setconst("#l2usgal",  0.26418);
	for (i = 0, lib_val = WAM_LIB_0; CLPR_sptab[i].str; i++, lib_val++) {
		CLPR_addspred(CLPR_sptab[i].str, CLPR_sptab[i].arity,
		         CLPR_sptab[i].func, CLPR_sptab+i, lib_val);
	}
	if (lib_val >= INIT_FUNC_ID) CLPR_fatal("64209");

}

/*-------------- utilities -------------*/

float CLPR_int_to_float(CLPR_int i);

static char malloc_space[PRE_MALLOC_MEM];
static char *cur_malloc = malloc_space;
static char *malloc_high = &malloc_space[PRE_MALLOC_MEM - 128];

void CLPR_init_malloc(void)
{
CLPR_unsigned pad;
	pad=((CLPR_unsigned) cur_malloc) % sizeof(double); 
/*  djgpp seems to have problems with compiling this line */
/***
	if ((pad=((CLPR_unsigned) cur_malloc)) % sizeof(double) != 0)
***/
	if (pad > 16) 
		CLPR_warning("Strange alignment found %d (*** May be a bug ***)\n",pad);
	if (pad) cur_malloc += sizeof(double) - pad;
}

char *CLPR_emalloc(CLPR_int n)
{
char *p;
register char *tmp;
register CLPR_int j;
CLPR_int pad;

	if ((CLPR_int) (cur_malloc + n) < (CLPR_int) malloc_high) {
/***
#ifdef HP835
		if (pad = n % sizeof(double)) n += sizeof(double) - pad;
#else
		if (pad = n % sizeof(CLPR_int)) n += sizeof(CLPR_int) - pad;
#endif
***/
		/* always align on double is safer */
		pad=((CLPR_unsigned) cur_malloc) % sizeof(double);
		if (pad) p = cur_malloc + sizeof(double) - pad; 
		else p = cur_malloc;

		/*** Change to initial padding as end padding should really
			be correct if sizeof is doing its job ***/
		cur_malloc = p + n;
		return(p);
	} 
	if ((p = (char *) malloc(n+8 /** TEMPORARY **/)) == NULL) 
		CLPR_fatal("Not enough memory (calloc)");
pad=((CLPR_unsigned) p) % sizeof(double);
if (pad) {
#ifdef HP835
if (pad) fprintf(stderr,"STRANGE ** malloc (realign) %x pad %d\n",p,pad);
#endif
p += sizeof(double) - pad; 
}
	return(p);
}

char *CLPR_char_malloc(CLPR_int n)
{
char *p;
register char *tmp;
register CLPR_int j;
CLPR_int pad;

	if ((CLPR_int) (cur_malloc + n) < (CLPR_int) malloc_high) {
		p = malloc_high - n + 64;
		malloc_high -= n;
		return(p);
	} 
	if ((p = (char *) malloc(n)) == NULL) 
		CLPR_fatal("Not enough memory (calloc)");
	return(p);
}

#ifdef unix
extern char *sbrk(CLPR_int);
#endif

/*
On some Unix systems, e.g. Linux, we need to allocate memory
with sbrk() rather than malloc(), otherwise our scheme of
using the top four bits of pointers as tags won't work.
*/
  
CLPR_int *CLPR_alloc_mem(CLPR_int n)
{
    char *p;

#if defined(unix) && !defined(NO_SBRK)
    p = (char *) sbrk(0);
    if ((((CLPR_unsigned) p) % sizeof(double)) != 0) {
/***
        CLPR_warning("sbrk unaligned at %x, realigned", p);
***/
        sbrk(sizeof(double) - (((CLPR_unsigned) p) % sizeof(double)));
    }
    if ((CLPR_int) (p = sbrk(n*sizeof(CLPR_int))) == -1) p = (char *) NULL;
#else
    p = (char *) calloc(1, sizeof(CLPR_int)*n);
#endif
	if (p == NULL) CLPR_fatal("Not enough memory to initialise");
    return (CLPR_int *) p;
}


float CLPR_int_to_float(CLPR_int i)
{
union convert {	CLPR_int i; float f;	} tmp;
	tmp.i = i;
	return tmp.f;
}	 
 
CLPR_int CLPR_float_to_int(float f)
{
union convert {	CLPR_int i; float f;	} tmp;
	tmp.f = f;
	return tmp.i;
}	 
 
void CLPR_print_pre_malloc_stats(void)
{
	printf("Pre-Malloc space: %d used (max %d)\n",
		(((CLPR_int) (cur_malloc)) - ((CLPR_int) malloc_space)) / 4,
		PRE_MALLOC_MEM);
}

static CLPR_int test_buf[2];
void CLPR_test_system_defaults(void)
{
CLPR_int *p;
	if (MAXPOSINT < 0) CLPR_init_fatal("MAXPOSINT assumed to be 2^31-1");
	p = (CLPR_int *) -1;
	if (((CLPR_int) p) > 0) CLPR_init_fatal("problem with pointer conversion");
	CLPR_test_pmask();
	CLPR_check_solver_sizes();  
}

void CLPR_test_pmask(void)
{
	CLPR_int i,j,k;
	CLPR_int local_buf[4];
	CLPR_int *malloc_buf;
	CLPR_int *ptr[3];
	CLPR_unsigned amask[3];
	CLPR_int flag = 0;
	char *msg;

	ptr[0] = (CLPR_int *) malloc(8);
	ptr[1] = test_buf;
	ptr[2] = local_buf;
	
	amask[0] = intptr(ptr[0]) & MASK4;
	amask[1] = intptr(ptr[1]) & MASK4;
	amask[2] = intptr(ptr[2]) & MASK4;

	for (i=0; i < 3; i++)
		if (((CLPR_int) amask[i]) < 0) 
			CLPR_init_fatal("Pointer topmost bit being used,\ncannot install: (%x)\n", amask[i]);
		else if (amask[i] != amask[0]) {
			flag = 1;
			break;
		}	
	if (flag) {
		fprintf(stderr, "Inconsistent masks\n");
		for (i=0; i < 3; i++)
			fprintf(stderr, "amask[%d] = %x\n", i, amask[i]);
		CLPR_init_fatal("Funny pointers",0);
	}
	if (pmask(0) == amask[0]) return;
	msg = CLPR_emalloc(80);
	if (amask[0] != 0)
		sprintf(msg,"#define pmask(i) ((((CLPR_int) i) & 0x0fffffff) | 0x%x)",
			amask[0]);
	else sprintf(msg,"#define pmask(i) (((CLPR_int) i) & 0x0fffffff)");
	CLPR_init_fatal("pmask() in emul.h defined incorrectly, redefine as \n%s",msg);
}

void CLPR_init_fatal(char *s, ...)
{
char buf[256];
va_list l;
	fprintf(stderr, "Installation error: ");
	va_start(l, s);
	vsprintf(buf, s, l);
	va_end(l);
	fprintf(stderr, "%s\n", buf);
	fprintf(stderr, "See chapter 7 of CLP(R) manual.\n");
	exit(1);
}
