/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: lib.c                                                *
 ***************************************************************/

#ifdef SYS5
#include <sys/types.h>
#include <sys/times.h>
#endif

#ifdef BSD
#ifdef OS2V2	/* OS2 not using BSD timings */
#include <sys/types.h>
#include <sys/times.h>
#else			/* BSD */
#include <sys/time.h>
#endif
#include <sys/resource.h>
#include <strings.h>
#endif

#ifdef OS2V2
#include <sys/timeb.h>
#endif

#include <unistd.h>

#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "standard.h"
#include "config.h"
#include "parse.h"
#include "lib.h"
#include "symbols.h"
#include "compile.h"
#include "emul.h"
#include "opcodes.h"
#include "main.h"

#define U_SEC (1000000.0)
#if defined(BSD) && !defined(OS2V2)
static struct timeval basetime, curtime;
#endif
#if defined(SYS5) || defined(OS2V2)
static struct tms basetime, curtime;
#endif

typedef struct io_rec {
	struct io_rec *next;	
	char *s;		
	bool mode;		
	FILE *fio;		
	bool dual;		
} IO_REC;

char *CLPR_CLPRLIB; 
bool CLPR_is_clam_file(char *s);
CLPR_int CLPR_efile[2] = { addtag(TAG_STR, QUERY), addtag(TAG_CON, END) };
IO_REC *CLPR_files_open = (IO_REC *) NULL;	
FILE *CLPR_see_input;
CLPR_int CLPR_see_input_t = addtag(TAG_CON, USER);

char *CLPR_fmt_printf(FILE *F, char *s, CLPR_int **t);
CLPR_int *CLPR_findbind(CLPR_int *x);
FILE *CLPR_file_is_open(char *name, char mode, IO_REC *fl);
PTERM *heap_to_pterm();
CLPR_int *CLPR_pterm_to_heap(PTERM *t);
CLPR_int *CLPR_end_of_constraints(CLPR_int *t);
CLPR_int *CLPR_make_new_rule_head(CLPR_int *t, CLPR_int **vars);
FILE *CLPR_my_open(char *name, char *mode);
char *CLPR_command_interpreter(void);
CLPR_int CLPR_round_int(double f);
CLPR_int CLPR_find_heap_addr(CLPR_int v);
CLPR_int CLPR_do_printf(FILE *F, CLPR_int *t, char *fmt, CLPR_int flush_flag);
CLPR_int CLPR_do_assert_rule(CLPR_int *t, CLPR_int start);

extern char *CLPR_CLP_SUFFIX;
extern RULE_ptr CLPR_current_rule;
extern FILE *CLPR_infile, *CLPR_outfile, *CLPR_input_fd, *CLPR_error_stream;
extern CLPR_int CLPR_infile_t, CLPR_outfile_t;
extern CLPR_int CLPR_hascut;
extern VREC *CLPR_varlist;
extern CLPR_unsigned *CLPR_code;
extern CLPR_int CLPR_pcode, CLPR_epcode, CLPR_endcode[], CLPR_startcode[];
extern FUNCTREC CLPR_functors[];
extern CLPR_int *CLPR_curar, *CLPR_stacktop, *CLPR_heaptop, *CLPR_retreg, **CLPR_trail, CLPR_trtop;
extern CLPR_int CLPR_solver_id;
extern CLPR_int CLPR_CODEGEN_DEBUG;
extern CLPR_int CLPR_TRACE, CLPR_CLAM_TRACE;
extern ASMREC CLPR_asmlist[];
extern SPNODE *CLPR_spred[];
extern CLPR_int CLPR_funct_vars_top;
extern CLPR_int CLPR_flatten_vars_top;
extern CLPR_int CLPR_IMPLICIT, CLPR_PARTIAL_IMPLICIT;
extern CLPR_int CLPR_WARN_ABORT;
extern CLPR_int CLPR_WARNING_FLAG;
extern bool CLPR_have_runable;
extern CLPR_int CLPR_CODE_SZ;
extern CLPR_int CLPR_STYLE_CHECK_FLAG;
extern CLPR_int CLPR_REDEFINE_FLAG;
extern struct HASHNODE *CLPR_hash_func_tab[];

extern PTERM *CLPR_lookup_name(char *sym, CLPR_int quick_cmp);
extern PTERM *CLPR_lookup_var(char *var, CLPR_int quick_cmp);
extern char *CLPR_maketag(CLPR_int n, char *extra, char *str, CLPR_int arity);
extern RULE_ptr CLPR_p_first_rule(CLPR_int name);
extern RULE_ptr CLPR_p_last_rule(CLPR_int name);
extern RULE_ptr CLPR_store_rule(PTERM *to, PTERM *t, CLPR_int clobber);
extern PTERM *CLPR_new_pt_var(CLPR_int val, char *s);
extern PTERM *CLPR_new_pterm(void);
extern PTERM *CLPR_new_pt_emptylist(void);
extern PTERM *CLPR_new_real(double val);
extern PTERM *CLPR_new_p_func2(CLPR_int val, PTERM *arg1, PTERM *arg2);
extern RULE_ptr CLPR_new_rule(void);
extern CLPR_int *CLPR_get_rule_constraints(CLPR_int *term);
extern HASH_P_NODE_ptr CLPR_new_hash_p_node(CLPR_int name);
extern char *getenv(const char *);
extern char *CLPR_emalloc(CLPR_int n);
extern char *mktemp(char *);
extern struct HASHNODE *CLPR_insert_func_htable(char *s);
/*---------------------------------------------------------------------------*/

SPNODE CLPR_sptab[] = {
	{ CLPR_do_history, 0, "history" , 0},
	{ CLPR_do_history1, 1, "history" , 0},
	{ CLPR_do_tell, 1, "tell" , 0},
	{ CLPR_do_telling, 1, "telling" , 0},
	{ CLPR_do_told, 0, "told" , 0},
	{ CLPR_do_see, 1, "see" , 0},
	{ CLPR_do_seeing, 1, "seeing" , 0},
	{ CLPR_do_seen, 0, "seen" , 0},
	{ CLPR_do_consult, 1, "consult" , 0},
	{ CLPR_do_reconsult, 1, "reconsult" , 0},
	{ CLPR_do_listing, 0, "ls" , 0},
	{ CLPR_do_listing1, 1, "ls" , 0},
	{ CLPR_do_listing, 0, "listing" , 0},
#ifdef CTRACE
	{ CLPR_do_listing1, 1, "listing" , 0},
#endif
	{ CLPR_do_clisting1, 1, "cls" , 0},
	{ CLPR_do_assertz, 1, "assert" , 0},
	{ CLPR_do_asserta, 1, "asserta" , 0},
	{ CLPR_do_assertz, 1, "assertz" , 0},
	{ CLPR_do_fassertz, 1, "fassert" , 0},
	{ CLPR_do_fasserta, 1, "fasserta" , 0},
	{ CLPR_do_fassertz, 1, "fassertz" , 0},
	{ CLPR_do_retractall, 0, "retractall" , 0},
	{ CLPR_do_nl, 0, "nl" , 0},
	{ CLPR_do_flush, 0, "flush" , 0},
	{ CLPR_do_oracle, 3, "oracle" , 0},
	{ CLPR_do_pipe, 1, "pipe" , 0},
	{ CLPR_do_fork, 0, "fork" , 0},
	{ CLPR_do_edit, 1, "edit" , 0},
	{ CLPR_do_more, 1, "more" , 0},
	{ CLPR_do_halt, 0, "halt" , 0},
	{ CLPR_do_abort, 0, "abort" , 0},
	{ CLPR_do_sh, 0, "sh" , 0},
	{ CLPR_do_csh, 0, "csh" , 0},
	{ CLPR_do_nonground, 1, "nonground" , 0},
	{ CLPR_do_ground, 1, "ground" , 0},
	{ CLPR_do_nonvar, 1, "nonvar" , 0},
	{ CLPR_do_var, 1, "var" , 0},
	{ CLPR_do_univ, 2, "=.." , 0},
	{ CLPR_do_occurs, 2, "CLPR_occurs" , 0},
	{ CLPR_do_eqeq, 2, "==" , 0},
	{ CLPR_do_atom, 1, "atom" , 0},
	{ CLPR_do_atomic, 1, "atomic" , 0},
	{ CLPR_do_functor1, 1, "functor" , 0},
	{ CLPR_do_arithmetic, 1, "arithmetic" , 0},
	{ CLPR_do_rand, 1, "rand" , 0},
	{ CLPR_do_srand, 1, "srand" , 0},
	{ CLPR_do_floor, 2, "floor", 2},
	{ CLPR_do_op3, 3, "op", 0},
	{ CLPR_do_prot, 2, "prot", 0},
	{ CLPR_do_hide, 2, "hide", 0},
	{ CLPR_do_warning, 1, "CLPR_warning", 0},
	{ CLPR_do_codegen_debug, 0, "codegen_debug", 0},
	{ CLPR_do_codegen_nodebug, 0, "codegen_nodebug", 0},
	{ CLPR_do_spy, 0, "spy", 0},
	{ CLPR_do_spy2, 2, "$spy", 0},
	{ CLPR_do_nospy, 0, "nospy", 0},
	{ CLPR_do_nospy2, 2, "$nospy", 0},
	{ CLPR_do_trace, 0, "trace", 0},
	{ CLPR_do_notrace, 0, "notrace", 0},
	{ CLPR_do_implicit, 0, "implicit", 0},
	{ CLPR_do_no_implicit, 0, "no_implicit", 0},
	{ CLPR_do_partial_implicit, 0, "partial_implicit", 0},
#ifdef CTRACE
	{ CLPR_do_clam_trace, 0, "ctrace", 0},
	{ CLPR_do_clam_notrace, 0, "noctrace", 0},
#endif
	{ CLPR_do_ztime, 0, "ztime", 0},
	{ CLPR_do_ctime, 1, "ctime", 0},
	{ CLPR_do_compile, 2, "CLPR_compile", 0},
#ifdef STATS
	{ CLPR_do_e_stats, 0, "e_stats" , 0},
	{ CLPR_do_s_stats, 0, "s_stats" , 0},
	{ CLPR_do_stats, 0, "stats" , 0},
#endif
	{ CLPR_do_dump3, 3, "dump", 0},
	{ CLPR_do_dump2, 2, "dump", 0},
	{ CLPR_do_dynamic, 2, "dynamic", 0},
	{ CLPR_do_functor3, 3, "functor", 0},
	{ CLPR_do_arg, 3, "arg", 0},
	{ CLPR_do_style_check, 1, "$style_check", 0},
	{ CLPR_do_clear_style, 0, "$clear_style_check", 0}, /** defined in codegen.c **/
	{ CLPR_do_counter_value, 2, "counter_value", 0},
	{ CLPR_do_set_counter, 2, "set_counter", 0},
	{ CLPR_do_add_counter, 2, "add_counter", 0},
	{ CLPR_do_atom_printf, 3, "printf_to_atom", 0},
	{ CLPR_do_new_constant, 2, "new_constant", 0},
/*****************************************************/
/*****
	{ CLPR_do_replace_arg, 3, "replace_arg", 0},
	{ CLPR_do_asm_code_1, 1, "$$CLPR_code", 0},
	{ CLPR_do_asm_code_2, 2, "$$CLPR_code", 0},
	{ CLPR_do_asm_code_3, 3, "$$CLPR_code", 0},
	{ CLPR_do_asm_code_4, 4, "$$CLPR_code", 0},
	{ CLPR_do_asm_label, 1, "$$label", 0},
	{ CLPR_do_asm_list, 2, "$$list", 0},
	{ CLPR_do_asm_call, 1, "$$call", 0},
	{ CLPR_do_asm_resolve, 0, "$$CLPR_resolve", 0},
	{ CLPR_do_dump_tableaus, 0, "dump_tableaus" , 0},
*****/
	{ 0, 0, 0, 0}
};

#define MAX_VAR_IN_HEAP 1024
static CLPR_int pterm_var_cnt = -1;
static CLPR_int var_heapval[MAX_VAR_IN_HEAP + 1];
static CLPR_int hterm_var_cnt = 0;
static CLPR_int heap_varval[MAX_VAR_IN_HEAP + 1];

static PTERM *ht_to_pterm(CLPR_int *t, CLPR_int quoted, CLPR_int with_dump);
static CLPR_int heapify_pterm(PTERM *t, CLPR_int *hp, CLPR_int quoted);
static void term_printf (FILE *F, char *fmt, CLPR_int fmt_ty,
			CLPR_int **arg_list);

/*---------------------------------------------------------------------------*/


extern void CLPR_warn_err (char *s, ...);
extern CLPR_int CLPR_make_a_con (CLPR_int *t, CLPR_int val);
extern CLPR_int CLPR_hashbuiltina (char *s, CLPR_int n);
extern CLPR_int CLPR_myisdigit (char c);
extern CLPR_int CLPR_is_ground (CLPR_int v, double *c_ptr);
extern void CLPR_print_var2pos (FILE *f, CLPR_int *v);
extern void CLPR_printterm (FILE *f, CLPR_int *t);
extern void CLPR_dump1 (FILE *f, CLPR_int *vars, char *names);
extern CLPR_int CLPR_dump2 (FILE *f, CLPR_int *vars, CLPR_int *names);
extern CLPR_int CLPR_dump3 (CLPR_int *vars, CLPR_int *user_vars, char *into);
extern void CLPR_new_input (FILE *stream, char *name);
extern CLPR_int CLPR_unify (CLPR_int *t1, CLPR_int *t2);
extern CLPR_int CLPR_close_input (FILE *stream);
CLPR_int CLPR_internal_consult (char *s);
extern void CLPR_readfile_low (CLPR_int clobber);
extern void CLPR_compile (char clobber);
extern void CLPR_resolve (CLPR_int warn);
extern void CLPR_unreadchar (register char c);
extern char CLPR_readchar (void);
CLPR_int CLPR_internal_reconsult (char *s);
extern CLPR_int CLPR_getmaxfunc (void);
extern CLPR_int CLPR_p_compiled (CLPR_int name);
extern CLPR_int CLPR_p_protected (CLPR_int name);
extern void CLPR_print_rule (PTERM *t);
extern CLPR_int CLPR_p_start_hdrcode (CLPR_int name);
extern CLPR_int CLPR_p_end_hdrcode (CLPR_int name);
extern void CLPR_print_asm (CLPR_int start, CLPR_int end);
extern CLPR_int CLPR_is_pivot_var (CLPR_int v);
extern CLPR_int CLPR_is_funct_var (CLPR_int *t);
extern CLPR_int CLPR_heap_to_pivot_var (CLPR_int *t);
extern CLPR_int CLPR_heap_to_free_var (CLPR_int *t);
CLPR_int CLPR_get_evaled_func (CLPR_int v);
extern CLPR_int CLPR_p_dynamic (CLPR_int name);
extern void CLPR_init_lable (void);
extern void CLPR_assign_p_last_rule (CLPR_int name, RULE_ptr rule);
extern void CLPR_compile_rule (PTERM *t, CLPR_int num);
extern void CLPR_assign_p_first_rule (CLPR_int name, RULE_ptr rule);
extern void CLPR_assign_p_start_hdrcode (CLPR_int name, CLPR_int i);
extern void CLPR_emit_dynamic_try (CLPR_int val);
extern void CLPR_assign_p_end_hdrcode (CLPR_int name, CLPR_int i);
extern void CLPR_assign_p_compiled (CLPR_int name, CLPR_int flag);
extern void CLPR_assign_p_last_compiled_rule (CLPR_int name, RULE_ptr rule);
extern void CLPR_cleanup_dump (void);
extern void CLPR_delete_all_rules (CLPR_int val);
extern void CLPR_assign_p_cut (CLPR_int name, CLPR_int flag);
extern void CLPR_assign_p_spy (CLPR_int name, CLPR_int flag);
extern void CLPR_delete_lable (CLPR_int val);
extern CLPR_int CLPR_makenum (CLPR_int v, double c);
extern CLPR_int CLPR_make_a_num (CLPR_int *t, double val);
extern CLPR_int CLPR_make_a_nil (CLPR_int *t);
extern CLPR_int CLPR_make_a_new_list (CLPR_int *t);
extern CLPR_int CLPR_wake_runable (void);
extern CLPR_int CLPR_make_a_new_fun (CLPR_int *t, CLPR_int val, CLPR_int arity);
CLPR_int CLPR_occurs (CLPR_int *vspot, CLPR_int *t);
extern CLPR_int CLPR_equal_param (CLPR_int v1, CLPR_int v2);
extern CLPR_int CLPR_do_op (char *sym, char *style, CLPR_int pri);
extern void CLPR_assign_p_protected (CLPR_int name, CLPR_int flag);
extern void CLPR_printhist (CLPR_int n);
extern CLPR_int CLPR_p_spy (CLPR_int name);
extern void CLPR_assign_p_dynamic (CLPR_int name, CLPR_int flag);
extern void CLPR_print_ineq_tableau (void);
extern void CLPR_check_tableaus (void);
CLPR_int CLPR_count_list (CLPR_int *l, CLPR_int *np);
extern CLPR_int *CLPR_make_cons (CLPR_int *t1, CLPR_int *t2);
void CLPR_emit_instruction (char *s, CLPR_int *t1, CLPR_int *t2, CLPR_int *t3);
extern void CLPR_store_lable (CLPR_int val, char *s);
extern CLPR_int CLPR_find_loc_for_lable (CLPR_int val, char *s);
extern CLPR_int CLPR_rungoal (CLPR_int ipc, CLPR_int interactive);
extern CLPR_int CLPR_find_asm_by_mnemonic (char *s);
extern void CLPR_emit_word (CLPR_int w);
extern CLPR_int CLPR_f_is_switch_inst (CLPR_int i);
extern void CLPR_emit_addr (char *s, CLPR_int val);
extern void CLPR_setconst (char *sym, double num);

FILE *CLPR_file_ready(char *name, char mode)
{
FILE *f;
IO_REC *t;
	if (f = CLPR_file_is_open(name, mode, CLPR_files_open)) return f;
	else {
		t = (IO_REC *) CLPR_emalloc(sizeof(IO_REC));
		if (mode == IO_READ) {
			if ((f = CLPR_my_open(name, "r")) == (FILE *) NULL) {
				CLPR_warn_err("Could not read from file %s", name);
				return (FILE *) NULL;
			}
		} else {
			if ((f = CLPR_my_open(name, "w")) == (FILE *) NULL) {
				CLPR_warn_err("Could not write to file %s", name);
				return (FILE *) NULL;
			}
		}
		t->fio = f;
		t->s = name;
		t->mode = mode;
		t->next = CLPR_files_open;
		t->dual = FALSE;
		CLPR_files_open = t;
		return f;
	}
}
 
FILE *CLPR_file_is_open(char *name, char mode, IO_REC *fl)
{
	if (fl == (IO_REC *) NULL) return FALSE;
	if (!strcmp(name, fl->s)) {
		if (fl->mode != mode && !fl->dual) {
			fprintf(CLPR_error_stream, "%s already engaged\n", name);
			CLPR_fatal("... bye");
		}
		if (fl->mode != mode) return CLPR_file_is_open(name, mode, fl->next);
		return fl->fio;	
	} else
		return CLPR_file_is_open(name, mode, fl->next);
}
 
IO_REC *CLPR_file_close(char *name, IO_REC *fl)
{
	if (!strcmp("<stdin>", name)) return fl;
	if (fl == (IO_REC *) NULL) {
		fprintf(stderr, "cannot close file %s\n", name);
		CLPR_fatal("... bye");
	}
	if (!strcmp(fl->s, name)) {
		fflush(fl->fio);
		fclose(fl->fio);
		return fl->next;
	} else {
		fl->next = CLPR_file_close(name, fl->next);
		return fl;
	}
}
 
void CLPR_file_allclose(IO_REC *fl)
{
	if (fl == (IO_REC *) NULL) return;
	fflush(fl->fio);
	fclose(fl->fio);
	CLPR_file_allclose(fl->next);
}
 
void CLPR_files_shutdown(void)
{
	CLPR_file_allclose(CLPR_files_open);
}

extern FILE *CLPR_TEMP_FILE;
extern char *CLPR_TEMP_FNAME;

char *CLPR_special_buffer;
/* temporary code */
CLPR_int CLPR_do_atom_printf(CLPR_int *t1, CLPR_int *t2, CLPR_int *t3)
{
CLPR_int l;
char *s;
	t2 = CLPR_findbind(t2);
	if (!is_con(t2)) {
		CLPR_warn_err("second argument of printf_to_atom/3 must be an atom/string");
		return FALSE;
	}
	t3 = CLPR_findbind(t3);
	if (!(is_nil(t3) || is_cons(t3))) {
		CLPR_warn_err("second argument of printf_to_atom/3 must be a list");
		return FALSE;
	}
	if (!CLPR_TEMP_FILE) { 
	/* just choose a fixed name to avoid any problems with DOS etc */
#if defined(MSDOS) || defined(OS2V2) || !defined(unix)
		if (!(CLPR_TEMP_FILE = CLPR_my_open(DOS_TMP_FILE, "w")))
			CLPR_fatal("couldn't open temporary file");
#else /* On Unix */
		CLPR_TEMP_FNAME = mktemp(CLPR_TEMP_FNAME);		
		if (!(CLPR_TEMP_FILE = CLPR_my_open(CLPR_TEMP_FNAME, "w")))
			CLPR_fatal("couldn't open temporary file");
		unlink(CLPR_TEMP_FNAME);
#endif
		CLPR_special_buffer = CLPR_emalloc(BUFSIZ + 16);
		setbuf(CLPR_TEMP_FILE, CLPR_special_buffer);
	}
	rewind(CLPR_TEMP_FILE);
	CLPR_do_printf(CLPR_TEMP_FILE, t3, string_of(mask(*t2)), FALSE);
	putc('\0', CLPR_TEMP_FILE);
	if ((l=strlen(CLPR_special_buffer)) >= BUFSIZ) CLPR_fatal("66012");
	s = CLPR_emalloc(l+1);
	strcpy(s, CLPR_special_buffer);
	return CLPR_make_a_con(t1, CLPR_hashbuiltina(s, 0));
}

void CLPR_clean_temp_file(void) /* remove the temp file */
{
	if (CLPR_TEMP_FILE) {
		fclose(CLPR_TEMP_FILE);
#if defined(MSDOS) || defined(OS2V2) || !defined(unix)
		unlink(DOS_TMP_FILE);
#endif
	}
}


CLPR_int CLPR_do_printf(FILE *F, CLPR_int *t, char *fmt, CLPR_int flush_flag)
{
CLPR_int j;
char c;
CLPR_int *arg;

	/* 2nd arg can be a var; but needs to be set to a list */
	arg = CLPR_findbind(t);
	if (!(is_nil(arg) || is_cons(arg))) {
		CLPR_warn_err("second argument of printf/2 must be a list");
		return FALSE;
	}
	for(; c= *fmt; fmt++) {
		if (c == '\\') 
			if (CLPR_myisdigit(c= * ++fmt)) {
				for (j = 0, c = 0; j < 3 && CLPR_myisdigit(*fmt) && *fmt != '9'; 
					j++, fmt++) 
						c = c * 8 + *fmt - '0';
				putc(c, F);
				fmt--;
			} else switch (c) {
				case 'n': putc('\n', F); break;
				case 't': putc('\t', F); break;
				case 'r': putc('\r', F); break;
				case 'b': putc('\b', F); break;
				case 'f': putc('\f', F); break;
				default:  putc(c, F);    break;
			}
		else if (c == '%') fmt = CLPR_fmt_printf(F, fmt, &arg);
		else putc(c, F);
	}
	if (flush_flag) fflush(F);
	return TRUE;
}
 
#define STR_FMT	1
#define REAL_FMT 2
#define INT_FMT 3
#define DEFAULT_FMT 4
 
char *CLPR_fmt_printf(FILE *F, char *s, CLPR_int **t)
{
char *s1;
char format[MAX_MESSAGE];
	s1 = s + 1;
	if (*s1 == '-') s1++;
	while (CLPR_myisdigit(*s1)) s1++;
	if (*s1 == '.') {
		s1++;
		while (CLPR_myisdigit(*s1)) s1++;
	}
	if (*s1 == 'l') s1++;
	if (s1 - s < MAX_MESSAGE - 1) {
		strncpy(format, s, s1 - s+1);
		format[s1 - s + 1] = '\0';
	} else CLPR_fatal("format string too long");
	switch (*s1) {
	case 'f':
	case 'e':
	case 'g': term_printf(F, format, REAL_FMT, t); break;
	case 'd':
	case 'o':
	case 'x':
	case 'c':
	case 'u': term_printf(F, format, INT_FMT, t); break;
	case 's': term_printf(F, format, STR_FMT, t); break;
	case '%': fprintf(F, "%%"); break;
	default:
		term_printf(F, format, DEFAULT_FMT, t);
		s1--;
		return(s1);
	}
	return(s1);
}

extern CLPR_int CLPR_slack_id;
 
static void
term_printf(FILE *F, char *fmt, CLPR_int fmt_ty, CLPR_int **arg_list)
{
CLPR_int *t;
double val;
CLPR_int ground, i;
	*arg_list = CLPR_findbind(*arg_list);
	if (CLPR_tag(**arg_list) == CLPR_TAG_CONS) {
		t = CLPR_findbind((CLPR_int *)pmask(**arg_list));
		*arg_list= CLPR_findbind((CLPR_int *)(pmask(**arg_list)) + 1);
	} else {
		CLPR_warn_err("bad second argument to printf/2");
		return;
	}
	if (CLPR_tag(*t) == TAG_PAR) {
		if (CLPR_is_ground(mask(*t), &val)) {
			if (is_zero(val)) val = 0.0;
			if (fmt_ty == STR_FMT) 
				CLPR_runtime_error("bad second argument to printf/2");
			else if (fmt_ty == REAL_FMT) fprintf(F, fmt, val);
			else if (fmt_ty == INT_FMT) 
				fprintf(F, fmt, CLPR_round_int(val));
			else fprintf(F, "%g", val);
		} else {
			i = mask(*t);
			if (i < CLPR_slack_id) fprintf(F, "_t%d", i);
			else fprintf(F, "_S%d", i);
		}
		return;
	}
	if (is_unbound_var(*t) || is_daemon(t)) {
		CLPR_print_var2pos(F, t);
		return;
	}
	if (is_con(t) || is_nil(t) || is_func(t) || is_cons(t)) {
		if (fmt_ty == DEFAULT_FMT) CLPR_printterm(F, t);
		else if ((is_con(t) || is_nil(t)) && fmt_ty == STR_FMT)
			fprintf(F, fmt, string_of(mask(*t)));
		else CLPR_warn_err("inappropriate format in printf/2");
		return;
	}
	CLPR_fatal("Term_printf");
}

CLPR_int CLPR_do_dump(FILE *f, CLPR_int *vars, char *names)
{
	if (is_nil(vars)) return TRUE;
	if (CLPR_tag(*vars) != CLPR_TAG_CONS) CLPR_runtime_error("bad argument to dump/1");
	CLPR_dump1(f, vars, names);
	return TRUE;

}

CLPR_int CLPR_do_dump3(CLPR_int *t1, CLPR_int *t2, CLPR_int *t3)
{
	return CLPR_dump3(t1,t2,(char *)t3);
}

CLPR_int CLPR_do_dump2(CLPR_int *t1, CLPR_int *t2)
{
	return CLPR_dump2(CLPR_outfile, t1, t2);
}

extern FILE *CLPR_cur_stream;
CLPR_int CLPR_do_read(FILE *f, CLPR_int *t)
{
PTERM *tt;
FILE *old_cur_stream;
CLPR_int nt;
/*
	old_cur_stream = CLPR_infile;
*/
	old_cur_stream = CLPR_cur_stream;
	if (CLPR_see_input == (FILE *) NULL) {
		CLPR_warn_err("read/1: no input stream");
		return FALSE;
	}
	t = CLPR_findbind(t);
	if (!(is_unbound_var(*t) || is_daemon(t))) {
		CLPR_warn_err("bad argument to read/1");
		return FALSE;
	}
	if (feof(CLPR_see_input)) make_bound_var(&nt, CLPR_efile);
	else {
		CLPR_new_input(CLPR_see_input, string_of(mask(CLPR_see_input_t)));
		if (tt = (PTERM *) CLPR_readterm()) {
			nt = (CLPR_int) CLPR_pterm_to_heap(CLPR_quote_expand(tt, FALSE));
		}
		else make_bound_var(&nt, (CLPR_int) CLPR_efile);
		CLPR_new_input(old_cur_stream, "");
	}
	*(CLPR_heaptop) = nt;
	return CLPR_unify(CLPR_heaptop++, t);
}

struct {
	FILE *f;
	CLPR_int t;
} CLPR_see_stack[MAX_STREAM];
CLPR_int CLPR_see_stack_top = 1;

CLPR_int CLPR_do_see(CLPR_int *t)
{
FILE *f;
	t = CLPR_findbind(t);
	if (!is_con(t)) {
		CLPR_warn_err("bad argument to see/1");
		return FALSE;
	}
	if (CLPR_see_stack_top >= MAX_STREAM) CLPR_fatal("see/1 stack blown");
	CLPR_see_stack[CLPR_see_stack_top].f = CLPR_see_input; 
	CLPR_see_stack[CLPR_see_stack_top++].t = CLPR_see_input_t; 
	if (strcmp(CLPR_functors[mask(*t)].pname, "user")) {
		if (!(f = CLPR_file_ready(CLPR_functors[mask(*t)].pname, IO_READ)))
			return FALSE;
/*
		CLPR_infile = f;
*/
		CLPR_see_input = f;
		CLPR_see_input_t = *t;
		return TRUE;
	} else {
		CLPR_see_input = stdin;
		CLPR_see_input_t = addtag(TAG_CON, USER);
		return !strcmp(CLPR_functors[mask(*t)].pname, "user");
	}
}
 
CLPR_int CLPR_do_seeing(CLPR_int *t)
{
	*CLPR_heaptop = CLPR_see_input_t;
	return CLPR_unify(CLPR_heaptop++, t);
}
 
CLPR_int CLPR_do_seen(CLPR_int *t)
{
	if (CLPR_see_input != stdin) {
		CLPR_close_input(CLPR_see_input);
		CLPR_files_open = CLPR_file_close(CLPR_functors[mask(CLPR_see_input_t)].pname, CLPR_files_open);
	}
	if (--CLPR_see_stack_top <= 0) CLPR_fatal("34802");
	CLPR_see_input = CLPR_see_stack[CLPR_see_stack_top].f;
	CLPR_see_input_t = CLPR_see_stack[CLPR_see_stack_top].t;
	return TRUE;
}
 
struct {
	FILE *f;
	CLPR_int t;
} CLPR_tell_stack[MAX_MESSAGE];
CLPR_int CLPR_tell_stack_top = 1;

CLPR_int CLPR_do_tell(CLPR_int *t)
{
FILE *f;
	t = CLPR_findbind(t);
	if (!is_con(t)) {
		CLPR_warn_err("bad argument to tell/1");
		return FALSE;
	}
	if (CLPR_tell_stack_top >= MAX_MESSAGE) CLPR_fatal("tell/1 stack blown");
	fflush(CLPR_outfile);
	CLPR_tell_stack[CLPR_tell_stack_top].f = CLPR_outfile;
	CLPR_tell_stack[CLPR_tell_stack_top++].t = CLPR_outfile_t;
	if (strcmp(CLPR_functors[mask(*t)].pname, "user")) {
	   if (!(f = CLPR_file_ready(CLPR_functors[mask(*t)].pname, IO_WRITE))) 
			return FALSE;
		CLPR_outfile = f;
		CLPR_outfile_t = *t;
		return TRUE;
	} else {
		CLPR_outfile = stdout;
		CLPR_outfile_t = addtag(TAG_CON, USER);
		return !strcmp(CLPR_functors[mask(*t)].pname, "user");
	}
}
 
CLPR_int CLPR_do_telling(CLPR_int *t)
{
	*CLPR_heaptop = CLPR_outfile_t;
	return CLPR_unify(CLPR_heaptop++, t);
}
 
CLPR_int CLPR_do_told(void)
{
	if (--CLPR_tell_stack_top <= 0) CLPR_fatal("34802");
	if (CLPR_outfile != stdout)
		CLPR_files_open = CLPR_file_close(CLPR_functors[mask(CLPR_outfile_t)].pname, CLPR_files_open);
	CLPR_outfile = CLPR_tell_stack[CLPR_tell_stack_top].f;
	CLPR_outfile_t = CLPR_tell_stack[CLPR_tell_stack_top].t;
	return TRUE;
}

CLPR_int CLPR_do_consult(CLPR_int *t)
{
char *s;
	t = CLPR_findbind(t);
	if (!is_con(t) && !is_nil(t)) {
		CLPR_warn_err("bad argument to consult/1");
		return FALSE;
	}
	s = CLPR_functors[mask(*t)].pname;
	return CLPR_internal_consult(s);
}

CLPR_int CLPR_internal_consult(char *s)
{
CLPR_int *tmp, *nt;
FILE *o_cur_stream, *cur_fd;
	o_cur_stream = CLPR_input_fd;
	if (CLPR_is_clam_file(s)) return FALSE;
	if (strcmp(s, "user")) {
		if (!(CLPR_input_fd = CLPR_my_open(s, "r"))) {
			CLPR_warn_err("Could not read file %s\n", s);
			return FALSE;
		}
		CLPR_new_input(CLPR_input_fd, s);
	} else {
		CLPR_new_input(stdin, "<stdin>");
		CLPR_input_fd = stdin;
	}
	cur_fd = CLPR_input_fd;
	CLPR_readfile_low(FALSE);
	CLPR_compile(FALSE);
	CLPR_resolve(TRUE);
	if (cur_fd != stdin) {
		fclose(cur_fd);
		CLPR_close_input(cur_fd);
	} else {
		rewind(stdin);
		CLPR_unreadchar(' ');
		CLPR_readchar();
	}
	CLPR_input_fd = o_cur_stream;
	CLPR_new_input(o_cur_stream, "");
	return TRUE;
}

bool CLPR_is_clam_file(char *s)
{
char *t;
	if (strlen(s) < 6) return FALSE;
	t = s + (strlen(s) - 5);
	return (!strcmp(t, ".clam"));
}
	
CLPR_int CLPR_do_reconsult(CLPR_int *t)
{
char *s;
	t = CLPR_findbind(t);
	if (!is_con(t) && !is_nil(t)) {
		CLPR_warn_err("bad argument to reconsult/1");
		return FALSE;
	}
	s = CLPR_functors[mask(*t)].pname;
	return CLPR_internal_reconsult(s);
}

CLPR_int CLPR_internal_reconsult(char *s)
{
FILE *o_cur_stream, *cur_fd;
CLPR_int i = 2;
CLPR_int *tmp, *nt;
 
	o_cur_stream = CLPR_input_fd;
	if (CLPR_is_clam_file(s)) return FALSE;
	if (strcmp(s, "user")) {
		if (!(CLPR_input_fd = CLPR_my_open(s, "r"))) return FALSE;
		CLPR_new_input(CLPR_input_fd, s);
	} else {
		CLPR_new_input(stdin, "<stdin>");
		CLPR_input_fd = stdin;
	}
	cur_fd = CLPR_input_fd;
	CLPR_readfile_low(TRUE);
#ifdef FREE_CODE_SPACE
	CLPR_reclaim_code_space();
#endif
	CLPR_compile(TRUE);
	CLPR_resolve(TRUE);
	if (cur_fd != stdin) {
		fclose(cur_fd);
		CLPR_close_input(cur_fd);
	}
	CLPR_input_fd = o_cur_stream;
	CLPR_new_input(o_cur_stream, "");
	return TRUE;
}

CLPR_int CLPR_do_listing(void)
{
	return CLPR_do_listing1(NULL);
}

CLPR_int CLPR_do_listing1(CLPR_int *t)
{
bool success = FALSE, printed = FALSE;
RULE *r;
CLPR_int i;
char *nm;
	if (t) {
		t = CLPR_findbind(t);
		if (CLPR_tag(*t) != TAG_CON) CLPR_runtime_error("bad argument to listing/1");
		nm = CLPR_functors[mask(*t)].pname;
	} else nm = (char *) NULL;
	for(i = 1; i < CLPR_getmaxfunc(); i++) {
		if (CLPR_p_first_rule(i) && CLPR_p_compiled(i) && !CLPR_p_protected(i) &&
		   ((!nm || !strcmp(nm, string_of(i))))) {
			r = CLPR_p_first_rule(i);
			while (r) {
				if (!printed) fprintf(CLPR_outfile, "\n");
				CLPR_print_rule(r->t);
				success = printed = TRUE;
				r = r->next;
			}
			printed = FALSE;
		} else if (CLPR_p_compiled(i) && !CLPR_p_protected(i) && !CLPR_p_first_rule(i) &&
				(!nm || !strcmp(nm, CLPR_functors[i].pname)))
			fprintf(CLPR_outfile, "%s/%d: source unavailable.\n",
				CLPR_functors[i].pname, CLPR_functors[i].arity);
	}
	return success;
}

CLPR_int CLPR_do_clisting1(CLPR_int *t)
{
CLPR_int i, j, k, flag;
char *s;
RULE_ptr r;

	if (t) {
		t = CLPR_findbind(t);
		if (CLPR_tag(*t) != TAG_CON) CLPR_runtime_error("bad argument to clisting/1");
		s = CLPR_functors[mask(*t)].pname;
	} else CLPR_fatal("88912");
	flag = FALSE;

	for (i = 1; i < CLPR_getmaxfunc(); i++) {
		if (CLPR_p_compiled(i) &&
			!strcmp(CLPR_functors[i].pname, s) && 
		    !CLPR_p_protected(i)) {
			flag = TRUE;
			j = CLPR_p_start_hdrcode(i);
			k = CLPR_p_end_hdrcode(i);
			if (j <= k) CLPR_print_asm(j, k);
			for (r = CLPR_p_first_rule(i); r; r = r->next) {
				if (r->startcode) CLPR_print_asm(r->startcode, r->endcode);
			}
		}
	}
	return flag;
}

void CLPR_init_ht_to_pterm(void)
{
	hterm_var_cnt = 0; /*** not 0 ***/
}

static PTERM *ht_to_pterm(CLPR_int *t, CLPR_int quoted, CLPR_int with_dump)
{
PTERM *nt;
CLPR_int i, j, val;
double rval;
char *buf;
PTERM *prev, *cur;
	t = CLPR_findbind(t);
	if (CLPR_tag(*t) == TAG_PAR) { 
		i = mask(*t);
		if (CLPR_is_ground(i, &rval)) return CLPR_new_real(rval);
		buf = (char *) malloc(20);
		if (with_dump) {
			if (!(j = CLPR_is_pivot_var(i))) CLPR_fatal("71231");
			j = j + CLPR_funct_vars_top + 2*CLPR_flatten_vars_top;
			sprintf(buf, "_%d", j);
			return CLPR_new_pt_var(j, buf);
		} else {
			if (!(j = CLPR_find_heap_addr(-i))) {
				if (hterm_var_cnt >= MAX_VAR_IN_HEAP) 			
			 		CLPR_fatal("too many variables in term to assert");
				heap_varval[++hterm_var_cnt] = -i;
				sprintf(buf, "_%d", i);
				return CLPR_new_pt_var(hterm_var_cnt, buf);
			} else {
				sprintf(buf, "_%d", j);
				return CLPR_new_pt_var(j, buf);
			}
		}
	} else if (CLPR_tag(*t) == TAG_CON) {
		nt = CLPR_new_pterm();
		nt->ty = FUNC;
		nt->narg = 0;
		nt->str = string_of(*t);
		nt->val = mask(*t);
		return nt;
	} else if (CLPR_tag(*t) == CLPR_TAG_NIL) 
		return CLPR_new_pt_emptylist();
	else if (is_unbound_var(*t) || is_daemon(t)) {
		buf = (char *) malloc(20);
		if (with_dump) {
			if (!(j = CLPR_is_funct_var(t))) {
				if (j = CLPR_heap_to_pivot_var(t)) j += CLPR_funct_vars_top + 2*CLPR_flatten_vars_top;
				else if (!(j = CLPR_heap_to_free_var(t))) {
					CLPR_printterm(stdout,t);
					CLPR_fatal("37832");
				}
			}
			sprintf(buf, "_%d", j);
			return CLPR_new_pt_var(j, buf);
		} else {
			if ((j = CLPR_find_heap_addr((CLPR_int) t))) {
				sprintf(buf, "_%d", j);
				return CLPR_new_pt_var(j, buf);
			}
			if (hterm_var_cnt >= MAX_VAR_IN_HEAP) 			
		 		CLPR_fatal("too many variables in term to assert");
			heap_varval[++hterm_var_cnt] = (CLPR_int) t;
			sprintf(buf, "_%d", hterm_var_cnt);
			return CLPR_new_pt_var(hterm_var_cnt, buf);
		}
	} else if (CLPR_tag(*t) == TAG_STR) {
		val = mask(*t);
		if (is_val_hat(val)) {
/* BUG */
			if (val == HAT_QUOTEVAL && !quoted) return ht_to_pterm(t + 1, TRUE,with_dump);
			if (!quoted) val = CLPR_get_evaled_func(val);
		}
		prev = (PTERM *) NULL;
		nt = CLPR_new_pterm();
		nt->ty = FUNC;
		nt->narg = arity_of(val);
		nt->str = string_of(val);
		nt->val = val;
		for (i = nt->narg; i >= 1; i--) {
			cur = ht_to_pterm(t + i, quoted, with_dump);
			cur->next = prev;
			prev = cur;
		}
		nt->first = prev;
		nt->next = (PTERM *) NULL;
		return nt;
	} else if (CLPR_tag(*t) == CLPR_TAG_CONS) {
		nt = CLPR_new_pterm();
		nt->ty = FUNC;
		nt->narg = 2;
		nt->str = string_of(DOT);
		nt->val = DOT;
		nt->first = ht_to_pterm(head_of(t), quoted, with_dump);
		nt->next = (PTERM *) NULL;
		nt->first->next = ht_to_pterm(tail_of(t), quoted, with_dump);
		nt->first->next->next = (PTERM *) NULL;
		return nt;
	} else CLPR_fatal("815381");
}

CLPR_int CLPR_find_heap_addr(CLPR_int v)
{
CLPR_int i;
	for (i = 1; i <= hterm_var_cnt; i++) if (heap_varval[i] == v) return i;
	return 0;
}


CLPR_int CLPR_do_fasserta(CLPR_int *t)
{
CLPR_int i;
PTERM *nt;
RULE_ptr r, first_rule;
CLPR_int first_code, first_flag;
	t = CLPR_findbind(t);
	if (!is_con(t) && !is_func(t)) {
		CLPR_warn_err("bad argument to assert/1 or assertz/1");
		return FALSE;
	}
	CLPR_init_ht_to_pterm();
	nt = ht_to_pterm(t, FALSE, FALSE);
	i = (head_of_rule(nt))->val;
	if (CLPR_p_protected(i)) {
		CLPR_warn_err("assert failed: %s/%d is protected",
			string_of(i), arity_of(i));
		return FALSE;
	}
	if (!CLPR_p_dynamic(i)) {
		CLPR_warn_err("assert failed: %s/%d is not dynamic",
			string_of(i), arity_of(i));
		return FALSE;
	}
	CLPR_pcode = CLPR_epcode;
	CLPR_init_lable();
	if (CLPR_p_compiled(i)) {
		first_rule = CLPR_p_first_rule(i);
		r = CLPR_new_rule();
		r->next = first_rule;
		r->t = nt;
		r->wt = nt;
		if ((r->id = first_rule->id - 1) == 0) r->id = -1;
		first_flag = FALSE;
	} else { 
		first_rule = (RULE *) NULL;
		r = CLPR_new_rule();
		r->next = (RULE *) NULL;
		r->t = nt;
		r->wt = nt;
		r->id = 1;
		CLPR_assign_p_last_rule(i, r);
		first_flag = TRUE;
	}
	r->startcode = CLPR_pcode;
	CLPR_current_rule = r;
	check_codespace(CLPR_pcode);
	CLPR_compile_rule(r->wt, -1);
	check_codespace(CLPR_pcode);
	r->endcode = CLPR_pcode - 1;
	CLPR_assign_p_first_rule(i, r);
	if (first_flag) {
		CLPR_assign_p_start_hdrcode(i, CLPR_pcode);
		CLPR_emit_dynamic_try(i);
		CLPR_assign_p_end_hdrcode(i, CLPR_pcode - 1);
		CLPR_assign_p_compiled(i, TRUE);
	}
	if (!first_rule) CLPR_assign_p_last_compiled_rule(i, r);
	CLPR_epcode = CLPR_pcode;
	CLPR_resolve(TRUE);
	return TRUE;
}

CLPR_int CLPR_do_fassertz(CLPR_int *t)
{
CLPR_int i;
PTERM *nt;
RULE_ptr r, prev_rule;
CLPR_int first_flag;
	t = CLPR_findbind(t);
	if (!is_con(t) && !is_func(t)) {
		CLPR_warn_err("bad argument to assert/1 or assertz/1");
		return FALSE;
	}
	CLPR_init_ht_to_pterm();
	nt = ht_to_pterm(t, FALSE, FALSE);
	i = (head_of_rule(nt))->val;
	if (CLPR_p_protected(i)) {
		CLPR_warn_err("assert failed: %s/%d is protected",
			CLPR_functors[i].pname, CLPR_functors[i].arity);
		return FALSE;
	}
	if (!CLPR_p_dynamic(i)) {
		CLPR_warn_err("assert failed: %s/%d is not dynamic",
			CLPR_functors[i].pname, CLPR_functors[i].arity);
		return FALSE;
	}
	CLPR_pcode = CLPR_epcode;
	CLPR_init_lable();
	if (CLPR_p_compiled(i)) {
		prev_rule = CLPR_p_last_rule(i);
		r = CLPR_store_rule(nt, nt, FALSE);
		first_flag = FALSE;
	} else { 
		prev_rule = (RULE *) NULL;
		r = CLPR_store_rule(nt, nt, FALSE);
		CLPR_assign_p_first_rule(i, r);
		first_flag = TRUE;
	}
	r->startcode = CLPR_pcode;
	CLPR_current_rule = r;
	check_codespace(CLPR_pcode);
	CLPR_compile_rule(r->wt, -1);
	check_codespace(CLPR_pcode);
	r->endcode = CLPR_pcode - 1;
	if (first_flag) {
		CLPR_assign_p_start_hdrcode(i, CLPR_pcode);
		CLPR_emit_dynamic_try(i);
		CLPR_assign_p_end_hdrcode(i, CLPR_pcode-1);
		CLPR_assign_p_compiled(i, TRUE);
	}
	CLPR_assign_p_last_rule(i, r);
	CLPR_assign_p_last_compiled_rule(i, r);
	CLPR_epcode = CLPR_pcode;
	CLPR_resolve(TRUE);
	return TRUE;
}


CLPR_int CLPR_do_asserta(CLPR_int *t)
{
	return CLPR_do_assert_rule(t, TRUE);
}

CLPR_int CLPR_do_assertz(CLPR_int *t)
{
	return CLPR_do_assert_rule(t, FALSE);
}

CLPR_int CLPR_do_assert_rule(CLPR_int *t, CLPR_int start)
{
CLPR_int i;
PTERM *nt;
RULE_ptr r, first_rule;
CLPR_int first_code;
CLPR_int *dump_constraints;
PTERM *c, *clist, *slist;
CLPR_int *head, *body, *lastc, *new_term;
CLPR_int **new_vars;
CLPR_int last_flag = FALSE, first_flag = FALSE;

	t = CLPR_findbind(t);
	if (!is_con(t) && !is_func(t)) {
		CLPR_warn_err("bad argument to assert/1 or assertz/1");
		return FALSE;
	}
	dump_constraints = CLPR_get_rule_constraints(t);
	CLPR_init_ht_to_pterm();
	if (is_nil(dump_constraints)) {
		new_term = t;
		nt = ht_to_pterm(new_term, FALSE, TRUE);
		goto ADD;
	}
	lastc = CLPR_end_of_constraints(dump_constraints);
	if (mask(*t) == IMPL) {
		make_bound_var(lastc + 2, t + 2);
		nt = CLPR_new_p_func2(IMPL,
			ht_to_pterm(t + 1, FALSE, TRUE), 
			ht_to_pterm(dump_constraints, FALSE, TRUE));
	} else { 
		make_bound_var(lastc, lastc[1]);
		nt = CLPR_new_p_func2(IMPL,
			ht_to_pterm(t, FALSE, TRUE), 
			ht_to_pterm(dump_constraints, FALSE, TRUE));
	}	
	make_bound_var(CLPR_heaptop++, dump_constraints);
	CLPR_heaptop++;	
ADD:
	CLPR_cleanup_dump();
	i = (head_of_rule(nt))->val;
	if (CLPR_p_protected(i)) {
		CLPR_warn_err("assert failed: %s/%d is protected",
			string_of(i), arity_of(i));
		return FALSE;
	}
	if (!CLPR_p_dynamic(i)) {
		CLPR_warn_err("assert failed: %s/%d is not dynamic",
			string_of(i), arity_of(i));
		return FALSE;
	}
	CLPR_pcode = CLPR_epcode;
	CLPR_init_lable();
	if (start) {
		r = CLPR_new_rule();
		r->t = r->wt = nt;
		if (CLPR_p_compiled(i)) {
			first_rule = CLPR_p_first_rule(i);
			r->next = first_rule;
			/* avoid special value of 0 */
			if ((r->id = first_rule->id - 1) == 0) r->id = -1;	
		} else {
			r->next = (RULE_ptr) NULL;
			r->id = 1;
			CLPR_assign_p_last_rule(i, r);
			last_flag = TRUE;
			first_flag = TRUE;	/* first time needs headers */
		}
	} else {
		r = CLPR_store_rule(nt, nt, FALSE);
		if (!CLPR_p_compiled(i)) {
			CLPR_assign_p_first_rule(i, r);
			first_flag = TRUE;
		}
		last_flag = TRUE;
	}
	r->startcode = CLPR_pcode;
	CLPR_current_rule = r;
	check_codespace(CLPR_pcode);
	CLPR_compile_rule(r->wt, -1);
	check_codespace(CLPR_pcode);
	r->endcode = CLPR_pcode - 1;
	if (start) CLPR_assign_p_first_rule(i, r);
	else CLPR_assign_p_last_rule(i, r);
	if (first_flag) {
		CLPR_assign_p_start_hdrcode(i, CLPR_pcode);
		CLPR_emit_dynamic_try(i);
		CLPR_assign_p_end_hdrcode(i, CLPR_pcode - 1);
		CLPR_assign_p_compiled(i, TRUE);
	}
	if (last_flag) CLPR_assign_p_last_compiled_rule(i, r);
	CLPR_epcode = CLPR_pcode;
	CLPR_resolve(TRUE);
	return TRUE;
}

CLPR_int CLPR_do_retractall(void)
{
CLPR_int i, k;
char *s;
	for(i = 1; i < CLPR_getmaxfunc(); i++) {
		if (CLPR_p_compiled(i) && !CLPR_p_protected(i)) {
			CLPR_delete_all_rules(i);
			CLPR_assign_p_first_rule(i, NULL);
			CLPR_assign_p_compiled(i, FALSE);
			CLPR_assign_p_cut(i, FALSE);
			CLPR_assign_p_spy(i, FALSE);
			CLPR_delete_lable(i);
		}
	}
	return TRUE;
}

CLPR_int CLPR_do_flush(void)
{
	fflush(CLPR_outfile);
	return TRUE;
}

#if defined(MSDOS) || defined(OS2V2)
CLPR_int CLPR_do_oracle(t1, t2, t3)
CLPR_int *t1, *t2, *t3;
{
	CLPR_warn_err("No oracle/3 on DOS/OS2");
	return FALSE;
}
#else
CLPR_int CLPR_do_oracle(CLPR_int *t1, CLPR_int *t2, CLPR_int *t3)
{
char *s, *p1, *p2;
int fildes1[2];
int fildes2[2];
pid_t p_pid;
IO_REC *tt;
	t1 = CLPR_findbind(t1);
	if (!is_con(t1)) {
		CLPR_warn_err("bad first argument to oracle/3");
		return FALSE;
	}
	t2 = CLPR_findbind(t2);
	if (!is_con(t2)) {
		CLPR_warn_err("bad second argument to oracle/3");
		return FALSE;
	}
	t3 = CLPR_findbind(t3);
	if (!is_con(t3)) {
		CLPR_warn_err("bad third argument to oracle/3");
		return FALSE;
	}
	s = CLPR_functors[mask(*t1)].pname;
	p1 = CLPR_functors[mask(*t2)].pname;
	p2 = CLPR_functors[mask(*t3)].pname;
	if (CLPR_file_is_open(p1, IO_READ, CLPR_files_open) ||
        CLPR_file_is_open(p1, IO_WRITE, CLPR_files_open) ||
        CLPR_file_is_open(p2, IO_READ, CLPR_files_open) ||
        CLPR_file_is_open(p2, IO_WRITE, CLPR_files_open)) CLPR_fatal("oracle/3 failed");
	if (pipe(fildes1) || pipe (fildes2)) CLPR_fatal("oracle/3 failed");
	if ((p_pid = fork ()) == 0) {
	   close (fildes1[1]);
	   close (0);
	   dup(fildes1[0]);
	   close (fildes1[0]);
	   close (fildes2[0]);
	   close (1);
	   dup(fildes2[1]);
	   close (fildes2[1]);
	   execlp (s, s, 0);
	}
	if (p_pid == -1) CLPR_fatal("fork creation unsuccessful in oracle/3");
	close (fildes1[0]);
	close (fildes2[1]);
	tt = (IO_REC *) CLPR_emalloc(sizeof(IO_REC));
	tt->fio = fdopen(fildes1[1], "w");
	tt->s = p1;
	tt->mode = IO_WRITE;
	tt->next = CLPR_files_open;
	CLPR_files_open = tt;
	tt = (IO_REC *) CLPR_emalloc(sizeof(IO_REC));
	tt->fio = fdopen(fildes2[0], "r");
	tt->s = p2;
	tt->mode = IO_READ;
	tt->next = CLPR_files_open;
	CLPR_files_open = tt;
	return TRUE;
}
#endif

#if defined(MSDOS) || defined(OS2V2)
CLPR_int CLPR_do_pipe(t)
CLPR_int *t;
{
	CLPR_warn_err("No pipe/1 on DOS/OS2");
	return FALSE;
}
#else
CLPR_int CLPR_do_pipe(CLPR_int *t)
{
char *s;
int fildes[2];
IO_REC *tt;
	t = CLPR_findbind(t);
	if (!is_con(t)) CLPR_runtime_error("bad argument to pipe/1");
	s = CLPR_functors[mask(*t)].pname;
	if (CLPR_file_is_open(s, IO_READ, CLPR_files_open) || 
	    CLPR_file_is_open(s, IO_WRITE, CLPR_files_open)) 	CLPR_runtime_error("pipe/1 failed");
	if (pipe(fildes)) CLPR_runtime_error("pipe/1 failed");
	tt = (IO_REC *) CLPR_emalloc(sizeof(IO_REC));
	tt->fio = fdopen(fildes[0], "r");
	tt->s = s;
	tt->mode = IO_READ;
	tt->next = CLPR_files_open;
	tt->dual = TRUE;
	CLPR_files_open = tt;
	tt = (IO_REC *) CLPR_emalloc(sizeof(IO_REC));
	tt->fio = fdopen(fildes[1], "w");
	tt->s = s;
	tt->mode = IO_WRITE;
	tt->next = CLPR_files_open;
	tt->dual = TRUE;
	CLPR_files_open = tt;
	return TRUE;
}
#endif

#if defined(MSDOS) || defined(OS2V2)
CLPR_int CLPR_do_fork()
{
	CLPR_warn_err("No fork/0 on DOS/OS2");
	return FALSE;
}
#else
CLPR_int CLPR_do_fork(void)
{
	if (fork()) return TRUE;
	else return FALSE;
}
#endif

CLPR_int CLPR_do_edit(CLPR_int *t)
{
char *s, *p, *command, cmd[MAX_MESSAGE];
	t = CLPR_findbind(t);
	if (!is_con(t)) CLPR_runtime_error("bad argument to edit/1");
	s = CLPR_functors[mask(*t)].pname;
#if defined(MSDOS) || defined(OS2V2)
	if (!(p = (char *) getenv("EDITOR"))) p = "edit";
	sprintf(cmd, "%s /c %s %s", CLPR_command_interpreter(), p, s);
#else
	if (!(p = (char *) getenv("EDITOR"))) p = "vi";
	sprintf(cmd, "%s %s", p, s);
#endif
	if (strlen(cmd) >= MAX_MESSAGE - 1) CLPR_runtime_error("editor name too long"); 
	if (system(cmd)) return FALSE;
	return CLPR_internal_reconsult(s);
}

CLPR_int CLPR_do_more(CLPR_int *t)
{
char *s, *p, cmd[MAX_MESSAGE];
	t = CLPR_findbind(t);
	if (!is_con(t)) CLPR_runtime_error("bad argument to more/1");
	s = CLPR_functors[mask(*t)].pname;
#if defined(MSDOS) || defined(OS2V2)
	if (!(p = getenv("PAGER"))) p = "more";
	sprintf(cmd, "%s /c %s <%s", CLPR_command_interpreter(), p, s);
#else
	if (!(p = getenv("PAGER"))) p = "more";
	sprintf(cmd, "%s %s", p, s);
#endif
	if (strlen(cmd) >= MAX_MESSAGE - 1) CLPR_runtime_error("more name too long"); 
	return (!system(cmd));
}

CLPR_int CLPR_do_halt(void)
{
	CLPR_files_shutdown();
	exit(0);
}

CLPR_int CLPR_do_abort(void)
{
	CLPR_interrupt(0);
}

CLPR_int CLPR_do_sh(void)
{
char *s;
#if defined(MSDOS) || defined(OS2V2)
	return system(CLPR_command_interpreter());
#else
	return system("/bin/sh");
#endif
}
 
CLPR_int CLPR_do_csh(void)
{
char *s;
#if defined(MSDOS) || defined(OS2V2)
	return system(CLPR_command_interpreter());
#else
	return system("/bin/csh");
#endif
}

CLPR_int CLPR_seed = 16;
void CLPR_cmd_ranseed(CLPR_int i)
{
	CLPR_seed = i;
}

CLPR_int CLPR_do_srand(CLPR_int *t)
{
double val;
	t = CLPR_findbind(t);
	if (!is_num(t, &val)) {
		CLPR_warn_err("bad argument to srand/1");
		return FALSE;
	}
	CLPR_seed = (CLPR_int) val;
	return TRUE;
}
 
/* Linear Congruential Pseudo Random Number Generator (Knuth II) */
CLPR_int CLPR_do_rand(CLPR_int *t)
{
CLPR_int *tmp;
double r;
	t = CLPR_findbind(t);
	if (!(is_unbound_var(*t) || is_daemon(t))) {
		CLPR_warn_err("bad argument to rand/1");
		return FALSE;
	}
	CLPR_seed = (CLPR_seed * (1 + 4 * 37 * 109) + 9999) % (32 * 1024);
	r = (((double) CLPR_seed / (double) (32 * 1024)));
	CLPR_makenum(++CLPR_solver_id, r);
	*CLPR_heaptop = addtag(TAG_PAR, CLPR_solver_id);
	tmp = CLPR_heaptop++;
	CLPR_unify(tmp, t);
	return TRUE;
}

CLPR_int CLPR_do_ground(CLPR_int *t)
{
CLPR_int k;
double val;
	t = CLPR_findbind(t);
	if (is_unbound_var(*t) || is_daemon(t)) return FALSE;
	switch ((CLPR_unsigned) CLPR_tag(*t)) {
	case TAG_CON:
	case CLPR_TAG_NIL: return TRUE;
	case TAG_PAR: return (CLPR_is_ground(mask(*t), &val) ? TRUE : FALSE);
	case TAG_STR:		
		for (k = 1; k <= CLPR_functors[mask(*t)].arity; k++) 
			if (!CLPR_do_ground(t + k)) return FALSE;
		return TRUE;
	case CLPR_TAG_CONS:
		if (!CLPR_do_ground((CLPR_int *) pmask(*t))) return FALSE;
		if (!CLPR_do_ground((CLPR_int *) (pmask(*t)) + 1)) return FALSE;
		return TRUE;
	} 
	CLPR_fatal("89422");
}
	
CLPR_int CLPR_do_nonground(CLPR_int *t)
{
	return (CLPR_do_ground(t) ? FALSE : TRUE);
}
 
CLPR_int CLPR_do_var(CLPR_int *t)
{
double dummy;
	t = CLPR_findbind(t);
	return (is_unbound_var(*t) || is_daemon(t) || 
		(is_param(t) && !is_num(t, &dummy)));
}

CLPR_int CLPR_do_nonvar(CLPR_int *t)
{
	return (!CLPR_do_var(t));
}

CLPR_int CLPR_do_arithmetic(CLPR_int *t)
{
	t = CLPR_findbind(t);
	return is_param(t);
}
 
CLPR_int CLPR_do_functor1(CLPR_int *t)
{
	t = CLPR_findbind(t);
	return (is_func(t) || is_cons(t));
}

CLPR_int CLPR_do_atom(CLPR_int *t)
{
	t = CLPR_findbind(t);
	return (is_con(t) || is_nil(t));
}

CLPR_int CLPR_do_atomic(CLPR_int *t)
{
double val;
	t = CLPR_findbind(t);
	return (is_con(t)|| is_nil(t) || is_num(t, &val));
}


CLPR_int CLPR_do_functor3(CLPR_int *t1, CLPR_int *t2, CLPR_int *t3)  /*** functor(term, name, arity) modes: +?? ?++ ***/ 
                  
{
CLPR_int val, val1, arity;
char *name;
double f_arity;
	t1 = CLPR_findbind(t1);
	if (!is_unbound_var(*t1)) {
		if (is_func(t1) || is_con(t1)) {
			val = mask(*t1);
			arity = arity_of(val);
			name = string_of(val);
			if (!CLPR_make_a_con(t2, CLPR_hashbuiltina(name, 0))) return FALSE;
			if (!CLPR_make_a_num(t3, (double) arity)) return FALSE;
			return TRUE;
		} else if (is_cons(t1)) {
			if (!CLPR_make_a_con(t2, CLPR_hashbuiltina(".", 0))) return FALSE;
			if (!CLPR_make_a_num(t3, 2.0)) return FALSE;
			return TRUE;
		} else if (is_nil(t1)) {
			if (!CLPR_make_a_nil(t2)) return FALSE;
			if (!CLPR_make_a_num(t3, 0.0)) return FALSE;
			return TRUE;
		} else return FALSE;
	}
	t2 = CLPR_findbind(t2);
	t3 = CLPR_findbind(t3);
	if (is_unbound_var(t2) || (!is_con(t2) && !is_nil(t2))) return FALSE;
	if (is_unbound_var(t3) || !is_num(t3, &f_arity)) return FALSE;

	val1 = mask(*t2);
	arity = (CLPR_int) f_arity;
	if (arity == 0) 
		if (is_con(t2))
			return CLPR_make_a_con(t1, CLPR_hashbuiltina(string_of(val1), arity));
		else return CLPR_make_a_nil(t1);
	if (arity == 2 && strcmp(string_of(val1), ".") == 0) {
		if (!CLPR_make_a_new_list(t1)) return FALSE;
		if (CLPR_have_runable && !CLPR_wake_runable()) return FALSE;
	} else {
		if (!CLPR_make_a_new_fun(t1, CLPR_hashbuiltina(string_of(val1), arity), arity))
			return FALSE;
		if (CLPR_have_runable && !CLPR_wake_runable()) return FALSE;
	}
	return TRUE;
}


CLPR_int CLPR_do_arg(CLPR_int *t1, CLPR_int *t2, CLPR_int *t3)  /*** oops:
					arg(+i, +func, arg) previously was
					arg(func, i, arg) modes: ++? ***/
                  
{
double f_arity;
CLPR_int *arg, *arg1, n, arity;
	t1 = CLPR_findbind(t1);
	if (is_unbound_var(*t1) || !is_num(t1, &f_arity)) return FALSE;
	t2 = CLPR_findbind(t2);
	if (is_unbound_var(*t2)) return FALSE;
	arity = (CLPR_int) f_arity;
	if (is_func(t2)) {
		n = arity_of(mask(*t2));
		if (arity < 1 || arity > n) return FALSE;
		return CLPR_unify(t2 + arity, t3);
	} else if (is_cons(t2)) {
		if (arity < 1 || arity > 2) return FALSE;
		return CLPR_unify(((CLPR_int *) pmask(*t2))+arity-1, t3);
	}
	return FALSE;
}

/* arg(func, i, arg) modes: ++? */
CLPR_int CLPR_do_replace_arg(CLPR_int *t1, CLPR_int *t2, CLPR_int *t3)
{
double f_arity;
CLPR_int *arg, *arg1, n, arity;
	t1 = CLPR_findbind(t1);
	if (is_unbound_var(*t1)) return FALSE;
	t2 = CLPR_findbind(t2);
	if (is_unbound_var(*t2) || !is_num(t2, &f_arity)) return FALSE;
	arity = (CLPR_int) f_arity;
	t3 = CLPR_findbind(t3);
	if (is_func(t1)) {
		n = arity_of(mask(*t1));
		arg = t1 + arity;
		if (arity < 1 || arity > n) return FALSE;
		if (is_unbound_var(*arg)) return CLPR_unify(arg, t3);
		else {
			CLPR_push_destructive_trail(arg, *arg);
			if (is_unbound_var(*t3)) {
				make_unbound_var(arg);
				return CLPR_unify(arg, t3);
			} else make_bound_var(arg, t3);
		}
		return TRUE;
	} else if (is_cons(t1)) {
		if (arity < 1 || arity > 2) return FALSE;
		arg = ((CLPR_int *) pmask(*t1))+arity-1;
		if (is_unbound_var(*arg)) return CLPR_unify(arg, t3);
		else {
			CLPR_push_destructive_trail(arg, *arg);
			if (is_unbound_var(*t3)) {
				make_unbound_var(arg);
				return CLPR_unify(arg, t3);
			} else make_bound_var(arg, t3);
		}
		return TRUE;
	}
	return FALSE;
}

CLPR_int CLPR_do_occurs(CLPR_int *t1, CLPR_int *t2)
{
	t1 = CLPR_findbind(t1);
	t2 = CLPR_findbind(t2);
	if (!(is_unbound_var(*t1) || is_daemon(t1) || is_param(t1))) {
		CLPR_warn_err("bad first argument to CLPR_occurs/2");
		return FALSE;
	}
	return CLPR_occurs(t1, t2);
}
 
CLPR_int CLPR_occurs(CLPR_int *vspot, CLPR_int *t)
{
CLPR_int j;
	if (is_unbound_var(*t) || is_daemon(t)) {
		if (t == vspot && !is_param(vspot)) return TRUE;
		else return FALSE;
	} else if (is_param(t)) {
		if (is_param(vspot) && CLPR_equal_param(mask(*t), mask(*vspot))) 
			return TRUE;
		else return FALSE;
	} else if (is_func(t)) {
		for (j = 1; j <= CLPR_functors[mask(*t)].arity; j++)
			if (CLPR_occurs(vspot, CLPR_findbind(t + j))) return TRUE;
	} else if (is_cons(t)) {
		if (CLPR_occurs(vspot, CLPR_findbind(head_of(t)))) return TRUE;
		if (CLPR_occurs(vspot, CLPR_findbind(tail_of(t)))) return TRUE;
	}
	return FALSE;
}

CLPR_int CLPR_do_eqeq(CLPR_int *t1, CLPR_int *t2)
{
double v1, v2;
CLPR_int i, n;

	t1 = CLPR_findbind(t1);
	t2 = CLPR_findbind(t2);
	if (is_unbound_var(*t1)) {
		if (t1 == t2 && is_unbound_var(*t2)) return TRUE;
	} else if (is_con(t1)) {
		if (is_con(t2) && mask(*t1) == mask(*t2)) return TRUE;
	} else if (is_nil(t1)) {
		if (is_nil(t2)) return TRUE;
	} else if (is_cons(t1)) {
		if (is_cons(t2)) {
			if (CLPR_do_eqeq(head_of(t1), head_of(t2)) &&
					CLPR_do_eqeq(tail_of(t1), tail_of(t2)))
				return TRUE;
		}
	} else if (is_param(t1)) {
		if (CLPR_is_ground(mask(*t1), &v1)) {
			if (is_num(t2, &v2) && is_equal(v1, v2)) return TRUE;
		} else if (is_param(t2) && CLPR_equal_param(mask(*t1), mask(*t2)))
			return TRUE;
	} else if (is_daemon(t1)) {
		if (t1 == t2 && is_daemon(t2)) return TRUE;
	} else if (is_func(t1)) {
		if (!is_func(t2) || mask(*t1) != mask(*t2)) return FALSE;
		n = arity_of(mask(*t1));
		for (i = 1; i <= n; i++)
			if (!CLPR_do_eqeq(t1 + i, t2 + i)) return FALSE;
		return TRUE;
	} else CLPR_fatal("71881");
	return FALSE;
}
	
CLPR_int CLPR_do_floor(CLPR_int *t1, CLPR_int *t2)
{
double c1, c2;
	t1 = CLPR_findbind(t1);
	if (!is_num(t1, &c1)) {
		CLPR_warn_err("bad first argument to floor/2");
		return FALSE;
	}
	c1 = floor(c1);
	t2 = CLPR_findbind(t2);
	return CLPR_make_a_num(t2, c1);
}

CLPR_int CLPR_do_op3(CLPR_int *t1, CLPR_int *t2, CLPR_int *t3)
{
double val;
char *opname, *style;
	t1 = CLPR_findbind(t1);
	if (!is_num(t1, &val)) {
		CLPR_warn_err("bad first argument to op/3");
		return FALSE;
	}
	if ((val != ((CLPR_int) val))
			|| val < MIN_OP_PRECEDENCE-1 || val > MAX_OP_PRECEDENCE) {
		CLPR_warn_err("bad first argument to op/3");	
		return FALSE;
	}
	t2 = CLPR_findbind(t2);
	if (!is_con(t2)) {
		CLPR_warn_err("bad second argument to op/3");
		return FALSE;
	}
	t3 = CLPR_findbind(t3);
	if (!is_con(t3)) {
		CLPR_warn_err("bad third argument to op/3");
		return FALSE;
	}
	opname = CLPR_functors[mask(*t3)].pname;
	style = CLPR_functors[mask(*t2)].pname;
	if (CLPR_do_op(opname, style, (CLPR_int) val)) return TRUE;
	else {
		CLPR_warn_err("bad second argument to op/3");
		return FALSE;
	}
}

CLPR_int CLPR_do_hide(CLPR_int *t1, CLPR_int *t2)
{
	CLPR_runtime_error("hide/2 not yet implemented");
}

CLPR_int CLPR_do_prot(CLPR_int *t1, CLPR_int *t2)
{
double val;
CLPR_int i;
char *nm;
CLPR_int arity;
RULE *r;
	t1 = CLPR_findbind(t1);
	if (!is_con(t1)) {
		CLPR_warn_err("bad first argument to prot/2");
		return FALSE;
	}
	nm = CLPR_functors[mask(*t1)].pname;
	t2 = CLPR_findbind(t2);
	if (!is_num(t2, &val)) {
		CLPR_warn_err("bad second argument to prot/2");
		return FALSE;
	}
	arity = val;  
	for (i = 1; i < CLPR_getmaxfunc(); i++)
		if (CLPR_p_compiled(i)) {
			if (!strcmp(CLPR_functors[i].pname, nm) && CLPR_functors[i].arity == arity) {
				CLPR_assign_p_protected(i, TRUE);
				return TRUE;
		}
	}
	return FALSE;
}

CLPR_int CLPR_do_ztime(void)
{
#if defined(SYS5) || defined(OS2V2) || defined(BSD)
	get_user_time(&basetime);
#endif
	return TRUE;
}

CLPR_int CLPR_do_ctime(CLPR_int *t)
{
long t1, t2;
double t3;
CLPR_int *tmp;
	t = CLPR_findbind(t);
	if (!(is_unbound_var(*t) || is_daemon(t))) {
		CLPR_warn_err("bad argument to ctime/1\n");
		return FALSE;
	}
#if defined(SYS5) || defined(OS2V2)
	get_user_time(&curtime);
	if ((t3 = (double) (curtime.tms_utime-basetime.tms_utime)) < 0.0) t3 = 0.0;
	t3 = t3 / (double) HZ;
#endif
#if defined(BSD) && !defined(OS2V2)
	get_user_time(&curtime);
	t1 = curtime.tv_sec - basetime.tv_sec;
	t2 = curtime.tv_usec - basetime.tv_usec;
	t3 = ((double) t1) + ((double) t2) / U_SEC;
#endif
	tmp = CLPR_heaptop;
	CLPR_makenum(++CLPR_solver_id, t3);
	*(CLPR_heaptop++) = addtag(TAG_PAR, CLPR_solver_id);
	CLPR_unify(t, tmp);
	return TRUE;
}

#if defined(SYS5) || defined(OS2V2)
void get_user_time(time)
struct tms *time;
{
	times(time);
}
#endif
#if defined(BSD) && !defined(OS2V2)
void get_user_time(time)
struct timeval *time;
{
struct rusage usg;
	getrusage(RUSAGE_SELF, &usg);
	time->tv_sec = usg.ru_utime.tv_sec;
	time->tv_usec = usg.ru_utime.tv_usec;
}
#endif

CLPR_int CLPR_do_history(void)
{
	CLPR_printhist(0);
	return TRUE;
}

CLPR_int CLPR_do_history1(CLPR_int *t)
{
double val;
	t = CLPR_findbind(t);
	if (!is_num(t, &val)) {
		CLPR_warn_err("bad argument to history/1");
		return FALSE;
	}
	CLPR_printhist((CLPR_int) val);
	return TRUE;
}

CLPR_int CLPR_do_nl(void)
{
	fprintf(CLPR_outfile, "\n");
	return TRUE;
}

CLPR_int CLPR_do_debugging(void)
{
CLPR_int i;
RULE *r;
	fprintf(CLPR_outfile, "\n");
	for(i = 1; i < CLPR_getmaxfunc(); i++) {
		if (CLPR_p_spy(i) && CLPR_p_compiled(i))
			fprintf(CLPR_outfile, "%s/%d\n", CLPR_functors[i].pname, CLPR_functors[i].arity);
	}
	fprintf(CLPR_outfile, "\n");
	return TRUE;
}

CLPR_int CLPR_do_codegen_debug(void)
{
	CLPR_CODEGEN_DEBUG = TRUE;
	return TRUE;
}

CLPR_int CLPR_do_codegen_nodebug(void)
{
	CLPR_CODEGEN_DEBUG = FALSE;
	return TRUE;
}

CLPR_int CLPR_do_spy(void)
{
CLPR_int i;
	for(i = 1; i < CLPR_getmaxfunc(); i++)
		if (CLPR_p_compiled(i) && !CLPR_p_protected(i) && !CLPR_p_spy(i))
			 CLPR_assign_p_spy(i, TRUE);
	return TRUE;
}

CLPR_int CLPR_do_nospy(void)
{
CLPR_int i;
	for(i = 1; i < CLPR_getmaxfunc(); i++)
		if (CLPR_p_compiled(i) && CLPR_p_spy(i)) CLPR_assign_p_spy(i, FALSE);
	return TRUE;
}

CLPR_int CLPR_do_spy2(CLPR_int *t1, CLPR_int *t2)
{
double val;
CLPR_int i, arity;
char *s;
	t1 = CLPR_findbind(t1);
	if (!is_con(t1)) {
		CLPR_warn_err("bad first argument to spy/2");
		return FALSE;
	}
	s = CLPR_functors[mask(*t1)].pname;
	t2 = CLPR_findbind(t2);
	if (!is_num(t2, &val)) {
		CLPR_warn_err("bad second argument to spy/2"); 
		return FALSE;
	}
	arity = (CLPR_int) val;
	for(i = 1; i < CLPR_getmaxfunc(); i++) {
		if (CLPR_p_compiled(i) &&
		    !strcmp(CLPR_functors[i].pname, s) && 
		    CLPR_functors[i].arity == arity &&
		    !CLPR_p_protected(i)) {
				CLPR_assign_p_spy(i, TRUE);
				return TRUE;
		}
	}
	return FALSE;
}

CLPR_int CLPR_do_nospy2(CLPR_int *t1, CLPR_int *t2)
{
double val;
CLPR_int i, arity;
char *s;
	t1 = CLPR_findbind(t1);
	if (!is_con(t1)) CLPR_warn_err("bad first argument to spy/2");
	s = CLPR_functors[mask(*t1)].pname;
	t2 = CLPR_findbind(t2);
	if (!is_num(t2, &val)) CLPR_warn_err("bad second argument to spy/2"); 
	arity = (CLPR_int) val;
	for(i = 1; i < CLPR_getmaxfunc(); i++) {
		if (CLPR_p_compiled(i) &&
		    !strcmp(CLPR_functors[i].pname, s) && 
		    CLPR_functors[i].arity == arity &&
		    !CLPR_p_protected(i)) {
				CLPR_assign_p_spy(i, FALSE);
				return TRUE;
		}
	}
	return FALSE;
}

CLPR_int CLPR_do_warning(CLPR_int *t)
{
	t = CLPR_findbind(t);
	if (!is_con(t)) {
		CLPR_warn_err("bad argument to CLPR_warning/1");
		return FALSE;
	}
	if (!strcmp(string_of(mask(*t)), "abort")) CLPR_WARN_ABORT = TRUE;
	else if (!strcmp(string_of(mask(*t)), "continue")) CLPR_WARN_ABORT = FALSE;
	else if (!strcmp(string_of(mask(*t)), "warning_on")) CLPR_WARNING_FLAG = TRUE;
	else if (!strcmp(string_of(mask(*t)), "warning_off")) CLPR_WARNING_FLAG = FALSE;
	else if (!strcmp(string_of(mask(*t)), "redefine_on")) CLPR_REDEFINE_FLAG = TRUE;
	else if (!strcmp(string_of(mask(*t)), "redefine_off")) CLPR_REDEFINE_FLAG = FALSE;
	else {
		CLPR_warn_err("bad argument to CLPR_warning/1");
		return FALSE;
	}
	return TRUE;
}

CLPR_int CLPR_do_implicit(void)
{
	CLPR_IMPLICIT = TRUE;
	CLPR_PARTIAL_IMPLICIT = FALSE;
	return TRUE;
}

CLPR_int CLPR_do_no_implicit(void)
{
	CLPR_IMPLICIT = FALSE;
	CLPR_PARTIAL_IMPLICIT = FALSE;
	return TRUE;
}

CLPR_int CLPR_do_partial_implicit(void)
{
	CLPR_IMPLICIT = TRUE;
	CLPR_PARTIAL_IMPLICIT = TRUE;
	return TRUE;
}

CLPR_int CLPR_do_trace(void)
{
	CLPR_TRACE = TRUE;
	return TRUE;
}

CLPR_int CLPR_do_notrace(void)
{
	CLPR_TRACE = FALSE;
	return TRUE;
}

CLPR_int CLPR_do_clam_trace(void)
{
	CLPR_CLAM_TRACE = TRUE;
	return TRUE;
}

CLPR_int CLPR_do_clam_notrace(void)
{
	CLPR_CLAM_TRACE = FALSE;
	return TRUE;
}

CLPR_int CLPR_do_compile(CLPR_int *t1, CLPR_int *t2)
{
	CLPR_runtime_error("CLPR_compile/2 not yet implemented");
}

CLPR_int CLPR_do_dynamic(CLPR_int *t1, CLPR_int *t2)
{
double arity;
CLPR_int val;
PTERM *nt;
char *str;
	t1 = CLPR_findbind(t1);
	t2 = CLPR_findbind(t2);
	if (!is_con(t1)) 
		CLPR_warn_err("bad first argument to dynamic/2 -- constant required");
	if (!is_num(t2, &arity))
		CLPR_warn_err("bad second argument to dynamic/2 -- number required");
	str = string_of(mask(*t1));
	val = CLPR_hashbuiltina(str, (CLPR_int) arity);
	if (CLPR_p_compiled(val) && !CLPR_p_dynamic(val)) 
		CLPR_warning("%s/%d already compiled, too late to make dynamic (ignored)",
			string_of(val), arity_of(val));
	else CLPR_assign_p_dynamic(val, TRUE);
	return TRUE;
}

CLPR_int CLPR_do_e_stats(void)
{
#ifdef STATS
	CLPR_print_e_stats();
#else
	fprintf(CLPR_outfile, "No stats - recompile with -DSTATS");
#endif
	return TRUE;
}

CLPR_int CLPR_do_s_stats(void)
{
#ifdef STATS
	CLPR_print_s_stats();
#else
	fprintf(CLPR_outfile, "No stats - recompile with -DSTATS");
#endif
	return TRUE;
}

CLPR_int CLPR_do_stats(void)
{
#ifdef STATS
	CLPR_print_stats();
#else
	fprintf(CLPR_outfile, "No stats - recompile with -DSTATS");
#endif
	return TRUE;
}

CLPR_int CLPR_do_dump_tableaus(void)
{
	CLPR_print_ineq_tableau();
	CLPR_check_tableaus();
	return TRUE;
}

CLPR_int CLPR_get_quoted_func(CLPR_int v)
{
	switch (v) {
		case PLUSVAL: return HAT_PLUSVAL;
		case U_MINUSVAL: return HAT_U_MINUSVAL;
		case MINUSVAL: return HAT_MINUSVAL;
		case MULTVAL: return HAT_MULTVAL;
		case DIVVAL: return HAT_DIVVAL;
		case SINVAL: return HAT_SINVAL;
		case COSVAL: return HAT_COSVAL;
		case ABSVAL: return HAT_ABSVAL;
		case MINVAL: return HAT_MINVAL;
		case MAXVAL: return HAT_MAXVAL;
		case POWVAL: return HAT_POWVAL;
		case EVALVAL: return HAT_EVALVAL;
		case QUOTEVAL: return HAT_QUOTEVAL;
		default: CLPR_fatal("63211");
	}
}

CLPR_int CLPR_get_evaled_func(CLPR_int v)
{
	switch (v) {
		case HAT_PLUSVAL: return PLUSVAL;
		case HAT_U_MINUSVAL: return U_MINUSVAL;
		case HAT_MINUSVAL: return MINUSVAL;
		case HAT_MULTVAL: return MULTVAL;
		case HAT_DIVVAL: return DIVVAL;
		case HAT_SINVAL: return SINVAL;
		case HAT_COSVAL: return COSVAL;
		case HAT_ABSVAL: return ABSVAL;
		case HAT_MINVAL: return MINVAL;
		case HAT_MAXVAL: return MAXVAL;
		case HAT_POWVAL: return POWVAL;
		case HAT_EVALVAL: return EVALVAL;
		case HAT_QUOTEVAL: return QUOTEVAL;
		default: CLPR_fatal("63212");
	}
}

static CLPR_int heapify_pterm(PTERM *t, CLPR_int *hp, CLPR_int quoted)
{
CLPR_int v, n, *start, *s, heap_func_val;
CLPR_int *t1;
PTERM *arg;
	switch (t->ty) {
	case REAL:
		CLPR_makenum(++CLPR_solver_id, t->rval);
		return addtag(TAG_PAR, CLPR_solver_id);
	case VAR:
		if (!hp) CLPR_fatal("81811");
		if (t->val >= MAX_VAR_IN_HEAP) CLPR_fatal("too many variables in term");
		if (var_heapval[t->val]) return var_heapval[t->val];
		else {
			var_heapval[t->val] = (CLPR_int) hp;
			if (t->val > pterm_var_cnt) pterm_var_cnt = t->val;
		}
		return 0;
	case FUNC:
		if (t->val == EMPTYLIST) return addtag(CLPR_TAG_NIL, EMPTYLIST);
		if (t->val == DOT) {
			start = s = CLPR_heaptop;
			CLPR_heaptop += 2;
			*s = heapify_pterm(t->first, s, quoted);
			s++;
			*s = heapify_pterm(t->first->next, s, quoted);
			return addtag(CLPR_TAG_CONS, (CLPR_int) start);
		}
		if (!t->narg) return addtag(TAG_CON, t->val);
		if (is_arith_term(t)) {
			heap_func_val = CLPR_get_quoted_func(t->val);
		} else if (is_arith_hat(t)) {
			if (quoted) {
				t1 = CLPR_heaptop;
				CLPR_heaptop += 2;
				make_bound_var(t1, addtag(TAG_STR, HAT_QUOTEVAL));
				make_bound_var(t1 + 1, heapify_pterm(t, 0, FALSE));	
				return (CLPR_int) t1;
			} else heap_func_val = t->val;			
		} else heap_func_val = t->val;
		start = s = CLPR_heaptop;
		CLPR_heaptop += t->narg + 1;
		*s++ = addtag(TAG_STR, heap_func_val);
		for (n = 1, arg = t->first; n <= t->narg; n++, s++, arg = arg->next) 
			*s = heapify_pterm(arg, s, quoted);
		return (CLPR_int) start;
	default: CLPR_fatal("08085");
	}
}

CLPR_int *CLPR_pterm_to_heap(PTERM *t)
{
CLPR_int i;
	for (i = 0; i <= pterm_var_cnt; i++) var_heapval[i] = 0;
	pterm_var_cnt = -1;
	return (CLPR_int *) heapify_pterm(t, CLPR_heaptop, TRUE);
}

CLPR_int *CLPR_pterm_to_quoted_heap(PTERM *t, CLPR_int use_collected_vars)
{
CLPR_int i;
	if (!use_collected_vars) {
		for (i = 0; i <= pterm_var_cnt; i++) var_heapval[i] = 0;
		pterm_var_cnt = -1;
	}
	return (CLPR_int *) heapify_pterm(t, CLPR_heaptop, TRUE);
}

CLPR_int *CLPR_func_univ(CLPR_int *t, CLPR_int n);
/* =..(Term, List */
CLPR_int CLPR_do_univ(CLPR_int *t1, CLPR_int *t2)
{
CLPR_int *l, *nt, *tt, *new_list;
CLPR_int *principle;
CLPR_int v1, v2, n, i;
double num;
	t1 = CLPR_findbind(t1);
	t2 = CLPR_findbind(t2);
	v1 = is_unbound_var(*t1) || is_daemon(t1);
	v2 = is_unbound_var(*t2) || is_daemon(t2);
	if (!v2 && !is_cons(t2)) {
		CLPR_warn_err("bad 2nd argument to =../2");
		return FALSE;
	}
	if (v1 && v2) {
		CLPR_warn_err("insufficiently instantiated arguments to =../2");
		return FALSE;
	}
	if (v1) {
		l = t2;
		if (!CLPR_count_list(l, &n) || n < 1) {
			CLPR_warn_err("bad 2nd argument to =../2");
			return FALSE;
		}
		principle = CLPR_findbind(head_of(l));
		if (n == 1) {
			if (is_nil(principle)) return CLPR_make_a_nil(t1);
			if (is_con(principle)) return CLPR_make_a_con(t1, mask(*principle));
			if (is_num(principle, &num)) return CLPR_make_a_num(t1, num);
			CLPR_warn_err("bad 2nd argument to =../2");
			return FALSE;
		} else if (!is_con(principle)) {
			CLPR_warn_err("bad 2nd argument to =../2");
			return FALSE;
		}
		n--;
		nt = CLPR_heaptop;
		if (n == 2 && !strcmp(string_of(mask(*principle)), ".")) {
			*CLPR_heaptop = addtag(CLPR_TAG_CONS, CLPR_heaptop + 1);
			tt = CLPR_findbind(tail_of(l));
			*(CLPR_heaptop + 1) = (CLPR_int) CLPR_findbind(head_of(tt));
			tt = CLPR_findbind(tail_of(tt));
			*(CLPR_heaptop + 2) = (CLPR_int) CLPR_findbind(head_of(tt));
			CLPR_heaptop += 3;
		} else {
			*CLPR_heaptop = addtag(TAG_STR,
				CLPR_hashbuiltina(CLPR_functors[mask(*CLPR_findbind((CLPR_int *)pmask(*l)))].pname, n));
			for (i = 1, tt = CLPR_findbind(tail_of(l)); i <= n;
					i++, tt = CLPR_findbind(tail_of(tt)))
				make_bound_var(nt + i, CLPR_findbind(head_of(tt)));
			CLPR_heaptop = nt + n + 1;
		}
		if (!CLPR_unify(t1, nt)) return FALSE;
		if (CLPR_have_runable && !CLPR_wake_runable()) return FALSE;
		return TRUE;
	}
	new_list = CLPR_findbind(CLPR_make_cons(NULL, NULL));
	switch ((CLPR_unsigned) CLPR_tag(*t1)) {
	case CLPR_TAG_NIL:
	case TAG_CON:
	case TAG_PAR:
		n = 0;
		break;
	case TAG_STR:
		n = arity_of(mask(*t1));
		break;
	case CLPR_TAG_CONS:
		n = 2;
		break;
	default:
		CLPR_fatal("83563");
	}
	principle = head_of(new_list);
	if (n == 0) {
		*(tail_of(new_list)) = addtag(CLPR_TAG_NIL, EMPTYLIST);
		if (is_nil(t1)) CLPR_make_a_nil(principle);
		else if (is_num(t1, &num)) CLPR_make_a_num(principle, num); 
		else if (is_con(t1)) CLPR_make_a_con(principle, mask(*t1));
	} else if (is_cons(t1)) {
		*principle = addtag(TAG_CON, CLPR_hashbuiltina(".", 0));
		tt = CLPR_findbind(CLPR_make_cons(head_of(t1), 
				CLPR_make_cons(tail_of(t1), (CLPR_int *) addtag(CLPR_TAG_NIL, EMPTYLIST))));
		principle[1] = (CLPR_int) tt;
	} else {
		*principle = addtag(TAG_CON, 
			CLPR_hashbuiltina(string_of(mask(*t1)), 0));
		*(tail_of(new_list)) = addtag(CLPR_TAG_CONS, CLPR_func_univ(t1, n));
	}
	if (!CLPR_unify(new_list, t2)) return FALSE;
	if (CLPR_have_runable && !CLPR_wake_runable()) return FALSE;
	return TRUE;
}
 
CLPR_int CLPR_count_list(CLPR_int *l, CLPR_int *np)
{
CLPR_int *ll;
CLPR_int n;
	for (ll = CLPR_findbind(l), n = 0; ll; ll = CLPR_findbind((CLPR_int *)(pmask(*ll))+1)) {
		if (CLPR_tag(*ll) == CLPR_TAG_NIL) {
			*np = n;
			return TRUE;
		} else if (CLPR_tag(*ll) != CLPR_TAG_CONS) return FALSE;
		n++;
	}
	return FALSE;
}
 
CLPR_int *CLPR_func_univ(CLPR_int *t, CLPR_int n)
{
CLPR_int i, *l, *start;
	l = start = CLPR_heaptop;
	CLPR_heaptop += 2;

	for (i = 1; i <= n; i++, l = CLPR_heaptop, CLPR_heaptop += 2) {
		*l = (CLPR_int) (t + i);
		*(l + 1) = addtag(CLPR_TAG_CONS, CLPR_heaptop);
	}
	*(l - 1) = addtag(CLPR_TAG_NIL, EMPTYLIST);
	CLPR_heaptop = l;
	return start;
}


CLPR_int CLPR_do_asm_code_1(CLPR_int *t)
{
char *s;
	t = CLPR_findbind(t);
	if (!is_con(t)) CLPR_runtime_error("bad argument to $$label/1");
	s = CLPR_functors[mask(*t)].pname;
        CLPR_emit_instruction(s,NULL,NULL,NULL);
        return TRUE;
}

CLPR_int CLPR_do_asm_code_2(CLPR_int *t, CLPR_int *t1)
{
char *s;
	t = CLPR_findbind(t);
	if (!is_con(t)) CLPR_runtime_error("bad argument to $$label/1");
	s = CLPR_functors[mask(*t)].pname;
        CLPR_emit_instruction(s,t1,NULL,NULL);
        return TRUE;
}

CLPR_int CLPR_do_asm_code_3(CLPR_int *t, CLPR_int *t1, CLPR_int *t2)
{
char *s;
	t = CLPR_findbind(t);
	if (!is_con(t)) CLPR_runtime_error("bad argument to $$label/1");
	s = CLPR_functors[mask(*t)].pname;
        CLPR_emit_instruction(s,t1,t2,NULL);
        return TRUE;
}

CLPR_int CLPR_do_asm_code_4(CLPR_int *t, CLPR_int *t1, CLPR_int *t2, CLPR_int *t3)
{
char *s;
	t = CLPR_findbind(t);
	if (!is_con(t)) CLPR_runtime_error("bad argument to $$label/1");
	s = CLPR_functors[mask(*t)].pname;
        CLPR_emit_instruction(s,t1,t2,t3);
        return TRUE;
}

CLPR_int CLPR_do_asm_label(CLPR_int *t)
{
char *s;
	t = CLPR_findbind(t);
	if (!is_con(t)) CLPR_runtime_error("bad argument to $$label/1");
	s = CLPR_functors[mask(*t)].pname;
        CLPR_pcode = CLPR_epcode;
        CLPR_store_lable(mask(*t),s); 
        CLPR_epcode = CLPR_pcode;
        return TRUE;
}

CLPR_int CLPR_do_asm_list(CLPR_int *t1, CLPR_int *t2)
{
CLPR_int start, end;
char *s;
CLPR_int i;        
	t1 = CLPR_findbind(t1);
	if (!is_con(t1)) CLPR_runtime_error("bad 1st argument to $$list/2");
	s = CLPR_functors[mask(*t1)].pname;
        start = CLPR_find_loc_for_lable(mask(*t1),s); 
        if (start == -1) CLPR_runtime_error("can't find 1st label in $$list/2");
	t2 = CLPR_findbind(t2);
	if (!is_con(t2)) CLPR_runtime_error("bad argument to $$label/1");
	s = CLPR_functors[mask(*t2)].pname;
        end = CLPR_find_loc_for_lable(mask(*t2),s); 
        if (i == -1) CLPR_runtime_error("can't find 2nd label in $$list/2");
        CLPR_print_asm(start,end-1);
        return TRUE;
}

CLPR_int CLPR_do_asm_call(CLPR_int *t)
{
char *s;
CLPR_int i;
	t = CLPR_findbind(t);
	if (!is_con(t)) CLPR_runtime_error("bad argument to $$call/1");
	s = CLPR_functors[mask(*t)].pname;
        i = CLPR_find_loc_for_lable(mask(*t),s);
        if (i == -1) CLPR_runtime_error("can't find label to call in $$call/1");
        return CLPR_rungoal(i, FALSE);
}

CLPR_int CLPR_do_asm_resolve(void)
{
CLPR_resolve(TRUE);
return TRUE;
}

void CLPR_emit_instruction(char *s, CLPR_int *t1, CLPR_int *t2, CLPR_int *t3)
{
CLPR_int *largs[3];
CLPR_int i, j, *k, opc;
char *mnmonic;
CLPR_int arity;
double tf, *CLPR_tfp;
double val;
CLPR_int savepcode;

	savepcode = CLPR_pcode;
	CLPR_pcode = CLPR_epcode;
	if (t1) largs[0] = CLPR_findbind(t1); else largs[0] = (CLPR_int *) NULL;
	if (t2) largs[1] = CLPR_findbind(t2); else largs[1] = (CLPR_int *) NULL;
	if (t3) largs[2] = CLPR_findbind(t3); else largs[2] = (CLPR_int *) NULL;
    if ((i = CLPR_find_asm_by_mnemonic(s)) == -1) {
		fprintf(stderr, "(%s) ", s);
		CLPR_runtime_error("unknown opcode in $$CLPR_code");
	}
    opc = CLPR_asmlist[i].num;
    arity = CLPR_asmlist[i].nargs;
        CLPR_emit_word(opc);
    if (CLPR_f_is_switch_inst(opc)) 
		CLPR_runtime_error("switches unimplemented in $$CLPR_code");
        /*** maybe handle some other special cases here ??? ***/
        /*** why not reintegrate this using an SPRED type ??? ***/
    if (opc == SPCALL) {
		if (!(is_func(largs[0]) || 
		    is_con(largs[0])) ||
		    !CLPR_spred[mask(*(largs[0]))])
		CLPR_runtime_error("bad system predicate in $$CLPR_code");
		CLPR_emit_word(mask(*(largs[0])));
		goto END_CODE;
    }
    CLPR_tfp = &tf;
    for (j = 1; j <= arity; j++) {
        k = largs[j - 1];
        switch(CLPR_asmlist[i].type[j - 1]) {
            case ADDRESS:
                k = CLPR_findbind(k);
                if (!is_con(k)) CLPR_runtime_error("bad argument to $CLPR_code");
                CLPR_emit_addr(CLPR_functors[mask(*k)].pname,mask(*k));
                break;
            case SINT:
            case BYTE:
            case VARIABLE:
            case REG:
				k = CLPR_findbind(k);
				if (!is_num(k, &val)) CLPR_runtime_error("bad argument to $$CLPR_code");
				CLPR_emit_word((CLPR_int)val);
				break;
            case FLEXARG:
				CLPR_runtime_error("unimplemented arg to $$CLPR_code");
				break;
            case FUNCTOR:
				k = CLPR_findbind(k);
				if (!is_func(k)) CLPR_runtime_error("bad argument to$$CLPR_code");
				CLPR_emit_word(mask(*k));
                break;
            case CONSTANT:
				k = CLPR_findbind(k);
				if (!is_con(k)) CLPR_runtime_error("bad argument to $$CLPR_code");
				CLPR_emit_word(mask(*k));
                break;
            case FLOAT:
				k = CLPR_findbind(k);
				if (!is_num(k, &val)) CLPR_runtime_error("bad argument to $$CLPR_code");
				tf = val;
				CLPR_emit_word(*((CLPR_int *) CLPR_tfp));
				CLPR_emit_word(*(((CLPR_int *) CLPR_tfp) + 1));
				break;
            default:
				CLPR_fatal("79965a");
        }
    }
END_CODE:
        CLPR_epcode = CLPR_pcode;
        CLPR_pcode = savepcode;
}

CLPR_int *CLPR_end_of_constraints(CLPR_int *t)
{
	CLPR_int *prev;
	for (prev = t = CLPR_findbind(t);
		 is_func(t) && mask(*t) == COMMA; prev = t, t = CLPR_findbind(t+2));
	if (is_nil(t)) return prev;
	CLPR_fatal("71341");
}

CLPR_int *CLPR_make_new_rule_head(CLPR_int *t, CLPR_int **vars)
{
	CLPR_int i, n, *new;

	t = CLPR_findbind(t);
	new = CLPR_heaptop;
	make_bound_var(CLPR_heaptop, *t);
	n = arity_of(mask(*t));
	for (i = 1; i <= n; i++)
		make_bound_var(CLPR_heaptop + i, vars[i-1]);
	CLPR_heaptop += n+1;
	return new;
}

CLPR_int CLPR_do_style_check(CLPR_int *t)
{
char *s;
CLPR_int flag, bit;

	t = CLPR_findbind(t);
	if (!is_con(t)) {
		CLPR_warn_err("bad argument to $style_check");
		return FALSE;
	}
	s = string_of(*t);
	if (!strcmp(s, "none")) 
		CLPR_STYLE_CHECK_FLAG = 0;
	else if (!strcmp(s, "all"))
		CLPR_STYLE_CHECK_FLAG = (CONTIG_STYLE | SINGLE_STYLE | OVERLOAD_STYLE);
	else {
		if (!strncmp(s, "no", 2)) {
			flag = FALSE;
			s += 2;
		} else flag = TRUE;
		if (!strcmp(s, "single_var")) bit = SINGLE_STYLE;
		else if (!strcmp(s, "discontiguous")) bit = CONTIG_STYLE;
		else if (!strcmp(s, "name_overload")) bit = OVERLOAD_STYLE;
		else {
			CLPR_warn_err("bad argument to $style_check");
			return FALSE;
		}
		if (flag) CLPR_STYLE_CHECK_FLAG |= bit;
		else CLPR_STYLE_CHECK_FLAG &= (~bit);
	}
	return TRUE;
}

CLPR_int CLPR_do_counter_value(CLPR_int *t1, CLPR_int *t2)
{
	t1 = CLPR_findbind(t1);
	if (!is_con(t1)) {
		CLPR_warn_err("bad first argument to counter_value/2");
		return FALSE;
	}
	return CLPR_make_a_num(t2, CLPR_new_hash_p_node(mask(*t1))->counter);
}

CLPR_int CLPR_do_set_counter(CLPR_int *t1, CLPR_int *t2)
{
double val;
	t1 = CLPR_findbind(t1); t2 = CLPR_findbind(t2);
	if (!is_con(t1)) {
		CLPR_warn_err("bad first argument to set_counter/2");
		return FALSE;
	}
	if (!is_num(t2, &val)) {
		CLPR_warn_err("bad second argument to set_counter/2");
		return FALSE;
	}
	CLPR_new_hash_p_node(mask(*t1))->counter = val;
	return TRUE;
}


CLPR_int CLPR_do_add_counter(CLPR_int *t1, CLPR_int *t2)
{
double val;
	t1 = CLPR_findbind(t1); t2 = CLPR_findbind(t2);
	if (!is_con(t1)) {
		CLPR_warn_err("bad first argument to add_counter/2");
		return FALSE;
	}
	if (!is_num(t2, &val)) {
		CLPR_warn_err("bad second argument to add_counter/2");
		return FALSE;
	}
	CLPR_new_hash_p_node(mask(*t1))->counter += val;
	return TRUE;
}

CLPR_int CLPR_do_new_constant(CLPR_int *t1, CLPR_int *t2)
{
double val;
char *name;
struct HASHNODE *hnode;
char buffer[256], buf1[50];

	t1 = CLPR_findbind(t1);
	if (!is_con(t1)) {
		CLPR_warn_err("bad first argument to new_constant/2");
		return FALSE;
	}
	t2 = CLPR_findbind(t2);
	if (!is_num(t2, &val)) {
		CLPR_warn_err("bad second argument to new_constant/2");
		return FALSE;
	}
	buffer[0] = '#';
	name = string_of(mask(*t1));
	if (strlen(name) > 254) CLPR_fatal("42578");
	strcpy(buffer+1, name);
	hnode = CLPR_insert_func_htable(buffer);
	if (hnode->extra) {
		sprintf(buf1, "%g", val);
		CLPR_warning("constant %s has been changed to %s", buffer, buf1);
		/** return FALSE; **/
	}
	CLPR_setconst(buffer, val);
	return TRUE;
}	

FILE *CLPR_my_open(char *name, char *mode)
{
char *suf,*start,fname[MAX_FNAME_LEN+1];
FILE *fil;
CLPR_int l;

	if ((fil = fopen(name, mode)) != (FILE *) NULL) return fil;
#ifdef BSD
	if (index(mode, 'w') != NULL || index(mode, 'a') != NULL) 
		return (FILE *) NULL;
#else
	if (strchr(mode, 'w') != NULL || strchr(mode, 'a') != NULL) 
		return (FILE *) NULL;
#endif
	suf=CLPR_CLP_SUFFIX;
	while (*suf) {
		strcpy(fname, name);
		if (*suf == ':') suf++;
		for (l=0,start=suf; *suf != '\0' && *suf != ':'; suf++,l++);
		strncat(fname, start, l);
		if ((fil = fopen(fname, mode)) != (FILE *) NULL) return fil;
	}
	return (FILE *) NULL;
}

/* Only for MSDOS or OS2V2*/
char *CLPR_command_interpreter(void)
{
	char *s;
	if (!(s = getenv("COMSPEC"))) s = "COMMAND.COM";
	return s;
}

CLPR_int CLPR_round_int(double f)
{
	CLPR_int i;

	i = f + 0.5;
	if (((double) i) - f == 0.5 && (i & 1)) i--;
	return i;
}
