/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: meta.c                                               *
 ***************************************************************/


#include "standard.h"
#include "config.h"
#include "parse.h"
#include "symbols.h"
#include "lib.h"
#include "compile.h"
#include "opcodes.h"
#include "emul.h"
#include "solver.h"
#include "main.h"

#define has_a_hat(t) ((mask(*t) >= ARITH_FUNC_4) && (mask(*t) <= ARITH_FUNC_5))

extern FUNCTREC CLPR_functors[];

CLPR_int *CLPR_findbind(CLPR_int *x);
CLPR_int *CLPR_make_delay_cell(CLPR_int *t1, CLPR_int *t2);
CLPR_int *CLPR_already_has_eval(CLPR_int *t);
CLPR_int *CLPR_check_eval_var(CLPR_int *t);

extern CLPR_int *CLPR_make_cons(CLPR_int *t1, CLPR_int *t2);
extern NL_EQN_ptr CLPR_solve_eval(CLPR_int *t, CLPR_int *v);
extern CLPR_int *CLPR_curar, *CLPR_stacktop, *CLPR_heaptop, *CLPR_retreg, **CLPR_trail, CLPR_trtop, *CLPR_lstack;
extern CLPR_int CLPR_solver_id;
extern CLPR_int *CLPR_heap, *CLPR_safeheap, *CLPR_lastcp;
extern CLPR_int *CLPR_runable;
extern bool CLPR_have_runable;

/*---------------------------------------------------------------------------*/


CLPR_int CLPR_suspend_eval (CLPR_int *t1, CLPR_int *t2);
extern CLPR_int CLPR_unify (CLPR_int *t1, CLPR_int *t2);
CLPR_int CLPR_setup_eqn (CLPR_int *t1, CLPR_int *t);
void CLPR_make_daemon (CLPR_int *t, CLPR_int *d);
void CLPR_add_runable (CLPR_int *t);
extern void CLPR_push_change_nl_eqn (CLPR_int old_type, CLPR_int pos, NL_EQN_ptr nl_eqn0);
CLPR_int CLPR_set_lhs_arith (CLPR_int *t);
extern void CLPR_init_eqn (void);
extern CLPR_int CLPR_s_evar (double c, CLPR_int v);
extern CLPR_int CLPR_s_econ (double c);
extern CLPR_int CLPR_s_mult (double c, CLPR_int v0, CLPR_int v1, CLPR_int v2);
extern CLPR_int CLPR_s_min (CLPR_int v0, CLPR_int v1, CLPR_int v2);
extern CLPR_int CLPR_s_max (CLPR_int v0, CLPR_int v1, CLPR_int v2);
extern CLPR_int CLPR_s_sin (CLPR_int v0, CLPR_int v1);
extern CLPR_int CLPR_s_cos (CLPR_int v0, CLPR_int v1);
extern CLPR_int CLPR_s_abs (CLPR_int v0, CLPR_int v1);

CLPR_int CLPR_s_eval(CLPR_int *t1, CLPR_int *t2) /*** t1 = eval(t2) ***/
             
{
CLPR_int *t, *t3, *t4, *base, n, i;

	CLPR_deref(t1);
	t1 = CLPR_check_eval_var(t1);
	t = CLPR_check_eval_var(CLPR_findbind(t2));
	if (is_unbound_var(*t) || is_daemon(t)) return CLPR_suspend_eval(t1, t);
	switch ((CLPR_unsigned) CLPR_tag(*t)) {
	case TAG_NUM:
	case CLPR_TAG_NIL:
	case TAG_CON:
	case TAG_PAR: return CLPR_unify(t1, t);
	case TAG_STR:
		if (has_a_hat(t)) return CLPR_setup_eqn(t1, t);
		else {
			base = CLPR_heaptop;
			*base = *t;
			n = arity_of(*t);
			CLPR_heaptop = CLPR_heaptop + n + 1;
			for (i = 1; i <= n; i++) {
				make_unbound_var(base + i);
				if (!CLPR_s_eval(base + i, t + i)) return FALSE;
			}
			return CLPR_unify(t1, base);
		}
	case CLPR_TAG_CONS:
		base = CLPR_heaptop; 
		CLPR_heaptop = CLPR_heaptop + 3;
		*base = addtag(CLPR_TAG_CONS, base + 1);
		make_unbound_var(base + 1);
		make_unbound_var(base + 2);
		if (!CLPR_s_eval(base + 1, head_of(t))) return FALSE;
		if (!CLPR_s_eval(base + 2,  tail_of(t))) return FALSE;
		return CLPR_unify(t1, base);
	case TAG_DAEMON:
		CLPR_fatal("meta_5");
	default:
		printf("%x, tag %x\n", *t, CLPR_tag(*t));
		CLPR_fatal("meta_1");
	}
}

CLPR_int *CLPR_check_eval_var(CLPR_int *t) /*** args to eval must have real locations ***/
       
{
CLPR_int *tmp;
	if (is_heap_ptr(t)) return t;
	tmp = CLPR_heaptop++;
	if (is_unbound_var(*t)) {
		make_bound_var(t, tmp);
		make_unbound_var(tmp);
		push_check_trail(BTK_BIND, t);
	} else make_bound_var(tmp, *t);
	return tmp;
}

CLPR_int CLPR_suspend_eval(CLPR_int *t1, CLPR_int *t2)
{
CLPR_int *v, *d;
CLPR_int result;
	if (v = CLPR_already_has_eval(t2)) { 
		result = CLPR_unify(v, t1); 
		return result;
	}
	d = CLPR_make_delay_cell(t1, t2); /*** suspend t1 = eval(t2) ***/
	CLPR_make_daemon(t2, d);
	return TRUE;
}

void CLPR_make_daemon(CLPR_int *t, CLPR_int *d)
{
CLPR_int *tmp, *base;
	if (*t == TAG_DAEMON) {
CLPR_fatal("think again!");
		tmp = (CLPR_int *) *(t + 1);
		base = CLPR_heaptop;
		CLPR_heaptop = CLPR_heaptop + 2;
		make_bound_var(t, base);
		push_check_trail(BTK_DBIND, t); /*** remove a DAEMON tag ***/
		*base = TAG_DAEMON;
		make_bound_var(base + 1, CLPR_make_cons(d, tmp));
	} else {
		base = CLPR_heaptop;
		CLPR_heaptop = CLPR_heaptop + 2;
		make_bound_var(t, base);
		push_check_trail(BTK_BIND, t); /*** add a (DAEMON) binding ***/
		*base = TAG_DAEMON;
		make_bound_var(base + 1, d);
	}
}


CLPR_int *CLPR_make_delay_cell(CLPR_int *t1, CLPR_int *t2)
{
CLPR_int *base;
	base = CLPR_heaptop;
	CLPR_heaptop = CLPR_heaptop + 2;
	*base = addtag(TAG_MISC, EVALVAL);
	make_bound_var(base + 1, addtag(TAG_MISC, CLPR_solve_eval(t1, t2)));
	return base;
}

CLPR_int *CLPR_already_has_eval(CLPR_int *t)
{
CLPR_int *tmp, *d;
NL_EQN_ptr nl_eqn;
	t = CLPR_findbind(t);
	if (CLPR_tag(*t) != TAG_DAEMON) return NULL;
	for (tmp = CLPR_findbind(t + 1); CLPR_tag(*tmp) == CLPR_TAG_CONS; 
		 tmp = CLPR_findbind(((CLPR_int *)(pmask(*tmp))) + 1)) {
			d = CLPR_findbind((CLPR_int *)pmask(*tmp));
			if (mask(*d) == EVALVAL) {
				nl_eqn = (NL_EQN_ptr) pmask(*(d + 1));
				return (CLPR_int *) nl_eqn->params1->id0;
			}
	}
	if (mask(*tmp) == EVALVAL) {
		nl_eqn = (NL_EQN_ptr) pmask(*(tmp + 1));
		return (CLPR_int *) nl_eqn->params1->id0;
	}
	return NULL;
}

void CLPR_make_runable(CLPR_int *t)
{
CLPR_int *tmp, *d;
	for (tmp = CLPR_findbind(t); CLPR_tag(*tmp) == CLPR_TAG_CONS; 
		 tmp = CLPR_findbind((CLPR_int *)pmask(*tmp) + 1)) {
			d = CLPR_findbind((CLPR_int *)pmask(*tmp));
			CLPR_add_runable(d);
	}
	CLPR_add_runable(tmp);
}

void CLPR_add_runable(CLPR_int *t)
{
CLPR_int *base;
	if (mask(*t) == EVALVAL) {
			base = CLPR_heaptop;
			CLPR_heaptop = CLPR_heaptop + 6;
			*base = addtag(CLPR_TAG_CONS, base + 1);
			make_bound_var(base + 1, base + 3);
			make_bound_var(base + 2, CLPR_runable);
			*(base + 3) = *t;
			*(base + 4) = *(t + 1);
			*(base + 5) = *(t + 2);
			CLPR_runable = base;
			CLPR_have_runable = TRUE;
	}
}

CLPR_int CLPR_wake_runable(void)
{
CLPR_int *tmp, *d, *r, *t1, *t2;
NL_EQN_ptr nl_eqn;
	CLPR_have_runable = FALSE;
	r = CLPR_runable;
	CLPR_runable = (CLPR_int *) addtag(CLPR_TAG_NIL, EMPTYLIST);
	for (tmp = CLPR_findbind(r); CLPR_tag(*tmp) == CLPR_TAG_CONS; 
		 tmp = CLPR_findbind(((CLPR_int *)(pmask(*tmp))) + 1)) {
			d = CLPR_findbind((CLPR_int *)pmask(*tmp));
			if (mask(*d) == EVALVAL) {
				nl_eqn = (NL_EQN_ptr) pmask(*(d + 1));
				t1 = (CLPR_int *) nl_eqn->params1->id0;			
				t2 = (CLPR_int *) nl_eqn->params1->id1;
				CLPR_push_change_nl_eqn(SOLVER_EVAL, 1, nl_eqn);
				nl_eqn->type = WOKEN;
				CLPR_nlin_count--;
				if (!CLPR_s_eval(t1, t2)) return FALSE;
			}
	}
	return TRUE;
}

CLPR_int CLPR_daemons_remaining(void) /*** bug? ***/
{
CLPR_int *h;
	for (h = CLPR_heap; h != CLPR_heaptop; h++) {
		if (CLPR_tag(*h) == TAG_DAEMON) return TRUE;
	}
	return FALSE;
}

CLPR_int CLPR_setup_eqn(CLPR_int *t1, CLPR_int *t)
{
CLPR_int *base, *base1;
	switch (mask(*t)) {
	case HAT_PLUSVAL:
	case HAT_MINUSVAL:
	case HAT_MULTVAL:
	case HAT_DIVVAL:
	case HAT_MINVAL:
	case HAT_MAXVAL:
	case HAT_POWVAL:
		base = CLPR_heaptop;
		base1 = base + 1;
		CLPR_heaptop = CLPR_heaptop + 2;
		make_unbound_var(base);
		make_unbound_var(base1);
		if (!CLPR_s_eval(base, t + 1)) return FALSE;
		if (!CLPR_s_eval(base1, t + 2)) return FALSE;
		t1 = CLPR_findbind(t1);
		if (!CLPR_set_lhs_arith(t1)) return FALSE;
		check_tag_par(base);
		check_tag_par(base1);
		switch (mask(*t)) {
		case HAT_PLUSVAL:
			CLPR_init_eqn();
			CLPR_s_evar(-1.0, mask(*t1));
			CLPR_s_evar(1.0, mask(*base));
			CLPR_s_evar(1.0, mask(*base1));
			return CLPR_s_econ(0.0);
		case HAT_MINUSVAL:
			CLPR_init_eqn();
			CLPR_s_evar(-1.0, mask(*t1));
			CLPR_s_evar(1.0, mask(*base));
			CLPR_s_evar(-1.0, mask(*base1));
			return CLPR_s_econ(0.0);
		case HAT_MULTVAL:
			return CLPR_s_mult(1.0, mask(*t1), mask(*base), mask(*base1));
		case HAT_DIVVAL:
			return CLPR_s_mult(1.0, mask(*base), mask(*t1), mask(*base1));
		case HAT_MINVAL:
			return CLPR_s_min(mask(*t1), mask(*base), mask(*base1));
		case HAT_POWVAL:
			return CLPR_s_max(mask(*t1), mask(*base), mask(*base1));
		default:
			CLPR_fatal("meta_3");
		}
	case HAT_U_MINUSVAL:
	case HAT_SINVAL:
	case HAT_COSVAL:
	case HAT_EVALVAL:
	case HAT_ABSVAL:
		base = CLPR_heaptop++;
		make_unbound_var(base);
		if (!CLPR_s_eval(base, t + 1)) return FALSE;
		switch (mask(*t)) {
		case HAT_U_MINUSVAL:
			t1 = CLPR_findbind(t1);
			if (!CLPR_set_lhs_arith(t1)) return FALSE;
			check_tag_par(base);
			CLPR_init_eqn();
			CLPR_s_evar(1.0, mask(*t1));
			CLPR_s_evar(1.0, mask(*base));
			return CLPR_s_econ(0.0);
		case HAT_SINVAL:
			t1 = CLPR_findbind(t1);
			if (!CLPR_set_lhs_arith(t1)) return FALSE;
			check_tag_par(base);
			return CLPR_s_sin(mask(*t1),mask(*base));
		case HAT_COSVAL:
			t1 = CLPR_findbind(t1);
			if (!CLPR_set_lhs_arith(t1)) return FALSE;
			check_tag_par(base);
			return CLPR_s_cos(mask(*t1),mask(*base));
		case HAT_ABSVAL:
			t1 = CLPR_findbind(t1);
			if (!CLPR_set_lhs_arith(t1)) return FALSE;
			check_tag_par(base);
			return CLPR_s_abs(mask(*t1), mask(*base));
		case HAT_EVALVAL:
			base1 = CLPR_heaptop++;
			make_unbound_var(base1);
			if (!CLPR_s_eval(base1, t + 1)) return FALSE;
			return CLPR_s_eval(t1, base1);
		default:
			CLPR_fatal("meta_4");
		}
	case HAT_QUOTEVAL: return CLPR_unify(t1, t + 1);
	default: CLPR_fatal("meta_2");
	}
	return TRUE;
 FAIL:
	return FALSE; /* type clash */
}

CLPR_int CLPR_set_lhs_arith(CLPR_int *t)
{
	if (is_unbound_var(*t)) { 
		*t = addtag(TAG_PAR, ++CLPR_solver_id); 
		push_check_trail(BTK_BIND, t); 
	} else if (is_daemon(t)) {
		make_bound_var(t, addtag(TAG_PAR, ++CLPR_solver_id));
		push_check_trail(BTK_DBIND, t);
		CLPR_make_runable(t + 1);
	} else if (!is_param(t)) return FALSE;
	return TRUE;
}
