/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: parse.c                                              *
 ***************************************************************/

#include <setjmp.h>

#include "standard.h"
#include "parse.h"
#include "symbols.h"
#include "main.h"
#include "scan.h"
#include "table.h"

#define NULLPT ((PTERM *) 0)

CLPR_int CLPR_debug_flg;

PTERM *CLPR_get_list(void);
PTERM *CLPR_get_nfix(PTERM *arg);
PTERM *CLPR_get_args(struct HASHNODE *hnode, PTERM *func);

static PTERM *make_op_func(struct HASHNODE *hnode, PTERM *t, CLPR_int narg, PTERM **exp);

extern char *CLPR_emalloc(CLPR_int n);
extern char *str_space();

extern char CLPR_tok_c;
extern CLPR_int CLPR_tok_type;
extern PTERM* CLPR_tok_ptr;
extern struct HASHNODE *CLPR_tok_hnode;
extern char CLPR_ch;
extern char *CLPR_cur_filename;
extern FILE *CLPR_cur_stream;
extern FILE *CLPR_outfile;
extern CLPR_int CLPR_debug_flg;
extern jmp_buf CLPR_err_readterm;
extern CLPR_int CLPR_comma_flg, CLPR_bar_flg;
extern PTERM *CLPR_null_list;
extern char *CLPR_cons_str;
extern PTERM *CLPR_new_pterm(void);
extern PTERM *CLPR_narg_pterm(struct HASHNODE *hnode, PTERM *t, CLPR_int narg);
extern PTERM *CLPR_new_pt_emptylist(void);
extern struct HASHNODE CLPR_hash_func_tab[];

/*---------------------------------------------------------------------------*/
/***
	Parse a prolog term with priority n (lower n, higer priority);
	Usually start with CLPR_get_pterm(MAXPRI)
***/

PTERM *CLPR_get_pterm(CLPR_int n)
{
PTERM *t, *exp[2];
CLPR_int oldpri, pri, left, right;
CLPR_int binflg;
CLPR_int old_comma, old_bar;
struct HASHNODE *hnode, *hnode1;

	switch (CLPR_tok_type) {
	case NAME:
		exp[0] = CLPR_tok_ptr;
		hnode = CLPR_tok_hnode;
		CLPR_scan();
		if (CLPR_is_op_hnode(hnode, PREFIX, &pri, &left, &right)) {
			if (CLPR_tok_type == REAL && exp[0]->val == U_MINUSVAL) {
				exp[0] = CLPR_tok_ptr;
				CLPR_tok_ptr->rval = -CLPR_tok_ptr->rval;
				CLPR_scan();
				break;
			}
			if (n >= pri) {
				t = exp[0];
				if (!(exp[0] = CLPR_get_pterm(right))) 
					return(CLPR_narg_pterm(hnode, t, 0));
				else {
					exp[0] = make_op_func(hnode, t, 1, exp);
					break;
				}
			} else return(CLPR_narg_pterm(hnode,exp[0],0));
		} 
		if (CLPR_tok_type == PAREN && CLPR_tok_c == '(')
			exp[0] = CLPR_get_args(hnode, exp[0]);
		else if (CLPR_is_nfix_hnode(hnode))
			exp[0] = CLPR_get_nfix(exp[0]);
		else CLPR_narg_pterm(hnode,exp[0],0);
		break;
	case STR:
	case VAR:
	case REAL:
		exp[0] = CLPR_tok_ptr;
		CLPR_scan();
		break;
	case PAREN:
		if (CLPR_tok_c == '(') {
			old_comma = CLPR_comma_flg;
			CLPR_comma_flg = FALSE;
			CLPR_scan();
			exp[0] = CLPR_get_pterm(MAXPRI);
			CLPR_comma_flg = old_comma;
			if (CLPR_tok_type != PAREN || CLPR_tok_c != ')')
				CLPR_synerr("expected closing ')'");
			else CLPR_scan();
		} else if (CLPR_tok_c == '[') {
			old_bar=CLPR_bar_flg;
			old_comma=CLPR_comma_flg;
			CLPR_comma_flg=TRUE;
			CLPR_bar_flg=TRUE;
			exp[0]=CLPR_get_list();
			CLPR_bar_flg=old_bar;
			CLPR_comma_flg=old_comma;
			if (CLPR_tok_type != PAREN || CLPR_tok_c != ']') 
				CLPR_synerr("expected closing ]");
			CLPR_scan();
		} else return(NULLPT);
		break;
	default:
		return(NULLPT);
	}

	oldpri = 0;
	hnode = CLPR_tok_hnode;
	while (CLPR_tok_type == NAME &&
		((binflg = CLPR_is_op_hnode(hnode,BINARY,&pri,&left,&right))||
		CLPR_is_op_hnode(hnode, POSTFIX, &pri, &left, &right))) {
		if (n < pri) return(exp[0]);
		if (left >= oldpri) {
			t = CLPR_tok_ptr;
			CLPR_scan();
			if (binflg) {
				if (!(exp[1]=CLPR_get_pterm(right)))
					CLPR_synerr("expected right operand");
				exp[0]=make_op_func(hnode,t,2,exp);
			} else exp[0]=make_op_func(hnode,t,1,exp);
			oldpri = pri;
		} else break;
		hnode = CLPR_tok_hnode;
	}
	return(exp[0]);
}

/***
	Make a functor term for operators only,
	f is the functor, narg the no of arguments, 
	exp is an array (PTERM *exp[2]),
	mode is the type of operator (eg. POSTFIX)
***/

static PTERM *make_op_func(struct HASHNODE *hnode, PTERM *t, CLPR_int narg, PTERM **exp)
{
CLPR_int i;

	if (t->ty != FUNC || narg > 2) CLPR_fatal("87540");
	t->first = exp[0];
	if (narg == 2) exp[0]->next = exp[1];
	return CLPR_narg_pterm(hnode, t, narg);
}

PTERM *CLPR_get_args(struct HASHNODE *hnode, PTERM *func)
{
	CLPR_int i=0,n=0;
	PTERM *t,*prev;
	CLPR_int oldcomma;

	oldcomma=CLPR_comma_flg;
	CLPR_comma_flg=TRUE;
	CLPR_scan();
	if (!(t=CLPR_get_pterm(MAXPRI)))
		CLPR_synerr("expected first argument");
	func->first=t;
	for (prev=t,n=1; CLPR_is_punc(','); prev->next=t,prev=t,n++) {
		CLPR_scan();
		if (!(t=CLPR_get_pterm(MAXPRI)))
			CLPR_synerr("expected argument");
	}
	if (CLPR_tok_type != PAREN || CLPR_tok_c != ')')
		CLPR_synerr("expected ')' to close functor");
	CLPR_comma_flg=oldcomma;
	CLPR_scan();
	return(CLPR_narg_pterm(hnode,func,n));
}

/*
 * Parses n-fix terms
 */
PTERM *CLPR_get_nfix(PTERM *arg)
{
	CLPR_fatal("botch CLPR_get_nfix");
}

/*
 * 	Parses list structures, of the form
 *		[X]
 *		[X|Y]
 *		[X,Y] 
 *	[] & X.Y are not done here
 */
PTERM *CLPR_get_list(void)
{
	PTERM *t,*t1,*t2=NULLPT;

	CLPR_scan();
	if ((t1=CLPR_get_pterm(MAXPRI)) == NULL)
		CLPR_synerr("malformed list head");
	if (CLPR_is_punc('|')) {
		CLPR_scan();
		if ((t2=CLPR_get_pterm(MAXPRI)) == NULL)
			CLPR_synerr("malformed list tail");
	} else if (CLPR_is_punc(',')) t2=CLPR_get_list();
	else if (CLPR_tok_type != PAREN || CLPR_tok_c != ']') 
		CLPR_synerr("expected one of {'|',',',']'} in list");
	t = CLPR_new_pterm();
	t->str = CLPR_cons_str;
	t->narg = 2;
	t->val = DOT;
	t->ty = FUNC;
	t->first = t1;
	if (t2) t1->next = t2;
	else t1->next = CLPR_new_pt_emptylist();
	return(t);
}

/*
 *	top level term reader 
 *		returns NULLPT for syntax errors or EOF
 *	terms must always end with a '.'
 */
jmp_buf CLPR_err_readterm;	
PTERM *CLPR_readterm(void)
{
	PTERM *t;
	CLPR_comma_flg = CLPR_bar_flg = FALSE;
	if (setjmp(CLPR_err_readterm))
		return(NULLPT);
	CLPR_zero_vars();
	CLPR_scan();
	if (CLPR_tok_type == ENDFILE) return(NULLPT);
	t=CLPR_get_pterm(MAXPRI);
/*
printf("READ "); fflush(stdout); show_term(t); putchar('\n');
*/
	if (CLPR_is_punc('.'))
		if (t == NULLPT)
			CLPR_synerr("expected a term");
		else	return(t);
	else if (CLPR_tok_type == ENDFILE) {
		CLPR_synerr("unexpected end of file in term");
		return(NULLPT);
	} else {
		CLPR_synerr("unexpected symbol (expected '.')");
		return(NULLPT); /* quiet gcc */
	}
	/*
	 * Oops can't get here
	 */
}

/*
 * 	parses a goal/CLPR_query
 *		returns 0 on EOF
 *		2 on syntax error
 *		1 on success
 */
CLPR_int CLPR_get_goal(PTERM **t)
{
*t=CLPR_readterm();
if (CLPR_tok_type == ENDFILE)
	return 0;
/* GRUNGE */
if (*t == (PTERM *) 0) 
	return 2; 
else return 1;
}

