/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: pio.c                                                *
 ***************************************************************/

#include <ctype.h>
#include <string.h>

#include "standard.h"
#include "config.h"
#include "parse.h"
#include "symbols.h"
#include "lib.h"
#include "compile.h"
#include "emul.h"
#include "main.h"

extern FILE *CLPR_outfile;
extern CLPR_int CLPR_debug_flg;
extern FUNCTREC CLPR_functors[];
extern CLPR_int *CLPR_heap, *CLPR_heaptop, *CLPR_lstack, *CLPR_stacktop;
extern CLPR_int *CLPR_findbind(CLPR_int *x);


void CLPR_show_pterm (PTERM *t);
void CLPR_print_term (PTERM *t, CLPR_int priority);
void CLPR_print_list (PTERM *t);
extern CLPR_int CLPR_get_op_priority (char *name, CLPR_int narg, CLPR_int *left, CLPR_int *right);
void CLPR_print_hterm (FILE *f, CLPR_int *t, CLPR_int precedence);
extern CLPR_int CLPR_is_ground (CLPR_int v, double *c_ptr);
void CLPR_print_var2pos (FILE *f, CLPR_int *v);
void CLPR_print_cons (FILE *f, CLPR_int *t);

void CLPR_display_pterm(FILE *F, PTERM *t)
{
	CLPR_show_pterm(t);
}

void CLPR_show_pterm(PTERM *t)
{
CLPR_int n,i;
PTERM *arg;

	if (!t) {
		fprintf(CLPR_outfile,"NULL\n");return;
	}
	n=t->narg;
	if (t->ty <= 0) {
		fprintf(CLPR_outfile,"term str %s narg %d\n",t->str,t->narg);
		CLPR_fatal("narg prob");
	}
	if (n < 0)
		CLPR_fatal("narg < 0 prob");
	if (t->ty==REAL) {
		if (!is_zero(t->rval)) fprintf(CLPR_outfile,"%g",t->rval);
		else fprintf(CLPR_outfile,"0");
		return;
	} else if (t->ty==STR) {
		fprintf(CLPR_outfile,"\"%s\"",t->str);
		return;
	} else if (t->ty == VAR) {
		fprintf(CLPR_outfile,"%s_%d",t->str,t->val);
		return;
	}
	if (CLPR_debug_flg > 1)
		fprintf(CLPR_outfile,"%s{t %p,%d,%d,%d}",t->str,t,t->ty,t->val,t->narg);
	else if (isalpha(t->str[0]))	
	    fprintf(CLPR_outfile,"%s",t->str);
	else
	    fprintf(CLPR_outfile,"`%s'",t->str);
	if (t->narg > 0) {
		/*** fprintf(CLPR_outfile,"[%d](",t->val); ***/
		fprintf(CLPR_outfile, "("); 
		for (arg=t->first;n > 0;n--) {
			CLPR_show_pterm(arg);
			if (n > 1) fprintf(CLPR_outfile,", ");
			arg=arg->next;
		}
		fprintf(CLPR_outfile,")");
	}
}

void CLPR_print_rule(PTERM *t)
{
	if (!strcmp(t->str,":-")) {
		CLPR_print_term(t->first, MAX_OP_PRECEDENCE+1);
		fprintf(CLPR_outfile,":-\n");
		t = t->first->next;
		goto TAIL;
	} else CLPR_print_term(t, MAX_OP_PRECEDENCE+1);
	fprintf(CLPR_outfile,".\n");
	return;

TAIL:
	if (!strcmp(t->str,",")) {
		fprintf(CLPR_outfile,"\t");
		CLPR_print_term(t->first, MAX_OP_PRECEDENCE+1);
		fprintf(CLPR_outfile,",\n");
		t = t->first->next;
		goto TAIL;
	}
	fprintf(CLPR_outfile,"\t");
	CLPR_print_term(t, MAX_OP_PRECEDENCE+1);
	fprintf(CLPR_outfile,".\n");
}

void CLPR_print_goal(PTERM *t)
{
TAIL:
	if (!strcmp(t->str,",")) {
		CLPR_print_term(t->first, MAX_OP_PRECEDENCE+1);
		fprintf(CLPR_outfile,", ");
		t = t->first->next;
		goto TAIL;
	}
	CLPR_print_term(t, MAX_OP_PRECEDENCE+1);
	fprintf(CLPR_outfile,".\n");
}


void CLPR_print_term(PTERM *t, CLPR_int priority)
{
CLPR_int i,n;
PTERM *arg;
CLPR_int current, left, right;

	switch (t->ty) {
	case REAL:
		if (!is_zero(t->rval)) fprintf(CLPR_outfile,"%g",t->rval);
		else fprintf(CLPR_outfile,"0");
		return;
	case STR:
		fprintf(CLPR_outfile,"\"%s\"",t->str);
		return;
	case VAR:
		fprintf(CLPR_outfile,"%s",t->str);
		return;
	case FUNC:
		if (is_list(t)) {
			CLPR_print_list(t);
			return;
		}
		current = CLPR_get_op_priority(t->str, t->narg, &left, &right);
		switch (CLPR_functors[t->val].special) {
		case PREFIX:
			if (current > priority) fprintf(CLPR_outfile, "(");
			fprintf(CLPR_outfile,"%s ",string_of(t->val));
			CLPR_print_term(t->first, right);
			if (current > priority) fprintf(CLPR_outfile, ")");
			return;
		case BINARY:
			if (current > priority) fprintf(CLPR_outfile, "(");
			CLPR_print_term(t->first, left);
			fprintf(CLPR_outfile," %s ",string_of(t->val));
			CLPR_print_term(t->first->next, right);
			if (current > priority) fprintf(CLPR_outfile, ")");
			return;
		case POSTFIX:
			if (current > priority) fprintf(CLPR_outfile, "(");
			CLPR_print_term(t->first, left);
			fprintf(CLPR_outfile," %s",string_of(t->val));
			if (current > priority) fprintf(CLPR_outfile, ")");
			return;
		case NONE:
			if (!t->narg) {
				fprintf(CLPR_outfile,"%s",string_of(t->val));
				return;
			} else fprintf(CLPR_outfile,"%s(",string_of(t->val));
			for (n=t->narg,arg=t->first; n > 0; n--,arg=arg->next) {
				CLPR_print_term(arg, MAX_OP_PRECEDENCE+1);
				if (n > 1) fprintf(CLPR_outfile,", ");
			}
			fprintf(CLPR_outfile,")");
			return;
		default:
			CLPR_fatal("botch case 1 in CLPR_print_term");
		}
	default:
		CLPR_fatal("botch case 2 in CLPR_print_term");
	}
}

void CLPR_print_list(PTERM *t)
{
	if (is_nullist(t)) {
		fprintf(CLPR_outfile,"[]");
		return;
	}
	fprintf(CLPR_outfile,"[");
MORE:
	CLPR_print_term(t->first, MAX_OP_PRECEDENCE+1);
	if (!is_list(t->first->next)) {
		fprintf(CLPR_outfile," | ");
		CLPR_print_term(t->first->next, MAX_OP_PRECEDENCE+1);
		fprintf(CLPR_outfile,"]");
		return;
	}
	if (!is_nullist(t->first->next)) {
		t=t->first->next;
		fprintf(CLPR_outfile,", ");
		goto MORE;
	}

	fprintf(CLPR_outfile,"]");
}

extern CLPR_int CLPR_slack_id;

extern CLPR_int *CLPR_stack;

void CLPR_printterm(FILE *f, CLPR_int *t)
{
	CLPR_print_hterm(f, t, MAX_OP_PRECEDENCE+1);
}

void CLPR_print_hterm(FILE *f, CLPR_int *t, CLPR_int precedence)
{
double val;
CLPR_int j, v;
char *name;
CLPR_int arity, current, left, right;

	t = CLPR_findbind(t);
	if (is_con(t))	fprintf(f,"%s", CLPR_functors[mask(*t)].pname);	
	else if (is_nil(t))	fprintf(f,"[]");	
	else if (is_param(t)) {
		j = mask(*t);
		if (CLPR_is_ground(j, &val)) {
			if (!is_zero(val)) fprintf(f, "%g", val);
			else fprintf(f, "0");
		} else {
			if (j < CLPR_slack_id) fprintf(f, "_t%d", j);
			else fprintf(f, "_S%d", j);
		}
	} else if (is_unbound_var(*t) || is_daemon(t)) CLPR_print_var2pos(f, t);
	else if (is_cons(t)) {
		fprintf(f,"["); CLPR_print_cons(f, t); fprintf(f,"]");
	} else {
		name = string_of(*t);
		arity = arity_of(*t);
		current = CLPR_get_op_priority(name, arity, &left, &right);
		switch (CLPR_functors[mask(*t)].special)	{
		case PREFIX:
			if (current > precedence) fprintf(f, "(");
			fprintf(f,"%s ", name);
			CLPR_print_hterm(f, t+1, right);
			if (current > precedence) fprintf(f, ")");
			break;
		case POSTFIX:
			if (current > precedence) fprintf(f, "(");
			CLPR_print_hterm(f, t+1, left);
			fprintf(f," %s", name);
			if (current > precedence) fprintf(f, ")");
			break;
		case BINARY:
			if (current > precedence) fprintf(f, "(");
			CLPR_print_hterm(f, t+1, left);
			fprintf(f," %s ", name);
			CLPR_print_hterm(f, t+2, right);
			if (current > precedence) fprintf(f, ")");
			break;
		case NONE:
			fprintf(f,"%s(", name);
			for (j = 1; j <= arity; j++)	{
				if (j != 1) fprintf(f,", ");
				CLPR_printterm(f, t + j);
			}
			fprintf(f,")");
			break;
		  default:
			CLPR_fatal("no such prefix operator (CLPR_printterm)");
		}
fflush(f);
	}
}

void CLPR_print_cons(FILE *f, CLPR_int *t)
{
CLPR_int *tmp;
	CLPR_printterm(f, (CLPR_int *)pmask(*t));
	tmp = (CLPR_int *) pmask(*t)+1;
	tmp = CLPR_findbind(tmp);
	if (is_unbound_var(*tmp) || is_daemon(tmp)) {
		if (!mask(*tmp)) {
			fprintf(f,"| ");
			CLPR_print_var2pos(f, tmp);
			return;
		}
	}
	if (is_nil(tmp)) return;
	if (is_cons(tmp)) {
		fprintf(f, ", ");
		CLPR_print_cons(f, tmp);
		return;
	}
	fprintf(f, " | ");
	CLPR_printterm(f, tmp);
}

void CLPR_print_var2pos(FILE *f, CLPR_int *v)
{
	if (is_heap_ptr(v)) fprintf(f, "_h%d", var2pos(v, CLPR_heap));
	else if (is_stack_ptr(v)) fprintf(f, "_s%d", var2pos(v, CLPR_lstack));
	else {
		printf("***** v = %p ***** stack=%p heaptop=%p stacktop=%p\n", 
			v, CLPR_stack, CLPR_heaptop, CLPR_stacktop);
		fflush(stdout);
		CLPR_fatal("56912");
	}
}
