/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: CLPR_scan.c                                               *
 ***************************************************************/

#include <setjmp.h>
#include <string.h>

#include "standard.h"
#include "config.h"
#include "parse.h"
#include "symbols.h"
#include "table.h"
#include "main.h"
#include "scan.h"

#define MAX_BUF	    (511)          /* max line length */

#define class(bits, no) ((bits << 4)| no)
#define CLASS_NO_MASK   (0xF)

#define CASE 0       
#if CASE
#define mask_class_no(x)		((x) & CLASS_NO_MASK)
#define mask_class_bits(x)		((x) & ~CLASS_NO_MASK)
#else
#define mask_class_no(x)		(x)
#define mask_class_bits(x)		((x) & ~CLASS_NO_MASK)
#endif

#define UC   class(0x1,0)	  /* Upper case alphabetic */
#define UL   class(0x2,1)	  /* Underline */
#define LC   class(0x4,2)	  /* Lower case alphabetic */
#define N    class(0x8,3)	  /* Digit */
#define QT   class(0x10,4)	  /* Single quote */
#define DC   class(0x20,5)    /* Double quote */
#define SY   class(0x40,6)	  /* Symbol character */
#define SL   class(0x80,7)	  /* Solo character */
#define BK   class(0x100,8)	  /* Brackets & friends */
#define BS   class(0x200,9)	  /* Blank space */
#define ENDF class(0x400,10)  /* EOF */

struct {
	FILE *stream;
	char filename[64];
	char line_buf[MAX_BUF];
	CLPR_int	used, lineno, old_len, line_pos;
	char ch;
} CLPR_stream_tab[MAX_STREAM+5];

static CLPR_int cur_stream_no = -1;

char *CLPR_cur_filename;
FILE *CLPR_cur_stream;
char CLPR_ch,CLPR_close_quote;
CLPR_int CLPR_tok_type;
char CLPR_tok_c;
PTERM *CLPR_tok_ptr;

static char buffer[MAX_BUF];
CLPR_int CLPR_COMMENT_NEST = 1;
static char unread_buf[MAX_BUF];
static CLPR_int unread_cnt = 0;
static char line_buf[MAX_BUF];
static CLPR_int old_len = 0, line_pos = 0;
static CLPR_int 	lineno = 1;

CLPR_int CLPR_comma_flg = FALSE;   /* ',' is a functor, not arg separator */
CLPR_int CLPR_bar_flg = FALSE;     /* TRUE in [A | B], otherwise treated as functor */

/*----------------------------------------------------- ascii ---------------*/

CLPR_int CLPR_ascii_classes[] = {

/* nul soh stx etx eot enq ack bel  bs  ht  nl  vt  np  cr  so  si */
    ENDF, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,

/* dle dc1 dc2 dc3 dc4 nak syn etb can  em sub esc  fs  gs  rs  us */
    BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,

/*  sp   !   "   #   $   %   &   '   (   )   *   +   ,   -   .   /  */
    BS, SL, DC,  SL, LC, SL, SY, QT, BK, BK, SY, SY, SL, SY, SY, SY,

/*  0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ? */
    N,  N,  N,  N,  N,  N,  N,  N,  N,  N, SY, SL, SY, SY, SY, SY,

/*  @   A   B   C   D   E   F   G   H   I   J   K   L   M   N   O */
   SY, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC,

/*  P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]   ^   _ */
   UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, BK, SY, BK, SY, UL,

/*  `   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o */
   SY, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC,

/*  p   q   r   s   t   u   v   w   x   y   z   {   |   }   ~  del */
   LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, SY, SL, SY, SY,  BS };

 
/*----------------------------------------------------- ebcdic --------------*/

CLPR_int CLPR_ebcdic_classes[] = {
/* 0 */
/* NUL SOH SIX ETX SEL HT  RNL DEL GE  SPS RPT VT  FF  CR  SO  SI    */
   ENDF, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
/* 16 */
/* DLE DC1 DC2 DC3 RES NL  BS  POC CAN EM  UBS CU1 IFS IGS IRS IUS   */
   BS, BS, BS, BS, BS, BS, BS, BS, BS, ENDF, BS, BS, BS, BS, BS, BS,
/* 32 */
/* DS  SOS FS  WUS BYP LF  ETB ESC RES RES SM  FMT RES ENQ ACK BEL   */
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
/* 48 */
/* RES RES SYN IR  PP  TRN NBS EOT SBS IT  RFF CU3 DC4 NAK RES SUB   */
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
/* 64 */
/* SP  RSP RES RES RES RES RES RES RES RES c   .   <   (   +   |     */
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, SY, SY, SY, BK, SY, SL,
/* 80 */
/* &   RES RES RES RES RES RES RES RES RES !   $   *   )   ;   not   */
   SY, BS, BS, BS, BS, BS, BS, BS, BS, BS, SL, LC, SY, BK, SL, SY,
/* 96 */
/* -   /   NON NON NON NON NON NON NON NON |   ,   %   _   >   ?     */
   SY, SY, BS, BS, BS, BS, BS, BS, BS, BS, SL, SL, SL, UL, SY, SY,
/* 112 */
/* NON NON NON NON NON NON NON NON NON `   :   #   @   '   =   "     */
   BS, BS, BS, BS, BS, BS, BS, BS, BS, SY, SY, SL, SY, QT, SY, DC,
/* 128 */
/* NON a   b   c   d   e   f   g   h   i   NON NON NON NON NON NON   */
   BS, LC, LC, LC, LC, LC, LC, LC, LC, LC, BS, BS, BS, BS, BS, BS,
/* 144 */
/* NON j   k   l   m   n   o   p   q   r   NON NON NON NON NON NON   */
   BS, LC, LC, LC, LC, LC, LC, LC, LC, LC, BS, BS, BS, BS, BS, BS,
/* 160 */
/* NON ~   s   t   u   v   w   x   y   z   NON NON NON [ NON NON   */
   BS, SY, LC, LC, LC, LC, LC, LC, LC, LC, BS, BS, BS, BK, BS, BS,
/* 176 */
/* NON NON NON NON NON NON NON NON NON NON NON NON NON ] NON NON   */
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BK, BS, BS,
/* 192 */	
/* {   A   B   C   D   E   F   G   H   I   NON NON NON NON NON NON   */
   SY, UC, UC, UC, UC, UC, UC, UC, UC, UC, BS, BS, BS, BS, BS, BS,
/* 208 */
/* }   J   K   L   M   N   O   P   Q   R   NON NON NON NON NON NON   */
   SY, UC, UC, UC, UC, UC, UC, UC, UC, UC, BS, BS, BS, BS, BS, BS,
/* 224 */
/* \   NON S   T   U   V   W   X   Y   Z   NON NON NON NON NON NON   */
   SY, UC, UC, UC, UC, UC, UC, UC, UC, UC, BS, BS, BS, BS, BS, BS,
/* 240 */
/* 0   1   2   3   4   5   6   7   8   9   |   NON NON NON NON EO    */
   N , N , N , N , N , N , N , N , N , N , SL, BS, BS, BS, BS, BS };
 
#ifdef EBCDIC
CLPR_int *CLPR_chtype = CLPR_ebcdic_classes;
#else
CLPR_int *CLPR_chtype = CLPR_ascii_classes;
#endif

char CLPR_peek(void);
char CLPR_readchar(void);

extern char *CLPR_emalloc(CLPR_int n);
extern char *str_space();

extern CLPR_int CLPR_name_hash;
extern struct HASHNODE *CLPR_tok_hnode;
extern FILE *CLPR_outfile;

extern CLPR_int CLPR_debug_flg;
extern jmp_buf CLPR_err_readterm;

extern PTERM *CLPR_null_list;
extern char *CLPR_cons_str;

extern struct HASHNODE *CLPR_hnode_emptylist;
extern PTERM *CLPR_null_list;

/*---------------------------------------------------------------------------*/


void CLPR_unreadchar (register char c);

void CLPR_init_scanner(void)
{
CLPR_int i;
	for (i = 0; i < MAX_STREAM; i++) CLPR_stream_tab[i].used = FALSE;
}

CLPR_int CLPR_close_input(FILE *stream)
{
CLPR_int i;
	for (i = 0; i < MAX_STREAM; i++) 
		if (CLPR_stream_tab[i].used && CLPR_stream_tab[i].stream == stream) {
			CLPR_stream_tab[i].used = FALSE;
			return TRUE;
		}
	return FALSE;
}

void CLPR_new_input(FILE *stream, char *name)
{
CLPR_int found, i, free_slot;

	for (found = FALSE, free_slot = -1, i=0; i < MAX_STREAM; i++) {
		if (CLPR_stream_tab[i].used) {
			if ( CLPR_stream_tab[i].stream == stream) {
				found = TRUE;
				break;
			} 
		} else 
			free_slot = i; 
	}
	if (stream != CLPR_cur_stream && cur_stream_no >= 0 && CLPR_stream_tab[cur_stream_no].used) {
		CLPR_stream_tab[cur_stream_no].lineno = lineno;
		CLPR_stream_tab[cur_stream_no].old_len = old_len;
		CLPR_stream_tab[cur_stream_no].line_pos = line_pos;
		CLPR_stream_tab[cur_stream_no].ch = CLPR_ch;
		if (line_pos)
			strncpy(CLPR_stream_tab[cur_stream_no].line_buf, line_buf, line_pos);
		else
			strncpy(CLPR_stream_tab[cur_stream_no].line_buf, line_buf, old_len);
	}
	if (found) {
		cur_stream_no = i;
		CLPR_cur_stream = CLPR_stream_tab[i].stream;
		CLPR_cur_filename = CLPR_stream_tab[i].filename;
		lineno = CLPR_stream_tab[i].lineno;
		old_len = CLPR_stream_tab[i].old_len;
		line_pos = CLPR_stream_tab[i].line_pos;
		CLPR_ch = CLPR_stream_tab[i].ch;
		if (line_pos) strncpy(line_buf, CLPR_stream_tab[i].line_buf, old_len);
		else strncpy(line_buf, CLPR_stream_tab[i].line_buf, old_len);
	} else {
		if (free_slot == -1) CLPR_fatal("no more i/o streams");
		CLPR_stream_tab[free_slot].stream = CLPR_cur_stream = stream;
		strncpy(CLPR_stream_tab[free_slot].filename, CLPR_cur_filename = name, 63);
		CLPR_stream_tab[free_slot].used = TRUE;
		CLPR_stream_tab[free_slot].ch = ' ';
		cur_stream_no = free_slot;
		old_len = 0;
		line_pos = 0;
		lineno = 1;
		CLPR_unreadchar(' ');
		CLPR_readchar();
	}
}

char CLPR_readchar(void) /*** returns '\0' on EOF ***/
{
register CLPR_int c;
	if (unread_cnt > 0) c = unread_buf[--unread_cnt];
	else if ((c = getc(CLPR_cur_stream)) == EOF)  c = '\0';
	if (c == '\n') {
		lineno++;
		old_len = line_pos;
		line_pos = 0;
	} else 
	if (line_pos < MAX_BUF) line_buf[line_pos++] = c;
	else {
		old_len = line_pos;
		line_pos = 0;
		line_buf[line_pos++] = c;
	}
	CLPR_ch = c;
	return c;
}

void CLPR_unreadchar(register char c) /*** NULL for EOF ***/
                
{
	if (c != '\0') {
		if (unread_cnt >= MAX_BUF) CLPR_fatal("too many pushbacks");
		unread_buf[unread_cnt++]=c;
		if (c == '\n') lineno--;
		if (line_pos) line_pos--;
	}
}

char CLPR_peek(void) /*** just look, not touch ***/
{
register char c1, c2;
	c1 = CLPR_ch;
	c2 = CLPR_readchar();
	CLPR_unreadchar(c2);
	CLPR_ch = c1;
	return c2;
}

/*** stores a NAME in s, NAME is either variable or symbolic ***/
CLPR_int CLPR_get_name(register char *s)
{
register char c;
register CLPR_int hash_sum;
	hash_sum = 0;
	c = CLPR_ch;
	do {
		*s++ = c;
		hash_sum += c; 
		c = CLPR_readchar();
	} while (CLPR_chtype[c] & mask_class_bits(UC | UL | LC | N));
	*s = '\0';
	return hash_sum;
}

/************************************************************
 *	top level call in scanner; returns CLPR_tok_type (eg. VAR);
 *	CLPR_tok_ptr contains a pointer to the PTERM constructed;
 *	CLPR_tok_c contains the character scanned for those characters 
 *	    which are single (SL);
 ************************************************************/

CLPR_int CLPR_scan(void)
{
register char c;
char *s;
double tens, intpart, fracpart, num;
CLPR_int sign = 1, esign, expart;
CLPR_int level;
CLPR_int hash_sum;

START:
	c = CLPR_ch;
	while (CLPR_chtype[c] == BS) c = CLPR_readchar();

	/********************************/
	switch(mask_class_no(CLPR_chtype[c])) {
	/********************************/

	case mask_class_no(UC):
	case mask_class_no(UL):
		CLPR_tok_ptr = CLPR_lookup_var(buffer, CLPR_get_name(buffer));
		return(CLPR_tok_type = VAR);
	case mask_class_no(LC):
		CLPR_tok_ptr = CLPR_lookup_name(buffer, CLPR_get_name(buffer)); 
		return(CLPR_tok_type = NAME);
	case mask_class_no(SY):
		s = buffer;
		if (CLPR_ch == '.' &&
				(CLPR_chtype[CLPR_peek()] == BS || CLPR_chtype[CLPR_peek()] == ENDF)) {
			CLPR_readchar();
			CLPR_tok_c = '.';
			return(CLPR_tok_type = PUNC);
		} else if (c == '/' && CLPR_peek() == '*') {
			level = 0;
		START_COMMENT:
				CLPR_readchar();
				if (CLPR_COMMENT_NEST) level++;
			WITHIN_COMMENT:
				if (CLPR_COMMENT_NEST)
					while ((c = CLPR_readchar()) != '/' && 
						c != '*' && CLPR_chtype[c] != ENDF);
				else
					while ((c = CLPR_readchar()) != '*' && 
						CLPR_chtype[c] != ENDF);
				
				if (CLPR_ch == '*' && CLPR_peek() == '/') {
					CLPR_readchar();
					CLPR_readchar();
					if (!CLPR_COMMENT_NEST || --level == 0) goto START;
				} else if (CLPR_COMMENT_NEST && CLPR_ch == '/' && CLPR_peek() == '*')
					goto START_COMMENT;
				else if (CLPR_chtype[CLPR_ch] == ENDF)
					CLPR_fatal("unexpected end of file in comment");
				goto WITHIN_COMMENT;
		} 
		c = CLPR_ch;
		hash_sum = 0;
		do {
			if (c == '/') {
				if (CLPR_peek() == '*') {
					/* BUG: oops
					CLPR_unreadchar('/');
					*/
					break;
				} 
			}
			*s++ =c;
			hash_sum += c;
		} while (CLPR_chtype[(c=CLPR_readchar())] == SY);
		*s = '\0';
		CLPR_tok_ptr = CLPR_lookup_name(buffer, hash_sum); 
		CLPR_tok_type = NAME;
		return(CLPR_tok_type = NAME);
	case mask_class_no(N):
		intpart = fracpart = 0;
		expart = 0; esign = 1;
		c = CLPR_ch;
		do {
			intpart = intpart*10+(c-'0');
		} while (CLPR_chtype[(c = CLPR_readchar())] == N);
		if (c == '.' && CLPR_chtype[CLPR_peek()] == N) {
			tens = 0.1;	
			c = CLPR_readchar();
			do {
				fracpart += tens*(c-'0');
				tens *= 0.1;
			} while (CLPR_chtype[(c = CLPR_readchar())] == N);
		}
		if (c == 'e') {
			if (CLPR_readchar() == '-') {
				esign = -1;
				CLPR_readchar();
			} else {
				if (CLPR_ch == '+') CLPR_readchar();	
				esign = 1;
			}
			if (CLPR_chtype[CLPR_ch] == N) 
				do 
					expart = expart*10+(CLPR_ch-'0');
				while (CLPR_chtype[CLPR_readchar()] == N);
		}
		num = sign*(intpart+fracpart)*pow(10.0,(double) esign*expart);
		CLPR_tok_ptr = CLPR_new_real(num);
		return(CLPR_tok_type = REAL);
	case mask_class_no(QT):
	case mask_class_no(DC):
		c = CLPR_close_quote = CLPR_ch;
		hash_sum = 0;
		for (s = buffer; (c = CLPR_readchar()) != CLPR_close_quote && CLPR_chtype[c] != ENDF;
			 	*s++ = c, hash_sum += c);
		*s='\0';
		CLPR_readchar();
		if (CLPR_chtype[c] == ENDF) 
			CLPR_synerr("unexpected end of file within quotes");
/*
		else if (CLPR_close_quote == '"') {
			CLPR_tok_ptr = CLPR_new_str(buffer);
			return(CLPR_tok_type = STR);
		} else	{
*/
		else { /* strings disabled - converted to literals */
			CLPR_tok_ptr=CLPR_lookup_name(buffer, hash_sum);
			return(CLPR_tok_type = NAME);
		}
	case mask_class_no(SL):
		if (c == '%') {
			while ((c=CLPR_readchar()) != '\n' && (CLPR_chtype[c] != ENDF));
			if (c == '\n') {
				CLPR_readchar();
				goto START;
			} else return(ENDFILE);
		} else if (CLPR_ch == '#' && CLPR_chtype[CLPR_peek()] == LC) {
			CLPR_tok_ptr=CLPR_lookup_const(buffer, CLPR_get_name(buffer));
			if (!CLPR_tok_ptr) CLPR_synerr("unknown special constant");
			return(CLPR_tok_type = REAL);
		}
		CLPR_readchar();
		CLPR_tok_c = c;
		if (c == ',' && CLPR_comma_flg) return(CLPR_tok_type=PUNC);
		if (c == '|' && CLPR_bar_flg) return(CLPR_tok_type=PUNC);
		buffer[0]=c;
		buffer[1]='\0';
		CLPR_tok_ptr=CLPR_lookup_name(buffer, (CLPR_int) c);
		return(CLPR_tok_type=NAME);
	case mask_class_no(BK):
		CLPR_tok_c = c;
		c = CLPR_readchar();
		if (CLPR_tok_c == '[' && c == ']') {
			CLPR_readchar();
			CLPR_tok_ptr = CLPR_new_pt_emptylist();
			CLPR_tok_hnode = CLPR_hnode_emptylist;
			return(CLPR_tok_type = NAME);
		}
		return(CLPR_tok_type = PAREN);
	case mask_class_no(ENDF):
		return(CLPR_tok_type = ENDFILE);
	default:
		CLPR_synerr("illegal character");
		return 0; /* quiet gcc */
	}
}

CLPR_int CLPR_myisdigit(char c)
{
	return(CLPR_chtype[c] == N);
}

CLPR_int CLPR_is_punc(char punc)
{
	if (CLPR_tok_type == PUNC && CLPR_tok_c == punc)
		return(TRUE);
	else	return(FALSE);
}

void CLPR_synerr(char *s) /*** if error, jump to CLPR_readterm in parser ***/
        
{
CLPR_int i;
	if (line_pos == 0 && old_len > 0) {
		printf("Syntax error, %s:%d : %s\n", CLPR_cur_filename, lineno - 1, s);
		for (i = 0; i < old_len; i++) putchar(line_buf[i]);
	} else {
		printf("Syntax error, %s:%d : %s\n",CLPR_cur_filename,lineno,s);
		for (i = 0; i < line_pos; i++) putchar(line_buf[i]);
	}
	printf("<--\n");
	while (!CLPR_is_punc('.') && CLPR_tok_type != ENDFILE) CLPR_scan();
	longjmp(CLPR_err_readterm,TRUE);
}

CLPR_int CLPR_sum_str(register char *s)
{
register CLPR_int c;
	c = 0;
	while (*s) c+= *s++;
	return c;
}

void CLPR_readline(char *s, CLPR_int n)
{
CLPR_int i;

	n--;
	CLPR_readchar();
	for (i=0; CLPR_ch != '\n' && i < n; i++) {
		*s++ = CLPR_ch;
		CLPR_readchar();
	}
	if (i == n) {
		while (CLPR_ch != '\n' && CLPR_ch)
			CLPR_readchar();
	}
	*s = '\0';
}

CLPR_int CLPR_the_lineno(void)
{
	if (CLPR_ch == '\n') return lineno-1;
	else return lineno;
}

char *CLPR_the_filename(void)
{
	return CLPR_cur_filename;
}
