/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: solver.c                                             *
 ***************************************************************/

#include <stdlib.h>
#include "standard.h"
#include "config.h"
#include "emul.h"
#include "solver.h"
#include "opcodes.h"
#include "main.h"

#define MIN_SOLVER_SZ 256

#define check_solver_id(i) { \
	if (i >= CLPR_slack_id) CLPR_fatal("too many solver variables"); \
}

EQN_ptr CLPR_eqn_scratch;
static CLPR_int new_solver_var;
static double new_solver_coef;

/*-------------------------------------------------- solver globals ---------*/

TAB_HDR_NODE_ptr *CLPR_tableau;
CLPR_int CLPR_SOLVER_SZ = DEF_SOLVER_SZ;

SOLVER_NODE_ptr CLPR_sms;
CLPR_int CLPR_sms_blocks;
CLPR_int CLPR_solver_id, CLPR_slack_id;
CLPR_int CLPR_nlin_count;
CLPR_int CLPR_s_count[MAX_SOLVER_COUNTER];
CLPR_int 
	CLPR_malloc_eqn_count, CLPR_malloc_eqn_par_count, 
	CLPR_malloc_entry_count, CLPR_malloc_occ_count, CLPR_malloc_nl_eqn_count, 
	CLPR_malloc_nl_occ_count, CLPR_malloc_nl_par1_count, CLPR_malloc_nl_par2_count, 
	CLPR_malloc_hdr_count, CLPR_malloc_koh_count, CLPR_malloc_list_count, CLPR_free_eqn_count,
	CLPR_free_eqn_par_count, CLPR_free_entry_count, CLPR_free_occ_count, CLPR_free_nl_eqn_count, 
	CLPR_free_nl_occ_count, CLPR_free_nl_par1_count, CLPR_free_nl_par2_count, 
	CLPR_free_hdr_count, CLPR_free_koh_count, CLPR_free_list_count;

extern NL_EQN_ptr CLPR_nl_eqn_top;
extern FILE *CLPR_outfile, *CLPR_error_stream;
extern CLPR_int CLPR_savep, CLPR_stamp;

/*----------------------------------------------- entry points --------------*/


CLPR_int CLPR_solve_eqn_1 (CLPR_int v, double c);
extern CLPR_int CLPR_solve_parametric_eqn (EQN_ptr eqn, CLPR_int call_ineq_subs);
extern void CLPR_backup_null_eqn (CLPR_int id);
extern void CLPR_insert_occ (CLPR_int type, CLPR_int i, CLPR_int v);
extern CLPR_int CLPR_wakeup (CLPR_int v, double c);
extern void CLPR_add_entry (CLPR_int type, EQN_ptr eqn, CLPR_int v, double c);
CLPR_int CLPR_greater0 (CLPR_int strict, double c);
CLPR_int CLPR_free_eqn_nodes (EQN_ptr eqn);
CLPR_int CLPR_is_ground (CLPR_int v, double *c_ptr);
CLPR_int CLPR_solve_eqn_2 (CLPR_int v0, CLPR_int v1, double c);
extern CLPR_int CLPR_solve_mult (CLPR_int v0, CLPR_int v1, CLPR_int v2, double c);
extern CLPR_int CLPR_solve_minmax (CLPR_int type, CLPR_int v0, CLPR_int v1, CLPR_int v2, double c);
extern CLPR_int CLPR_solve_power (CLPR_int type, CLPR_int v0, CLPR_int v1, CLPR_int v2, double c);
extern CLPR_int CLPR_solve_absolute (CLPR_int type, CLPR_int v0, CLPR_int v1);
extern CLPR_int CLPR_solve_sin (CLPR_int type, CLPR_int v0, CLPR_int v1);
extern CLPR_int CLPR_solve_cos (CLPR_int type, CLPR_int v0, CLPR_int v1);
extern void CLPR_warn_err (char *s, ...);
extern CLPR_int CLPR_solve_arcsin (CLPR_int v0, CLPR_int v1);
extern CLPR_int CLPR_solve_arccos (CLPR_int v0, CLPR_int v1);
void CLPR_malloc_warning (CLPR_int n, char *s);
CLPR_int CLPR_deref_param (CLPR_int v);

void CLPR_init_eqn(void)
{
	malloc_EQN(CLPR_eqn_scratch);
	CLPR_eqn_scratch->next = NULL;
	CLPR_eqn_scratch->time_stamp = CLPR_stamp;
	CLPR_eqn_scratch->id = 0;
	CLPR_eqn_scratch->constant = 0.0;
	CLPR_eqn_scratch->params = NULL;
	new_solver_var = 0;
	new_solver_coef = 0.0;
}

CLPR_int CLPR_makenum(CLPR_int v, double c) /*** v = c, v is brand new ***/
      
         
{
	check_solver_id(v);
	if (!CLPR_solve_eqn_1(v, c)) CLPR_fatal("19753");
	return TRUE;
}

CLPR_int CLPR_s_econ(double c) /*** CLPR_eqn_scratch = c ***/
         
{
register ENTRY_ptr entry;
TAB_HDR_NODE_ptr hdr;
double factor;
CLPR_int v, flag;

	CLPR_eqn_scratch->constant += c;
	if (is_zero(new_solver_coef)) {
		if (entry = CLPR_eqn_scratch->next) {
			flag = CLPR_solve_parametric_eqn(CLPR_eqn_scratch, TRUE);
			return flag;
		}
		flag = is_zero(CLPR_eqn_scratch->constant);
		free_EQN(CLPR_eqn_scratch);
		return flag;
	} 
	stats_count(4);
	v = CLPR_eqn_scratch->id = new_solver_var;
	malloc_HDR(hdr);
	CLPR_tableau[v] = hdr;
	CLPR_backup_null_eqn(v);
	entry = CLPR_eqn_scratch->next;
	hdr->eqn = CLPR_eqn_scratch;
	factor = -1.0 / new_solver_coef;
	CLPR_eqn_scratch->constant *= factor;
	while (entry) {
		entry->coef *= factor;
		CLPR_insert_occ(EQ, v, entry->id);
		entry = entry->next;
	}
	if (CLPR_eqn_scratch->next) return TRUE;
	if (!CLPR_wakeup(v, CLPR_eqn_scratch->constant)) {
		free_EQN(CLPR_eqn_scratch);
		free_HDR(hdr);
		CLPR_tableau[v] = NULL;
		return FALSE;
	}
	return TRUE;
}				

CLPR_int CLPR_s_evar(double c, CLPR_int v) /*** CLPR_eqn_scratch + c*v ***/
         
      
{
	check_solver_id(v);
	if (is_zero(new_solver_coef)) {
		if (!CLPR_tableau[v]) {
			new_solver_var = v;
			new_solver_coef = c;
		} else CLPR_add_entry(EQ, CLPR_eqn_scratch, v, c);
	} else {
		if (v != new_solver_var) CLPR_add_entry(EQ, CLPR_eqn_scratch, v, c);
		else new_solver_coef += c;
	}
	return TRUE;
}

CLPR_int CLPR_s_ge0(double c)
{
	return CLPR_greater0(FALSE, c);
}

CLPR_int CLPR_s_gt0(double c)
{
	return CLPR_greater0(TRUE, c);
}

CLPR_int CLPR_greater0(CLPR_int strict, double c)
{
double c2;
EQN_PARAMS_ptr params;
CLPR_int flag;
	malloc_EQN_PARAMS(params);
	stats_count(125);
	CLPR_eqn_scratch->params = params;
	params->basic = 0;
	params->epsilon = (strict ? -1.0 : 0.0);
	if (!is_zero(new_solver_coef)) 
		CLPR_add_entry(INEQ, CLPR_eqn_scratch, new_solver_var, new_solver_coef); 
	if (!CLPR_eqn_scratch->next) {
		stats_count(135);
		CLPR_free_eqn_nodes(CLPR_eqn_scratch);
		c2 = CLPR_eqn_scratch->constant;
		if (strict) return (c2 > c + CLPR_EPS);
		else return (c2 + CLPR_EPS >= c);
	} else {
		CLPR_eqn_scratch->constant -= c;
		if (CLPR_solve_inequality(strict, CLPR_eqn_scratch))
			return CLPR_check_implicits();
		else return FALSE;
	}
}

CLPR_int CLPR_s_mult(double c, CLPR_int v0, CLPR_int v1, CLPR_int v2) /*** v0 = c * v1 * v2 ***/
         
               
{
CLPR_int foo;
double c1, c2;
	check_solver_id(v0);
	check_solver_id(v1);
	check_solver_id(v2);
	if (CLPR_is_ground(v1, &c1)) {
		if (CLPR_is_ground(v2, &c2)) return CLPR_solve_eqn_1(v0, c * c1 * c2); 
		else return CLPR_solve_eqn_2(v0, v2, c * c1);
	}
	if (CLPR_is_ground(v2, &c2)) return CLPR_solve_eqn_2(v0, v1, c * c2);
	else return CLPR_solve_mult(v0, v1, v2, c);
}

CLPR_int CLPR_s_max(CLPR_int v0, CLPR_int v1, CLPR_int v2)
{
double c1, c2;
	check_solver_id(v0);
	check_solver_id(v1);
	check_solver_id(v2);
	if (CLPR_is_ground(v1, &c1)) {
		if (CLPR_is_ground(v2, &c2)) return CLPR_solve_eqn_1(v0, (c1 > c2 ? c1 : c2));
		return CLPR_solve_minmax(SOLVER_MAX1, v0, v1, v2, c1);
	}
	if (CLPR_is_ground(v2, &c2)) return CLPR_solve_minmax(SOLVER_MAX2, v0, v1, v2, c2);
	return CLPR_solve_minmax(SOLVER_MAX, v0, v1, v2, 0.0);
}
 
CLPR_int CLPR_s_min(CLPR_int v0, CLPR_int v1, CLPR_int v2)
{
double c1, c2;
	check_solver_id(v0);
	check_solver_id(v1);
	check_solver_id(v2);
	if (CLPR_is_ground(v1, &c1)) {
		if (CLPR_is_ground(v2, &c2)) return CLPR_solve_eqn_1(v0, (c1 < c2 ? c1 : c2));
		return CLPR_solve_minmax(SOLVER_MIN1, v0, v1, v2, c1);
	}
	if (CLPR_is_ground(v2, &c2)) return CLPR_solve_minmax(SOLVER_MIN2, v0, v1, v2, c2);
	return CLPR_solve_minmax(SOLVER_MIN, v0, v1, v2, 0.0);
}

CLPR_int CLPR_s_pow(CLPR_int v0, CLPR_int v1, CLPR_int v2) /*** v0 = pow(v1, v2) ***/
               
{
double c0, c1, c2;
	check_solver_id(v0); 
	check_solver_id(v1); 
	check_solver_id(v2);
	if (CLPR_is_ground(v0, &c0)) {
/*
		if (CLPR_is_ground(v2, &c2)) { WRONG
			if (is_zero(c2)) return CLPR_solve_eqn_1(v0, 1.0);
			if (is_equal(c2, 1.0)) return CLPR_solve_eqn_1(v1, c0);
			return CLPR_solve_eqn_1(v1, pow(c0, 1.0 / c2));
		}
*/
		if (CLPR_is_ground(v1, &c1)) return CLPR_solve_eqn_1(v2, log(c0) / log(c1));
		if (CLPR_is_ground(v2, &c2)) {
			if (is_equal(c2, 1.0)) return CLPR_solve_eqn_1(v1, c0);
			else if (is_equal(c2, 0.0)) return CLPR_solve_eqn_1(v1, 1.0);
			return CLPR_solve_power(SOLVER_POW02, v0, v1, v2, c0);
		}
		return CLPR_solve_power(SOLVER_POW0, v0, v1, v2, c0);
	}
	if (CLPR_is_ground(v1, &c1)) {
		if (CLPR_is_ground(v2, &c2)) return CLPR_solve_eqn_1(v0, pow(c1, c2));
	/*	if (is_zero(c1)) return CLPR_solve_eqn_1(v0, 0.0);WRONG */
		if (is_equal(c1, 1.0)) return CLPR_solve_eqn_1(v0, 1.0);
		return CLPR_solve_power(SOLVER_POW1, v0, v1, v2, c1);
	}
	if (CLPR_is_ground(v2, &c2)) {
		if (is_zero(c2)) return CLPR_solve_eqn_1(v0, 1.0);
		if (is_equal(c2, 1.0)) return CLPR_solve_eqn_2(v0, v1, 1.0);
		return CLPR_solve_power(SOLVER_POW2, v0, v1, v2, c2);
	}
	return CLPR_solve_power(SOLVER_POW, v0, v1, v2, 0.0);
}

CLPR_int CLPR_s_abs(CLPR_int v0, CLPR_int v1)
{
double c;
	/*** printf(">>> X%d = abs(X%d)\n", v0, v1); ***/
	check_solver_id(v0);
	check_solver_id(v1);
	if (CLPR_is_ground(v1, &c)) return CLPR_solve_eqn_1(v0, fabs(c));
	if (CLPR_is_ground(v0, &c)) {
		if (is_zero(c)) return CLPR_solve_eqn_1(v1, 0.0);
		else return CLPR_solve_absolute(SOLVER_ABS0, v0, v1);
	}
	return CLPR_solve_absolute(SOLVER_ABS, v0, v1);
}

CLPR_int CLPR_s_sin(CLPR_int v0, CLPR_int v1)
{
double c;
	check_solver_id(v0);
	check_solver_id(v1);
	if (CLPR_is_ground(v1, &c)) {
		if (c != 0.0) return CLPR_solve_eqn_1(v0, sin(c));
		else return CLPR_solve_eqn_1(v0, 0.0);
	}
	if (CLPR_is_ground(v0, &c)) return CLPR_solve_sin(SOLVER_SIN0, v0, v1);
	else return CLPR_solve_sin(SOLVER_SIN, v0, v1);
}

CLPR_int CLPR_s_cos(CLPR_int v0, CLPR_int v1)
{
double c;
	check_solver_id(v0);
	check_solver_id(v1);
	if (CLPR_is_ground(v1, &c)) {
		if (c != 0.0) return CLPR_solve_eqn_1(v0, cos(c));
		else return CLPR_solve_eqn_1(v0, 1.0);
	}
	if (CLPR_is_ground(v0, &c)) return CLPR_solve_cos(SOLVER_COS0, v0, v1);
	else return CLPR_solve_cos(SOLVER_COS, v0, v1);
}

CLPR_int CLPR_s_arcsin(CLPR_int v0, CLPR_int v1)
{
double c;
	check_solver_id(v0);
	check_solver_id(v1);
	if (CLPR_is_ground(v1, &c)) {
		if (c >= -1 && c <= 1) {
			if (c != 0.0) return CLPR_solve_eqn_1(v0, asin(c));
			else return CLPR_solve_eqn_1(v0, 0.0);
		} else if (c >= 1 && c <= 1+CLPR_EPS) return CLPR_solve_eqn_1(v0, PI_2_VAL);
		else if (c <= -1 && c >= -1-CLPR_EPS) return CLPR_solve_eqn_1(v0, -PI_2_VAL);
		else {
			CLPR_warn_err("Out of range value outside [-1,1] to arcsin");
			return FALSE;
		}
	}
	if (CLPR_is_ground(v0, &c)) {
		if (c >= -PI_2_VAL && c <= PI_2_VAL) {
			if (c != 0.0) return CLPR_solve_eqn_1(v1, sin(c));
			else return CLPR_solve_eqn_1(v1, 0.0);
		} else if (c > 0 && c <= PI_2_VAL+CLPR_EPS) return CLPR_solve_eqn_1(v1, 1.0);
		else if (c < 0 && c >= -PI_2_VAL-CLPR_EPS) return CLPR_solve_eqn_1(v1, -1.0);
		else {
			CLPR_warn_err("Out of range value outside [-pi/2, pi/2] to arcsin");
			return FALSE;
		}
	}
	return CLPR_solve_arcsin(v0, v1);
}

CLPR_int CLPR_s_arccos(CLPR_int v0, CLPR_int v1)
{
double c;
	check_solver_id(v0);
	check_solver_id(v1);
	if (CLPR_is_ground(v1, &c)) {
		if (c >= -1 && c <= 1) {
			if (c != 1.0) return CLPR_solve_eqn_1(v0, acos(c));
			else return CLPR_solve_eqn_1(v0, 0.0);
		} else if (c >= 1 && c <= 1+CLPR_EPS) return CLPR_solve_eqn_1(v0, 0.0);
		else if (c <= -1 && c >= -1-CLPR_EPS) return CLPR_solve_eqn_1(v0, PI_VAL);
		else {
			CLPR_warn_err("Out of range value outside [-1, 1] to arccos");
			return FALSE;
		}
	}
	if (CLPR_is_ground(v0, &c)) {
		if (c >= 0 && c <= PI_VAL) {
			if (c != 0.0) return CLPR_solve_eqn_1(v1, cos(c));
			else return CLPR_solve_eqn_1(v1, 1.0);
		} else if (c > 0 && c <= PI_VAL+CLPR_EPS) return CLPR_solve_eqn_1(v1, -1.0);
		else if (c < 0 && c >= -CLPR_EPS) return CLPR_solve_eqn_1(v1, 1.0);
		else {
			CLPR_warn_err("Out of range value outside [0, pi] to arccos");
			return FALSE;
		}
	}
	return CLPR_solve_arccos(v0, v1);
}

/*---------------------------------------------------------------------------*/

CLPR_int CLPR_solve_eqn_1(CLPR_int v, double c) /*** v = c ***/
       
         
{
TAB_HDR_NODE_ptr hdr;
EQN_ptr eqn;
CLPR_int flag;
	check_solver_id(v);
	malloc_EQN(eqn);
	eqn->next = NULL;
	eqn->id = 0;
	eqn->time_stamp = CLPR_stamp;
	eqn->params = NULL;
	eqn->constant = c;
	if (!(hdr = CLPR_tableau[v])) {
		stats_count(4);
		malloc_HDR(hdr);
		CLPR_tableau[eqn->id = v] = hdr;
		CLPR_backup_null_eqn(v);
		hdr->eqn = eqn;
		if (!CLPR_wakeup(v, c)) {
			free_EQN(eqn);
			free_HDR(hdr);
			CLPR_tableau[v] = NULL;
			return FALSE;
		}
		return TRUE;
	}
	eqn->id = 0;
	CLPR_add_entry(EQ, eqn, v, -1.0);
	if (eqn->next) return CLPR_solve_parametric_eqn(eqn, TRUE);
	flag = is_zero(eqn->constant);
	free_EQN(eqn);
	return flag;
}

CLPR_int CLPR_solve_eqn_2(CLPR_int v0, CLPR_int v1, double c) /*** v0 = c * v1 ***/
            
         
{
TAB_HDR_NODE_ptr hdr;
EQN_ptr eqn;
double c2;
CLPR_int flag;
	if (is_zero(c)) return CLPR_solve_eqn_1(v0, 0.0);
	if (v0 == v1) 
		return (is_equal(c, 1.0) ? TRUE : CLPR_solve_eqn_1(v0, 0.0)); 
	check_solver_id(v0);
	check_solver_id(v1);
	malloc_EQN(eqn);
	eqn->next = NULL;
	eqn->id = 0;
	eqn->time_stamp = CLPR_stamp;
	eqn->params = NULL;
	eqn->constant = 0.0;
	if (!(hdr = CLPR_tableau[v0])) {
		stats_count(4);
		malloc_HDR(hdr);
		CLPR_tableau[eqn->id = v0] = hdr;
		CLPR_backup_null_eqn(v0);
		hdr->eqn = eqn;
		CLPR_add_entry(EQ, eqn, v1, c);
		if (eqn->next) return TRUE;
		if (!CLPR_wakeup(v0, eqn->constant)) {
			free_EQN(eqn);
			free_HDR(hdr);
			CLPR_tableau[v0] = NULL;
			return FALSE;
		}
		return TRUE;
	}
	if (!(hdr = CLPR_tableau[v1])) { /*** v1 = (1/c) * v0 ***/
		stats_count(4);
		malloc_HDR(hdr);
		CLPR_tableau[eqn->id = v1] = hdr;
		CLPR_backup_null_eqn(v1);
		hdr->eqn = eqn;
		CLPR_add_entry(EQ, eqn, v0, 1.0 / c);
		if (eqn->next) return TRUE;
		if (!CLPR_wakeup(v1, eqn->constant)) {
			free_EQN(eqn);
			free_HDR(hdr);
			CLPR_tableau[v1] = NULL;
			return FALSE;
		}
		return TRUE;
	}
	CLPR_add_entry(EQ, eqn, v0, -1.0); /*** 0 = -v0 + c * v1 ***/
	CLPR_add_entry(EQ, eqn, v1, c);
	if (eqn->next) return CLPR_solve_parametric_eqn(eqn, TRUE);
	flag = is_zero(eqn->constant);
	free_EQN(eqn);
	return flag;
}

/*----------------------------------------------- aux stuff -----------------*/

void CLPR_check_solver_sizes(void)
{
CLPR_int i;
	i = sizeof(SOLVER_NODE);
	if (i != sizeof(ENTRY)) CLPR_fatal("ENTRY node not aligned (see solver.h)");
	if (i != sizeof(OCC)) CLPR_fatal("OCC node not aligned (see solver.h)");
	if (i != sizeof(EQN)) CLPR_fatal("EQN node not aligned (see solver.h)");
	if (i != sizeof(EQN_PARAMS)) 
		CLPR_fatal("EQN_PARAMS node not aligned (see solver.h)");
	if (i != sizeof(NL_PARAMS1)) 
		CLPR_fatal("NL_PARAMS1 node not aligned (see solver.h)");
	if (i != sizeof(NL_PARAMS2)) 
		CLPR_fatal("NL_PARAMS2 node not aligned (see solver.h)");
	if (i != sizeof(NL_OCC)) 
		CLPR_fatal("NL_OCC node not aligned (see solver.h)");
	if (i != sizeof(TAB_HDR_NODE)) 
		CLPR_fatal("TAB_HDR_NODE node not aligned (see solver.h)");
	if (i != sizeof(KOHLER_NODE)) 
		CLPR_fatal("KOHLER_NODE node not aligned (see solver.h)");
}

void CLPR_init_solver(void)
{
CLPR_int i;
SOLVER_NODE_ptr tmp1, tmp2;
	if (CLPR_SOLVER_SZ < MIN_SOLVER_SZ) 
		CLPR_fatal("solver size (DEF_SOLVER_SZ) too small");
/*******
	CLPR_tableau = (TAB_HDR_NODE_ptr *) calloc(1, CLPR_SOLVER_SZ * sizeof(CLPR_int));
*******/
	CLPR_tableau = (TAB_HDR_NODE_ptr *) CLPR_alloc_mem(CLPR_SOLVER_SZ);
	if (!CLPR_tableau) CLPR_fatal("not enough memory to initialize clpr (calloc)");
	CLPR_solver_id = 0;
	CLPR_slack_id = CLPR_SOLVER_SZ;
	CLPR_nlin_count = 0;
	CLPR_sms_blocks = 1;
/*******
	CLPR_sms = (SOLVER_NODE_ptr) malloc(SOLVER_MALLOC_NODES*sizeof(SOLVER_NODE));
********/
	CLPR_sms = (SOLVER_NODE_ptr) CLPR_alloc_mem(
	    (SOLVER_MALLOC_NODES*sizeof(SOLVER_NODE))/sizeof(CLPR_int) + 1);
	if (!CLPR_sms) CLPR_fatal("Not enough memory to initialize solver");
	tmp1 = CLPR_sms;
	tmp2 = &tmp1[SOLVER_MALLOC_NODES - 1];
	do {
		tmp1->next = tmp1 + 1;
	} while (++tmp1 < tmp2);
	tmp1->next = NULL;
}

void CLPR_init_solver_goal(void)
{
CLPR_int i;
	CLPR_nl_eqn_top = NULL;
	if (CLPR_savep > 1) return;
/***
	for (i = 0; i < MAX_SOLVER_COUNTER; i++) CLPR_s_count[i] = 0;
***/
	CLPR_malloc_eqn_count = CLPR_malloc_eqn_par_count = CLPR_malloc_entry_count = 0;
	CLPR_malloc_occ_count = CLPR_malloc_nl_eqn_count = CLPR_malloc_nl_occ_count = 0;
	CLPR_malloc_nl_par1_count = CLPR_malloc_nl_par2_count = CLPR_malloc_hdr_count = 0;
	CLPR_malloc_koh_count = CLPR_free_eqn_count = CLPR_free_eqn_par_count = 0;
	CLPR_free_entry_count = CLPR_free_occ_count = CLPR_free_nl_eqn_count = 0;
	CLPR_free_nl_occ_count = CLPR_free_nl_par1_count = CLPR_free_nl_par2_count = 0;
	CLPR_free_hdr_count = CLPR_free_koh_count = 0;
}

SOLVER_NODE_ptr CLPR_make_new_solver_malloc_space(void)
{
SOLVER_NODE_ptr tmp0, tmp1, tmp2;

/***
	fprintf(CLPR_error_stream, "Making more solver malloc space ... ");
	fflush(stdout);
***/
	CLPR_sms_blocks++;
/********
	tmp0 = (SOLVER_NODE_ptr) malloc(SOLVER_MALLOC_NODES*sizeof(SOLVER_NODE));
********/
	tmp0 = (SOLVER_NODE_ptr) CLPR_alloc_mem(
	    (SOLVER_MALLOC_NODES*sizeof(SOLVER_NODE))/sizeof(CLPR_int) + 1);
	if (!tmp0) CLPR_fatal("Not enough memory (solver)");
	tmp1 = tmp0;
	tmp2 = &tmp1[SOLVER_MALLOC_NODES - 1];
	do {
		tmp1->next = tmp1 + 1;
	} while (++tmp1 < tmp2);
	tmp1->next = NULL;
/***
	fprintf(CLPR_error_stream, "done\n");
***/


	return tmp0; 
}

void CLPR_check_solver_malloc_space(void)
{
CLPR_int i;
	if (i = (CLPR_malloc_eqn_count - CLPR_free_eqn_count)) 
		CLPR_malloc_warning(i, "EQN");
	if (i = (CLPR_malloc_eqn_par_count - CLPR_free_eqn_par_count)) 
		CLPR_malloc_warning(i, "EQN_PARAMS");
	if (i = (CLPR_malloc_entry_count - CLPR_free_entry_count)) 
		CLPR_malloc_warning(i, "ENTRY");
	if (i = (CLPR_malloc_occ_count - CLPR_free_occ_count)) 
		CLPR_malloc_warning(i, "OCC");
	if (i = (CLPR_malloc_nl_eqn_count - CLPR_free_nl_eqn_count)) 
		CLPR_malloc_warning(i, "NL_EQN");
	if (i = (CLPR_malloc_nl_par1_count - CLPR_free_nl_par1_count)) 
		CLPR_malloc_warning(i, "NL_PARAMS1");
	if (i = (CLPR_malloc_nl_par2_count - CLPR_free_nl_par2_count)) 
		CLPR_malloc_warning(i, "NL_PARAMS2");
	if (i = (CLPR_malloc_hdr_count - CLPR_free_hdr_count)) 
		CLPR_malloc_warning(i, "HDR");
	if (i = (CLPR_malloc_koh_count - CLPR_free_koh_count)) 
		CLPR_malloc_warning(i, "DUMP");
}

CLPR_int CLPR_count_solver_malloc_space(void)
{
CLPR_int count;
SOLVER_NODE_ptr sn;
	count = 0;
	sn = CLPR_sms;
	while (sn) {
		count++;
		sn = sn->next;
	}
	return count;
}

void CLPR_malloc_warning(CLPR_int n, char *s)
{
	printf("*** Warning: %d %s nodes lost\n", n, s);
}

/*---------------------------------------------------------------------------*/

CLPR_int CLPR_is_ground(CLPR_int v, double *c_ptr)
{
EQN_ptr eqn;
	if (!CLPR_tableau[v] || !(eqn = CLPR_tableau[v]->eqn) || eqn->next) return FALSE;
	if (c_ptr) *c_ptr = eqn->constant;
	return TRUE;
}

CLPR_int CLPR_free_eqn_nodes(EQN_ptr eqn)
{
register ENTRY_ptr entry;
ENTRY_ptr entry99;
CLPR_int count;
	if (!eqn) return 0;
	count = 0;
	if (eqn->params) {
		free_EQN_PARAMS(eqn->params);
		count++;
	}
	entry = eqn->next;
	free_EQN(eqn);
	count++;
	while (entry) {
		count++;
		entry99 = entry->next;
		free_ENTRY(entry);
		entry = entry99;
	}
	return count;
}

void CLPR_free_nlin_nodes(NL_EQN_ptr nl_eqn)
{
	if (!nl_eqn) return;
	if (nl_eqn->params1) free_NL_PARAMS1(nl_eqn->params1);
	if (nl_eqn->params2) free_NL_PARAMS2(nl_eqn->params2);
	free_NL_EQN(nl_eqn);
}

CLPR_int CLPR_free_occ_nodes(OCC_ptr occ)
{
OCC_ptr occ99;
CLPR_int count;
	count = 0;
	while (occ) {
		count++;
		occ99 = occ->next;
		free_OCC(occ);
		occ = occ99;
	}
	return count;
}

CLPR_int CLPR_free_nl_occ_nodes(NL_OCC_ptr occ)
{
NL_OCC_ptr occ99;
CLPR_int count;
	count = 0;
	while (occ) {
		count++;
		occ99 = occ->next;
		free_NL_OCC(occ);
		occ = occ99;
	}
	return count;
}

CLPR_int CLPR_equal_param(CLPR_int v1, CLPR_int v2)
{
	return (CLPR_deref_param(v1) == CLPR_deref_param(v2));
}

CLPR_int CLPR_deref_param(CLPR_int v)
{
TAB_HDR_NODE_ptr hdr;
EQN_ptr eqn;
	if (!(hdr = CLPR_tableau[v])) return v;
	if (!(eqn = hdr->eqn)) return v;
	if (is_zero(eqn->constant)) {
		if (!eqn->next) return v;
		if (!eqn->next->next && is_equal(eqn->next->coef, 1.0))
			return eqn->next->id;
	} 
	return v;
}
