/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: solver.ineq.c                                        *
 ***************************************************************/

#include "standard.h"
#include "config.h"
#include "emul.h"
#include "solver.h"
#include "main.h"

#define zero_format(eqn) \
	(is_zero((eqn)->constant) && is_zero((eqn)->params->epsilon))

#define neg_format(eqn) \
	((eqn->constant > CLPR_EPS) || \
		(is_zero(eqn->constant) && eqn->params->epsilon > CLPR_EPS))

#define flip_eqn(eqn) { \
	CLPR_mult_eqn(INEQ, eqn, -1.0); \
}

#define check_slack_id(i) { \
	if (i <= CLPR_solver_id) CLPR_fatal("too many solver variables"); \
}

static CLPR_int basic_var;
static EQN_ptr exit_eqn;
static double pivot_number;
static ENTRY_ptr pivot_entry;
static EQN_ptr art_eqn;
static ENTRY_ptr prev_pivot_entry;

extern CLPR_int CLPR_stamp;
extern CLPR_int CLPR_TRACE;

CLPR_int CLPR_IMPLICIT = TRUE;
CLPR_int CLPR_PARTIAL_IMPLICIT = FALSE;

CLPR_int CLPR_implicit_eqn_buf[MAX_IMPLICIT];
CLPR_int CLPR_implicit_top = 0;
static CLPR_int depth_implicit = 0;

extern CLPR_int CLPR_CLAM_TRACE;
/*---------------------------------------------------------------------------*/


void CLPR_save_orig_ineq (EQN_ptr eqn);
CLPR_int CLPR_destroyable (EQN_ptr eqn);
CLPR_int CLPR_some_implicits (EQN_ptr eqn);
CLPR_int CLPR_full_implicits (EQN_ptr eqn);
CLPR_int CLPR_solve_ineq_tableau (CLPR_int art_row, CLPR_int *destroyed);
static void find_basic_var (void);
void CLPR_find_exit_eqn (void);
void CLPR_pivot (CLPR_int flag);
CLPR_int CLPR_check_implicits (void);
void CLPR_delete_row_occs (EQN_ptr eqn);
CLPR_int CLPR_find_implicit_row (CLPR_int v, double *coef, EQN_ptr *eqn);
void CLPR_delete_var (CLPR_int v);

/*** eqn is read-only ***/
CLPR_int CLPR_solve_inequality(CLPR_int strict, EQN_ptr eqn)
{
EQN_ptr eqn2;
EQN_PARAMS_ptr params;
ENTRY_ptr entry;
TAB_HDR_NODE_ptr hdr;
CLPR_int flag, dummy;

	stats_count(180);
	CLPR_slack_id--;
	check_slack_id(CLPR_slack_id);
	CLPR_backup_null_ineq(CLPR_slack_id);
	malloc_EQN(eqn2);
	malloc_EQN_PARAMS(params);
	stats_count(126);
	eqn2->next = NULL;
	eqn2->time_stamp = CLPR_stamp;
	eqn2->id = CLPR_slack_id;
	eqn2->constant = eqn->constant;
	eqn2->params = params;
	params->basic = 0;
	params->epsilon = (strict ? -1.0 : 0.0);
	if (!(hdr = CLPR_tableau[CLPR_slack_id])) {
		malloc_HDR(hdr);
		CLPR_tableau[CLPR_slack_id] = hdr;
	}
	if (CLPR_IMPLICIT) CLPR_save_orig_ineq(eqn);
	hdr->eqn = eqn2;
	entry = eqn->next;
	while (entry) {
		CLPR_add_entry(INEQ, eqn2, entry->id, entry->coef);
		entry = entry->next;
	}
	CLPR_add_entry(INEQ, eqn2, CLPR_slack_id, -1.0);

	if (CLPR_IMPLICIT && CLPR_destroyable(eqn2)) {
		if (CLPR_PARTIAL_IMPLICIT) return CLPR_some_implicits(eqn2);
		else return CLPR_full_implicits(eqn2);
	}
	if (neg_format(eqn2)) 
		CLPR_mult_eqn(INEQ, eqn2, -1.0);
	if (CLPR_solve_ineq_tableau(CLPR_slack_id, &dummy)) {
		return TRUE;
	} else return FALSE;
}

CLPR_int CLPR_solve_ineq_tableau(CLPR_int art_row, CLPR_int *destroyed)
{
TAB_HDR_NODE_ptr hdr;

	stats_count(112);
	*destroyed = FALSE;
	art_eqn = CLPR_tableau[art_row]->eqn;
	loop {
		find_basic_var();
		if (!basic_var) return FALSE;
		else if (basic_var == -1) {
			*destroyed = TRUE;
			if (CLPR_PARTIAL_IMPLICIT) return CLPR_some_implicits(art_eqn);
			else return CLPR_full_implicits(art_eqn);
		}
		CLPR_find_exit_eqn();

		if (exit_eqn->time_stamp < CLPR_stamp) {
			CLPR_backup_ineq(exit_eqn);
			exit_eqn->time_stamp = CLPR_stamp;
		}
		hdr = CLPR_tableau[exit_eqn->params->basic];
		if (hdr) hdr->basic_eqn = 0;
		exit_eqn->params->basic = basic_var;
		if (!(hdr = CLPR_tableau[basic_var])) {
			malloc_HDR(hdr);
			CLPR_tableau[basic_var] = hdr;
		}
		hdr->basic_eqn = exit_eqn->id;
		if ((pivot_number = pivot_entry->coef) > CLPR_EPS) {
			CLPR_mult_eqn(INEQ, exit_eqn, 1.0 / pivot_number);
			CLPR_pivot(1);
		} else {
			CLPR_mult_eqn(INEQ, exit_eqn, -1.0 / pivot_number);
			CLPR_pivot(-1);
		}
		if (exit_eqn == art_eqn) return TRUE;
	}
}

/*** Used for the equation case where we avoid doing the final pivot  ***/
/*** Thats the only difference from CLPR_solve_ineq_tableau                ***/

CLPR_int CLPR_solve_ineq_subs(CLPR_int art_row, CLPR_int *destroyed)
{
TAB_HDR_NODE_ptr hdr;

	stats_count(112);
	*destroyed = FALSE;
	art_eqn = CLPR_tableau[art_row]->eqn;
	loop {
		find_basic_var();
		if (!basic_var) return FALSE;
		else if (basic_var == -1) {
			*destroyed = TRUE;
			if (CLPR_PARTIAL_IMPLICIT) return CLPR_some_implicits(art_eqn);
			else return CLPR_full_implicits(art_eqn);
		}
		stats_count(103);
		CLPR_find_exit_eqn();
		if (exit_eqn == art_eqn) return TRUE;

		if (exit_eqn->time_stamp < CLPR_stamp) {
			CLPR_backup_ineq(exit_eqn);
			exit_eqn->time_stamp = CLPR_stamp;
		}
		hdr = CLPR_tableau[exit_eqn->params->basic];
		if (hdr) hdr->basic_eqn = 0;
		exit_eqn->params->basic = basic_var;
		if (!(hdr = CLPR_tableau[basic_var])) {
			malloc_HDR(hdr);
			CLPR_tableau[basic_var] = hdr;
		}
		hdr->basic_eqn = exit_eqn->id;
		if ((pivot_number = pivot_entry->coef) > CLPR_EPS) {
			CLPR_mult_eqn(INEQ, exit_eqn, 1.0 / pivot_number);
			CLPR_pivot(1);
		} else {
			CLPR_mult_eqn(INEQ, exit_eqn, -1.0 / pivot_number);
			CLPR_pivot(-1);
		}
	}
}

/************************************************************
 * input: basic_var, pivot_entry, prev_pivot_entry, exit_eqn;
 ************************************************************/
void CLPR_pivot(CLPR_int flag)
{
register OCC_ptr occ;
OCC_ptr occ99;
EQN_ptr eqn;
TAB_HDR_NODE_ptr hdr;
CLPR_int i;
double c;

	if (!(hdr = CLPR_tableau[basic_var])) return;
	occ = hdr->ineq_occ; 
	while (occ) {
		occ99 = occ->next;
		if ((i = occ->id) != exit_eqn->id) {
			eqn = CLPR_tableau[i]->eqn;
			if (eqn->time_stamp < CLPR_stamp) {
				CLPR_backup_ineq(eqn);
				eqn->time_stamp = CLPR_stamp;
			}
			c = CLPR_delete_entry(INEQ, eqn, basic_var);

			/*** CLPR_tmp_delete_entry(exit_eqn, pivot_entry) ***/
			if (prev_pivot_entry) prev_pivot_entry->next = pivot_entry->next;
			else exit_eqn->next = pivot_entry->next;
			
			if (flag > 0) CLPR_add_eqns(INEQ, eqn, exit_eqn, -c);
			else CLPR_add_eqns(INEQ, eqn, exit_eqn, c);

			/*** CLPR_tmp_undelete_entry(exit_eqn) ***/
			if (prev_pivot_entry) {
				pivot_entry->next = prev_pivot_entry->next;
				prev_pivot_entry->next = pivot_entry;
			} else {
				pivot_entry->next = exit_eqn->next;
				exit_eqn->next = pivot_entry;
			}
			if (neg_format(eqn)) CLPR_mult_eqn(INEQ, eqn, -1.0);
		}
		occ = occ99;
	}
}

/*******************************************************
 * Find_basic_var()
 *		input: art_eqn
 *      output: basic_var, pivot_entry, prev_pivot_entry;
 *******************************************************/

void find_basic_var(void) /*** return -1 for implicits, 0 for fail ***/
{
register ENTRY_ptr entry, prev_entry;

AGAIN:
	prev_pivot_entry = pivot_entry = prev_entry = NULL;
	entry = art_eqn->next;
	while (entry) {
		if (entry->id < CLPR_slack_id) {
			prev_pivot_entry = prev_entry;
			pivot_entry = entry;
			basic_var = entry->id;
			return;
		}
		if (entry->coef > CLPR_EPS) {
			prev_pivot_entry = prev_entry;
			pivot_entry = entry;
		}
		prev_entry = entry;
		entry = entry->next;
	}
	if (pivot_entry) basic_var = pivot_entry->id;
	else if (CLPR_IMPLICIT) basic_var = (zero_format(art_eqn) ? -1 : 0);
	else if (zero_format(art_eqn)) {
		flip_eqn(art_eqn);
		goto AGAIN;
	} else basic_var = 0;
}

/****************************************************************
 * Find_exit_eqn()
 *       input: art_eqn, basic_var, pivot_entry, prev_pivot_entry
 *      output: exit_eqn, pivot_entry, prev_pivot_entry
 ***************************************************************/

void CLPR_find_exit_eqn(void)
{
register ENTRY_ptr entry, prev_entry;
register OCC_ptr occ, occ99;
EQN_ptr eqn;
EQN_PARAMS_ptr art_params;
double min_c, min_eps, c, c1, e;

	pivot_number = pivot_entry->coef;
	exit_eqn = art_eqn;
	if (basic_var < CLPR_slack_id) return;
	art_params = art_eqn->params;
/*** Tricky: only if we don't use implicits ***/
	if (!CLPR_IMPLICIT && is_zero(art_eqn->constant) 
			&& is_zero(art_params->epsilon)) 
		return;
	/* used to be 0 but need to make this be preferred */
	art_params->basic = CLPR_SOLVER_SZ;
	min_c = (c = art_eqn->constant) / pivot_number;
	min_eps = (art_params->epsilon) / pivot_number;

	occ = CLPR_tableau[basic_var]->ineq_occ;
	while (occ) {
		occ99 = occ->next;
		eqn = CLPR_tableau[occ->id]->eqn;
		if (eqn->params->basic < CLPR_slack_id) goto CONTINUE;
		if (eqn == art_eqn) goto CONTINUE;

		/*** c1 = (entry = CLPR_find_entry(eqn, basic_var))->coef; ***/
		entry = eqn->next;
		prev_entry = NULL;
		loop {
			if (entry->id == basic_var) break;
			prev_entry = entry;
			entry = entry->next;
		}
		if ((c1 = entry->coef) < CLPR_EPS) goto CONTINUE;
		c = (eqn->constant / c1);
		e = (eqn->params->epsilon / c1);

		/* NOTE: **************************/
		/* CHECK BUG(?) should CLPR_EPS be used in comparisons */
		if (c > min_c || 
		   (is_equal(c, min_c) && e > min_eps) ||
		   (is_equal(c, min_c) && is_equal(e, min_eps) && 
					eqn->params->basic < exit_eqn->params->basic)) {
/************
		if (c > min_c+CLPR_EPS || 
		   (is_equal(c, min_c) && e > min_eps+CLPR_EPS) ||
		   (is_equal(c, min_c) && is_equal(e, min_eps) && 
					eqn->params->basic < exit_eqn->params->basic)) {
************/
			exit_eqn = eqn;
			pivot_entry = entry;
			prev_pivot_entry = prev_entry;
			min_c = c;
			min_eps = e;
		}
		CONTINUE: occ = occ99;
	}
	art_params->basic = 0;
}

/*** subs v0 in CLPR_tableau[] by eqn0 ***/
CLPR_int CLPR_ineq_substitute(CLPR_int v0, EQN_ptr eqn0)
{
register ENTRY_ptr entry, entry99;
register OCC_ptr occ, occ99;
EQN_ptr eqn;
CLPR_int i, flag, destroyed;
double c;

	stats_count(104);
	if (i = CLPR_tableau[v0]->basic_eqn) {
		stats_count(105);
		eqn = CLPR_tableau[i]->eqn;
		if (eqn->time_stamp < CLPR_stamp) {
			CLPR_backup_ineq(eqn);
			eqn->time_stamp = CLPR_stamp;
		}
		eqn->params->basic = CLPR_tableau[v0]->basic_eqn = 0;
		c = CLPR_delete_entry(INEQ, eqn, v0);
		eqn->constant += c * eqn0->constant;
		entry = eqn0->next;
		while (entry) {
			entry99 = entry->next;
			CLPR_add_entry(INEQ, eqn, entry->id, c * entry->coef);
			entry = entry99;
		}
		if (neg_format(eqn)) CLPR_mult_eqn(INEQ, eqn, -1.0);
		else if (CLPR_IMPLICIT && zero_format(eqn)) {
			if (CLPR_solve_ineq_subs(i, &destroyed)) {
				if (destroyed) return TRUE;
				flip_eqn(CLPR_tableau[i]->eqn);
				return CLPR_solve_ineq_tableau(i, &destroyed);
			} else return FALSE;
		}
		return CLPR_solve_ineq_tableau(i, &destroyed);
	}
	occ = CLPR_tableau[v0]->ineq_occ;
	while (occ) { /*** no var in eqn0 is basic ***/
		occ99 = occ->next;
		eqn = CLPR_tableau[occ->id]->eqn;
		if (eqn->time_stamp < CLPR_stamp) {
			CLPR_backup_ineq(eqn);
			eqn->time_stamp = CLPR_stamp;
		}
		c = CLPR_delete_entry(INEQ, eqn, v0);
		CLPR_add_eqns(INEQ, eqn, eqn0, c);
		if (neg_format(eqn)) CLPR_mult_eqn(INEQ, eqn, -1.0);
		occ = occ99;
	}
	return TRUE;
}

/*--------------------------------------------------- implicit eqns ---------*/

void CLPR_save_orig_ineq(EQN_ptr eqn)	/*** save a COPY ***/
            
{
	if (CLPR_tableau[CLPR_slack_id]->next) {
		CLPR_free_eqn_nodes((EQN_ptr) CLPR_tableau[CLPR_slack_id]->next);
	}
	CLPR_tableau[CLPR_slack_id]->next = (TAB_HDR_NODE_ptr) eqn;
}

CLPR_int CLPR_destroyable(EQN_ptr eqn)
{
ENTRY_ptr entry;
CLPR_int sign;
	if (!zero_format(eqn)) return FALSE;
	entry = eqn->next;
	if (entry->coef > CLPR_EPS) sign = 1; else sign = -1;
	for (entry = eqn->next; entry ; entry = entry->next) {
		if ((entry->id < CLPR_slack_id) || (sign * entry->coef < -CLPR_EPS)) 
			return FALSE;
	}
	return TRUE;
}

CLPR_int CLPR_some_implicits(EQN_ptr eqn)
{
register ENTRY_ptr entry, entry99;
register OCC_ptr occ, occ99;
EQN_ptr eqn2;
CLPR_int i, v, flag;
double c;
CLPR_int current_id;

	if (eqn->time_stamp < CLPR_stamp) /* assumed no backup needed */ CLPR_fatal("78412");
	if (!(entry = eqn->next)) return TRUE;
	current_id = eqn->id;
	while (entry) {
		entry99 = entry->next;
		v = entry->id;
		occ = CLPR_tableau[v]->ineq_occ;
		while (occ) {
			occ99 = occ->next;
			if ((i = occ->id) != current_id) {
				eqn2 = CLPR_tableau[i]->eqn;
				if (eqn2->time_stamp < CLPR_stamp) {
					CLPR_backup_ineq(eqn2);
					eqn2->time_stamp = CLPR_stamp;
				}
				CLPR_delete_entry(INEQ, eqn2, v);
			}
			occ = occ99;
		}
		if (CLPR_implicit_top >= MAX_IMPLICIT-1) CLPR_fatal("too many implicits");
		CLPR_implicit_eqn_buf[CLPR_implicit_top++] = v;
		CLPR_delete_entry(INEQ, eqn, v);
		entry = entry99;
	}
	if (eqn->next != NULL) CLPR_fatal("99667");

	/* Only send n-1 of the equations to the solver */
    /* We dont send the last such equation since it may not exist */

	CLPR_implicit_top--;
	return TRUE;
}

CLPR_int CLPR_parametrify_eqn(EQN_ptr eqn)
{
EQN_ptr imp_scratch;
ENTRY_ptr entry;

	malloc_EQN(imp_scratch);
	imp_scratch->next = NULL;
	imp_scratch->id = 0;
	imp_scratch->constant = eqn->constant;
	imp_scratch->time_stamp = CLPR_stamp;
	imp_scratch->params = NULL;
	entry = eqn->next;
	while (entry) {
		CLPR_add_entry(EQ, imp_scratch, entry->id, entry->coef);
		entry = entry->next;
	}
	if (!imp_scratch->next) {
		if (is_zero(imp_scratch->constant)) {
			free_EQN(imp_scratch);	
			return CLPR_check_implicits();
		} else CLPR_fatal("boo boo");
	}
	return CLPR_solve_parametric_eqn(imp_scratch, FALSE);
}

CLPR_int CLPR_check_implicits(void) /*** one implicit at a time - recursively ***/
{
CLPR_int v;
EQN_ptr eqn;
	if (CLPR_implicit_top <= 0) return TRUE;
	v = CLPR_implicit_eqn_buf[--CLPR_implicit_top];
	if (!(eqn = (EQN_ptr) CLPR_tableau[v]->next)) CLPR_fatal("287233");
	return CLPR_parametrify_eqn(eqn);
}

CLPR_int CLPR_full_implicits(EQN_ptr eqn)
{
ENTRY_ptr entry, entry99, first_entry;
EQN_ptr art_eqn;
CLPR_int v, destroyed;
double coef;
CLPR_int i;

	if (eqn->time_stamp < CLPR_stamp) {
		/*** assumed it didn't need to be backed up ***/
		CLPR_fatal("78413");
	}
	CLPR_delete_row_occs(eqn);
	first_entry = entry = eqn->next;
	eqn->next = NULL;     /*** Make it invisible in the tableau ***/
	for (; entry; entry = entry->next) {
		v = entry->id;
		if (!CLPR_find_implicit_row(v, &coef, &art_eqn)) { 
			CLPR_delete_var(v);
			if (CLPR_implicit_top >= MAX_IMPLICIT-1) CLPR_fatal("too many implicits");
			CLPR_implicit_eqn_buf[CLPR_implicit_top++] = v;
		} else {
			/*** substitute out for v everywhere ***/
			if (art_eqn->time_stamp < CLPR_stamp) {
				CLPR_backup_ineq(art_eqn);
				art_eqn->time_stamp = CLPR_stamp;
			}
			if (CLPR_tableau[art_eqn->params->basic]) 
				CLPR_tableau[art_eqn->params->basic]->basic_eqn = 0;
			art_eqn->params->basic = v;
			CLPR_mult_eqn(INEQ, art_eqn, 1.0 / coef);
			exit_eqn = art_eqn;
			CLPR_pivot(1);
			CLPR_delete_entry(INEQ, art_eqn, v);
			art_eqn->params->basic = 0;
			if (!CLPR_solve_ineq_tableau(art_eqn->id, &destroyed)) {
				CLPR_fatal("93368");
			}
			if (CLPR_implicit_top >= MAX_IMPLICIT-1) CLPR_fatal("too many implicits");
			CLPR_implicit_eqn_buf[CLPR_implicit_top++] = v;			
		}
	}
	/* Only send n-1 of the equations to the solver */
    /* We dont send the last such equation since it may not exist */
	/* Tricky with the recursion but it should work */

	CLPR_implicit_top--;
	entry = first_entry;
	while (entry) {
		entry99 = entry->next;
		free_ENTRY(entry);
		entry = entry99;
	}
	depth_implicit--;
	return TRUE;
}


void CLPR_delete_var(CLPR_int v) /*** Delete a non-basic variable from the tableau ***/
      
{
register OCC_ptr occ, occ99;
EQN_ptr eqn2;
CLPR_int i;

	occ = CLPR_tableau[v]->ineq_occ;
	while (occ) {
		occ99 = occ->next;
		i = occ->id;
		eqn2 = CLPR_tableau[i]->eqn;
		if (eqn2->time_stamp < CLPR_stamp) {
			CLPR_backup_ineq(eqn2);
			eqn2->time_stamp = CLPR_stamp;
		}
		CLPR_delete_entry(INEQ, eqn2, v);
		occ = occ99;
	}
}

/*** Delete all the occs corresponding to this row ***/
void CLPR_delete_row_occs(EQN_ptr eqn)
{
register ENTRY_ptr entry;
CLPR_int id;
	id = eqn->id;
	entry = eqn->next;
	while (entry) {
		CLPR_delete_occ(INEQ, id, entry->id);
		entry = entry->next;
	}
}

/****************************************************************
*  Find a suitable row for finding more implicits if there is on
*    output: basic_var, pivot_entry, prev_pivot_entry
****************************************************************/

CLPR_int CLPR_find_implicit_row(CLPR_int v, double *coef, EQN_ptr *eqn)
{
register OCC_ptr occ;
register ENTRY_ptr entry, entry99;
EQN_ptr teqn;
double tcoef;
CLPR_int i, flag;

	/*** destroyed variable is now basic ***/
	if (i = CLPR_tableau[v]->basic_eqn) {
		*coef = 1.0;
		*eqn = CLPR_tableau[i]->eqn;
		basic_var = v;
		entry99 = NULL;
		entry = (*eqn)->next;
		while(entry) {
			if (entry->id == v) {
				pivot_entry = entry;
				prev_pivot_entry = entry99;
			}
			entry99 = entry;
			entry = entry->next;
		}
		return TRUE;
	}
	occ = CLPR_tableau[v]->ineq_occ;
	while (occ) {
		i = occ->id;
		teqn = CLPR_tableau[i]->eqn;
		if (zero_format(teqn)) {
			entry99 = NULL;
			entry = teqn->next;
			flag = TRUE; 
			while(entry) {
				if (entry->id == v) {
					tcoef = entry->coef;
					pivot_entry = entry;
					prev_pivot_entry = entry99;
				}
				if (entry->id < CLPR_slack_id) { flag = FALSE; break; }
				entry99 = entry;
				entry = entry->next;
			}
			if (flag && tcoef < -CLPR_EPS) {
				if (is_zero(tcoef)) CLPR_fatal("3847348");
				*coef = tcoef; *eqn = teqn; 
				basic_var = v;
				return TRUE;
			}
		}	
		occ = occ->next;
	}
	return FALSE;
}

/* NOTE: TEST routines not used *******************/

double CLPR_div_by(double x, double r, double c)
{
double q, rem, y;
/*
	q = x*r;
	rem = q*c - x;
	y = q + rem*r;
	if (!is_equal(x/c,y)) CLPR_fatal("DIVISION");
	return y;
*/
	return x/c;
/*
	return x*r;
*/
}

void CLPR_norm_eqn(CLPR_int type, EQN_ptr eqn, double factor)
{
register ENTRY_ptr entry;
double rec;
	if (is_equal(factor, 1.0)) return;
	if (is_equal(factor, -1.0)) CLPR_fatal("-1 factor 13333");
	rec = 1/factor;
	eqn->constant = CLPR_div_by(eqn->constant, rec, factor);
	entry = eqn->next;
	while (entry) {
		entry->coef = CLPR_div_by(entry->coef, rec, factor);
		entry = entry->next;
	}
	if (type == INEQ)
		eqn->params->epsilon = CLPR_div_by(eqn->params->epsilon, rec, factor);
}

/*** NOTE: delete ***/
/******
void scale_eqn(type, eqn)
CLPR_int type;
EQN_ptr eqn;
{
ENTRY_ptr entry;
double min, max;
CLPR_int e1, e2, e;

	min=1.0e300; max = 0.0;
	for (entry=eqn->next; entry; entry=entry->next) {
		if (fabs(entry->coef) < min) min=fabs(entry->coef);
		if (fabs(entry->coef) > max) max=fabs(entry->coef);
	}
	frexp(min, &e1);  frexp(max, &e2);
	e = (e1+e2)/2;
	if (e > 0) e = 0;
	eqn->constant = ldexp(eqn->constant,-e);
	entry = eqn->next;
	while (entry) {
		entry->coef = ldexp(entry->coef, -e);
		entry = entry->next;
	}
	if (type == INEQ) eqn->params->epsilon = ldexp(eqn->params->epsilon,-e);
}
******/
