/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: solver.nlin.c                                        *
 ***************************************************************/

#include "standard.h"
#include "config.h"
#include "emul.h"
#include "solver.h"
#include "main.h"

NL_EQN_ptr CLPR_nl_eqn_top = NULL;
extern CLPR_int CLPR_stamp, CLPR_trtop, **CLPR_trail;

/*---------------------------------------------------------------------------*/


CLPR_int CLPR_wakeup_mult (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_pow (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_pow0 (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_pow1 (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_pow2 (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_pow02 (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_min (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_min1 (double c2, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_min2 (double c1, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_max (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_max1 (double c2, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_max2 (double c1, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_abs (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_abs0 (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_sin (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_sin0 (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_cos (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_cos0 (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_arcsin (double c, NL_OCC_ptr occ);
CLPR_int CLPR_wakeup_arccos (double c, NL_OCC_ptr occ);
extern CLPR_int CLPR_solve_eqn_1 (CLPR_int v, double c);
extern CLPR_int CLPR_solve_eqn_2 (CLPR_int v0, CLPR_int v1, double c);
extern CLPR_int CLPR_is_ground (CLPR_int v, double *c_ptr);
extern void CLPR_warn_err (char *s, ...);

NL_EQN_ptr CLPR_solve_eval(CLPR_int *t, CLPR_int *v) /*** t = eval(v), CLPR_wakeup() in meta.c ***/
           
{
NL_EQN_ptr nl_eqn;
NL_PARAMS1_ptr params1;
NL_PARAMS2_ptr params2;
	CLPR_nlin_count++;
	malloc_NL_EQN(nl_eqn);
	malloc_NL_PARAMS1(params1);
	malloc_NL_PARAMS2(params2);
	nl_eqn->next = CLPR_nl_eqn_top;
	nl_eqn->time_stamp = CLPR_stamp;
	nl_eqn->type = SOLVER_EVAL;
	nl_eqn->partner = NULL;
	nl_eqn->params1 = params1;
	nl_eqn->params2 = params2;
	params1->id0 = (CLPR_int) t;
	params1->id1 = (CLPR_int) v;
	params1->id2 = 0;
	params1->mark = 0;
	params2->occ0 = params2->occ1 = params2->occ2 = NULL;
	CLPR_nl_eqn_top = nl_eqn;
	push_trail(BTK_NLIN, nl_eqn);
	return nl_eqn;
}

CLPR_int CLPR_solve_mult(CLPR_int v0, CLPR_int v1, CLPR_int v2, double c) /*** v0 = c * v1 * v2 ***/
               
         
{
NL_EQN_ptr nl_eqn;
	stats_count(200);
	nl_eqn = CLPR_push_nl_eqn(SOLVER_MULT, v0, v1, v2, c, NULL);
	nl_eqn->params2->occ0 = NULL;
	nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_MULT, v1, nl_eqn, 1);
	nl_eqn->params2->occ2 = CLPR_insert_nl_occ(SOLVER_MULT, v2, nl_eqn, 2);
	return TRUE;
}

CLPR_int CLPR_solve_minmax(CLPR_int type, CLPR_int v0, CLPR_int v1, CLPR_int v2, double c)
{
NL_EQN_ptr nl_eqn;
	nl_eqn = CLPR_push_nl_eqn(type, v0, v1, v2, c, NULL);
	switch (type) {
		case SOLVER_MIN: 
		case SOLVER_MAX: 
			nl_eqn->params2->occ0 = NULL;
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(type, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = CLPR_insert_nl_occ(type, v2, nl_eqn, 2);
			break;
		case SOLVER_MIN1: 
		case SOLVER_MAX1: 
			nl_eqn->params2->occ0 = NULL;
			nl_eqn->params2->occ1 = NULL;
			nl_eqn->params2->occ2 = CLPR_insert_nl_occ(type, v2, nl_eqn, 2);
			break;
		case SOLVER_MIN2: 
		case SOLVER_MAX2: 
			nl_eqn->params2->occ0 = NULL;
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(type, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = NULL;
	}
	return TRUE;
}

CLPR_int CLPR_solve_power(CLPR_int type, CLPR_int v0, CLPR_int v1, CLPR_int v2, double c)
{
NL_EQN_ptr nl_eqn;
	nl_eqn = CLPR_push_nl_eqn(type, v0, v1, v2, c, NULL);
	switch (type) {
		case SOLVER_POW:
			nl_eqn->params2->occ0 = CLPR_insert_nl_occ(SOLVER_POW, v0, nl_eqn, 0);
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_POW, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = CLPR_insert_nl_occ(SOLVER_POW, v2, nl_eqn, 2);
			break;
		case SOLVER_POW0:
			nl_eqn->params2->occ0 = NULL;
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_POW0, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = CLPR_insert_nl_occ(SOLVER_POW0, v2, nl_eqn, 2);
			break;
		case SOLVER_POW1:
			nl_eqn->params2->occ0 = CLPR_insert_nl_occ(SOLVER_POW1, v0, nl_eqn, 0);
			nl_eqn->params2->occ1 = NULL;
			nl_eqn->params2->occ2 = CLPR_insert_nl_occ(SOLVER_POW1, v2, nl_eqn, 2);
			break;
		case SOLVER_POW2:
			nl_eqn->params2->occ0 = CLPR_insert_nl_occ(SOLVER_POW2, v0, nl_eqn, 0);
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_POW2, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = NULL;
			break;
		case SOLVER_POW02:
			nl_eqn->params2->occ0 = NULL;
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_POW02, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = NULL;
			break;
		default: CLPR_fatal("61934");
	}
	return TRUE;
}

CLPR_int CLPR_solve_absolute(CLPR_int type, CLPR_int v0, CLPR_int v1)
{
NL_EQN_ptr nl_eqn;
	nl_eqn = CLPR_push_nl_eqn(type, v0, v1, 0, 0.0, NULL);
	switch (type) {
		case SOLVER_ABS:
			nl_eqn->params2->occ0 = CLPR_insert_nl_occ(SOLVER_ABS, v0, nl_eqn, 0);
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_ABS, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = NULL;
			break;
		case SOLVER_ABS0:
			nl_eqn->params2->occ0 = NULL;
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_ABS0, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = NULL;
	}
	return TRUE;
}

CLPR_int CLPR_solve_sin(CLPR_int type, CLPR_int v0, CLPR_int v1)
{
NL_EQN_ptr nl_eqn;
	nl_eqn = CLPR_push_nl_eqn(type, v0, v1, 0, 0.0, NULL);
	switch (type) {
		case SOLVER_SIN:
			nl_eqn->params2->occ0 = CLPR_insert_nl_occ(SOLVER_SIN, v0, nl_eqn, 0);
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_SIN, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = NULL;
			break;
		case SOLVER_SIN0:
			nl_eqn->params2->occ0 = NULL;
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_SIN0, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = NULL;
			break;
	}
	return TRUE;
}

CLPR_int CLPR_solve_cos(CLPR_int type, CLPR_int v0, CLPR_int v1)
{
NL_EQN_ptr nl_eqn;
	nl_eqn = CLPR_push_nl_eqn(type, v0, v1, 0, 0.0, NULL);
	switch (type) {
		case SOLVER_COS:
			nl_eqn->params2->occ0 = CLPR_insert_nl_occ(SOLVER_COS, v0, nl_eqn, 0);
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_COS, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = NULL;
			break;
		case SOLVER_COS0:
			nl_eqn->params2->occ0 = NULL;
			nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_COS0, v1, nl_eqn, 1);
			nl_eqn->params2->occ2 = NULL;
			break;
	}
	return TRUE;
}

CLPR_int CLPR_solve_arcsin(CLPR_int v0, CLPR_int v1) /*** v0 = c * v1 * v2 ***/
           
{
NL_EQN_ptr nl_eqn;
	nl_eqn = CLPR_push_nl_eqn(SOLVER_ARCSIN, v0, v1, 0, 0.0, NULL);
	nl_eqn->params2->occ2 = NULL;
	nl_eqn->params2->occ0 = CLPR_insert_nl_occ(SOLVER_ARCSIN, v0, nl_eqn, 0);
	nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_ARCSIN, v1, nl_eqn, 1);
	return TRUE;
}

CLPR_int CLPR_solve_arccos(CLPR_int v0, CLPR_int v1) /*** v0 = c * v1 * v2 ***/
           
{
NL_EQN_ptr nl_eqn;
	nl_eqn = CLPR_push_nl_eqn(SOLVER_ARCCOS, v0, v1, 0, 0.0, NULL);
	nl_eqn->params2->occ2 = NULL;
	nl_eqn->params2->occ0 = CLPR_insert_nl_occ(SOLVER_ARCCOS, v0, nl_eqn, 0);
	nl_eqn->params2->occ1 = CLPR_insert_nl_occ(SOLVER_ARCCOS, v1, nl_eqn, 1);
	return TRUE;
}

/*-----------------------------------------------nl_eqn/nl_occ handling -----*/

void CLPR_push_change_nl_eqn(CLPR_int old_type, CLPR_int pos, NL_EQN_ptr nl_eqn0)
{
NL_EQN_ptr nl_eqn;
	stats_count(206);
	malloc_NL_EQN(nl_eqn);
	nl_eqn->next = CLPR_nl_eqn_top;
	nl_eqn->time_stamp = pos; /*** overload ***/
	nl_eqn->type = old_type;
	nl_eqn->partner = nl_eqn0;
	nl_eqn->params1 = NULL;
	nl_eqn->params2 = NULL;
	CLPR_nl_eqn_top = nl_eqn;
	push_trail(BTK_NLIN, nl_eqn);
}

NL_EQN_ptr CLPR_push_nl_eqn(CLPR_int type, CLPR_int v0, CLPR_int v1, CLPR_int v2, double c, NL_EQN_ptr nl_eqn0)
{
NL_EQN_ptr nl_eqn;
NL_PARAMS1_ptr params1;
NL_PARAMS2_ptr params2;
	stats_count(204);
	CLPR_nlin_count++;
	malloc_NL_EQN(nl_eqn);
	malloc_NL_PARAMS1(params1);
	malloc_NL_PARAMS2(params2);
	nl_eqn->next = CLPR_nl_eqn_top;
	nl_eqn->time_stamp = CLPR_stamp;
	nl_eqn->type = type;
	nl_eqn->partner = nl_eqn0;
	nl_eqn->params1 = params1;
	nl_eqn->params2 = params2;
	params1->id0 = v0; 
	params1->id1 = v1; 
	params1->id2 = v2;
	params1->coef = c;
	params1->mark = 0;
	params2->occ0 = params2->occ1 = params2->occ2 = NULL;
	CLPR_nl_eqn_top = nl_eqn;
	push_trail(BTK_NLIN, nl_eqn);
	return nl_eqn;
}

NL_OCC_ptr CLPR_insert_nl_occ(CLPR_int type, CLPR_int v, NL_EQN_ptr nl_eqn, CLPR_int pos)
{
NL_OCC_ptr occ, first_occ;
TAB_HDR_NODE_ptr hdr;
	stats_count(207);
	if (!(hdr = CLPR_tableau[v])) {
		malloc_HDR(hdr);
		CLPR_tableau[v] = hdr;
	}
	first_occ = hdr->nlin_occ;
	malloc_NL_OCC(occ);
	occ->next = first_occ;
	occ->prev = NULL;
	occ->type = type;
	occ->id = v;
	occ->nl_eqn = nl_eqn;
	occ->pos = pos;
	if (first_occ) first_occ->prev = occ;
	hdr->nlin_occ = occ;
	return occ;
}	

void CLPR_delete_nl_occ(NL_OCC_ptr occ)
{
NL_OCC_ptr prev, next;
TAB_HDR_NODE_ptr hdr;

	stats_count(208);
	prev = occ->prev;
	next = occ->next;
	hdr = CLPR_tableau[occ->id];
	if (prev) prev->next = next;
	else hdr->nlin_occ = next;
	if (next) next->prev = prev;
	else 
		if (null_hdr(hdr)) {
			free_HDR(hdr);
			CLPR_tableau[occ->id] = NULL;
		}
	free_NL_OCC(occ);
}

/*------------------------------------------------ wakeup -------------------*/

CLPR_int CLPR_wakeup(CLPR_int v, double c)
{
NL_OCC_ptr occ;
CLPR_int flag;
	stats_count(214);
	if (!CLPR_tableau[v]) return TRUE;
	occ = CLPR_tableau[v]->nlin_occ;
	flag = TRUE;
	while (occ) {
		stats_count(261);
		switch (occ->nl_eqn->type) {
			case WOKEN:        return TRUE;
			case SOLVER_MULT:  flag = CLPR_wakeup_mult(c, occ); break;
			case SOLVER_POW:   flag = CLPR_wakeup_pow(c, occ); break; 
			case SOLVER_POW0:  flag = CLPR_wakeup_pow0(c, occ); break; 
			case SOLVER_POW1:  flag = CLPR_wakeup_pow1(c, occ); break; 
			case SOLVER_POW2:  flag = CLPR_wakeup_pow2(c, occ); break;  
			case SOLVER_POW02: flag = CLPR_wakeup_pow02(c, occ); break;  
			case SOLVER_MIN:   flag = CLPR_wakeup_min(c, occ); break;	
			case SOLVER_MIN0:  CLPR_fatal("63960");
			case SOLVER_MIN1:  flag = CLPR_wakeup_min1(c, occ); break;	
			case SOLVER_MIN2:  flag = CLPR_wakeup_min2(c, occ); break;	
			case SOLVER_MAX:   flag = CLPR_wakeup_max(c, occ); break;	
			case SOLVER_MAX0:  CLPR_fatal("63961");
			case SOLVER_MAX1:  flag = CLPR_wakeup_max1(c, occ); break;	
			case SOLVER_MAX2:  flag = CLPR_wakeup_max2(c, occ); break;	
			case SOLVER_ABS:   flag = CLPR_wakeup_abs(c, occ); break;
			case SOLVER_ABS0:  flag = CLPR_wakeup_abs0(c, occ); break;
			case SOLVER_SIN:   flag = CLPR_wakeup_sin(c, occ); break;
			case SOLVER_SIN0:  flag = CLPR_wakeup_sin0(c, occ); break;
			case SOLVER_COS:   flag = CLPR_wakeup_cos(c, occ); break;
			case SOLVER_COS0:  flag = CLPR_wakeup_cos0(c, occ); break;
			case SOLVER_ARCSIN: flag = CLPR_wakeup_arcsin(c, occ); break;
			case SOLVER_ARCCOS: flag = CLPR_wakeup_arccos(c, occ); break;
			default:          CLPR_fatal("63962");
		}
		if (!flag) return FALSE;
		occ = occ->next;
	}
	return TRUE;
}

CLPR_int CLPR_wakeup_mult(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
NL_PARAMS1_ptr params1;
CLPR_int v0, v1, v2;
double c99;
	stats_count(209);
	nl_eqn = occ->nl_eqn;
	params1 = nl_eqn->params1;
	v0 = params1->id0; 
	v1 = params1->id1; 
	v2 = params1->id2;
	c99 = params1->coef;
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	if (occ->pos == 1) {
		CLPR_push_change_nl_eqn(SOLVER_MULT, 1, nl_eqn);
		CLPR_delete_nl_occ(nl_eqn->params2->occ2);
		if (v1 == v2) return CLPR_solve_eqn_1(v0, c99 * c * c);
		return CLPR_solve_eqn_2(v0, v2, c99 * c); 
	}
	CLPR_push_change_nl_eqn(SOLVER_MULT, 2, nl_eqn);
	CLPR_delete_nl_occ(nl_eqn->params2->occ1);
	if (v1 == v2) return CLPR_solve_eqn_1(v0, c99 * c * c);
	return CLPR_solve_eqn_2(v0, v1, c99 * c); 
}

CLPR_int CLPR_wakeup_pow(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn, nl_eqn99;
CLPR_int pos;
	stats_count(210);
	nl_eqn = occ->nl_eqn;
	CLPR_push_change_nl_eqn(SOLVER_POW, (pos = occ->pos), (nl_eqn = occ->nl_eqn));
	switch (pos) {
		case 2:
			if (is_zero(c)) {
				nl_eqn->type = WOKEN;
				CLPR_nlin_count--;
				CLPR_delete_nl_occ(nl_eqn->params2->occ0);
				CLPR_delete_nl_occ(nl_eqn->params2->occ1);
				return CLPR_solve_eqn_1(nl_eqn->params1->id0, 1.0);
			}
			if (is_equal(c, 1.0)) {
				nl_eqn->type = WOKEN;
				CLPR_nlin_count--;
				CLPR_delete_nl_occ(nl_eqn->params2->occ0);
				CLPR_delete_nl_occ(nl_eqn->params2->occ1);
				return 
					CLPR_solve_eqn_2(
						nl_eqn->params1->id0, nl_eqn->params1->id1, 1.0);
			}
			nl_eqn->type = SOLVER_POW2;
			nl_eqn->params1->coef = c;
			return TRUE;
		case 1:
			if (is_equal(c, 1.0)) {
				nl_eqn->type = WOKEN;
				CLPR_nlin_count--;
				CLPR_delete_nl_occ(nl_eqn->params2->occ0);
				CLPR_delete_nl_occ(nl_eqn->params2->occ2);
				return CLPR_solve_eqn_1(nl_eqn->params1->id0, c); /* 0 = 0^0 ? */
			}
			nl_eqn->type = SOLVER_POW1;
			nl_eqn->params1->coef = c;
			return TRUE;
		case 0:
			if (is_zero(c)) {
				nl_eqn->type = WOKEN;
				CLPR_nlin_count--;
				CLPR_delete_nl_occ(nl_eqn->params2->occ1);
				CLPR_delete_nl_occ(nl_eqn->params2->occ2);
				return CLPR_solve_eqn_1(nl_eqn->params1->id1, 0.0);
			}
			nl_eqn->type = SOLVER_POW0;
			nl_eqn->params1->coef = c;
			return TRUE;
		default:
			CLPR_fatal("56180");		
	}
}

CLPR_int CLPR_wakeup_pow0(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
CLPR_int pos;
double c99;
	stats_count(210);
	nl_eqn = occ->nl_eqn;
	c99 = nl_eqn->params1->coef;
	pos = occ->pos;
	if (pos == 2 && !is_zero(c) && !is_equal(c, 1.0)) {
		CLPR_push_change_nl_eqn(SOLVER_POW0, pos, nl_eqn);
		nl_eqn->type = SOLVER_POW02; /* special case */
		return TRUE;
	}
	CLPR_push_change_nl_eqn(SOLVER_POW0, pos, nl_eqn);
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	if (pos == 1) {
		CLPR_delete_nl_occ(nl_eqn->params2->occ2);
		return CLPR_solve_eqn_1(nl_eqn->params1->id2, log(c99)/log(c));
	} else {
		if (is_zero(c)) return is_equal(c99, 1.0);
		CLPR_delete_nl_occ(nl_eqn->params2->occ1);
		if (is_equal(c, 1.0)) 
			return CLPR_solve_eqn_1(nl_eqn->params1->id1, c99);
		CLPR_fatal("78912");
	}
}

CLPR_int CLPR_wakeup_pow1(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
NL_PARAMS1_ptr params1;
CLPR_int pos;
double c99;
	stats_count(210);
	nl_eqn = occ->nl_eqn;
	c99 = nl_eqn->params1->coef;
	CLPR_push_change_nl_eqn(SOLVER_POW1, (pos = occ->pos), nl_eqn);
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	if (pos == 2) {
		CLPR_delete_nl_occ(nl_eqn->params2->occ0);
		return CLPR_solve_eqn_1(nl_eqn->params1->id0, pow(c99, c));
	} else {
		CLPR_delete_nl_occ(nl_eqn->params2->occ2);
		return CLPR_solve_eqn_1(nl_eqn->params1->id2, log(c) / log(c99));
	}
}
			
CLPR_int CLPR_wakeup_pow2(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
CLPR_int pos;
double c99;
	stats_count(210);
	nl_eqn = occ->nl_eqn;
	c99 = nl_eqn->params1->coef;
	pos = occ->pos;
	if (pos == 0) {
		CLPR_push_change_nl_eqn(SOLVER_POW2, pos, nl_eqn);
		nl_eqn->type = SOLVER_POW02; /* special case */
		return TRUE;
	} else {
		CLPR_push_change_nl_eqn(SOLVER_POW2, pos, nl_eqn);
		nl_eqn->type = WOKEN;
		CLPR_nlin_count--;
		CLPR_delete_nl_occ(nl_eqn->params2->occ0);
		return CLPR_solve_eqn_1(nl_eqn->params1->id0, pow(c, c99));
	}
}
			
CLPR_int CLPR_wakeup_pow02(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
CLPR_int pos, v0, v2;
double c0, c2;
	stats_count(210);
	nl_eqn = occ->nl_eqn;
	CLPR_push_change_nl_eqn(SOLVER_POW02, (pos = occ->pos), nl_eqn);
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	v0 = nl_eqn->params1->id0;
	if (!CLPR_is_ground(v0, &c0)) CLPR_fatal("61821");
	v2 = nl_eqn->params1->id2;
	if (!CLPR_is_ground(v2, &c2)) CLPR_fatal("61821");
	return (is_equal(c0, pow(c, c2)));
}

CLPR_int CLPR_wakeup_min(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
CLPR_int pos;
	stats_count(211);
	nl_eqn = occ->nl_eqn;
	CLPR_push_change_nl_eqn(SOLVER_MIN, (pos = occ->pos), nl_eqn);
	if (pos == 1) nl_eqn->type = SOLVER_MIN1; 
	else nl_eqn->type = SOLVER_MIN2; 
	nl_eqn->params1->coef = c;
	return TRUE;
}

CLPR_int CLPR_wakeup_max(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
CLPR_int pos;
	stats_count(211);
	nl_eqn = occ->nl_eqn;
	CLPR_push_change_nl_eqn(SOLVER_MAX, (pos = occ->pos), nl_eqn);

	if (pos == 1) nl_eqn->type = SOLVER_MAX1; 
	else nl_eqn->type = SOLVER_MAX2; 
	nl_eqn->params1->coef = c;
	return TRUE;
}

CLPR_int CLPR_wakeup_min1(double c2, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
double c1;
	stats_count(211);
	nl_eqn = occ->nl_eqn;
	c1 = nl_eqn->params1->coef;
	CLPR_push_change_nl_eqn(SOLVER_MIN1, 1, nl_eqn);
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	return CLPR_solve_eqn_1(nl_eqn->params1->id0, (c1 < c2 ? c1 : c2));
}

CLPR_int CLPR_wakeup_max1(double c2, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
double c1;
	stats_count(211);
	nl_eqn = occ->nl_eqn;
	c1 = nl_eqn->params1->coef;
	CLPR_push_change_nl_eqn(SOLVER_MAX1, 1, nl_eqn);
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	return CLPR_solve_eqn_1(nl_eqn->params1->id0, (c1 > c2 ? c1 : c2));
}

CLPR_int CLPR_wakeup_min2(double c1, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
double c2;
	stats_count(211);
	nl_eqn = occ->nl_eqn;
	c2 = nl_eqn->params1->coef;
	CLPR_push_change_nl_eqn(SOLVER_MIN2, 2, nl_eqn);
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	return CLPR_solve_eqn_1(nl_eqn->params1->id0, (c1 < c2 ? c1 : c2));
}

CLPR_int CLPR_wakeup_max2(double c1, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
double c2;
	stats_count(211);
	nl_eqn = occ->nl_eqn;
	c2 = nl_eqn->params1->coef;
	CLPR_push_change_nl_eqn(SOLVER_MAX2, 2, nl_eqn);
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	return CLPR_solve_eqn_1(nl_eqn->params1->id0, (c1 > c2 ? c1 : c2));
}

CLPR_int CLPR_wakeup_abs(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
CLPR_int pos;
	stats_count(212);
	nl_eqn = occ->nl_eqn;
	CLPR_push_change_nl_eqn(SOLVER_ABS, (pos = occ->pos), nl_eqn);
	if (pos == 1) {
		nl_eqn->type = WOKEN;
		CLPR_nlin_count--;
		CLPR_delete_nl_occ(nl_eqn->params2->occ0);
		return CLPR_solve_eqn_1(nl_eqn->params1->id0, fabs(c));
	}
	if (c > CLPR_EPS) {
		nl_eqn->type = SOLVER_ABS0;
		nl_eqn->params1->coef = c;
		return TRUE;
	}
	if (is_zero(c)) {
		nl_eqn->type = WOKEN;
		CLPR_nlin_count--;
		CLPR_delete_nl_occ(nl_eqn->params2->occ1);
		return CLPR_solve_eqn_1(nl_eqn->params1->id1, 0.0);
	}
	return FALSE;
}

CLPR_int CLPR_wakeup_abs0(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
	stats_count(213);
	nl_eqn = occ->nl_eqn;
	CLPR_push_change_nl_eqn(SOLVER_ABS0, 1, nl_eqn);
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	return (is_equal(nl_eqn->params1->coef, fabs(c)));
}

CLPR_int CLPR_wakeup_sin(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
CLPR_int pos;
	stats_count(213);
	nl_eqn = occ->nl_eqn;
	CLPR_push_change_nl_eqn(SOLVER_SIN, (pos = occ->pos), nl_eqn);
	if (pos == 1) {
		nl_eqn->type = WOKEN;
		CLPR_nlin_count--;
		CLPR_delete_nl_occ(nl_eqn->params2->occ0);
		if (c != 0.0) return CLPR_solve_eqn_1(nl_eqn->params1->id0, sin(c));
		else return CLPR_solve_eqn_1(nl_eqn->params1->id0, 0.0);
	}
	if (c < -1.0 || c > 1.0) return FALSE;
	nl_eqn->type = SOLVER_SIN0;
	nl_eqn->params1->coef = c;
	return TRUE;
}

CLPR_int CLPR_wakeup_sin0(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
	stats_count(213);
	nl_eqn = occ->nl_eqn;
	CLPR_push_change_nl_eqn(SOLVER_SIN0, 1, nl_eqn);
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	return (is_equal(nl_eqn->params1->coef, sin(c)));
}

CLPR_int CLPR_wakeup_cos(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
CLPR_int pos;
	stats_count(213);
	nl_eqn = occ->nl_eqn;
	CLPR_push_change_nl_eqn(SOLVER_COS, (pos = occ->pos), nl_eqn);
	if (pos == 1) {
		nl_eqn->type = WOKEN;
		CLPR_nlin_count--;
		CLPR_delete_nl_occ(nl_eqn->params2->occ0);
		if (c != 0.0) return CLPR_solve_eqn_1(nl_eqn->params1->id0, cos(c));
		else return CLPR_solve_eqn_1(nl_eqn->params1->id0, 1.0);
	}
	if (c < -1.0 || c > 1.0) return FALSE;
	nl_eqn->type = SOLVER_COS0;
	nl_eqn->params1->coef = c;
	return TRUE;
}

CLPR_int CLPR_wakeup_cos0(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
	stats_count(213);
	nl_eqn = occ->nl_eqn;
	CLPR_push_change_nl_eqn(SOLVER_COS0, 1, nl_eqn);
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	return (is_equal(nl_eqn->params1->coef, cos(c)));
}

CLPR_int CLPR_wakeup_arcsin(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
NL_PARAMS1_ptr params1;
CLPR_int v0, v1;
	nl_eqn = occ->nl_eqn;
	params1 = nl_eqn->params1;
	v0 = params1->id0; 
	v1 = params1->id1; 
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	if (occ->pos == 0) {
		CLPR_push_change_nl_eqn(SOLVER_ARCSIN, 0, nl_eqn);
		CLPR_delete_nl_occ(nl_eqn->params2->occ1);
		if (c >= -PI_2_VAL && c <= PI_2_VAL) {
			if (c != 0.0) return CLPR_solve_eqn_1(v1, sin(c));
			else return CLPR_solve_eqn_1(v1, 0.0);
		} else if (c < 0 && c >= -PI_2_VAL-CLPR_EPS) return CLPR_solve_eqn_1(v1, -1.0);
		else if (c > 0 && c <= PI_2_VAL+CLPR_EPS) return CLPR_solve_eqn_1(v1, 1.0);
		else {
			CLPR_warn_err("Out of range value outside [-pi/2, pi/2] to arcsin");
			return FALSE; 
		}
	}
	CLPR_push_change_nl_eqn(SOLVER_ARCSIN, 1, nl_eqn);
	CLPR_delete_nl_occ(nl_eqn->params2->occ0);
	if (c <= 1 && c >= -1 ) {
		if (c != 0.0) return CLPR_solve_eqn_1(v0, asin(c));
		else return CLPR_solve_eqn_1(v0, 0.0);
	}
	else if (c > 0 && c <= 1+CLPR_EPS) return CLPR_solve_eqn_1(v0, PI_2_VAL);
	else if (c < 0 && c >= -1-CLPR_EPS) return CLPR_solve_eqn_1(v0, -PI_2_VAL); 
	else {
		CLPR_warn_err("Out of range value outside [-1, 1] to arcsin");
		return FALSE; 
	}
}

CLPR_int CLPR_wakeup_arccos(double c, NL_OCC_ptr occ)
{
NL_EQN_ptr nl_eqn;
NL_PARAMS1_ptr params1;
CLPR_int v0, v1;
	nl_eqn = occ->nl_eqn;
	params1 = nl_eqn->params1;
	v0 = params1->id0; 
	v1 = params1->id1; 
	nl_eqn->type = WOKEN;
	CLPR_nlin_count--;
	if (occ->pos == 0) {
		CLPR_push_change_nl_eqn(SOLVER_ARCCOS, 0, nl_eqn);
		CLPR_delete_nl_occ(nl_eqn->params2->occ1);
		if (c >= 0 && c <= PI_VAL) {
			if (c != 0.0) return CLPR_solve_eqn_1(v1, cos(c));
			else return CLPR_solve_eqn_1(v1, 1.0);
		} else if (c < 0 && c >= -CLPR_EPS) return CLPR_solve_eqn_1(v1, 1.0);
		else if (c > 0 && c <= PI_VAL+CLPR_EPS) return CLPR_solve_eqn_1(v1, -1.0);
		else {
			CLPR_warn_err("Out of range value outside [0, pi] to arccos");
			return FALSE; 
		}
	}
	CLPR_push_change_nl_eqn(SOLVER_ARCCOS, 1, nl_eqn);
	CLPR_delete_nl_occ(nl_eqn->params2->occ0);
	if (c <= 1 && c >= -1 ) {
		if (c != 1.0) return CLPR_solve_eqn_1(v0, acos(c));
		else return CLPR_solve_eqn_1(v0, 0.0);
	} else if (c > 0 && c <= 1+CLPR_EPS) return CLPR_solve_eqn_1(v0, 0.0);
	else if (c < 0 && c >= -1-CLPR_EPS) return CLPR_solve_eqn_1(v0, PI_VAL); 
	else {
		CLPR_warn_err("Out of range value outside [-1, 1] to arccos");
		return FALSE; 
	}
}


