/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: table.c                                              *
 ***************************************************************/

#include <string.h>
#include <setjmp.h>

#include "standard.h"
#include "config.h"
#include "parse.h"
#include "symbols.h"
#include "lib.h"
#include "compile.h"
#include "emul.h"
#include "main.h"

#define MAXHASH_FUNC (1999)
#define MAXHASH_VAR (1999)
#define DYNAMIC_HASHTABLE (1)
#define INIT_HASH_SPACE	(2000)

struct HASHNODE *CLPR_hash_func_tab[MAXHASH_FUNC];
struct HASHNODE *CLPR_hash_var_tab[MAXHASH_VAR];
struct HASHNODE *CLPR_tok_hnode; /* current hash node returned by CLPR_lookup_name() */
static struct HASHNODE hash_space[INIT_HASH_SPACE];
struct HASHNODE *CLPR_hash_base = hash_space;

static CLPR_int hash_size = 0;
CLPR_int CLPR_name_hash = 0;
CLPR_int CLPR_hash_max = 0, CLPR_hash_cnt = 0, CLPR_hash_cnt1 = 0, CLPR_hash_cnt2 = 0, CLPR_hash_cnt3 = 0;
CLPR_int CLPR_hash_cnt4 = 0, CLPR_hash_cnt5 = 0, CLPR_hash_cnt6 = 0, CLPR_hash_cnt7 = 0; 	
CLPR_int CLPR_cur_func_id = INIT_FUNC_ID; /* lower bound for user-def vals */

PTERM *CLPR_new_pterm(void);
CLPR_int CLPR_hashbuiltina(char *s, CLPR_int n);

extern char CLPR_tok_c;
extern CLPR_int CLPR_tok_type;
extern PTERM* CLPR_tok_ptr;
extern char CLPR_ch;
extern char *CLPR_cur_filename;
extern FILE *CLPR_cur_stream;
extern FILE *CLPR_outfile;
extern CLPR_int CLPR_debug_flg;
extern CLPR_int CLPR_comma_flg, CLPR_bar_flg;
extern jmp_buf CLPR_err_readterm;
extern PTERM *CLPR_null_list;
extern char *CLPR_cons_str;
extern FUNCTREC CLPR_functors[];

extern char *CLPR_emalloc(CLPR_int n);
extern char *str_space();

/*---------------------------------------------------------------------------*/


extern CLPR_int CLPR_sum_str (register char *s);
void CLPR_numfunc (CLPR_int val, CLPR_int arity, char *s, CLPR_int special);

CLPR_int CLPR_hash(char *s, CLPR_int siz)
{
register CLPR_unsigned h;
	for (h = 0; *s; s++) h = ((h * h) << 2) + *s;
	if ((CLPR_int) h >= 0) return (CLPR_int) (h % siz);
	else return (-(CLPR_int) h) % siz;
}

struct HASHNODE *CLPR_new_hashnode(void)
{
struct HASHNODE *p;
HASH_AGAIN:
	if (hash_size < INIT_HASH_SPACE) p = &CLPR_hash_base[hash_size++];
	else {
#if DYNAMIC_HASHTABLE
		CLPR_hash_base = 
			(struct HASHNODE *) 
				CLPR_emalloc(sizeof(struct HASHNODE)*INIT_HASH_SPACE);
		hash_size = 0;
		goto HASH_AGAIN;
#else
		CLPR_fatal("no more HASHNODE space");
#endif
	}
	p->next = p->link = NULL;
	p->quick_hash = 0;
	p->t = NULL;
	p->op = NULL;
	p->extra = (char) FALSE;
	return(p);
}

/*	Maximum number of characters of string space	*/
#define MAX_STRTABLE	(4000) 
#define DYNAMIC_STRTABLE 1
/*	High address of string space	*/
static char str_table[MAX_STRTABLE];
static char *max_str_table= &str_table[MAX_STRTABLE-2];
/*	Current empty location in string table	*/
static char *str_ptr=str_table;

/*
 *	Puts s into the string table and returns the string table address
 */
char *CLPR_add_string(char *s)
{
	char *start;
	register char *s1;

PUT_STRING:
	s1 = s;
	start = str_ptr;
	while (*s1 && str_ptr < max_str_table) 
		*str_ptr++ = *s1++;
#if !DYNAMIC_STRTABLE
	if (*s1 && str_ptr >= max_str_table) 
		CLPR_fatal("internal error: ran out of string space");
#else
	if (*s1 && str_ptr >= max_str_table) {
		str_ptr = CLPR_emalloc(MAX_STRTABLE);
		max_str_table = str_ptr + MAX_STRTABLE - 2;
		goto PUT_STRING;
	}
#endif
		
	*str_ptr++ ='\0';
	return(start);
}

struct HASHNODE *CLPR_insert_name(struct HASHNODE **hash_tab, char *sym, CLPR_int quick_cmp, CLPR_int size)
{
CLPR_int h;
struct HASHNODE *cur, *p1;
	if ((cur = hash_tab[h=CLPR_hash(sym, size)]) != NULL)
		for (; cur != NULL; cur = cur->next) {
			if (cur->quick_hash == quick_cmp && !strcmp(cur->s, sym)) {
				return cur;
			}
		}
	cur = CLPR_new_hashnode(); 
	cur->quick_hash = quick_cmp;
	cur->s = CLPR_add_string(sym);
	cur->val = CLPR_cur_func_id++;
	if (hash_tab[h] == NULL) hash_tab[h] = cur;
	else {
	  	cur->next = hash_tab[h];
		hash_tab[h] = cur; 
	}
	return cur;
}

PTERM *CLPR_lookup_name(char *sym, CLPR_int quick_cmp)
{
struct HASHNODE *hnode;
PTERM *t;
	hnode = CLPR_insert_name(CLPR_hash_func_tab, sym, quick_cmp, MAXHASH_FUNC);
	t = CLPR_new_pterm();
	t->str = hnode->s;
	t->val = hnode->val;
	t->ty = FUNC;
	t->narg = -1;
	CLPR_tok_hnode = hnode;
	return t;
}

CLPR_int CLPR_no_var_pterm; /* number of useful variables in pterm */
static CLPR_int var_cnt = 0;	/* max variables in current term */
static struct HASHNODE *used_vars[MAXVAR]; /* used hnodes in current term */
static CLPR_int used_var_cnt = 0; /* unique value for variable names */
static CLPR_int var_hash_id = 1;

void CLPR_zero_vars(void) /* start new terms with no variables seen yet	*/
{
CLPR_int i;
	CLPR_no_var_pterm = var_cnt = 0;
	for (i = 0; i < used_var_cnt; i++) used_vars[i]->special = -1;
	used_var_cnt = 0;
}

static struct HASHNODE *insert_var(struct HASHNODE **hash_tab, char *sym, CLPR_int quick_cmp, CLPR_int size)
{
CLPR_int h;
struct HASHNODE *cur;
	if ((cur = hash_tab[h=CLPR_hash(sym, size)]) != NULL)
		for (; cur != NULL; cur = cur->next) {
			if (cur->quick_hash == quick_cmp && !strcmp(cur->s, sym)) {
				return cur;
			}
		}
	cur = CLPR_new_hashnode();
	cur->quick_hash = quick_cmp;
	cur->s = CLPR_add_string(sym);
	cur->val = var_hash_id++;
	cur->special = -1;
	if (hash_tab[h] == NULL) hash_tab[h] = cur;
	else {
		cur->next = hash_tab[h];
		hash_tab[h] = cur;
	}
	return cur;
}

/*
 *	Lookup the variable var and return a term to it
 *	variables which have been seen before get the same value
 */
PTERM *CLPR_lookup_var(char *var, CLPR_int quick_cmp)
{
CLPR_int i;
PTERM *p, *t;
struct HASHNODE *hnode;
CLPR_int hval, var_index;
	hnode = insert_var(CLPR_hash_var_tab, var, quick_cmp, MAXHASH_VAR);
	t = CLPR_new_pterm();
	t->ty = VAR;
	t->str = hnode->s;
	if (*var != '_' || var[1])    /* fast strcmp with "_" */
		if (hnode->special < 0) { /* first instance of var */
			CLPR_no_var_pterm++;
			hnode->var_index = t->val = var_cnt++;
			hnode->special = 1;
			used_vars[used_var_cnt++] = hnode;
			t->new_var = FIRST_VAR;
		} else {
			t->val = hnode->var_index;
			t->new_var = SEEN_VAR;
		}
	else {
		t->val = var_cnt++;
		t->new_var = ANONYMOUS_VAR;
	}
	return t;
}

void CLPR_setconst(char *sym, double num) /* set #<CONST> to predefined value */
          
           
{
CLPR_int i;
PTERM *p, *t;
struct HASHNODE *hnode;
CLPR_int hval, var_index;
	hnode = CLPR_insert_name(CLPR_hash_func_tab, sym, CLPR_sum_str(sym), MAXHASH_FUNC);
	hnode->rval = num;
	hnode->extra = REAL;
}

PTERM *CLPR_lookup_const(char *sym, CLPR_int quick_cmp)
{
PTERM *p;
struct HASHNODE *hnode;
	hnode = CLPR_insert_name(CLPR_hash_func_tab, sym, quick_cmp, MAXHASH_FUNC);
	if (hnode->extra != REAL) return (PTERM *) NULL;
	p = CLPR_new_pterm();
	p->ty = REAL;
	p->str = "$real";
	p->rval = hnode->rval;
	return p;
}

PTERM *CLPR_new_real(double val)
{
PTERM *t;
	t = CLPR_new_pterm();
	t->ty = REAL;
	t->rval = val;
	t->str = "$real";
	return(t);
}

PTERM *CLPR_new_str(char *s)
{
PTERM *t;
	t = CLPR_new_pterm();
	t->ty = STR;
	t->str = CLPR_add_string(s);
	return t;
}

PTERM *CLPR_new_pt_emptylist(void)
{
PTERM *t;
  	t = CLPR_new_pterm();
 	t->str = CLPR_null_list->str;
  	t->val = CLPR_null_list->val;
  	t->narg = 0;
  	t->ty = FUNC;
  	return t;
}

/*
 * 	add a new functor called "name" with "narg" arguments of type "type"
 *	with value val, used to predefine functors
 *	Not allowed to redefine a name with the same arity
 */
void CLPR_add_func(char *name, CLPR_int narg, CLPR_int val)
{
CLPR_int h, size, quick_cmp, found=FALSE;
struct HASHNODE *cur;
PTERM *p, *t;
struct OPNODE *op;
	quick_cmp = CLPR_sum_str(name);
	if ((cur = CLPR_hash_func_tab[h=CLPR_hash(name, MAXHASH_FUNC)]) != NULL)
		for (; cur != NULL; cur = cur->next) 
			if (cur->quick_hash == quick_cmp && !strcmp(cur->s, name)) {
				found = TRUE;
				break;
			}

	if (!found) {
		cur = CLPR_new_hashnode();
		cur->quick_hash = quick_cmp;
		cur->s = CLPR_add_string(name);
		cur->val = val;
		if (CLPR_hash_func_tab[h] == NULL)
			CLPR_hash_func_tab[h] = cur;
		else {
			cur->next = CLPR_hash_func_tab[h];
			CLPR_hash_func_tab[h] = cur;
		}
	}

	for (p = cur->t; p; p = p->next)
		if (p->narg == narg)
			if (p->val == val)
				CLPR_fatal("redefining initialised functor as the same");
			else {
				printf("name %s narg %d val %d ty %d\n", 
					name, p->narg, p->val, p->ty);
				CLPR_fatal("redefining initialised functor differently");
			}
	t = CLPR_new_pterm();
	t->narg = narg;
	t->val = val;
	t->next = cur->t;
	cur->t = t;

	if (!(op = cur->op)) CLPR_numfunc(val, narg, name, NONE);
	else if (narg == 2) {
		if (op->ops[BINARY].pri >= 0) CLPR_numfunc(val, narg, name, BINARY);
		else CLPR_fatal("CLPR_add_func special wrong");
	} else if (narg == 1) {
		if (op->ops[PREFIX].pri >= 0) CLPR_numfunc(val, narg, name, PREFIX);
		else if (op->ops[POSTFIX].pri >= 0) CLPR_numfunc(val, narg, name, POSTFIX);
		else CLPR_fatal("CLPR_add_func special wrong");
	} 
}

/*
 *	Add a new operator "sym" of type "ty" and priority "pri"
 *	left operands have priority <= left
 *	similarly for right operands
 */
void CLPR_addop(char *sym, CLPR_int ty, CLPR_int pri, CLPR_int left, CLPR_int right, CLPR_int readonly)
{
struct OPNODE *op, *opnode;
struct HASHNODE *hnode;
PTERM *t;
CLPR_int i, n = 0;
	if (ty == BINARY) n = 2;
	else if (ty == POSTFIX || ty == PREFIX) n = 1;
	else CLPR_fatal("21333");
	if (pri < 0) {
		CLPR_warning("priority of operator %s is negative", sym);
		return;
	}
	hnode = CLPR_insert_name(CLPR_hash_func_tab, sym, CLPR_sum_str(sym), MAXHASH_FUNC);
	if (!(op = hnode->op)) {
		opnode = (struct OPNODE *) CLPR_emalloc(sizeof(struct OPNODE));
		for (i=0; i < N_OP_KIND; i++)
			opnode->ops[i].pri = 
				opnode->ops[i].left =
					opnode->ops[PREFIX].right = -1;
		hnode->op = opnode;
		op = opnode;
	} else if (readonly && op->ops[ty].pri >= 0) {
		CLPR_warning("attempt to redefine %s (ignored)", sym);
		return;
	}
	i = CLPR_hashbuiltina(sym, n);
	if (pri == 0) {
		pri = left = right = -1;
		CLPR_functors[i].special = NONE;
	} else CLPR_numfunc(i, n, sym, ty);
	op->ops[ty].pri = pri;
	op->ops[ty].left = left;
	op->ops[ty].right = right;
}

CLPR_int CLPR_do_op(char *sym, char *style, CLPR_int pri)
{
	if (strcmp(style, "xfx") == 0)
		CLPR_addop(sym, BINARY, pri, pri - 1, pri - 1, FALSE);
	else if (strcmp(style, "xfy") == 0)
		CLPR_addop(sym, BINARY, pri, pri - 1, pri, FALSE);
	else if (strcmp(style, "yfx") == 0)
		CLPR_addop(sym, BINARY, pri, pri, pri - 1, FALSE);
	else if (strcmp(style, "yfy") == 0)
		CLPR_addop(sym, BINARY, pri, pri, pri, FALSE);
	else if (strcmp(style, "fx") == 0)
		CLPR_addop(sym, PREFIX, pri, 0, pri - 1, FALSE);
	else if (strcmp(style, "fy") == 0)
		CLPR_addop(sym, PREFIX, pri, 0, pri, FALSE);
	else if (strcmp(style, "xf") == 0)
        CLPR_addop(sym, POSTFIX, pri, pri - 1, 0, FALSE);
	else if (strcmp(style, "yf") == 0)
		CLPR_addop(sym, POSTFIX, pri, pri, 0, FALSE);
	else return FALSE;
	return TRUE;
}

/*
 *	get information about operator "sym" of type "ty"
 *	returns TRUE if operator exists and sets pri, left, right
 */
CLPR_int CLPR_is_op_hnode(struct HASHNODE *hnode, CLPR_int ty, CLPR_int *pri, CLPR_int *left, CLPR_int *right)
{
struct OPNODE *op;
	if (!(op = hnode->op)) return FALSE;
#ifdef ENABLE_HAT
	if (op->ops[ty].pri >= 0) {
#else
	if (op->ops[ty].pri >= 0 && !is_val_hat(hnode->val)) {
#endif
		*pri = op->ops[ty].pri;
		*left = op->ops[ty].left;
		*right = op->ops[ty].right;
		return(TRUE);
	}
	return(FALSE);
}

struct HASHNODE *CLPR_has_op(char *sym)
{
	struct HASHNODE *hnode;

	hnode = CLPR_insert_name(CLPR_hash_func_tab, sym, CLPR_sum_str(sym), MAXHASH_FUNC);
	if (!hnode->op) return NULL;
	else return hnode;
}

CLPR_int CLPR_get_op_priority(char *name, CLPR_int narg, CLPR_int *left, CLPR_int *right)
{
struct HASHNODE *hnode;
struct OPNODE *op;
CLPR_int i;

	hnode = CLPR_insert_name(CLPR_hash_func_tab, name, CLPR_sum_str(name), MAXHASH_FUNC);
	if (!(op = hnode->op) || narg < 1 || narg > 2) return 0;
	if (narg == 2) i = BINARY;
	else if (op->ops[PREFIX].pri) i = PREFIX;
	else i = POSTFIX;
	*left = op->ops[i].left;
	*right = op->ops[i].right;
	return op->ops[i].pri;
}
	
CLPR_int CLPR_is_nfix_hnode(struct HASHNODE *hnode)
{
  return FALSE;
}

PTERM *CLPR_narg_pterm(struct HASHNODE *hnode, PTERM *t, CLPR_int narg) /* correct val and narg? */
                       
         
         
{
PTERM *p, *p1;
CLPR_int val, found = FALSE;
struct OPNODE *op;
	if (t->ty != FUNC) CLPR_fatal("57233");
	t->narg = narg;
	for (p1 = hnode->t; p1; p1 = p1->next)
		if (p1->narg == narg) {
			found = TRUE;
			break;
		}
	if (found) t->val = p1->val;
	else {
		p = CLPR_new_pterm();
		p->narg = narg;
		p->next = hnode->t;
		hnode->t = p;
		if (hnode->t) t->val = p->val = CLPR_cur_func_id++;
		else t->val = p->val = hnode->val;
	}
	if (!(op=hnode->op)) CLPR_numfunc(t->val, narg, t->str, NONE);
	else if (narg == 2) {
		if (op->ops[BINARY].pri >= 0) CLPR_numfunc(t->val, narg, t->str, BINARY);
		else CLPR_numfunc(t->val, narg, t->str, NONE);
	} else if (narg == 1) {
		if (op->ops[PREFIX].pri >= 0) CLPR_numfunc(t->val, narg, t->str, PREFIX);
		else if (op->ops[POSTFIX].pri >= 0) 
			CLPR_numfunc(t->val, narg, t->str, POSTFIX);
		else CLPR_numfunc(t->val, narg, t->str, NONE);
	} else CLPR_numfunc(t->val, narg, t->str, NONE); 

	return t;
}

PTERM *CLPR_func_narg_pterm(PTERM *t, CLPR_int narg)
{
struct HASHNODE *hnode;
	hnode = CLPR_insert_name(CLPR_hash_func_tab, t->str, CLPR_sum_str(t->str), MAXHASH_FUNC);
	return(CLPR_narg_pterm(hnode, t, narg));
}

CLPR_int CLPR_hashbuiltina(char *s, CLPR_int n)
{
PTERM *t;
struct HASHNODE *hnode;
	hnode = CLPR_insert_name(CLPR_hash_func_tab, s, CLPR_sum_str(s), MAXHASH_FUNC);
	t = CLPR_new_pterm();
	t->ty = FUNC;
	t->str = s;
	CLPR_narg_pterm(hnode, t, n);
	return t->val;
}

CLPR_int CLPR_getmaxfunc(void)
{
	return (CLPR_cur_func_id + 1);
}


#define PTERM_BLOCK	(80)

static PTERM *free_list = (PTERM *) NULL;
static char *start_term = (char *) NULL;
static CLPR_int unused = 0;

PTERM *CLPR_new_pterm(void)
{
PTERM *t;
	if (free_list) {
		t = free_list;
		free_list = t->next;
	} else if (unused > 0) {
		unused--;
		t = (PTERM *) start_term;
		start_term += sizeof(PTERM);
	} else {
		unused = PTERM_BLOCK;
		start_term = CLPR_emalloc(PTERM_BLOCK * sizeof(PTERM));
		unused--;
		t = (PTERM *) start_term;
		start_term += sizeof(PTERM);
	}
	t->next = t->first = NULL;
	return(t);
}

void CLPR_free_pterm(PTERM *p)
{
PTERM *arg, *arg99;
	if (p) {
		if (is_p_functor(p) && p->narg > 0) {
			for (arg = p->first; arg; arg = arg99) {
				arg99 = arg->next;
				CLPR_free_pterm(arg);
			}
		}
		p->next = free_list;
		free_list = p;
	}
}

void CLPR_numfunc(CLPR_int val, CLPR_int arity, char *s, CLPR_int special)
{
	CLPR_functors[val].arity = arity;
	CLPR_functors[val].pname = s;
	CLPR_functors[val].special = special;
}

PTERM *CLPR_new_pt_var(CLPR_int val, char *s)
{
	PTERM *t;

	if (val < 0) CLPR_fatal("77319");
	t = CLPR_new_pterm();
	t->ty = VAR;
	t->val = val;
	t->str = s;
	return t;
}

PTERM *CLPR_new_p_func2(CLPR_int val, PTERM *arg1, PTERM *arg2)
{
	PTERM *nt;
	nt = CLPR_new_pterm();
	nt->ty = FUNC;
	nt->narg = 2;
	nt->str = string_of(val);
	nt->val = val;
	nt->first = arg1;
	arg1->next = arg2;
	return nt;
}

struct HASHNODE *CLPR_get_similar_func(char *name)
{
CLPR_int h, size, quick_cmp, found=FALSE;
struct HASHNODE *cur;
PTERM *p, *t;
struct OPNODE *op;

	quick_cmp = CLPR_sum_str(name);
	if ((cur = CLPR_hash_func_tab[h=CLPR_hash(name, MAXHASH_FUNC)]) != NULL)
		for (; cur != NULL; cur = cur->next) 
			if (cur->quick_hash == quick_cmp && !strcmp(cur->s, name)) {
				return cur;
			}
	return NULL;
}

struct HASHNODE *CLPR_insert_func_htable(char *s)
{
	return CLPR_insert_name(CLPR_hash_func_tab, s, CLPR_sum_str(s), MAXHASH_FUNC);
}
