%-----------------------------------------------------------------------------%
% Copyright (C) 1996-1997 University of Melbourne.
% This file may only be copied under the terms of the GNU Library General
% Public License - see the file COPYING.LIB in the Mercury distribution.
%-----------------------------------------------------------------------------%
%
% float_cfloat - part of the interface with the CLP(R) constraint solver. 
%
%	This module defines the operators which mix floats and cfloats.
%	(The reason this is in a different module than the stuff in `cfloat'
%	is because Mercury doesn't support overloading operators with
%	the same module, name, and arity.)
%
% Main author: fjh.
% Stability: low.
%
%-----------------------------------------------------------------------------%

:- module float_cfloat.

:- interface.
:- import_module cfloat, float.

%-----------------------------------------------------------------------------%

	% cfloat-float equality
	%
:- pred ==(float, cfloat).
:- mode ==(in, ca) is semidet.
:- mode ==(in, co) is det.

	% cfloat-float disequality
	%
:- pred \==(float, cfloat).
:- mode \==(in, ca) is semidet.
:- mode \==(in, co) is det.

	% addition
	%
:- func '+'(float, cfloat) = cfloat.
:- mode '+'(in, ca) = ca is semidet.
:- mode '+'(in, co) = ca is det.
:- mode '+'(in, ca) = co is det.
:- mode '+'(in, co) = co is det.

	% subtraction
	%
:- func '-'(float, cfloat) = cfloat.
:- mode '-'(in, ca) = ca is semidet.
:- mode '-'(in, co) = ca is det.
:- mode '-'(in, ca) = co is det.
:- mode '-'(in, co) = co is det.

	% multiplication
	%
:- func '*'(float, cfloat) = cfloat.
:- mode '*'(in, ca) = ca is semidet.
:- mode '*'(in, co) = ca is semidet.	% semidet since eg. X*0=1 fails
:- mode '*'(in, ca) = co is det.
:- mode '*'(in, co) = co is det.

	% division
	%
:- func '/'(float, cfloat) = cfloat.
:- mode '/'(in, ca) = ca is semidet.
:- mode '/'(in, co) = ca is semidet.	% semidet since eg. X/0=1 fails
:- mode '/'(in, ca) = co is semidet.	% semidet since eg. 1/0=Y fails
:- mode '/'(in, co) = co is semidet.	% XXX really det

	% X > Y
	%
:- pred '>'(float, cfloat).
:- mode '>'(in, ca) is semidet.
:- mode '>'(in, co) is det.

	% X >= Y
	%
:- pred '>='(float, cfloat).
:- mode '>='(in, ca) is semidet.
:- mode '>='(in, co) is det.

	% X < Y
	%
:- pred '<'(float, cfloat).
:- mode '<'(in, ca) is semidet.
:- mode '<'(in, co) is det.

	% X =< Y
	%
:- pred '=<'(float, cfloat).
:- mode '=<'(in, ca) is semidet.
:- mode '=<'(in, co) is det.

%-----------------------------------------------------------------------------%
%-----------------------------------------------------------------------------%

:- implementation.

X == Y :- cfloat__eq_float(Y, X).
X \== Y :- cfloat__diseq_float(Y, X).

X + Y = Z :- cfloat__plus_float(Y, X, Z).
X - Y = Z :- X == Cfloat_X, Cfloat_X - Y = Z.	% not maximally efficient
X * Y = Z :- cfloat__mult_float(Y, X, Z).
X / Y = Z :- X == Cfloat_X, Cfloat_X / Y = Z.	% not maximally efficient

X > Y :- cfloat__gt_float(Y, X).
X >= Y :- cfloat__ge_float(Y, X).
X < Y :- cfloat__lt_float(Y, X).
X =< Y :- cfloat__le_float(Y, X).

%-----------------------------------------------------------------------------%
:- end_module float_cfloat.
%-----------------------------------------------------------------------------%
