:- module laplace.

:- interface.
:- import_module io.

:- pred main(io__state::di, io__state::uo) is det.

:- implementation.
:- import_module cfloat, cfloat_float, float_cfloat, dump.
:- import_module list, int, require, string.

% Solve the Dirichlet problem for Laplace's equation using
% Leibman's five-point finit-differenc approximation. 

:- type vector == list(cfloat).
:- type matrix == list(vector).

:- inst vec_of_constrained == list_of_constrained.
:- mode vec_co == list_co.
:- mode vec_ca == list_ca.
:- mode vec_cg == list_cg.

:- inst mat_of_constrained == list(vec_of_constrained).
:- mode mat_co == free >> mat_of_constrained.
:- mode mat_ca == mat_of_constrained >> mat_of_constrained.
:- mode mat_cg == mat_of_constrained >> ground.

:- pred laplace(matrix::mat_ca) is semidet.
laplace([_, _]).
laplace([H1, H2, H3|T]):-
	laplace_vec(H1, H2, H3), 
	laplace([H2, H3|T]).

:- pred laplace_vec(vector::vec_ca, vector::vec_ca, vector::vec_ca) is semidet.
laplace_vec([_, _], [_, _], [_, _]).
laplace_vec([_TL, T, TR|T1], [ML, M, MR|T2], [_BL, B, BR|T3]):-
	B + T + ML + MR = 4.0 * M,
	% dump,
	laplace_vec([T, TR|T1], [M, MR|T2], [B, BR|T3]).

:- pred printmat(matrix::mat_ca, io__state::di, io__state::uo) is det.
printmat([]) 	--> 
	[].
printmat([H|T]) -->
	printvec(H), 
	printmat(T).

:- pred printvec(vector::vec_ca, io__state::di, io__state::uo) is det.
printvec([]) 	-->
	io__write_string("\n").
printvec([H|T]) -->
	{ cfloat__get_val(H, HVal) },
	{ string__format("%7.2f", [f(HVal)], HString) },
	io__write_string(HString), 
	printvec(T).

:- pred init_all(vector::vec_co, float::in) is det.
init_all(Vec, Val) :-
	init_all_2(Vec, Val, 11).

:- pred init_all_2(vector::vec_co, float::in, int::in) is det.
init_all_2(Vec, Val, N) :-	
	(
		N = 0
	->
		Vec = []
	;
		Vec = [H|T],
		H == Val,
		init_all_2(T, Val, N - 1)
	).

:- pred hundreds_at_ends(vector::vec_co) is det.
hundreds_at_ends([V|Vs]) :-	
	V == 100.0,
	hundred_2(Vs, 10).

:- pred hundred_2(vector::vec_co, int::in) is det.
hundred_2(Vec, N) :-
	(
		N = 1
	->
		Vec = [V],
		V == 100.0
	;
		Vec = [V|Vs],
		cfloat__init(V),
		hundred_2(Vs, N - 1)
	).

main -->
	{ init_all(Vec1, 0.0) },
	{ hundreds_at_ends(Vec2) },
	{ hundreds_at_ends(Vec3) },
	{ hundreds_at_ends(Vec4) },
	{ hundreds_at_ends(Vec5) },
	{ hundreds_at_ends(Vec6) },
	{ hundreds_at_ends(Vec7) },
	{ hundreds_at_ends(Vec8) },
	{ hundreds_at_ends(Vec9) },
	{ hundreds_at_ends(Vec10) },
	{ init_all(Vec11, 100.0) },
	{ X = 	[Vec1, Vec2, Vec3, Vec4, Vec5, Vec6, 
		Vec7, Vec8, Vec9, Vec10, Vec11] },

	    % [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], 
	    % [100, _, _, _, _, _, _, _, _, _, 100], 
	    % [100, _, _, _, _, _, _, _, _, _, 100], 
	    % [100, _, _, _, _, _, _, _, _, _, 100], 
	    % [100, _, _, _, _, _, _, _, _, _, 100], 
	    % [100, _, _, _, _, _, _, _, _, _, 100], 
	    % [100, _, _, _, _, _, _, _, _, _, 100], 
	    % [100, _, _, _, _, _, _, _, _, _, 100], 
	    % [100, _, _, _, _, _, _, _, _, _, 100], 
	    % [100, _, _, _, _, _, _, _, _, _, 100], 
	    % [100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100]

	( { laplace(X) } ->
		io__write_string("\n"),
		printmat(X)
	;
		{ error("laplace/1 failed") }
	).

% Answer:
%    0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00
%  100.00  51.11  32.52  24.56  21.11  20.12  21.11  24.56  32.52  51.11 100.00
%  100.00  71.91  54.41  44.63  39.74  38.26  39.74  44.63  54.41  71.91 100.00
%  100.00  82.12  68.59  59.80  54.97  53.44  54.97  59.80  68.59  82.12 100.00
%  100.00  87.97  78.03  71.00  66.90  65.56  66.90  71.00  78.03  87.97 100.00
%  100.00  91.71  84.58  79.28  76.07  75.00  76.07  79.28  84.58  91.71 100.00
%  100.00  94.30  89.29  85.47  83.10  82.30  83.10  85.47  89.29  94.30 100.00
%  100.00  96.20  92.82  90.20  88.56  88.00  88.56  90.20  92.82  96.20 100.00
%  100.00  97.67  95.59  93.96  92.93  92.58  92.93  93.96  95.59  97.67 100.00
%  100.00  98.89  97.90  97.12  96.63  96.46  96.63  97.12  97.90  98.89 100.00
%  100.00 100.00 100.00 100.00 100.00 100.00 100.00 100.00 100.00 100.00 100.00

:- pred main2(io__state::di, io__state::uo) is cc_multi.
main2 --> dump_one_solution(go2).

:- pred go2(list(cfloat)::list_co, list(string)::out) is semidet.
go2(		[B31,   M32,   M33,   B34,   B42,   B43,   
		 B12,   B13,   B21,   M22,   M23,   B24],
     		["B31", "M32", "M33", "B34", "B42", "B43",
     		 "B12", "B13", "B21", "M22", "M23", "B24"]) :-
	init(B11), init(B12), init(B13), init(B14),
	init(B21), init(M22), init(M23), init(B24),
	init(B31), init(M32), init(M33), init(B34),
	init(B41), init(B42), init(B43), init(B44),
	laplace([
		[B11, B12, B13, B14], 
		[B21, M22, M23, B24], 
		[B31, M32, M33, B34], 
		[B41, B42, B43, B44]
	]).

/*
go2 :-
	laplace([
		[_B11, B12, B13, _B14], 
		[B21, M22, M23, B24], 
		[B31, M32, M33, B34], 
		[_B41, B42, B43, _B44]
	]).
	dump([B31, M32, M33, B34, B42, B43, B12, B13, B21, M22, M23, B24]).
*/

% Answer:
%  B12 =  -B21 - 4*B31 + 16*M32 - 8*M33 + B34 - 4*B42 + B43
%  B13 =  -B24 + B31 - 8*M32 + 16*M33 - 4*B34 + B42 - 4*B43
%  M22 =  -B31 + 4*M32 - M33 - B42
%  M23 =  -M32 + 4*M33 - B34 - B43

