/***************************************************************
 *  CLP(R)                                                     *
 *  (c) Copyright International Business Machines Corporation  *
 *  (1989) 1991, 1992 All Rights Reserved                      *
 *  Version: 1.2                                               *
 *  File: main.c                                               *
 ***************************************************************

 RESEARCH SOFTWARE DISCLAIMER

    As unestablished, research software, this program is provided free of 
    charge on an "as is" basis without warranty of any kind, either 
    expressed or implied, including but not limited to implied warranties 
    of merchantability and fitness for a particular purpose.  IBM does not 
    warrant that the functions contained in this program will meet the user's
    requirements or that the operation of this program will be uninterrupted 
    or error-free.  Acceptance and use of this program constitutes the user's
    understanding that he will have no recourse to IBM for any actual or 
    consequential damages, including, but not limited to, lost profits or 
    savings, arising out of the use or inability to use this program.  Even 
    if the user informs IBM of the possibility of such damages, IBM expects 
    the user of this program to accept the risk of any harm arising out of 
    the use of this program, or the user shall not attempt to use this 
    program for any purpose.

 USER AGREEMENT

     BY ACCEPTANCE AND USE OF THIS EXPERIMENTAL PROGRAM
     THE USER AGREES TO THE FOLLOWING:

     a.  This program is provided for the user's personal, non-commercial, 
         experimental use and the user is granted permission to copy this
         program to the extent reasonably required for such use.

     b.  All title, ownership and rights to this program and any copies 
         remain with IBM, irrespective of the ownership of the media on 
         which the program resides.

     c.  The user is permitted to create derivative works to this program.  
         However, all copies of the program and its derivative works must
         contain the IBM copyright notice, the UNESTABLISHED SOFTWARE 
         DISCLAIMER and this USER AGREEMENT.

     d.  By furnishing this program to the user, IBM does NOT grant either 
         directly or by implication, estoppel, or otherwise any license
         under any patents, patent applications, trademarks, copyrights or 
         other rights belonging to IBM or to any third party, except as 
         expressly provided herein.

     e.  The user understands and agrees that this program and any derivative
         works are to be used solely for experimental uses and are not to be
         sold, distributed to a commercial organization, or be commercially 
         exploited in any manner.

     f.  IBM requests that the user supply to IBM a copy of any changes, 
         enhancements, or derivative works which the user may create.  The 
         user grants IBM and its subsidiaries an irrevocable, nonexclusive, 
         worldwide and royalty-free license to use, execute, reproduce, 
         display, perform, prepare derivative works based upon, and 
         distribute, (INTERNALLY AND EXTERNALLY) copies of any and all such 
         materials and derivative works thereof, and to sublicense others to 
         do any, some, or all of the foregoing, (including supporting
         documentation).

         Copies of these modifications should be sent to:

               SOFTWARE @ YKTVMV.BITNET

     g.  Technical data and/or commodities furnished under this agreement
         are subject to United States export/reexport control laws and
         regulations.  The user agrees to comply with such laws and 
         regulations, including complying with the terms of the U.S. 
         license authorizing IBM to furnish the technical data and/or 
         commodities to the user.
     
     h.  The user agrees not to export, directly or indirectly, any technical
         data provided by IBM or any direct product of such data to, or to the
         nationals of, Czechoslovakia, Estonia, Hungary, Iraq, Laos, Latvia,
         Libya, Lithuania, Mongolian Peoples Republic, North Korea, People's
         Republic of China, Poland, Romania, Union of Soviet Socialist
         Republics, and Vietnam.
     
     i.  It is a requirement that the assurance from the recipient of the
         tech data be received by IBM prior to the disclosure or export.  It
         need not be in the actual license/contract/or other agreement
         setting forth the deal itself-a side letter is perfectly acceptable.
         The executed copy of the assurance should be retained locally for
         file purposes.        

******************************************************************************/

#if defined(MSDOS) || defined(OS2V2)
#include <sys/signal.h>
#else
#include <signal.h>
#endif
#include <setjmp.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include "standard.h"
#include "config.h"
#include "parse.h"
#include "symbols.h"
#include "emul.h"
#include "main.h"

#define NARGS_MIN 1            /* minimum command line arguments */
#define NARGS_MAX 100          /* maximum command line arguments */
#define MAX_HISTORY 64 

extern CLPR_int CLPR_HEAP_SZ, CLPR_LSTACK_SZ, CLPR_TRAIL_SZ, CLPR_SOLVER_SZ, CLPR_CODE_SZ;

CLPR_int CLPR_START_USER_SESSION = FALSE;
CLPR_int CLPR_TRACE = FALSE;
CLPR_int CLPR_CLAM_TRACE = FALSE;
CLPR_int CLPR_WARN_ABORT = TRUE;
CLPR_int CLPR_WARNING_FLAG = TRUE;
CLPR_int CLPR_STYLE_CHECK_FLAG = ALL_STYLE;

CLPR_int CLPR_restart_ready = FALSE;     /* OK to do longjmp? */
CLPR_int CLPR_query = 0;                 /* unique stamp of this goal */
FILE *CLPR_infile, *CLPR_outfile;        /* input/output stream numbers */
FILE *CLPR_error_stream;
CLPR_int CLPR_infile_t;                  /* names of input file */
CLPR_int CLPR_outfile_t;                 /* names of output file */
FILE *CLPR_input_fd;                /* temporary input file */
char *CLPR_input_fname;             /* file to be reconsulted immediately */
CLPR_int CLPR_auto_pause_duration;       /* duration of the pause during tracing */
CLPR_int CLPR_REDEFINE_FLAG = TRUE;      /* warn if predicates are redefined */
FILE *CLPR_TEMP_FILE = (FILE *) NULL; /* special temporary file for atom_to_printf */
char *CLPR_TEMP_FNAME = "/tmp/clprbuf.XXXXXX"; /* Only on Unix */

#ifdef SYS5
char stdout_buf[BUFSIZ];
#endif

jmp_buf CLPR_int_restart;           /* Restart vector for errors & interrupt */
PTERM *CLPR_hist[MAX_HISTORY];      /* last n commands */
CLPR_int CLPR_histbase;                  /* current command line */
char *CLPR_cmd_p;                   /* Name of command running clpr */
bool CLPR_eflag;                    /* reported an error for this goal? */

extern FILE *CLPR_cur_stream;       /* parser input */
extern double CLPR_EPS;
extern CLPR_int CLPR_epcode;

extern char *CLPR_the_filename(void);
#ifndef GCC
extern char CLPR_peek(void);
#endif

void CLPR_interrupt(int) NO_RETURN;
void CLPR_bus_error(int) NO_RETURN;
void CLPR_seg_fault(int) NO_RETURN;

/*---------------------------------------------------------------------------*/

/***********************************************/
/*** Initialization in Interactive Main Loop ***/
/***********************************************/


extern void CLPR_test_system_defaults (void);
extern CLPR_int CLPR_myisdigit (char c);
extern void CLPR_cmd_ranseed (CLPR_int i);
extern void CLPR_init_scanner (void);
extern void CLPR_new_input (FILE *stream, char *name);
extern void CLPR_unreadchar (register char c);
extern void CLPR_init_all (void);
extern void CLPR_first_engine (void);
extern void CLPR_bootstrap (void);
extern CLPR_int isatty (CLPR_int);
extern char CLPR_readchar (void);
extern CLPR_int CLPR_internal_consult (char *s);
extern void CLPR_top_free_things (void);
extern void CLPR_files_shutdown (void);

static void usage_error (void);
static void final_clean (void);

CLPR_int CLPR_checkhist(PTERM *t, CLPR_int *np);
void CLPR_pushhist (PTERM *g);
extern CLPR_int CLPR_get_goal (PTERM **t);
extern void CLPR_print_goal (PTERM *t);
extern void CLPR_compile_goal (PTERM *t);

extern void CLPR_save_context (void);
extern CLPR_int CLPR_rungoal (CLPR_int ipc, CLPR_int interactive);
extern void CLPR_restore_context (void);
extern void CLPR_int_clean (void);
extern CLPR_int CLPR_the_lineno (void);
extern void CLPR_clean_temp_file (void);


int CLPR_main(int argc, char **argv)
{
char *opt_p;
CLPR_int i, pc, return_code;
volatile CLPR_int goal_number; /* must be volatile because of setjmp() */
PTERM *t;

	CLPR_error_stream = stderr;
	CLPR_outfile = stdout;
    CLPR_test_system_defaults();
	CLPR_input_fd = NULL;
#if !defined(MSDOS) /* OS2 has bus errors */
	signal(SIGBUS, CLPR_bus_error);
#endif
	signal(SIGSEGV, CLPR_seg_fault);
	argc--;
	argv++;
	while (argc && *argv[0] == '-') {
		opt_p = argv[0];
		while(*(++opt_p)) {
			switch(*opt_p) {
				case 'c':
					if (*(++opt_p) != 's') usage_error();
					argc--;
					argv++;
					if (argc && CLPR_myisdigit(**argv)) CLPR_CODE_SZ = atoi(*argv);
					else usage_error();
					goto GET_MORE_ARGS;
				case 'h':
					if (*(++opt_p) != 's') usage_error();
					argc--;
					argv++;
					if (argc && CLPR_myisdigit(**argv)) CLPR_HEAP_SZ = atoi(*argv);
					else usage_error();
					goto GET_MORE_ARGS;
				case 'l':
					if (*(++opt_p) != 's') usage_error();
					argc--;
					argv++;
					if (argc && CLPR_myisdigit(**argv)) CLPR_LSTACK_SZ = atoi(*argv);
					else usage_error();
					goto GET_MORE_ARGS;
				case 's':
					if (*(++opt_p) != 's') usage_error();
					argc--;
					argv++;
					if (argc && CLPR_myisdigit(**argv)) CLPR_SOLVER_SZ = atoi(*argv);
					else usage_error();
					goto GET_MORE_ARGS;
				case 't':
					if (*(++opt_p) != 's') usage_error();
					argc--;
					argv++;
					if (argc && CLPR_myisdigit(**argv)) CLPR_TRAIL_SZ = atoi(*argv);
					else usage_error();
					goto GET_MORE_ARGS;
				case 'z':
					if (*(++opt_p)) usage_error();
					argc--;
					argv++;
					if (argc && CLPR_myisdigit(**argv)) CLPR_EPS = atof(*argv);
					else usage_error();	
					goto GET_MORE_ARGS;
				case 'r':
					if (*(++opt_p)) usage_error();
					argc--;
					argv++;
					if (argc && CLPR_myisdigit(**argv)) CLPR_cmd_ranseed(atoi(*argv));
					else usage_error();	
					goto GET_MORE_ARGS;
				default: usage_error();
			}
		}
	GET_MORE_ARGS:
		argc--;
		argv++;
	}

	if (argc == 0) CLPR_input_fname = NULL;
	else if (argc == 1) CLPR_input_fname = *argv;
	else if (argc > 1) usage_error();

#ifdef SYS5
	setvbuf(stdout, stdout_buf, _IOLBF, BUFSIZ); /* Buffer output for pipes */
#endif

/*
	if (CLPR_input_fname == NULL) {
		CLPR_infile = stdin;
		CLPR_input_fd = stdin;
	}
*/
	CLPR_input_fd = stdin;

	CLPR_infile_t = addtag(TAG_CON, STDINVAL);
	CLPR_outfile_t = addtag(TAG_CON, USER);
	CLPR_init_scanner();
	CLPR_new_input(stdin,"<stdin>");
	CLPR_unreadchar(' ');
	CLPR_epcode = MAX_GOAL_CODE;
	fprintf(CLPR_outfile, "\nCLP(R) Version 1.2\n");
	fprintf(CLPR_outfile, 
		"(c) Copyright International Business Machines Corporation\n");
	fprintf(CLPR_outfile, "1989 (1991, 1992) All Rights Reserved\n"); 
	fflush(CLPR_outfile);

	/*********/
	CLPR_init_all();
	/*********/

	CLPR_first_engine();
	CLPR_bootstrap();
	CLPR_histbase = 0;
	goal_number = 0;

	if (CLPR_input_fname) {
		CLPR_new_input(stdin, "<stdin>");
		if (isatty(fileno(stdin)) && feof(stdin)) {
		    rewind(stdin);
		    CLPR_unreadchar(' ');
		    CLPR_readchar();
		}
		if (!CLPR_internal_consult(CLPR_input_fname)) {
			CLPR_warning("cannot open file %s", CLPR_input_fname);
		}
	}
	fflush(CLPR_outfile);

	CLPR_START_USER_SESSION = TRUE;
 
	loop {
		setjmp(CLPR_int_restart);
		CLPR_top_free_things();
		CLPR_restart_ready = TRUE;
		signal(SIGINT, CLPR_interrupt);
		CLPR_new_input(stdin, "<stdin>");
		if (isatty(fileno(stdin)) && feof(stdin)) {
		    rewind(stdin);
		    CLPR_unreadchar(' ');
		    CLPR_readchar();
		}
		CLPR_eflag = FALSE;
		printf("\n%d ?- ", goal_number + 1); 
		fflush(stdout);
		if (!(return_code = CLPR_get_goal(&t))) {
			printf("\nEnd of clp(R) Session\n");
			fflush(stdout);
			CLPR_files_shutdown();
			final_clean();
			exit(0);
		}
		if (return_code == 1) {
			if (CLPR_checkhist(t, &i)) {
				i = CLPR_histbase - i;
				if (i >= 0 && i < MAX_HISTORY && CLPR_hist[i]) t = CLPR_hist[i];
				else continue;
				CLPR_print_goal(t);
			}
			CLPR_compile_goal((PTERM *) CLPR_quote_expand(t,FALSE)); 
			goal_number++;
			CLPR_pushhist(t);
			CLPR_query++;
			CLPR_first_engine();
			/***************/
			CLPR_save_context();
			CLPR_rungoal(GOAL_START, TRUE);
			CLPR_restore_context();
			/***************/
			}
		else 
			putchar('\n');
	}
}

/*--------------------------------------------------- aux stuff -------------*/

void CLPR_pushhist(PTERM *g)
{
CLPR_int i;
	for (i = MAX_HISTORY - 1; i > 0; i--) CLPR_hist[i] = CLPR_hist[i - 1];
	CLPR_hist[0] = g;
	CLPR_histbase++;	
}

void CLPR_printhist(CLPR_int n)
{
CLPR_int i;
	n = abs(n);
	for (i = MAX_HISTORY - 1; i > 0; i--) 
		if (CLPR_hist[i] && ((i <= n) || (n == 0))) {
			fprintf(CLPR_outfile, "%d\t", CLPR_histbase - i);
			CLPR_print_goal(CLPR_hist[i]);
		}
}

CLPR_int CLPR_checkhist(PTERM *t, CLPR_int *np)
{
	if (t->ty == REAL) {
		*np = (CLPR_int)(t->rval);
		return TRUE;
	}
	return FALSE;
}

/*---------------------------------------------------------------------------*/

void CLPR_interrupt(int sig)
{
	signal(SIGINT, SIG_IGN);
	CLPR_int_clean();
	CLPR_new_input(stdin, "<stdin>");
	if (!CLPR_restart_ready) exit(1);
	longjmp(CLPR_int_restart, TRUE);
}

void CLPR_top_level(void)
{
	CLPR_interrupt(0);
}

void CLPR_bus_error(int sig)
{
	fprintf(CLPR_error_stream, "Bus error: no core dumped\n");
	exit(1);
}

void CLPR_seg_fault(int sig)
{
	fprintf(CLPR_error_stream, "Segmentation fault: no core dumped\n");
	fflush(CLPR_error_stream);
	exit(1);
}

static void usage_error(void)
{
	fprintf(CLPR_error_stream, "Usage: clpr [-{chlst}s <n>] ");
	fprintf(CLPR_error_stream, "[-z <real>] [-r <int>] [file]\n");
	fflush(CLPR_error_stream);
	exit(2);
}

void CLPR_warning(const char *s, ...)
{
char buf[256];
	if (CLPR_WARNING_FLAG) {
		va_list l;
		va_start(l, s);
		vsprintf(buf, s, l);
		va_end(l);
		fprintf(CLPR_error_stream, "Warning: %s\n", buf);
		fflush(CLPR_error_stream);
	}
}

void CLPR_warn_err(char *s, ...)
{
char buf[256];
	if (CLPR_WARNING_FLAG || CLPR_WARN_ABORT) {
		va_list l;
		va_start(l, s);
		vsprintf(buf, s, l);
		va_end(l);
		fprintf(CLPR_error_stream, "Error: %s\n", buf);
		fflush(CLPR_error_stream);
	}
	if (CLPR_WARN_ABORT) {
		fprintf(CLPR_error_stream, "Aborting\n");
		CLPR_interrupt(0);
	}
}

void CLPR_runtime_error(const char *s, ...)
{
char buf[256];
	va_list l;
	va_start(l, s);
	vsprintf(buf, s, l);
	va_end(l);
	fprintf(CLPR_error_stream, "Error: %s\n", buf);
	fflush(CLPR_error_stream);
	CLPR_top_level();
}

void CLPR_fatal(const char *s)
{
	fprintf(stderr, "Fatal Error: %s\n",s);
	fflush(stderr);
	final_clean();
	exit(1);
}


void CLPR_warn_header(void)
{
	fprintf(CLPR_error_stream, "Warning, %s:%d: ", CLPR_the_filename(),
		CLPR_the_lineno());
}

static void final_clean(void)
{
	CLPR_clean_temp_file();
}
